/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        hwiasync.h

    Purpose:

        Header file for hwiasync.c

--*/

#ifndef __hwiasync_h__
#define __hwiasync_h__

//
// ----------------------------------------------------------------------------
// A note on the structures defined in this file.
// ----------------------------------------------------------------------------
//
// ASYNC_PDU_HANDLE is the software reference for a buffer posted in hardware
// for async PDU. Both headers and data is referenced by the same structure,
// only that is_header will indicate if the handle is for a header or for data
//
// HWI_ASYNC_ENTRY is the software reference to the list that gets built at a
// particular index in the default ring. EP firmware may still be working on a 
// buffer at index idx when another buffer at index idx+i may complete. This is
// then an indication to the driver that the hardware has consumed buffers upto
// idx+i and software may post new buffers from idx to idx+i, provided the 
// buffers are posted in multiples of 8.
//
// Atleast 8 of the above structure is created. For crashdump mode, the driver
// _has_ to post atleast 8 buffers. Otherwise unsolicited PDUs will stall in pmem 
// and iSCSI Rx will not process any more PDUs including the solicited ones
// 
// HWI_ASYNC_PDU_CONTEXT is the main structure that holds all context info
// for async PDU. Information about the host write pointers, free entries
// and lists are all maintained here. This structure is extendible in the
// sense that the number of async_entry structures can be extended beyond
// the default allocation.
// To enable extensibility of this structure, it is allocated off a
// memory descriptor HWI_MEM_ASYNC_PDU_CONTEXT whose size can be configured 
// depending on the number of connections supported. This way crashdump
// drivers will only have the default allocations whereas normal drivers
// will allocate additional entries depending on the connection count.
//
// ----------------------------------------------------------------------------
// A note on the functions declared here.
// ----------------------------------------------------------------------------
//
// hwi_init_async_pdu_ctx is called during initialization. Once initialized,
// these buffers are posted to the hardware via hwi_post_async_buffers. When
// there is a completion, hwi_process_default_pdu_ring is called. 
// 
// This function will gather all the buffers (header+possibly multiple data) 
// belonging to a connection in hwi_gather_async_pdu. When all the buffers 
// have been gathered, the async message is forwarded to iSCSI in 
// hwi_fwd_async_msg. 
// 
// Sometimes iSCSI may not be able to process the message immediately since it 
// may be in the process of offloading the connection. So the buffers are 
// left in the wait_queue maintained for every connection. iSCSI can request 
// delivery of any pending message by calling hwi_deliver_pending_async_pdu. 
//
// When iSCSI has finished processing, the buffers are removed from the 
// wait_queue and added to the free_list. hwi_update_async_writables is then
// called to update 'writables', which is the number of buffers we know that 
// the hardware has consumed and therefore can accept. When the free_entries 
// and writables are atleast a multiple of 8, hwi_post_async_buffers is called 
// to post them to hardware.
//

typedef struct _ASYNC_PDU_HANDLE {

    ACIT_DC_LIST    link;
    ACIT_BUS_ADDRESS pa;
    PVOID           pbuffer;
    ACIT_UEXACT16   consumed;
    ACIT_UEXACT16   completed;
    ACIT_UEXACT8    index;
    ACIT_UEXACT8    is_header;
    ACIT_UEXACT16   cri;
    ACIT_UEXACT32   buffer_len;

} ASYNC_PDU_HANDLE, *PASYNC_PDU_HANDLE;

typedef struct _HWI_ASYNC_ENTRY
{
    //
    // This is the list of async PDUs that are waiting to be processed
    // while the connection is being offloaded. Buffers live in this list for
    // a brief duration before they get processed and posted back to hardware
    //

    // Note that we don't really need one wait_queue per async entry. We need
    // one entry per CRI. But considering the crashdump requirements, its easier 
    // to manage if this is tagged along with the async entry

    ACIT_UEXACT32   wait_entries;
    struct 
    {
        ACIT_UEXACT8    hdr_received;
        ACIT_UEXACT8    hdr_len;
        ACIT_UEXACT16   bytes_received;
        ACIT_UEXACT32   bytes_needed;
        ACIT_DC_LIST    list;
    } wait_queue;
    
    //
    // List of buffers pending with the hardware. It includes buffers that are
    // consumed, but not yet completed
    //
    ACIT_DC_LIST    header_busy_list;
    ACIT_DC_LIST    data_busy_list;

} HWI_ASYNC_ENTRY, *PHWI_ASYNC_ENTRY;

typedef struct _HWI_ASYNC_PDU_CONTEXT {

    ACIT_UEXACT32   buffer_size;

    struct {

        ACIT_BUS_ADDRESS pa_base;
        PACIT_VOID      ring_base;
        PACIT_VOID      va_base;
        PASYNC_PDU_HANDLE handle_base;

        ACIT_UEXACT32   host_write_ptr;
        ACIT_UEXACT32   ep_read_ptr;
        ACIT_UEXACT32   writables;

        ACIT_UEXACT32   free_entries;
        ACIT_UEXACT32   busy_entries;

        //
        // List of free async PDU buffers. Once iSCSI finishes processing an
        // async message, the buffers corresponding to the message are added
        // to this list and when they reach the threshold of 8 they get posted 
        // back to hardware
        //
        ACIT_DC_LIST    free_list;

    } async_header;

    struct {

        ACIT_BUS_ADDRESS pa_base;
        PACIT_VOID      ring_base;
        PACIT_VOID      va_base;
        PASYNC_PDU_HANDLE handle_base;

        ACIT_UEXACT32   host_write_ptr;
        ACIT_UEXACT32   ep_read_ptr;
        ACIT_UEXACT32   writables;

        ACIT_UEXACT32   free_entries;
        ACIT_UEXACT32   busy_entries;

        //
        // List of free async PDU buffers. Once iSCSI finishes processing an
        // async message, the buffers corresponding to the message are added
        // to this list and when they reach the threshold of 8 they get posted 
        // back to hardware
        //
        ACIT_DC_LIST    free_list;

    } async_data;

    //
    // Temp buffer to hold the default PDU data being dispatched to iSCSI
    ACIT_UEXACT8    buffer[64];

    // number of async_entry structures.
    ACIT_UEXACT32   num_entries;

    // This is a varying size list! Do not add anything after this entry!!
    HWI_ASYNC_ENTRY async_entry[BE_MIN_ASYNC_ENTRIES];
    
} HWI_ASYNC_PDU_CONTEXT, *PHWI_ASYNC_PDU_CONTEXT;

ACIT_VOID
hwi_init_async_pdu_ctx (
    PACIT_CONTROLLER_WS     pcontroller);

PACIT_DC_LIST
hwi_get_async_busy_list (
    PHWI_ASYNC_PDU_CONTEXT  pasync_ctx,
    ACIT_UEXACT32           is_header,
    ACIT_UEXACT32           host_write_ptr);

PPHYS_ADDR
hwi_get_ring_address (
    PHWI_ASYNC_PDU_CONTEXT  pasync_ctx,
    ACIT_UEXACT32           is_header,
    ACIT_UEXACT32           host_write_ptr);

PASYNC_PDU_HANDLE
hwi_get_async_handle (
    PACIT_CONTROLLER_WS     pcontroller,
    PHWI_ASYNC_PDU_CONTEXT  pasync_ctx,
    PI_T_DPDU_CQE           pdpdu_cqe,
    PACIT_UEXACT32          pcq_index);

ACIT_UEXACT32
hwi_gather_async_pdu (
    PACIT_CONTROLLER_WS     pcontroller,
    PASYNC_PDU_HANDLE       pasync_handle);

ACIT_UEXACT32
hwi_fwd_async_msg (
    PACIT_CONTROLLER_WS     pcontroller,
    PHWI_ASYNC_PDU_CONTEXT  pasync_ctx,
    ACIT_UEXACT16           cri);

ACIT_UEXACT32
hwi_deliver_pending_async_pdu (
    PACIT_CONTROLLER_WS     pcontroller, 
    ACIT_UEXACT32           cri);

ACIT_UEXACT32
hwi_free_async_msg (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cri);

ACIT_UEXACT32
hwi_update_async_writables (
    PHWI_ASYNC_PDU_CONTEXT  pasync_ctx, 
    ACIT_UEXACT32           is_header, 
    ACIT_UEXACT32           cq_index);

ACIT_UEXACT32
hwi_post_async_buffers (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           is_header);

ACIT_STATUS
hwi_process_default_pdu_ring (
    PACIT_CONTROLLER_WS     pcontroller,
    PI_T_DPDU_CQE           pdpdu_cqe);

ACIT_STATUS
hwi_flush_default_pdu_buffer (
    PACIT_CONTROLLER_WS     pcontroller,
    PI_T_DPDU_CQE           pdpdu_cqe);

#endif
