/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
//
#ifndef __be2_gen_erx_h__
#define __be2_gen_erx_h__

#include "be2_gen_check_defines.h"
#include "be2_gen_api_enum.h"
#include "be2_gen_misc.h"

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_ERX_MTU_Config
//
//   Configures the MTU for the ERX ring. This is actually the MAX Frame size
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// ring_id            - ERX ring ID. 
// mtu                - Maximum transmission unit value for the ERX ring. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_ERX_MTU_Config (
   IN void* function_context,
   IN u32   ring_id,
   IN u32   mtu )
{
    BE_GEN_CSR_WRITE(function_context, erx.mtu[ ring_id ], mtu);
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_ERX_Ring_Init
//
//   Programs an Ethernet receive ring into BladeEngine2. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// ring_id            - ERX ring ID. 
// p_phys_addr        - Pointer to an array of two physical addresses representing the ERX
//                      ring. 
// pd_id_reg          - Protection domain configuration register value for the ERX ring. 
// cq_id              - Completion queue ID for the ERX ring. 
// mtu                - Maximum transmission unit value for the ERX ring. 
// rss_interface      - RSS interface number (if an RSS ring). 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_ERX_Ring_Init (
   IN void* function_context,
   IN u32   ring_id,
   IN u64*  p_phys_addr,
   IN u32   pd_id_reg,
   IN u32   cq_id,
   IN u32   mtu,
   IN u8    rss_interface )
{
    int i;
    int j;
    u32 addr;

    // Program the physical addresses
    for( i = 0 ; i < 2 ; ++i ) {
        for( j = 0 ; j < 2 ; j++ ) {
            addr = (u32)(p_phys_addr[i] >> (j * 32));
            BE_GEN_CSR_WRITE(function_context, erx.addr_idx, ring_id * 4 + i * 2 + j);
            BE_GEN_CSR_WRITE(function_context, erx.base_addr, addr);
        }
    }
    
    // Program cqid
    BE_GEN_CSR_WRITE(function_context, erx.cqid[ ring_id ], cq_id);

    // Program pdid register
    BE_GEN_CSR_WRITE(function_context, erx.pdid[ ring_id ], pd_id_reg);
    
    // Program MTU
    BE2_Gen_ERX_MTU_Config(function_context, ring_id, mtu);

    if (ring_id > 35)
    {
        // An RSS ring
        ERX_RSS_ASSOCIATION_CSR rss_association;
        rss_association.dw = BE_GEN_CSR_READ(function_context, erx.rss_association);
        switch (rss_interface) 
        {
        case BE2_GEN_RSS_INTERFACE_0:
            rss_association.interface0 |= (1 << (ring_id - 36));
            rss_association.interface1 &= ~(1 << (ring_id - 36));
            break;

        case BE2_GEN_RSS_INTERFACE_1:
            rss_association.interface0 &= ~(1 << (ring_id - 36));
            rss_association.interface1 |= (1 << (ring_id - 36));
            break;

        case BE2_GEN_RSS_INTERFACE_NONE:
        default:
            rss_association.interface0 &= ~(1 << (ring_id - 36));
            rss_association.interface1 &= ~(1 << (ring_id - 36));
            break;
        }

        BE_GEN_CSR_WRITE (function_context, erx.rss_association, rss_association.dw);
    }
    return;
}

#endif  // __be2_gen_erx_h__

