/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
//
#ifndef __be2_gen_gpio_h__
#define __be2_gen_gpio_h__

#include "be2_gen_check_defines.h"
#include "be2_gen_api_enum.h"

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_GPIO_Set_Mode
//
//   This routine is used to set the GPI/GPO mode for a particular GPIO bit 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// gpio_bit           - Selects one of 64 available GPIO bits. 
// gpio_mode          - Selects a particular GPI or GPO mode. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_GPIO_Set_Mode (
   IN void* function_context,
   IN u8    gpio_bit,
   IN u8    gpio_mode )
{
    u32 gpiomap;
    u8 which_gpiomap;
    u8 mode_shift;

    //
    // There are 4 bits that determine the behavior mode of a particular GPIO bit.
    // As there are 64 GPIO bits, there are 8 "gpiomap" registers (each controlling 
    // the mode for 8 GPIO bits).
    //
    BE_GEN_ASSERT(gpio_bit < 64);

    // Determine the GPIOMAP register for the GPIO bit
    which_gpiomap = (gpio_bit & 0x3F) >> 3;
    
    // Number of bit positions to shift for the GPIO mode
    mode_shift = (gpio_bit & 0x7) * 4;

    // Update the mode for the GPIO bit
    gpiomap = BE_GEN_PCI0_READ(function_context, gpiomap[which_gpiomap]);
    gpiomap &= ~(0xF << mode_shift);
    gpiomap |= (gpio_mode << mode_shift);
    BE_GEN_PCI0_WRITE(function_context, gpiomap[which_gpiomap], gpiomap);

    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_GPO_Set_Data
//
//   This routine is used to set the data value for a particular GPO bit. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// gpo_bit            - Selects one of 64 available GPO bits. 
// gpo_value          - GPO bit's data value (0 or 1). 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_GPO_Set_Data (
   IN void*   function_context,
   IN u8      gpo_bit,
   IN boolean gpo_value )
{
    u32 gpo_data;
    u32 gpo_mask;

    //
    // Write the GPO data.
    // There are 64 gpio bits.
    // 0..31 use PCICFG_GPO_LOW_CSR,
    // 32..63 use PCICFG_GPO_HI_CSR
    //
    if (gpo_bit < 32) {
        gpo_data = BE_GEN_PCI0_READ(function_context, gpo_low);
    } else {
        BE_GEN_ASSERT(gpo_bit < 64);
        gpo_data = BE_GEN_PCI0_READ(function_context, gpo_hi);
    }

    // Mask for this bit position (0 - 31 normalized)
    gpo_mask = 1 << (gpo_bit & 0x1F);
    if (gpo_value) {
        gpo_data |= gpo_mask;
    } else {
        gpo_data &= ~gpo_mask;
    }

    if (gpo_bit < 32) {
        BE_GEN_PCI0_WRITE(function_context, gpo_low, gpo_data);
    } else {
        BE_GEN_PCI0_WRITE(function_context, gpo_hi, gpo_data);
    }

    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_GPI_Get_Data
//
//   This routine is used to get the data value for a particular GPI bit. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// gpi_bit            - Selects one of 64 available GPO bits. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE boolean
BE2_Gen_GPI_Get_Data (
   IN void* function_context,
   IN u8    gpi_bit )
{
    u32 gpi_data;
    u32 gpi_mask;

    //
    // Read the GPI data.
    // There are 64 gpio bits.
    // 0..31 use PCICFG_GPI_CSR[0],
    // 32..63 use PCICFG_GPI_CSR[1].
    //
    if (gpi_bit < 32) {
        gpi_data = BE_GEN_PCI0_READ(function_context, gpi[0]);
    } else {
        BE_GEN_ASSERT(gpi_bit < 64);
        gpi_data = BE_GEN_PCI0_READ(function_context, gpi[1]);
    }

    // Mask for this bit's position (0-31 normalized)
    gpi_mask = 1 << (gpi_bit & 0x1F);
    
    // Return bit's value
    return (!!(gpi_data & gpi_mask));
}

#endif  // __be2_gen_gpio_h__

