/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
// 
// @file
//      be2_gen_id_ranges.h
// 
// @brief
//     Provides ID ranges and conversion macros between the types of ID numbers.
//    
// 
#ifndef __be2_gen_id_ranges_h__
#define __be2_gen_id_ranges_h__

#ifndef BE_CONFIG
    #pragma message("WARNING: configuration not defined. Assuming BE_CONFIG=0 by default.")
    #define BE_CONFIG 0
#endif

#ifndef BE_GEN_ASSERT
    #error "Error: BE_GEN_ASSERT not defined."
#endif

#ifndef BE_GEN_STATIC_INLINE
    #error "Error: BE_GEN_STATIC_INLINE not defined."
#endif


/////////////////////////////////////////////////////////////////////////////// 
// NOTICE NOTICE NOTICE NOTICE NOTICE NOTICE NOTICE NOTICE NOTICE NOTICE NOTICE
//
// 12/06/2008 --- update by Vahid 
// ALL ID definitions within this file are being deprecated by the new 
// BE_CONFIG framework whereing each BE_CONFIG has an ID ranges SG file under 
// fw\config\.
// 
// The only thing defined within this file are macros to validate and convert 
// between different ID values (ie. ZID, RID, BID, QID, and CID).
//
// There are 9 generic ring types: 
// 1) Event queues (EQ)
// 2) Completion queues (CQ)
// 3) Managment queues (MCC)
// 4) Ethernet transmit queues (ETX)
// 5) Protocol send queues (SQ)
// 6) Large receive offload queues (LRO)
// 7) Protocol default receive queues (DQ)
// 8) Protocol preposted receive queues (RQ)
// 
// ID label reference:
// * ZID - zero-based ID. The ZID range is specific to a ULP. This range 
// always starts at zero for every ULP.
// * RID - relative zero-based ID. The RID range reresents the IDs available
// accross the ULPs. 
// * BID - butler ID. The BID range represents the IDs used when writing a 
// doorbell register.
// * QID - quadrant ID. The QID range represents the IDs used when forming the
// CRA (for Unified Table purposes).
// * CID - connection ID. The CID range represents the IDs used when accessing
// TCP context.
//
// Obviously all ID labels are not valid for each ring type. EQs do not have 
// TCP context, so the CID range for EQs is not applicable. In fact only 
// SQ and RQ rings are directly associated with TCP context. Note that for 
// legacy reasons we typically also allow ETX and LRO rings to be referenced 
// by a CID, despite that being a bad label for those IDs. The ETX CID is 
// really the BID... 
//
///////////////////////////////////////////////////////////////////////////////


// PRIVATE CONVERSION FUNCTIONS. DO NOT USE THESE DIRECTLY IN SOFTWARE.
#define ZID_TO_RID(_name_, _zid_)  ((_zid_) + _name_##_RID_START)
#define ZID_TO_BID(_name_, _zid_)  ((_zid_) + _name_##_BID_START)
#define ZID_TO_CID(_name_, _zid_)  ((_zid_) + _name_##_CID_START)
#define ZID_TO_QID(_name_, _zid_)  ((_zid_) + _name_##_QID_START)

#define RID_TO_ZID(_name_, _rid_)  ((_rid_) - _name_##_RID_START)
#define RID_TO_BID(_name_, _rid_)  ((_rid_) + _name_##_BID_START - _name_##_RID_START)
#define RID_TO_QID(_name_, _rid_)  ((_rid_) + _name_##_QID_START - _name_##_RID_START)
#define RID_TO_CID(_name_, _rid_)  ((_rid_) + _name_##_CID_START - _name_##_RID_START)

#define BID_TO_ZID(_name_, _bid_)  ((_bid_) - _name_##_BID_START)
#define BID_TO_RID(_name_, _bid_)  ((_bid_) + _name_##_RID_START - _name_##_BID_START)
#define BID_TO_QID(_name_, _bid_)  ((_bid_) + _name_##_QID_START - _name_##_BID_START)
#define BID_TO_CID(_name_, _bid_)  ((_bid_) + _name_##_CID_START - _name_##_BID_START)

#define QID_TO_ZID(_name_, _qid_)  ((_qid_) - _name_##_QID_START)
#define QID_TO_RID(_name_, _qid_)  ((_qid_) + _name_##_RID_START - _name_##_QID_START)
#define QID_TO_BID(_name_, _qid_)  ((_qid_) + _name_##_BID_START - _name_##_QID_START)
#define QID_TO_CID(_name_, _qid_)  ((_qid_) + _name_##_CID_START - _name_##_QID_START)

#define CID_TO_ZID(_name_, _cid_)  ((_cid_) - _name_##_CID_START)
#define CID_TO_RID(_name_, _cid_)  ((_cid_) + _name_##_RID_START - _name_##_CID_START)
#define CID_TO_BID(_name_, _cid_)  ((_cid_) + _name_##_BID_START - _name_##_CID_START)
#define CID_TO_QID(_name_, _cid_)  ((_cid_) + _name_##_QID_START - _name_##_CID_START)

#define VALID_ZID(_name_, _zid_) ((_zid_) && ((_zid_) <= _name_##_COUNT))
#define VALID_RID(_name_, _rid_) (((_rid_) >= _name_##_RID_START) && ((_rid_) < _name_##_RID_START + _name_##_COUNT))
#define VALID_BID(_name_, _bid_) (((_bid_) >= _name_##_BID_START) && ((_bid_) < _name_##_BID_START + _name_##_COUNT))
#define VALID_QID(_name_, _qid_) (((_qid_) >= _name_##_QID_START) && ((_qid_) < _name_##_QID_START + _name_##_COUNT))
#define VALID_CID(_name_, _cid_) (((_cid_) >= _name_##_CID_START) && ((_cid_) < _name_##_CID_START + _name_##_COUNT))


// Public ring type-specific conversion functions.
#define DECLARE_ID_CONVERSIONS(_name_)                                                                                                  \
    BE_GEN_STATIC_INLINE u32 _name_##_VALID_ZID (u32 zid) { return (u32)VALID_ZID(_name_,zid); }                                        \
    BE_GEN_STATIC_INLINE u32 _name_##_VALID_RID (u32 rid) { return (u32)VALID_RID(_name_,rid); }                                        \
    BE_GEN_STATIC_INLINE u32 _name_##_VALID_BID (u32 bid) { return (u32)VALID_BID(_name_,bid); }                                        \
    BE_GEN_STATIC_INLINE u32 _name_##_VALID_QID (u32 qid) { return (u32)VALID_QID(_name_,qid); }                                        \
    BE_GEN_STATIC_INLINE u32 _name_##_VALID_CID (u32 cid) { return (u32)VALID_CID(_name_,cid); }                                        \
    \
    BE_GEN_STATIC_INLINE u32 _name_##_ZID_TO_RID (u32 zid) { BE_GEN_ASSERT (_name_##_VALID_ZID(zid)); return ZID_TO_RID(_name_,zid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_ZID_TO_BID (u32 zid) { BE_GEN_ASSERT (_name_##_VALID_ZID(zid)); return ZID_TO_BID(_name_,zid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_ZID_TO_QID (u32 zid) { BE_GEN_ASSERT (_name_##_VALID_ZID(zid)); return ZID_TO_QID(_name_,zid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_ZID_TO_CID (u32 zid) { BE_GEN_ASSERT (_name_##_VALID_ZID(zid)); return ZID_TO_CID(_name_,zid); }  \
    \
    BE_GEN_STATIC_INLINE u32 _name_##_RID_TO_ZID (u32 rid) { BE_GEN_ASSERT (_name_##_VALID_RID(rid)); return RID_TO_ZID(_name_,rid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_RID_TO_BID (u32 rid) { BE_GEN_ASSERT (_name_##_VALID_RID(rid)); return RID_TO_BID(_name_,rid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_RID_TO_QID (u32 rid) { BE_GEN_ASSERT (_name_##_VALID_RID(rid)); return RID_TO_QID(_name_,rid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_RID_TO_CID (u32 rid) { BE_GEN_ASSERT (_name_##_VALID_RID(rid)); return RID_TO_CID(_name_,rid); }  \
    \
    BE_GEN_STATIC_INLINE u32 _name_##_BID_TO_ZID (u32 bid) { BE_GEN_ASSERT (_name_##_VALID_BID(bid)); return BID_TO_ZID(_name_,bid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_BID_TO_RID (u32 bid) { BE_GEN_ASSERT (_name_##_VALID_BID(bid)); return BID_TO_RID(_name_,bid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_BID_TO_QID (u32 bid) { BE_GEN_ASSERT (_name_##_VALID_BID(bid)); return BID_TO_QID(_name_,bid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_BID_TO_CID (u32 bid) { BE_GEN_ASSERT (_name_##_VALID_BID(bid)); return BID_TO_CID(_name_,bid); }  \
    \
    BE_GEN_STATIC_INLINE u32 _name_##_QID_TO_ZID (u32 qid) { BE_GEN_ASSERT (_name_##_VALID_QID(qid)); return QID_TO_ZID(_name_,qid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_QID_TO_RID (u32 qid) { BE_GEN_ASSERT (_name_##_VALID_QID(qid)); return QID_TO_RID(_name_,qid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_QID_TO_BID (u32 qid) { BE_GEN_ASSERT (_name_##_VALID_QID(qid)); return QID_TO_BID(_name_,qid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_QID_TO_CID (u32 qid) { BE_GEN_ASSERT (_name_##_VALID_QID(qid)); return QID_TO_CID(_name_,qid); }  \
    \
    BE_GEN_STATIC_INLINE u32 _name_##_CID_TO_ZID (u32 cid) { BE_GEN_ASSERT (_name_##_VALID_BID(cid)); return CID_TO_ZID(_name_,cid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_CID_TO_BID (u32 cid) { BE_GEN_ASSERT (_name_##_VALID_CID(cid)); return CID_TO_BID(_name_,cid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_CID_TO_QID (u32 cid) { BE_GEN_ASSERT (_name_##_VALID_CID(cid)); return CID_TO_QID(_name_,cid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_CID_TO_RID (u32 cid) { BE_GEN_ASSERT (_name_##_VALID_CID(cid)); return CID_TO_RID(_name_,cid); }  \


// ERX IDs (normal ERX and RSS rings)
#define ERX_COUNT (44)
#define ERX_ID_NUM (ERX_COUNT)
#define ERX_RID_START (0)
#define ERX_BID_START (0)
#define ERX_QID_START (0x1000)  // not valid: ERX does not use UT
#define ERX_CID_START (0x1000)  // not valid: CID concept is only applicable to protocol SQ/RQ
DECLARE_ID_CONVERSIONS(ERX)

// CQ IDs
#define CQ_COUNT (1024)
#define CQ_ID_NUM (CQ_COUNT)
#define CQ_RID_START (0)
#define CQ_BID_START (0)
#define CQ_QID_START (0)
#define CQ_CID_START (0x1000)   // not valid: CID concept is only applicable to protocol SQ/RQ
DECLARE_ID_CONVERSIONS(CQ)

// EQ IDs.
#define EQ_COUNT (192)
#define EQ_ID_NUM (EQ_COUNT)
#define EQ_RID_START (0)
#define EQ_BID_START (0)
#define EQ_QID_START (0)
#define EQ_CID_START (0x1000)   // not valid: CID concept is only applicable to protocol SQ/RQ
DECLARE_ID_CONVERSIONS(EQ)

// MCC IDs
#define MCC_COUNT (8)
#define MCC_ID_NUM (MCC_COUNT)
#define MCC_ZID_START (0)
#define MCC_RID_START (0)
#define MCC_BID_START (0)
#define MCC_QID_START (0x38)
#define MCC_CID_START (0x1000)  // not valid: CID concept is only applicable to protocol SQ/RQ
DECLARE_ID_CONVERSIONS(MCC)

// ETX IDs
DECLARE_ID_CONVERSIONS(TXULP0_ETX)
DECLARE_ID_CONVERSIONS(TXULP1_ETX)

// SQ IDs
DECLARE_ID_CONVERSIONS(TXULP0_SQ)
DECLARE_ID_CONVERSIONS(TXULP1_SQ)

// LRO IDs
DECLARE_ID_CONVERSIONS(RXULP0_LRO)
DECLARE_ID_CONVERSIONS(RXULP1_LRO)

// RQ IDs
DECLARE_ID_CONVERSIONS(RXULP0_RQ)
DECLARE_ID_CONVERSIONS(RXULP1_RQ)

// DQ IDs
DECLARE_ID_CONVERSIONS(RXULP0_DQ)
DECLARE_ID_CONVERSIONS(RXULP1_DQ)



///////////////////////////////////////////////////////////////////////////////
// EVERYTHING BELOW HERE IS BEING DEPRECATED. PLEASE MOVE TO USING THE NEW
// BE_CONFIG FRAMEWORK LITERALS!!!! THIS WILL BE DELETED EVENTUALLY!!!!!!!!!!!!
///////////////////////////////////////////////////////////////////////////////

#if (BE_CONFIG==0)

    // --- CID_START_ENUM ---
    #define ISCSI_CID_START                 (0)     // Start of connection ID range for ISCSI
    #define DEFPDU_CID_START                (0)     // Start of connection ID range for iSCSI default pdu queue
    #define TOEDQ_CID_START                 (568)   // Start of connection ID range for TOE dq
    #define ETHTOE_CID_START                (576)   // Start of connection ID range for TOE
    #define ETX_CID_START                   (1024)  // Start of connection ID range for ETX
    #define RDMA_CID_START                  (0)     // Start of connection ID range for RDMA
    #define FC_CID_START                    (0)     // Start of virtual device ID range for FCIP
    #define LRO_CID_START                   (ETX_CID_START) // Start of connection ID range for LRO (shares ETX CID range)
    #define DAL_CID_START                   (0)     // Start of connection ID range for DAL
    #define DAL_RX_CID_START                (0)      // Start of connection ID range for DAL Receive

    // --- MAX_CIDS_ENUM ---
    #define MAX_TOEDQ_CIDS                  (8)     // Maximum number of TOE DQ cids
    #define MAX_DEFPDU_CIDS                 (32)    // Maximum number of iSCSI DEF pdu cids
    #define MAX_ETX_CIDS                    (64)    // Maximum number of ETX connections
    #define MAX_ETHTOE_CIDS                 (448)   // Maximum number of TOE conections
    #define MAX_ISCSI_CIDS                  (512)   // Maximum number of ISCSI connections
    #define MAX_ISCSI_CIDS_CT               (512)   // Maximum number of ISCSI connections per chute
    #define MAX_RDMA_CIDS                   (0)     // Maximum number of RDMA connections
    #define MAX_FC_CIDS                     (0)     // Maximum number of FCIP virtual device ID
    #define MAX_LRO_CIDS                    (18)    // Maximum number of LRO connections
    #define MAX_DAL_CIDS                    (0)     // Maximum number of DAL connections

#elif (BE_CONFIG==1)

///////////////////////////////////////////////////////////////////////////////
// EVERYTHING BELOW HERE IS BEING DEPRECATED. PLEASE MOVE TO USING THE NEW
// BE_CONFIG FRAMEWORK LITERALS!!!! THIS WILL BE DELETED EVENTUALLY!!!!!!!!!!!!
///////////////////////////////////////////////////////////////////////////////

    // --- CID_START_ENUM ---
    #define ISCSI_CID_START                 (0)     // Start of connection ID range for ISCSI
    #define DEFPDU_CID_START                (0)     // Start of connection ID range for iSCSI default pdu queue
    #define RDMA_CID_START                  (64)    // Start of connection ID range for RDMA
    #define TOEDQ_CID_START                 (568)   // Start of connection ID range for TOE dq
    #define ETHTOE_CID_START                (576)   // Start of connection ID range for TOE
    #define ETX_CID_START                   (1024)  // Start of connection ID range for ETX
    #define FC_CID_START                    (0)     // Start of virtual device ID range for FCIP
    #define LRO_CID_START                   (0)     // Start of connecting ID range for LRO
    #define DAL_CID_START                   (0) // Start of connection ID range for DAL
    #define DAL_RX_CID_START                (0)      // Start of connection ID range for DAL Receive

    // --- MAX_CIDS_ENUM ---
    #define MAX_TOEDQ_CIDS                  (8)     // Maximum number of TOE DQ cids
    #define MAX_DEFPDU_CIDS                 (32)    // Maximum number of iSCSI DEF pdu cids
    #define MAX_ISCSI_CIDS                  (64)    // Maximum number of ISCSI connections
    #define MAX_ISCSI_CIDS_CT               (64)    // Maximum number of ISCSI connections per chute
    #define MAX_ETX_CIDS                    (64)    // Maximum number of ETX connections
    #define MAX_ETHTOE_CIDS                 (448)   // Maximum number of TOE conections
    #define MAX_RDMA_CIDS                   (512)   // Maximum number of RDMA connections
    #define MAX_FC_CIDS                     (0)     // Maximum number of FCIP virtual device ID
    #define MAX_LRO_CIDS                    (4)     // Maximum number of LRO connections
    #define MAX_DAL_CIDS                    (0)     // Maximum number of DAL connections

#elif (BE_CONFIG==2)

///////////////////////////////////////////////////////////////////////////////
// EVERYTHING BELOW HERE IS BEING DEPRECATED. PLEASE MOVE TO USING THE NEW
// BE_CONFIG FRAMEWORK LITERALS!!!! THIS WILL BE DELETED EVENTUALLY!!!!!!!!!!!!
///////////////////////////////////////////////////////////////////////////////

    // --- CID_START_ENUM ---
    #define DEFPDU_CID_START                (0)     // Start of connection ID range for iSCSI default pdu queue
    #define ISCSI_TGT_CID_START             (448)   // Start of connection ID range for ISCSI
    #define ISCSI_CID_START                 (ISCSI_TGT_CID_START)
    #define TOEDQ_CID_START                 (568)   // Start of connection ID range for TOE dq
    #define ETHTOE_CID_START                (960)   // Start of connection ID range for Ethernet/TOE
    #define ETHTOE_TGT_CID_START            (960)   // Start of connection ID range for Ethernet/TOE
    #define ETX_CID_START                   (1024)  // Start of connection ID range for ETX
    #define RDMA_CID_START                  (0)     // Start of connection ID range for RDMA
    #define FC_CID_START                    (0)     // Start of virtual device ID range for FCIP
    #define LRO_CID_START                   (0)     // Start of connecting ID range for LRO
    #define DAL_CID_START                   (0) // Start of connection ID range for DAL
    #define DAL_RX_CID_START                (0)      // Start of connection ID range for DAL Receive

    // --- MAX_CIDS_ENUM ---
    #define MAX_TOEDQ_CIDS                  (8)     // Maximum number of TOE DQ cids
    #define MAX_DEFPDU_CIDS                 (32)    // Maximum number of iSCSI DEF pdu cids
    #define MAX_ETHTOE_CIDS                 (64)    // Maximum number of TOE conections
    #define MAX_ETX_CIDS                    (64)    // Maximum number of ETX cids
    #define MAX_ISCSI_TGT_CIDS              (512)   // Maximum number of ISCSI connections
    #define MAX_ISCSI_CIDS                  (MAX_ISCSI_TGT_CIDS)
    #define MAX_ISCSI_CIDS_CT               (256)   // Maximum number of ISCSI connections per chute
    #define MAX_RDMA_CIDS                   (0)     // Maximum number of RDMA connections
    #define MAX_FC_CIDS                     (0)     // Maximum number of FCIP virtual device ID
    #define MAX_LRO_CIDS                    (0)     // Maximum number of LRO connections
    #define MAX_DAL_CIDS                    (0)     // Maximum number of DAL connections

#elif (BE_CONFIG==3)

///////////////////////////////////////////////////////////////////////////////
// EVERYTHING BELOW HERE IS BEING DEPRECATED. PLEASE MOVE TO USING THE NEW
// BE_CONFIG FRAMEWORK LITERALS!!!! THIS WILL BE DELETED EVENTUALLY!!!!!!!!!!!!
///////////////////////////////////////////////////////////////////////////////

    // --- CID_START_ENUM ---
    #define DEFPDU_CID_START                (0)     // Start of connection ID range for iSCSI default pdu queue
    #define ISCSI_TGT_CID_START             (0)     // Start of connection ID range for ISCSI
    #define ISCSI_CID_START                 (ISCSI_TGT_CID_START)
    #define TOEDQ_CID_START                 (568)   // Start of connection ID range for TOE dq
    #define ETHTOE_CID_START                (576)  // Start of connection ID range for TOE
    #define ETHTOE_TGT_CID_START            (1024)  // Start of connection ID range for Ethernet/TOE
    #define ETX_CID_START                   (1024)  // Start of connection ID range for ETX
    #define RDMA_CID_START                  (0)     // Start of connection ID range for RDMA
    #define FC_CID_START                    (0)     // Start of virtual device ID range for FCIP
    #define LRO_CID_START                   (0)     // Start of connecting ID range for LRO
    #define DAL_CID_START                   (0) // Start of connection ID range for DAL
    #define DAL_RX_CID_START                (0)      // Start of connection ID range for DAL Receive

    // --- MAX_CIDS_ENUM ---
    #define MAX_TOEDQ_CIDS                  (0)     // Maximum number of TOE DQ cids
    #define MAX_DEFPDU_CIDS                 (32)    // Maximum number of iSCSI DEF pdu cids
    #define MAX_ETHTOE_CIDS                 (0)     // Maximum number of Ethernet/TOE conections
    #define MAX_ETHTOE_TGT_CIDS             (0)     // Maximum number of Ethernet/TOE connections
    #define MAX_ETX_CIDS                    (64)    // Maximum number of ETX cids
    #define MAX_ISCSI_TGT_CIDS              (1024)  // Maximum number of ISCSI connections supported.
    #define MAX_ISCSI_CIDS                  (MAX_ISCSI_TGT_CIDS)
    #define MAX_ISCSI_CIDS_CT               (512)   // Maximum number of ISCSI connections per chute
    #define MAX_RDMA_CIDS                   (0)     // Maximum number of RDMA connections
    #define MAX_FC_CIDS                     (0)     // Maximum number of FCIP virtual device ID
    #define MAX_LRO_CIDS                    (0)     // Maximum number of LRO connections
    #define MAX_DAL_CIDS                    (0)     // Maximum number of DAL connections

#elif (BE_CONFIG==4)

///////////////////////////////////////////////////////////////////////////////
// EVERYTHING BELOW HERE IS BEING DEPRECATED. PLEASE MOVE TO USING THE NEW
// BE_CONFIG FRAMEWORK LITERALS!!!! THIS WILL BE DELETED EVENTUALLY!!!!!!!!!!!!
///////////////////////////////////////////////////////////////////////////////


    // --- CID_START_ENUM ---
    #define ISCSI_CID_START                 (0)     // Start of connection ID range for ISCSI
    #define DEFPDU_CID_START                (0)     // Start of connection ID range for iSCSI default pdu queue
    #define TOEDQ_CID_START                 (568)   // Start of connection ID range for TOE dq
    #define ETHTOE_CID_START                (576)   // Start of connection ID range for TOE
    #define ETX_CID_START                   (1024)  // Start of connection ID range for ETX
    #define RDMA_CID_START                  (0)     // Start of connection ID range for RDMA
    #define FC_CID_START                    (0)     // Start of virtual device ID range for FCIP
    #define LRO_CID_START                   (0)     // Start of connecting ID range for LRO
    #define DAL_CID_START                   (0) // Start of connection ID range for DAL
    #define DAL_RX_CID_START                (0)      // Start of connection ID range for DAL Receive

    // --- MAX_CIDS_ENUM ---
    #define MAX_TOEDQ_CIDS                  (8)     // Maximum number of TOE DQ cids
    #define MAX_DEFPDU_CIDS                 (32)    // Maximum number of iSCSI DEF pdu cids
    #define MAX_ETX_CIDS                    (64)    // Maximum number of ETX connections
    #define MAX_ETHTOE_CIDS                 (448)   // Maximum number of TOE conections
    #define MAX_ISCSI_CIDS                  (512)   // Maximum number of ISCSI connections
    #define MAX_ISCSI_CIDS_CT               (256)   // Maximum number of ISCSI connections per chute
    #define MAX_RDMA_CIDS                   (0)     // Maximum number of RDMA connections
    #define MAX_FC_CIDS                     (0)     // Maximum number of FCIP virtual device ID
    #define MAX_LRO_CIDS                    (0)     // Maximum number of LRO connections
    #define MAX_DAL_CIDS                    (0)     // Maximum number of DAL connections

#elif (BE_CONFIG==6)

///////////////////////////////////////////////////////////////////////////////
// EVERYTHING BELOW HERE IS BEING DEPRECATED. PLEASE MOVE TO USING THE NEW
// BE_CONFIG FRAMEWORK LITERALS!!!! THIS WILL BE DELETED EVENTUALLY!!!!!!!!!!!!
///////////////////////////////////////////////////////////////////////////////

    // --- CID_START_ENUM ---
    #define ISCSI_CID_START                 (0)     // Start of connection ID range for ISCSI
    #define DEFPDU_CID_START                (0)     // Start of connection ID range for iSCSI default pdu queue
    #define TOEDQ_CID_START                 (568)   // Start of connection ID range for TOE dq
    #define ETHTOE_CID_START                (576)   // Start of connection ID range for TOE
    #define ETX_CID_START                   (1024)  // Start of connection ID range for ETX
    #define RDMA_CID_START                  (0)     // Start of connection ID range for RDMA
    #define FC_CID_START                    (0)     // Start of virtual device ID range for FCIP
    #define LRO_CID_START                   (ETX_CID_START) // Start of connection ID range for LRO (shares ETX CID range)
    #define DAL_CID_START                   (ETX_CID_START - 32) // Start of connection ID range for DAL

    // --- MAX_CIDS_ENUM ---
    #define MAX_TOEDQ_CIDS                  (8)     // Maximum number of TOE DQ cids
    #define MAX_DEFPDU_CIDS                 (32)    // Maximum number of iSCSI DEF pdu cids
    #define MAX_ETX_CIDS                    (64)    // Maximum number of ETX connections
    #define MAX_ETHTOE_CIDS                 (448)   // Maximum number of TOE conections
    #define MAX_ISCSI_CIDS                  (512)   // Maximum number of ISCSI connections
    #define MAX_ISCSI_CIDS_CT               (512)   // Maximum number of ISCSI connections per chute
    #define MAX_RDMA_CIDS                   (0)     // Maximum number of RDMA connections
    #define MAX_FC_CIDS                     (0)     // Maximum number of FCIP virtual device ID
    #define MAX_LRO_CIDS                    (4)     // Maximum number of LRO connections
    #define MAX_DAL_CIDS                    (32)    // Maximum number of LRO connections

#elif (BE_CONFIG==7)

///////////////////////////////////////////////////////////////////////////////
// EVERYTHING BELOW HERE IS BEING DEPRECATED. PLEASE MOVE TO USING THE NEW
// BE_CONFIG FRAMEWORK LITERALS!!!! THIS WILL BE DELETED EVENTUALLY!!!!!!!!!!!!
///////////////////////////////////////////////////////////////////////////////

    // --- CID_START_ENUM ---
    #define ISCSI_CID_START                 (0)    // Start of connection ID range for ISCSI
    #define DEFPDU_CID_START                (0)     // Start of connection ID range for iSCSI default pdu queue
    #define TOEDQ_CID_START                 (632)   // Start of connection ID range for TOE dq
    #define ETHTOE_CID_START                (640)   // Start of connection ID range for TOE
    #define ETX_CID_START                   (1024)  // Start of connection ID range for ETX
    #define RDMA_CID_START                  (0)     // Start of connection ID range for RDMA
    #define FC_CID_START                    (0)     // Start of virtual device ID range for FCIP
    #define LRO_CID_START                   (ETX_CID_START)     // Start of connecting ID range for LRO
    #define DAL_CID_START                   (0)     // Start of connection ID range for DAL
    #define DAL_RX_CID_START                (0)     // Start of connection ID range for DAL Receive

    // --- MAX_CIDS_ENUM ---
    #define MAX_TOEDQ_CIDS                  (8)     // Maximum number of TOE DQ cids
    #define MAX_DEFPDU_CIDS                 (0)    // Maximum number of iSCSI DEF pdu cids
    #define MAX_ETX_CIDS                    (64)    // Maximum number of ETX connections
    #define MAX_ETHTOE_CIDS                 (384)   // Maximum number of TOE conections
    #define MAX_ISCSI_CIDS                  (0)   // Maximum number of ISCSI connections
    #define MAX_ISCSI_CIDS_CT               (0)   // Maximum number of ISCSI connections per chute
    #define MAX_RDMA_CIDS                   (0)     // Maximum number of RDMA connections
    #define MAX_FC_CIDS                     (1024)    // Maximum number of FCIP virtual device ID
    #define MAX_LRO_CIDS                    (0)     // Maximum number of LRO connections
    #define MAX_DAL_CIDS                    (0)     // Maximum number of DAL connections

#elif (BE_CONFIG==9)

    // --- CID_START_ENUM ---
    #define ISCSI_CID_START                 (0)     // Start of connection ID range for ISCSI
    #define DEFPDU_CID_START                (0)     // Start of connection ID range for iSCSI default pdu queue
    #define TOEDQ_CID_START                 (568)   // Start of connection ID range for TOE dq
    #define ETHTOE_CID_START                (576)   // Start of connection ID range for TOE
    #define ETX_CID_START                   (1024)  // Start of connection ID range for ETX
    #define RDMA_CID_START                  (0)     // Start of connection ID range for RDMA
    #define FC_CID_START                    (0)     // Start of virtual device ID range for FCIP
    #define LRO_CID_START                   (ETX_CID_START) // Start of connection ID range for LRO (shares ETX CID range)
    #define DAL_CID_START                   (0)     // Start of connection ID range for DAL
    #define DAL_RX_CID_START                (0)      // Start of connection ID range for DAL Receive

    // --- MAX_CIDS_ENUM ---
    #define MAX_TOEDQ_CIDS                  (8)     // Maximum number of TOE DQ cids
    #define MAX_DEFPDU_CIDS                 (32)    // Maximum number of iSCSI DEF pdu cids
    #define MAX_ETX_CIDS                    (64)    // Maximum number of ETX connections
    #define MAX_ETHTOE_CIDS                 (448)   // Maximum number of TOE conections
    #define MAX_ISCSI_CIDS                  (512)   // Maximum number of ISCSI connections
    #define MAX_ISCSI_CIDS_CT               (512)   // Maximum number of ISCSI connections per chute
    #define MAX_RDMA_CIDS                   (0)     // Maximum number of RDMA connections
    #define MAX_FC_CIDS                     (0)     // Maximum number of FCIP virtual device ID
    #define MAX_LRO_CIDS                    (18)    // Maximum number of LRO connections
    #define MAX_DAL_CIDS                    (0)     // Maximum number of DAL connections

#else
    #error Error: Invalid configuration (BE_CONFIG unrecognized).
#endif

///////////////////////////////////////////////////////////////////////////////
// EVERYTHING BELOW HERE IS BEING DEPRECATED. PLEASE MOVE TO USING THE NEW
// BE_CONFIG FRAMEWORK LITERALS!!!! THIS WILL BE DELETED EVENTUALLY!!!!!!!!!!!!
///////////////////////////////////////////////////////////////////////////////



/* Define the ethernet CIDs that have higher priority over normal ethernet transmit
 * rings. This really should be determined based on the context memory map, but
 * by default there are two ETX super rings (1024 and 1025). There's no need to tie
 * the ring ID to the ULP in BE2 (as we only expect to have ETX in one ULP). For now
 * just keep the old-BE1 naming convention and build on that.
 */
#define ETX_SUPER_CID_ULP1                  (ETX_CID_START)
#define ETX_SUPER_CID1                      (ETX_SUPER_CID_ULP1)
#define ETX_SUPER_CID2                      (ETX_SUPER_CID1 + 1)

// ------------------------------------------------------------------------
//  There are 4 ID numbers identified for each doorbell.
//  BID - Bulter ID.  This is the value in the doorbell -- although the doorbell
//        struct may define it as the CID or RID for legacy reasons.
//  RID - Resource ID.  This is always a zero based index used for fetching an
//        item from context memory.
//  QID - Queue ID.  This is the queue_id portion of the CRA for fetching a
//        ring from the unified table.
//  CID - Connection ID. This is the index of the TCP connection used by tpost/tpre.
//        It indexes the TCP_CONTEXT in context memory.  This is only valid for some
//        TXULP and RXULP doorbells.
//
//  Currently, all the mappings are simple linear translations.  This may change, so
//  the code below declares different functions for each mapping.  This gives the 
//  flexibility to create a more complicated mapping if necessary.
// ------------------------------------------------------------------------
#define OLD_VALID_RID(_name_, _rid_) (((_rid_) >= _name_##_RID_START) && ((_rid_) < _name_##_RID_START + _name_##_ID_NUM))
#define OLD_VALID_BID(_name_, _bid_) (((_bid_) >= _name_##_BID_START) && ((_bid_) < _name_##_BID_START + _name_##_ID_NUM))
#define OLD_VALID_QID(_name_, _qid_) (((_qid_) >= _name_##_QID_START) && ((_qid_) < _name_##_QID_START + _name_##_ID_NUM))
#define OLD_VALID_CID(_name_, _cid_) (((_cid_) >= _name_##_CID_START) && ((_cid_) < _name_##_CID_START + _name_##_ID_NUM))

#define ID_DECLARE_CONVERSIONS(_name_)                                                                                                  \
    BE_GEN_STATIC_INLINE u32 _name_##_VALID_RID (u32 rid) { return (u32)OLD_VALID_RID(_name_,rid); }                                    \
    BE_GEN_STATIC_INLINE u32 _name_##_VALID_BID (u32 bid) { return (u32)OLD_VALID_BID(_name_,bid); }                                    \
    BE_GEN_STATIC_INLINE u32 _name_##_VALID_QID (u32 qid) { return (u32)OLD_VALID_QID(_name_,qid); }                                    \
    BE_GEN_STATIC_INLINE u32 _name_##_VALID_CID (u32 cid) { return (u32)OLD_VALID_CID(_name_,cid); }                                    \
    \
    BE_GEN_STATIC_INLINE u32 _name_##_RID_TO_BID (u32 rid) { BE_GEN_ASSERT (_name_##_VALID_RID(rid)); return RID_TO_BID(_name_,rid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_RID_TO_QID (u32 rid) { BE_GEN_ASSERT (_name_##_VALID_RID(rid)); return RID_TO_QID(_name_,rid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_RID_TO_CID (u32 rid) { BE_GEN_ASSERT (_name_##_VALID_RID(rid)); return RID_TO_CID(_name_,rid); }  \
    \
    BE_GEN_STATIC_INLINE u32 _name_##_BID_TO_RID (u32 bid) { BE_GEN_ASSERT (_name_##_VALID_BID(bid)); return BID_TO_RID(_name_,bid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_BID_TO_QID (u32 bid) { BE_GEN_ASSERT (_name_##_VALID_BID(bid)); return BID_TO_QID(_name_,bid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_BID_TO_CID (u32 bid) { BE_GEN_ASSERT (_name_##_VALID_BID(bid)); return BID_TO_CID(_name_,bid); }  \
    \
    BE_GEN_STATIC_INLINE u32 _name_##_QID_TO_RID (u32 qid) { BE_GEN_ASSERT (_name_##_VALID_QID(qid)); return QID_TO_RID(_name_,qid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_QID_TO_BID (u32 qid) { BE_GEN_ASSERT (_name_##_VALID_QID(qid)); return QID_TO_BID(_name_,qid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_QID_TO_CID (u32 qid) { BE_GEN_ASSERT (_name_##_VALID_QID(qid)); return QID_TO_CID(_name_,qid); }  \
    \
    BE_GEN_STATIC_INLINE u32 _name_##_CID_TO_BID (u32 cid) { BE_GEN_ASSERT (_name_##_VALID_CID(cid)); return CID_TO_BID(_name_,cid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_CID_TO_QID (u32 cid) { BE_GEN_ASSERT (_name_##_VALID_CID(cid)); return CID_TO_QID(_name_,cid); }  \
    BE_GEN_STATIC_INLINE u32 _name_##_CID_TO_RID (u32 cid) { BE_GEN_ASSERT (_name_##_VALID_CID(cid)); return CID_TO_RID(_name_,cid); }  \

///////////////////////////////////////////////////////////////////////////////
// EVERYTHING BELOW HERE IS BEING DEPRECATED. PLEASE MOVE TO USING THE NEW
// BE_CONFIG FRAMEWORK LITERALS!!!! THIS WILL BE DELETED EVENTUALLY!!!!!!!!!!!!
///////////////////////////////////////////////////////////////////////////////



// ETX IDs
#define ETX_ID_NUM    (MAX_ETX_CIDS)
#define ETX_BID_START (ETX_CID_START)
#define ETX_RID_START (0)
#define ETX_QID_START (0)
#ifndef ETX_CID_START 
    #error You must have ETX_CID_START defined.
#else
    #if (ETX_BID_START != ETX_CID_START)
        #error Creepers, Batman.  Your cid ranges do not match.
    #endif
#endif
ID_DECLARE_CONVERSIONS(ETX)

///////////////////////////////////////////////////////////////////////////////
// EVERYTHING BELOW HERE IS BEING DEPRECATED. PLEASE MOVE TO USING THE NEW
// BE_CONFIG FRAMEWORK LITERALS!!!! THIS WILL BE DELETED EVENTUALLY!!!!!!!!!!!!
///////////////////////////////////////////////////////////////////////////////

// DAL RX IDs
#define DAL_ID_NUM    (MAX_DAL_CIDS)
#define DAL_BID_START (DAL_CID_START)
#define DAL_RID_START (0)
#define DAL_QID_START (DAL_CID_START)
ID_DECLARE_CONVERSIONS(DAL)
///////////////////////////////////////////////////////////////////////////////
// EVERYTHING BELOW HERE IS BEING DEPRECATED. PLEASE MOVE TO USING THE NEW
// BE_CONFIG FRAMEWORK LITERALS!!!! THIS WILL BE DELETED EVENTUALLY!!!!!!!!!!!!
///////////////////////////////////////////////////////////////////////////////
// LRO RX IDs
#define LRO_ID_NUM    (MAX_LRO_CIDS) // LRO uses CRA / QID duplication 
#define LRO_BID_START (LRO_CID_START)
#define LRO_RID_START (0)
#define LRO_QID_START (LRO_CID_START)
ID_DECLARE_CONVERSIONS(LRO)
///////////////////////////////////////////////////////////////////////////////
// EVERYTHING BELOW HERE IS BEING DEPRECATED. PLEASE MOVE TO USING THE NEW
// BE_CONFIG FRAMEWORK LITERALS!!!! THIS WILL BE DELETED EVENTUALLY!!!!!!!!!!!!
///////////////////////////////////////////////////////////////////////////////
// TOE TX IDs
#define TOE_TX_ID_NUM (MAX_ETHTOE_CIDS)
#define TOE_TX_BID_START (ETHTOE_CID_START)
#define TOE_TX_RID_START (0)
#define TOE_TX_QID_START (ETHTOE_CID_START)
#define TOE_TX_CID_START (ETHTOE_CID_START)
ID_DECLARE_CONVERSIONS(TOE_TX)
///////////////////////////////////////////////////////////////////////////////
// EVERYTHING BELOW HERE IS BEING DEPRECATED. PLEASE MOVE TO USING THE NEW
// BE_CONFIG FRAMEWORK LITERALS!!!! THIS WILL BE DELETED EVENTUALLY!!!!!!!!!!!!
///////////////////////////////////////////////////////////////////////////////
// TOE RX IDs (mirror of TOE TX)
#define TOE_RX_ID_NUM (MAX_ETHTOE_CIDS)
#define TOE_RX_BID_START (ETHTOE_CID_START)     // Single ULP butler answers both toe default (DEFQ) & preposted buffers (RQ). Need 1 range of BIDs.
#define TOE_RX_RID_START (0)
#define TOE_RX_QID_START (ETHTOE_CID_START)
#define TOE_RX_CID_START (ETHTOE_CID_START)
ID_DECLARE_CONVERSIONS(TOE_RX)
///////////////////////////////////////////////////////////////////////////////
// EVERYTHING BELOW HERE IS BEING DEPRECATED. PLEASE MOVE TO USING THE NEW
// BE_CONFIG FRAMEWORK LITERALS!!!! THIS WILL BE DELETED EVENTUALLY!!!!!!!!!!!!
///////////////////////////////////////////////////////////////////////////////
// TOE Default IDs
#define TOE_DEFQ_ID_NUM    (MAX_TOEDQ_CIDS)
#define TOE_DEFQ_BID_START (TOEDQ_CID_START)    // Single ULP butler answers both toe default (DEFQ) & preposted buffers (RQ). Need 1 range of BIDs.
#define TOE_DEFQ_RID_START (0)
#define TOE_DEFQ_QID_START (TOEDQ_CID_START)
#define TOE_DEFQ_CID_START (0x1000)     // not valid: CID concept is only applicable to protocol TX/RX rings
ID_DECLARE_CONVERSIONS(TOE_DEFQ)
///////////////////////////////////////////////////////////////////////////////
// EVERYTHING BELOW HERE IS BEING DEPRECATED. PLEASE MOVE TO USING THE NEW
// BE_CONFIG FRAMEWORK LITERALS!!!! THIS WILL BE DELETED EVENTUALLY!!!!!!!!!!!!
///////////////////////////////////////////////////////////////////////////////
// ISCSI transmit IDs
#define ISCSI_TX_ID_NUM    (MAX_ISCSI_CIDS)
#define ISCSI_TX_BID_START (ISCSI_CID_START)
#define ISCSI_TX_RID_START (0)
#define ISCSI_TX_QID_START (ISCSI_CID_START)
#define ISCSI_TX_CID_START (ISCSI_CID_START)
ID_DECLARE_CONVERSIONS(ISCSI_TX)
///////////////////////////////////////////////////////////////////////////////
// EVERYTHING BELOW HERE IS BEING DEPRECATED. PLEASE MOVE TO USING THE NEW
// BE_CONFIG FRAMEWORK LITERALS!!!! THIS WILL BE DELETED EVENTUALLY!!!!!!!!!!!!
///////////////////////////////////////////////////////////////////////////////
// ISCSI Default PDU IDs
#define ISCSI_DEFQ_ID_NUM    (32)
#define ISCSI_DEFQ_BID_START (ISCSI_CID_START)
#define ISCSI_DEFQ_RID_START (0)
#define ISCSI_DEFQ_QID_START (ISCSI_CID_START)
#define ISCSI_DEFQ_CID_START (0x1000)   // not valid: CID concept is only applicable to protocol TX/RX rings
ID_DECLARE_CONVERSIONS(ISCSI_DEFQ)
///////////////////////////////////////////////////////////////////////////////
// EVERYTHING BELOW HERE IS BEING DEPRECATED. PLEASE MOVE TO USING THE NEW
// BE_CONFIG FRAMEWORK LITERALS!!!! THIS WILL BE DELETED EVENTUALLY!!!!!!!!!!!!
///////////////////////////////////////////////////////////////////////////////
// RDMA QP IDs
#define RDMA_QP_ID_NUM    (MAX_RDMA_CIDS)
#define RDMA_QP_BID_START (RDMA_CID_START)
#define RDMA_QP_RID_START (0)
#define RDMA_QP_QID_START (RDMA_CID_START)
#define RDMA_QP_CID_START (RDMA_CID_START)
ID_DECLARE_CONVERSIONS(RDMA_QP)

// FCOE transmit IDs
#define FC_TX_ID_NUM    (MAX_FC_CIDS)
#define FC_TX_BID_START (FC_CID_START)
#define FC_TX_RID_START (0)
#define FC_TX_QID_START (FC_CID_START)
#define FC_TX_CID_START (FC_CID_START)
ID_DECLARE_CONVERSIONS(FC_TX)
///////////////////////////////////////////////////////////////////////////////
// EVERYTHING BELOW HERE IS BEING DEPRECATED. PLEASE MOVE TO USING THE NEW
// BE_CONFIG FRAMEWORK LITERALS!!!! THIS WILL BE DELETED EVENTUALLY!!!!!!!!!!!!
///////////////////////////////////////////////////////////////////////////////
// FCOE Default PDU IDs
#define FCOE_DEFQ_ID_NUM    (2)
#define FCOE_DEFQ_BID_START (FC_CID_START)
#define FCOE_DEFQ_RID_START (0)
#define FCOE_DEFQ_QID_START (FC_CID_START)
#define FCOE_DEFQ_CID_START (0x1000)   // not valid: CID concept is only applicable to protocol TX/RX rings
ID_DECLARE_CONVERSIONS(FCOE_DEFQ)

#endif // __be2_gen_id_ranges_h__

