/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
//
#ifndef __be2_gen_ut_h__
#define __be2_gen_ut_h__

#include "be2_gen_check_defines.h"
#include "be2_gen_api_enum.h"
#include "be2_gen_misc.h"

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_DT_Init
//
//   Program a Directory Table Entry. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// cra                - Cache request address. 
// host_memory        - TRUE for host memory physical address, FALSE for ddr physical
//                      address. 
// ats_flags          - Encoded PCI ATS flags. Valid values are defined in the
//                      DMA_ATS_ENUM enumeration. 
// p_phys_addr        - Phisical address representing the DT entry. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_DT_Init (
   IN void*   function_context,
   IN u32     cra,
   IN boolean host_memory,
   IN u8      ats_flags,
   IN u64*    p_phys_addr )
{
    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Access_DT
//
//   Access a Directory Table entry. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// index              - DT index to access. 
// p_phys_addr        - Pointer to phsyical address data. 
// write_flag         - TRUE for write access, FALSE for read access. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Access_DT (
   IN void*   function_context,
   IN u16     index,
   IN u64*    p_phys_addr,
   IN boolean write_flag )
{
    u32 *addr_lo = (u32 *)p_phys_addr;
    u32 *addr_hi = addr_lo + 1;
    BE_GEN_CSR_WRITE (function_context, ptc.dt_index, index);
    if (write_flag) 
    {
        BE_GEN_CSR_WRITE (function_context, ptc.dt_pa_lo, *addr_lo);
        BE_GEN_CSR_WRITE (function_context, ptc.dt_pa_hi, *addr_hi);
    }
    else
    {
        *addr_lo = BE_GEN_CSR_READ (function_context, ptc.dt_pa_lo);
        *addr_hi = BE_GEN_CSR_READ (function_context, ptc.dt_pa_hi);
    }
    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Invalidate_PTC_Entry
//
//   Invalidates a Protection Table Cache entry. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// cra                - Cache request address. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Invalidate_PTC_Entry (
   IN void* function_context,
   IN u32   cra )
{
    PTC_COMMAND_CSR ptc_command;
    
    // PTC expects the CRA to be right shifted by 5 bits
    BE_GEN_CSR_WRITE (function_context, ptc.cra, cra >> 5);

    // Issue invalidate command
    ptc_command.dw = 0;
    ptc_command.invalidate = TRUE;
    BE_GEN_CSR_WRITE (function_context, ptc.command, ptc_command.dw);    
 
    // Wait for command to finish
    while ((ptc_command.status == 0) && (ptc_command.err == 0))
    {
        // FIXME TODO @@@ May want to sleep here since this could take a while to complete
        ptc_command.dw = BE_GEN_CSR_READ(function_context, ptc.command);
    }
    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Invalidate_UC_Entry
//
//   Invalidates a Unified Cache entry. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// cra                - Cache request address. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Invalidate_UC_Entry (
   IN void* function_context,
   IN u32   cra )
{
    UC_COMMAND_CSR uc_command;
    UC_RESP0_CSR uc_resp0;
    
    // UC expects the CRA to be right shifted by 5 bits
    BE_GEN_CSR_WRITE (function_context, uc.cra, cra >> 5);

    // Issue invalidate command
    uc_command.dw = 0;
    uc_command.req_opcode = UC_COMMAND_INVL;
    uc_command.requester_id = UC_HOST;
    BE_GEN_CSR_WRITE (function_context, uc.command, uc_command.dw);    
 
    // Wait for command to finish
    uc_resp0.valid = 0;
    while (uc_resp0.valid == 0)
    {
        // FIXME TODO @@@ May want to sleep here since this could take a while to complete
        uc_resp0.dw = BE_GEN_CSR_READ(function_context, uc.resp0);
    }
    BE_GEN_ASSERT(uc_resp0.status == UC_STATUS_SUCCESS);
    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_DTE_PTE_Init
//
//   Initializer for Directory Table Entry or Page Table Entry. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// p_phys_addr        - Pointer to DTE or PTE data. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_DTE_PTE_Init (
   IN void* function_context,
   IN u64*  p_phys_addr )
{
    // FIXME TODO @@@ Need format of DTE and PTE defined in sg. Asked Farhan about this
    // on 10/23/2007. --Vahid
    p_phys_addr = 0;
}

#endif  // __be2_gen_ut_h__

