/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!

@file
    cq.cpp

@brief
    Implementation for CQ objects. These objects represent and manage on-chip
    Completion Queue structures.

@author jflores

@note
    jflores   8/03/2004    initial
    $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/drivers/beclib/cq_ll.c,v 1.23 2006/04/09 23:32:54 menz Exp $

*/

#include "pch.h"

/*!
@brief
    This routines searches for a CqId that is bound to a specified function object.
@param    
    function_object  - function object the CQ should be bound to.
@param    
    cq_id            - cq_id to find
@return   
    A pointer to the CQ_OBJECT that represents this CqId. NULL if not found
@note
    IRQL: <= DISPATCH_LEVEL
*/
STATIC
INLINE
PBE_CQ_OBJECT
be_cq_id_to_object (
    PBE_FUNCTION_OBJECT  function_object, 
    u32                  cq_id)
{
    PBE_CQ_OBJECT   cq_cur;
    PBE_CQ_OBJECT   cq_ret = NULL;

    _be_function_lock (function_object);

    SA_FOR_EACH_LIST_ENTRY (cq_cur, function_object->links.cq_list_head, BE_CQ_OBJECT, cq_list)
    {
        if (be_cq_get_id (cq_cur) == cq_id) {

            cq_ret = cq_cur;
            break;
        }
    }

    _be_function_unlock (function_object);

    if (!cq_ret) {
        TRACE (DL_ERR, "Failed to locate cq_id:%d for processing.", cq_id);
    }

    return  cq_ret;
}


//============================================================================
//                  P U B L I C  R O U T I N E S
//============================================================================


//---------------------------------------------------------------------------------------
// Function: be_cq_create
//
//   Creates a CQ. This uses a synchronous IOCTL.
//
// function_object     - 
// sgl                 - No virtual address required, except for MCC CQ.
// length              - Length is bytes of queue. Number of entries is power of 2, from
//                       256 to 1k. 
// solicited_eventable - If TRUE, only CQEs with solicited event bit cause EQE write. 
// no_delay            - TRUE means force interrupt immediately without waiting for any EQ
//                       delay. 
// cqe_dma_coalescing  - Each CQ supports coalescing the DMA of CQs into larger transfers.
//                       The maximum DMA transfer is up to 4 CQEs in a single DMA. When
//                       the CQ ring contains 'N' entries that the driver has not
//                       consumed, it will start coalescing future DMAs. The future DMA
//                       will not occur until the driver rings the CQ doorbell to reduce
//                       the outstanding number of CQ entries to less than 'N' -- or when
//                       CEV has buffered 4 CQEs to transfer. 
// 
//                       The value 'N' is
//                       configurable by this setting using the enumeration
//                       BE_CQ_DMA_COALESCING. Using a small value (i.e. 2) will allow
//                       more DMA coalescing. However, the driver must be careful about
//                       increased latencies. 
// eq_object           - Optional, if set this is an eventable CQ. 
// cq_object           - Created CQ object. 
// return status       - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_cq_create (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PSA_SGL             sgl,
   IN u32                 length,
   IN boolean             solicited_eventable,
   IN boolean             no_delay,
   IN u32                 cqe_dma_coalescing,
   IN PBE_EQ_OBJECT       eq_object,
   OUT PBE_CQ_OBJECT      cq_object
)
{
    BESTATUS        status = BE_SUCCESS;
    u32             num_entries_encoding;
    u32             num_entries = length / 16; // all CQEs are same size
    IOCTL_COMMON_CQ_CREATE *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    
    ASSERT (sgl);
    ASSERT (cq_object);
    ASSERT (length % 16 == 0);
    
    FUNCTION_ASSERT (function_object);    
    
    switch (num_entries) {
        case 256: num_entries_encoding  = CEV_CQ_CNT_256; break;
        case 512: num_entries_encoding  = CEV_CQ_CNT_512; break;
        case 1024: num_entries_encoding = CEV_CQ_CNT_1024; break;
        default:
            SA_BREAKPOINT_MSG ("bad cq num entries");
            return BE_STATUS_INVALID_PARAMETER;
    }

    // All cq entries all the same size.  Use iSCSI version as a test for the proper
    // sgl length.
    sa_zero_mem (cq_object, sizeof (*cq_object));

    cq_object->magic           = BE_CQ_MAGIC;
    cq_object->ref_count       = 0;
    cq_object->parent_function = function_object;    
    cq_object->eq_object       = eq_object;    
    cq_object->num_entries     = num_entries;
    cq_object->va              = sa_sgl_get_base_va (sgl); // save for MCC cq processing

    // map into UT.
    length = num_entries * 16; // all CQ entries are same size

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        ASSERT (wrb);
        TRACE (DL_ERR, "No free MCC WRBs in create EQ.");
        status = BE_STATUS_NO_MCC_WRB;
        goto Error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_CQ_CREATE);

    ioctl->params.request.num_pages = (UEXACT16) SA_PAGES_SPANNED (sa_sgl_get_offset (sgl), length);
    
    // context
    ioctl->params.request.context.valid     = TRUE;
    ioctl->params.request.context.Func      = be_function_get_function_number (function_object);
    ioctl->params.request.context.Eventable = (eq_object != NULL);
    ioctl->params.request.context.Armed     = TRUE;
    ioctl->params.request.context.EQID      = eq_object ? eq_object->eq_id : 0;
    ioctl->params.request.context.Count     = num_entries_encoding;
    ioctl->params.request.context.PD        = be_function_get_pd_number (function_object);
    ioctl->params.request.context.NoDelay   = no_delay;
    ioctl->params.request.context.SolEvent  = solicited_eventable;            
    ioctl->params.request.context.CoalescWM = cqe_dma_coalescing; // See BE_CQ_DMA_COALESCING enum
    ASSERT (ioctl->params.request.context.CoalescWM == cqe_dma_coalescing); // Detect truncation
    
    // Create a page list for the IOCTL.
    be_sgl_to_pa_list (sgl, 
                       ioctl->params.request.pages, 
                       SA_NUMBER_OF (ioctl->params.request.pages));

    // Post the Ioctl
    status = be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, ioctl);
    if (status != BE_SUCCESS) {
        TRACE (DL_ERR, "MCC to create CQ failed.");
        goto Error;
    }

    // Remember the CQ id.
    cq_object->cq_id = ioctl->params.response.cq_id;

    // insert this cq into eq_object reference
    if (eq_object) {
        _be_eq_add_cq (eq_object, cq_object);
    }

    // insert this cq into function_object
    _be_function_add_cq (function_object, cq_object);

    TRACE (DL_INFO, "cq created. function:%d pd:%d cqid:%d bytes:%d eqid:%d se:%d nd:%d",
           be_function_get_function_number (function_object),
           be_function_get_pd_number (function_object),
           cq_object->cq_id,
           num_entries * 16,
           eq_object ? be_eq_get_id (eq_object) : -1,
           (u32)solicited_eventable,
           (u32)no_delay
           );

Error:
    be_unlock_wrb_post (function_object);
    return status;
}


/*!

@brief
    Returns the CQ ID for CQ. This is the ID the chip references the queue as.

@param    
    cq_object            - CQ handle returned from cq_object_create.

@return   
    CQ ID
@note
    IRQL: any

*/
u32 
be_cq_get_id (PBE_CQ_OBJECT cq_object)
{
    CQ_ASSERT (cq_object);    
    return cq_object->cq_id;
}

/*!

@brief
    Sets the current callback to be invoked when a CQ entry is made available. 
    This need only be set if the owner of this CQ does not recieve the CQ 
    completions directly. i.e. it does not service interrupts.

@param    
    cq_object            - CQ handle returned from cq_object_create.
    
@param    
    callback            - function to invoke when the CQ needs processing.
    
@param    
    context             - opaque context to pass to callback.        

@return   
    ptr to the previous callback.
    
@note
    IRQL: any

    this routine is currently not synchronized w/ the DPC that may run the CQ.
    therefore it is the callers responsibility to ensure the CQ will not
    interrupt while this routine is executing.
*/
CQ_CALLBACK 
be_cq_object_set_callback (
    PBE_CQ_OBJECT   cq_object, 
    CQ_CALLBACK     callback, 
    PVOID           context)
{
    CQ_CALLBACK old_callback;

    CQ_ASSERT (cq_object);    

    old_callback = cq_object->callback;

    cq_object->callback          = callback;
    cq_object->callback_context  = context;

    return old_callback;
}

/*!

@brief
    This routine handles CQ processing on CQs that were 'unrecognized' by upper
    level drivers.  The only CQs that should be processed by this routine are
        * MCC CQ
        * iWARP CQ
        
    And of these, only the MCC CQ will be processed in this routine. The iWARP
    CQ will be delegated further for handling by the iWARP device driver

@param
    function_object      - handle to a function object                           
@param
    cq_id                - the cq that has just completed processing                         

@return
    

@note
    IRQL: DISPATCH_LEVEL only
    
    this routine should be called immediately when a CQ from a ULD is not 
    recognized.
*/     
void
be_cq_object_delegate_processing (
    PBE_FUNCTION_OBJECT     function_object, 
    u32                     cq_id)
{
    PBE_CQ_OBJECT   cq_object;

    FUNCTION_ASSERT (function_object);
    

    cq_object = be_cq_id_to_object (function_object, cq_id);

    if (cq_object) {

        ASSERT (cq_object->callback);
        cq_object->callback (function_object, cq_object, cq_object->callback_context);

    } else {

        TRACE (DL_WARN, "cq_id %d not found for delegation", cq_id);
    }

}



/*!

@brief
    References the given object. The object is guaranteed to remain active until
    the reference count drops to zero.

@param    
    cq_object            - CQ handle returned from cq_object_create.

@return   
    returns the current reference count on the object
    
@note
    IRQL: any

*/
u32
be_cq_object_reference (
    PBE_CQ_OBJECT cq_object)
{
    CQ_ASSERT (cq_object);    
    return  sa_atomic_increment (&cq_object->ref_count);
}


/*!

@brief
    Dereferences the given object. The object is guaranteed to remain active until
    the reference count drops to zero.

@param    
    cq_object            - CQ handle returned from cq_object_create.

@return   
    returns the current reference count on the object
    
@note
    IRQL: any

*/
u32
be_cq_object_dereference (
    PBE_CQ_OBJECT cq_object)
{
    CQ_ASSERT (cq_object);    
    return  sa_atomic_decrement (&cq_object->ref_count);
}



/*!

@brief
    Deferences the given object. Once the object's reference count drops to
    zero, the object is destroyed and all resources that are held by this object
    are released.  The on-chip context is also destroyed along with the queue
    ID, and any mappings made into the UT.

@param    
    cq_object            - CQ handle returned from cq_object_create.

@return   
    returns the current reference count on the object
    
@note
    IRQL: IRQL < DISPATCH_LEVEL
*/
BESTATUS
be_cq_destroy (
    PBE_CQ_OBJECT cq_object)
{
    BESTATUS status = 0;

    CQ_ASSERT (cq_object);
    FUNCTION_ASSERT (cq_object->parent_function);

    // Nothing should reference this CQ at this point.
    ASSERT (cq_object->ref_count == 0);
    
    // Send ioctl to destroy the CQ.
    status = be_function_ring_destroy (cq_object->parent_function, cq_object->cq_id, IOCTL_RING_TYPE_CQ);
    ASSERT (status == 0 || cq_object->parent_function->ue);

    // Remove reference if this is an eventable CQ.
    if (cq_object->eq_object) {
        _be_eq_remove_cq (cq_object->eq_object, cq_object);
    }

    // Remove from the function object.
    _be_function_remove_cq (cq_object->parent_function, cq_object);

    // Zero the software tracking object.
    sa_zero_mem (cq_object, sizeof (*cq_object));

    return BE_SUCCESS;
}






