/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
#include "pch.h"
/*!

@brief
    This routine serializes access to resources maintained through an EQ object.

@param    
    eq_object      - The event queue object to acquire the lock for.

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/

STATIC void
_be_eq_lock (
    IN      PBE_EQ_OBJECT eq_object)
{
    SA_NOT_USED (eq_object);
    // BUG - Work around storport driver dispatch issues.
    //SA_ASSERT_BLOCKING ();
    //be_lock_acquire (&eq_object->lock);
}

/*!

@brief
    This routine removes serialization done by EqObjectLock. 
@param    
    eq_object      - the event queue object to drop the lock for.
@return   
@note
    IRQL < DISPATCH_LEVEL    
*/
STATIC void
_be_eq_unlock (
    IN      PBE_EQ_OBJECT eq_object)
{
    SA_NOT_USED (eq_object);
    //be_lock_release (&eq_object->lock);
}


STATIC
INLINE
PBE_EQ_OBJECT
_be_eq_id_to_object (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      u32                  eq_id)
{
    PBE_EQ_OBJECT   eq_cur;
    PBE_EQ_OBJECT   eq_ret   = NULL;

    _be_function_lock (function_object);

    SA_FOR_EACH_LIST_ENTRY (eq_cur, function_object->links.eq_list_head, BE_EQ_OBJECT, eq_list)
    {
        if (be_eq_get_id (eq_cur) == eq_id) {

            eq_ret   = eq_cur;
            break;
        }
    }

    _be_function_unlock (function_object);

    return  eq_ret;
}



/*
    This routine creates an event queue based on the client completion 
    queue configuration information.                                       
    
    FunctionObject      - Handle to a function object
    EqBaseVa            - Base VA for a the EQ ring
    SizeEncoding        - The encoded size for the EQ entries. This value is 
                          either CEV_EQ_SIZE_4 or CEV_EQ_SIZE_16
    NumEntries          - CEV_CQ_CNT_* values.
    ppEqObject          - Internal EQ Handle returned.

return   
    BE_SUCCESS if successful, otherwise a useful error code is returned.
*/
BESTATUS
be_eq_create_internal (
    PBE_FUNCTION_OBJECT  function_object,
    PSA_SGL              sgl,
    u32                  eqe_size,
    u32                  num_entries,
    boolean              fake_eq,
    u32                  delay_multiplier,
    PBE_EQ_OBJECT        eq_object)
{
    BESTATUS                status  = BE_SUCCESS;
    u32 num_entries_encoding, eqe_size_encoding, length;
    IOCTL_COMMON_EQ_CREATE *ioctl = NULL;
    MCC_WRB* wrb = NULL;

    ASSERT (sgl);
    ASSERT (eq_object);

    FUNCTION_ASSERT (function_object);   

    // To avoid assuming the enumeration values are constant, I'm using a switch
    // statement instead of calculating this with sa_log2.
    switch (num_entries) {
        case 256:  num_entries_encoding = CEV_EQ_CNT_256;  break;
        case 512:  num_entries_encoding = CEV_EQ_CNT_512;  break;
        case 1024: num_entries_encoding = CEV_EQ_CNT_1024; break; 
        case 2048: num_entries_encoding = CEV_EQ_CNT_2048; break; 
        case 4096: num_entries_encoding = CEV_EQ_CNT_4096; break; 
        default:
            SA_BREAKPOINT_MSG ("illegal num entries");
            return BE_STATUS_INVALID_PARAMETER;
    }

    // To avoid assuming the enumeration values are constant, I'm using a switch
    // statement instead of calculating this with sa_log2.
    switch (eqe_size) {
        case 4:  eqe_size_encoding = CEV_EQ_SIZE_4; break;
        case 16: eqe_size_encoding = CEV_EQ_SIZE_16; break;
        default:
            SA_BREAKPOINT_MSG ("illegal eqe size");
            return BE_STATUS_INVALID_PARAMETER;
    }

    if ((eqe_size == 4 && num_entries < 1024) ||
        (eqe_size == 16 && num_entries == 4096))
    {
        TRACE (DL_ERR, "Bad EQ size. eqe_size:%d num_entries:%d", eqe_size, num_entries);
        SA_BREAKPOINT_MSG ("bad");
        return BE_STATUS_INVALID_PARAMETER;
    }

    // This is the maximum delay value.
    ASSERT (delay_multiplier < 1024);

    sa_zero_mem (eq_object, sizeof (*eq_object));

    eq_object->magic             = BE_EQ_MAGIC;
    eq_object->ref_count         = 0;
    eq_object->parent_function   = function_object;
    eq_object->eq_id             = 0xFFFFFFFF;        

    be_lock_init (&eq_object->lock);

    sa_initialize_list_head (&eq_object->cq_list_head);

    length = num_entries * eqe_size;        


    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        ASSERT (wrb);
        TRACE (DL_ERR, "No free MCC WRBs in create EQ.");
        status = BE_STATUS_NO_MCC_WRB;
        goto Error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_EQ_CREATE);

    ioctl->params.request.num_pages = (UEXACT16) SA_PAGES_SPANNED (sa_sgl_get_offset (sgl), length);

    // init the context
    ioctl->params.request.context.Func      = be_function_get_function_number (function_object);
    ioctl->params.request.context.valid     = TRUE;
    ioctl->params.request.context.Size      = eqe_size_encoding;
    ioctl->params.request.context.PD        = be_function_get_pd_number (function_object);
    ioctl->params.request.context.Count     = num_entries_encoding;
    ioctl->params.request.context.Armed     = FALSE;  // Let the caller ARM the EQ with the doorbell.
    ioctl->params.request.context.NoDelay   = 0; // NoDelay=1 is equivalent to DelayMult=0, so it is not used.
    ioctl->params.request.context.DelayMult = delay_multiplier; // Maximum interrupt rate by default.
    ioctl->params.request.context.PHASE     = 0; // Phase must be set with be_eq_modify_delay

    if (fake_eq) {
        ioctl->params.request.context.Cidx = 0;
        ioctl->params.request.context.Pidx = 2;
        ioctl->params.request.context.EPIdx = 2;
    }    

    // Create a page list for the IOCTL.
    be_sgl_to_pa_list (sgl, 
                       ioctl->params.request.pages, 
                       SA_NUMBER_OF (ioctl->params.request.pages));

    status = be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, ioctl);
    if (status != BE_SUCCESS) {
        TRACE (DL_ERR, "MCC to create EQ failed.");
        goto Error;
    }

    // Get the EQ id. The MPU allocates the IDs.
    eq_object->eq_id = ioctl->params.response.eq_id;

    // insert tracking object
    _be_function_add_eq (function_object, eq_object);

    TRACE (DL_INFO, "eq created. function:%d pd:%d eqid:%d bytes:%d eqeBytes:%d",
           be_function_get_function_number (function_object),
           be_function_get_pd_number (function_object),
           eq_object->eq_id,
           num_entries * eqe_size,
           eqe_size
           );

Error:
    be_unlock_wrb_post (function_object);
    return status;
}


//============================================================================
//                  P U B L I C  R O U T I N E S
//============================================================================

//---------------------------------------------------------------------------------------
// Function: be_eq_create
//
//   Creates an eq_object. This is a synchronous IOCTL.
//
// function_object  - 
// sgl              - No virtual address required.
// eqe_size         - EQ entry size in bytes. Either 4 or 16.
// num_entries      - Power of 2, fom 256 to 4k.
// delay_multiplier - EQ delay in arbitrary units. Use be_eq_delay_to_multiplier to
//                    convert a delay from microseconds to a multiplier. A value of 0 is
//                    no delay. The delay may be modified during runtime using
//                    be_eq_modify_delay. 
// eq_object        - Created EQ object.
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eq_create (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PSA_SGL             sgl,
   IN u32                 eqe_size,
   IN u32                 num_entries,
   IN u32                 delay_multiplier,
   OUT BE_EQ_OBJECT*      eq_object
)
{
    return be_eq_create_internal (function_object, 
                                  sgl, 
                                  eqe_size, 
                                  num_entries, 
                                  FALSE, 
                                  delay_multiplier,
                                  eq_object);    
}


//---------------------------------------------------------------------------------------
// Function: be_eq_fake_create
//
//   Creates an eq_object that does not have an associated memory ring. The object is used
//   to generate an interrupt from software. The EQ will interrupt the host every time the
//   EQ is rearmed, after the EQ delay expires. The EQ should always be rearmed with a
//   num_popped count of 0 in the doorbell register. The be_function_enable_interrupts
//   function will cause an immediate interrupt from this EQ. This is a synchronous IOCTL.
//
// function_object  - 
// delay_multiplier - EQ delay in arbitrary units. Use be_eq_delay_to_multiplier to
//                    convert a delay from microseconds to a multiplier. A value of 0 is
//                    no delay. The delay may be modified during runtime using
//                    be_eq_modify_delay. 
// eq_object        - Created EQ object.
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eq_fake_create (
   IN PBE_FUNCTION_OBJECT function_object,
   IN u32                 delay_multiplier,
   OUT BE_EQ_OBJECT*      eq_object
)
{
    SA_SGL sgl;
    
    sa_sgl_create_contiguous ((PVOID)0xbad0000, 
                              (SA_PHYSICAL_ADDRESS)0xbad00000, 
                              SA_PAGE_SIZE, 
                              &sgl);

    return be_eq_create_internal (function_object, 
                                  &sgl, 
                                  sizeof (EQ_ENTRY), 
                                  4096 / sizeof (EQ_ENTRY), 
                                  TRUE, 
                                  delay_multiplier,
                                  eq_object);    
}

/*!

@brief
    Returns the EQ ID for EQ. This is the ID the chip references the queue as.

    EqObject            - EQ Handle returned from EqObjectCreate.

@return   
    EQ ID
@note
    IRQL: any

*/
u32 
be_eq_get_id ( 
    IN      PBE_EQ_OBJECT eq_object)
{
    EQ_ASSERT (eq_object);    
    return eq_object->eq_id;
}


/*!

@brief
    Sets the current callback to be invoked when a EQ entry is made available. 
    This need only be set if the owner of this EQ does not recieve the EQ 
    completions directly. i.e. it does not service interrupts.

@param    
    eq_object            - EQ handle returned from eq_object_create.
    
@param    
    callback            - function to invoke when the EQ needs processing.
    
@param    
    context             - opaque context to pass to callback.    

@return   
    ptr to the previous callback.
    
@note
    IRQL: any

    this routine is currently not synchronized w/ the DPC that may run the EQ.
    therefore it is the callers responcibility to ensure the EQ will not
    interrupt while this routine is executing.
*/
EQ_CALLBACK 
be_eq_set_callback (
    IN      PBE_EQ_OBJECT   eq_object, 
    IN      EQ_CALLBACK     callback, 
    IN      PVOID           context)
{
    EQ_CALLBACK     old_callback;

    EQ_ASSERT (eq_object);
    
    old_callback = eq_object->callback;

    eq_object->callback          = callback;
    eq_object->callback_context   = context;

    return old_callback;
}

/*!

@brief
    This routine handles EQ processing on EQs that were 'unrecognized' by upper
    level drivers.  The only EQs that should be processed by this routine are
        * iWARP EQ
        
    The iWARP EQ will be delegated further for handling by the iWARP device driver

@param
    FunctionObject      - Handle to a function object                           
@param
    EqId                - The Eq that has just completed processing                         

@return
    
@note
    IRQL: DISPATCH_LEVEL only
    
    This routine should be called immediately when a EQ from a ULD is not 
    recognized.
*/     
void
be_eq_delegate_processing (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      u32                  eq_id)
{
    PBE_EQ_OBJECT   eq_object;

    FUNCTION_ASSERT (function_object);
    

    eq_object = _be_eq_id_to_object (function_object, eq_id);

    if (eq_object) {

        eq_object->callback (function_object, eq_object, eq_object->callback_context);

    } else {

        TRACE (DL_WARN, "eq_id %d not found for delegation", eq_id);
    }

}



/*!

@brief
    This routine adds a CQ as a dependent object on the given EQ. This will 
    cause the reference count to be incrmented and the CQ to be placed on
    the EQ's list of active CQ objects.

@param    
    eq_object            - EQ handle returned from eq_object_create.
@param    
    cq_object            - CQ handle that is dependent on this EQ

@return   
    EQ ID
@note
    IRQL: < DISPATCH_LEVEL

*/
void
_be_eq_add_cq (
    IN      PBE_EQ_OBJECT           eq_object, 
    IN      PBE_CQ_OBJECT           cq_object)
{
    EQ_ASSERT (eq_object);    

    CQ_ASSERT (cq_object);    

    _be_eq_lock (eq_object);

    be_eq_reference (eq_object);

    sa_insert_tail_list (&eq_object->cq_list_head, &cq_object->cqlist_for_eq);

    _be_eq_unlock (eq_object);
}


/*!

@brief
    References the given object. The object is guaranteed to remain active until
    the reference count drops to zero.

@param    
    eq_object            - CQ handle returned from cq_object_create.

@return   
    returns the current reference count on the object
    
@note
    IRQL: any

*/
u32
be_eq_reference (
    PBE_EQ_OBJECT eq_object)
{
    EQ_ASSERT (eq_object);    
    return  sa_atomic_increment (&eq_object->ref_count);
}


/*!

@brief
    Dereferences the given object. The object is guaranteed to remain active until
    the reference count drops to zero.

@param    
    eq_object            - CQ handle returned from cq_object_create.

@return   
    returns the current reference count on the object
    
@note
    IRQL: any

*/
u32
be_eq_dereference (
    PBE_EQ_OBJECT eq_object)
{
    EQ_ASSERT (eq_object);    
    return  sa_atomic_decrement (&eq_object->ref_count);
}


/*!
@brief
    This routine removes a CQ as a dependent object on the given EQ. This will 
    cause the reference count to be decrmented and the CQ to be removed from
    the EQ's list of active CQ objects.
@param    
    eq_object            - EQ handle returned from eq_object_create.
@param    
    cq_object            - CQ handle that was dependent on this EQ
@return   
    EQ ID
@note
    IRQL: < DISPATCH_LEVEL
*/
void
_be_eq_remove_cq (
    IN      PBE_EQ_OBJECT           eq_object, 
    IN      PBE_CQ_OBJECT           cq_object)
{
    EQ_ASSERT (eq_object);
    CQ_ASSERT (cq_object);

    be_eq_dereference (eq_object);
    
    _be_eq_lock (eq_object);
    sa_remove_entry_list (&cq_object->cqlist_for_eq);
    _be_eq_unlock (eq_object);
}



/*!
@brief
    Deferences the given object. Once the object's reference count drops to
    zero, the object is destroyed and all resources that are held by this object
    are released.  The on-chip context is also destroyed along with the queue
    ID, and any mappings made into the UT.
@param    
    eq_object            - EQ handle returned from eq_object_create.
@return   
    BE_SUCCESS if successfull, , otherwise a useful error code is returned.
@note
    IRQL: IRQL < DISPATCH_LEVEL
*/
BESTATUS 
be_eq_destroy (
   PBE_EQ_OBJECT eq_object )
{
    BESTATUS status = 0;

    EQ_ASSERT (eq_object);    

    ASSERT (eq_object->ref_count == 0);
    ASSERT (sa_is_list_empty (&eq_object->cq_list_head)); // no CQs should reference this EQ now

    // Send ioctl to destroy the EQ.
    status = be_function_ring_destroy (eq_object->parent_function, eq_object->eq_id, IOCTL_RING_TYPE_EQ);
    ASSERT (status == 0 || eq_object->parent_function->ue);

    // remove from the function
    _be_function_remove_eq (eq_object->parent_function,  eq_object);

    // zero tracking object
    sa_zero_mem (eq_object, sizeof (BE_EQ_OBJECT));

    return BE_SUCCESS;
}


//---------------------------------------------------------------------------------------
// Function: be_eq_modify_delay
//
//   Changes the EQ delay for a group of EQs.
//
// function_object           - 
// num_eq                    - The number of EQs in the eq_array to adjust. This also is
//                             the number of delay values in the eq_delay_array. 
// eq_array                  - Array of BE_EQ_OBJECT pointers to adjust. 
// eq_delay_multiplier_array - Array of "num_eq" timer delay multipliers. These are in
//                             arbitrary units. Use be_eq_delay_to_multiplier to convert
//                             from a microsecond delay to a multiplier. 
// eq_phase_array            - Array of "num_eq" phase values from 0 to 3. Each setting
//                             represents a 25 percent phase shift. 
//                             The phase is
//                             measured relative to other EQs to allow the driver to
//                             synchronize interrupts or stagger interrupts. 
// 
//                             The phase
//                             only comes into play when the EQs are operating with zero
//                             credits in their leaky bucket. During this condition, the
//                             interrupt rate is exactly (1,000,000 /
//                             delay_in_microseconds) interrupts/sec. Phase is only
//                             meaningful, among EQs with the same delay setting. 
// callback                  - optional
// callback_context          - optional
// queue_context             - Optional. Pointer to a previously allocated struct. If the
//                             MCC WRB ring is full, this structure is used to queue the
//                             operation. It will be posted to the MCC ring when space
//                             becomes available. All queued commands will be posted to
//                             the ring in the order they are received. It is always valid
//                             to pass a pointer to a generic BE_GENERIC_QUEUE_CONTEXT.
//                             However, the specific context structs are generally smaller
//                             than the generic struct. 
// return pend_status        - BE_SUCCESS (0) on success. BE_PENDING (postive value) if
//                             the IOCTL completion is pending. Negative error code on
//                             failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eq_modify_delay (
   IN PBE_FUNCTION_OBJECT               function_object,
   IN u32                               num_eq,
   IN BE_EQ_OBJECT**                    eq_array,
   IN u32*                              eq_delay_array,
   IN u32*                              eq_phase_array,
   IN MCC_WRB_CQE_CALLBACK              callback,
   IN PVOID                             callback_context,
   IN PBE_EQ_MODIFY_DELAY_QUEUE_CONTEXT queue_context
) 
{
    IOCTL_COMMON_MODIFY_EQ_DELAY *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    u32 i;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_MODIFY_EQ_DELAY);

    ASSERT (num_eq > 0);
    ASSERT (num_eq <= SA_NUMBER_OF (ioctl->params.request.delay));
    ioctl->params.request.num_eq = 0;

    for (i=0; i<num_eq; i++) {
    
        ASSERT (eq_array[i]);
        if (eq_array[i]) {
            ioctl->params.request.delay[i].eq_id = be_eq_get_id (eq_array[i]);

            ASSERT (eq_delay_array[i] < 1024);
            ioctl->params.request.delay[i].delay_multiplier = MIN (eq_delay_array[i], 1023);

            ASSERT (eq_phase_array[i] < 4);
            ioctl->params.request.delay[i].phase = MIN (eq_phase_array[i], 3);

            ioctl->params.request.num_eq++;
        }

    }

    // No need to post the IOCTL if 0 EQs are provided.
    if (ioctl->params.request.num_eq == 0) {
        status = BE_SUCCESS;
        goto Error;
    }

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (function_object, 
                                                          wrb, 
                                                          generic_context, 
                                                          callback, 
                                                          callback_context, 
                                                          ioctl);

Error:
    be_unlock_wrb_post (function_object);

    return status;


}


// Returns the maximum number of interrupts/sec allowed by CEV per EQ.
// This is used in the EQ delay calculation below.
STATIC INLINE u32
be_eq_max_interrupt_rate (
    IN BE_FUNCTION_OBJECT *function_object)
{
    FUNCTION_ASSERT (function_object);

    //
    // Assumes a constant register setting for 
    // cev.cev_max_sust_int_dly_cnt_csr.MAX_CSR_STROLL = 95
    //
    // CLOCK_PERIOD (ns) = 4 (asic) or 400 (fpga)
    //
    // max_interrupt_rate = 1000000000 / ((MAX_CSR_STROLL+1) * 4 * CLOCK_PERIOD)
    //
    if (be_function_is_asic (function_object)) { 
        return 651042; // asic
    } else {
        return 6510; // fpga
    }
}

// Converts the EQ delay in microseconds to the multiplier that is programmed
// into the EQ context. 
u32 
be_eq_delay_to_multiplier (
    IN BE_FUNCTION_OBJECT *function_object,
    IN u32                 microsecond_delay)
{
    const u32 round = 10;
    u32 max_interrupt_rate = be_eq_max_interrupt_rate (function_object);
    u32 multiplier;    

    if (microsecond_delay == 0) {
        
        // No EQ delay, corresponding to the maximum interrupt rate.
        multiplier = 0; 

    } else {

        u32 interrupt_rate = 1000000 / microsecond_delay;

        if (interrupt_rate == 0) {

            // Maximum delay, corresponding to the lowest interrupt rate.
            multiplier = 1023; 

        } else {     

            multiplier = (max_interrupt_rate - interrupt_rate) * round;
            multiplier /= interrupt_rate; // At this point multiplier is *round bigger than required.
            
            // Round the multiplier to the closest value.
            multiplier = (multiplier + round/2) / round;

            multiplier = MIN (multiplier, 1023);
        }

    }

    return multiplier;
}

// Converts the EQ delay multiplier in EQ context to a delay value in microseconds.
u32 
be_eq_multiplier_to_delay (
    IN BE_FUNCTION_OBJECT *function_object, 
    IN u32 multiplier)
{
    u32 max_interrupt_rate = be_eq_max_interrupt_rate (function_object);
    u32 microsecond_delay;

    if (multiplier == 0) {
        return 0; // special case for disabling the delay
    }

    BE_GEN_ASSERT (multiplier < 1024);
    multiplier = MIN (multiplier, 1023);

    microsecond_delay = (multiplier + 1) * 1000000 / max_interrupt_rate;

    return microsecond_delay;
}




