/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

#include "pch.h"



void be_eth_rq_create_internal_callback (PVOID context, BESTATUS status, MCC_WRB* wrb)
{
    PBE_ETHRQ_OBJECT eth_rq = (PBE_ETHRQ_OBJECT)context;
    ETHRQ_ASSERT (eth_rq);

    if (status != BE_SUCCESS) {
        
        TRACE (DL_ERR, "Create eth rq failed in internal callback.\n");

    } else {

        IOCTL_ETH_RX_CREATE *ioctl = NULL;

        ASSERT (wrb); // non embedded by definition

        ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (eth_rq->parent_function, wrb, ETH_RX_CREATE);
        
        // Save the ring ID for cleanup.
        eth_rq->rid = ioctl->params.response.id;
        eth_rq->rss_cpu_index = ioctl->params.response.cpu_id;

        be_cq_object_reference (eth_rq->cq_object);

        _be_function_add_eth_rq (eth_rq->parent_function, eth_rq);           
    
    }

    return;
}

/*!

@brief
    This routine creates a Ethernet receive ring.  
    
@param
    function_object      - handle to a function object                           
@param
    rq_base_va            - base VA for the default receive ring. this must be
                          exactly 8K in length and continguous physical memory.
@param
    cq_object            - handle to a previously created CQ to be associated
                          with the RQ.
@param
    pp_eth_rq             - pointer to an opqaue handle where an eth/TOE default 
                          receive object is returned.                          
@return
    BE_SUCCESS if successfull, , otherwise a useful BESTATUS error code is returned.
@note
    IRQL: < DISPATCH_LEVEL
    this function allocates a PBE_ETHRQ_OBJECT object.
    there must be no more than 1 of these per function object, unless the
    function object supports RSS (is networking and on the host).         
    the rq_base_va must point to a buffer of exactly 8K.
    the erx::host_cqid (or host_stor_cqid) register and erx::ring_page registers
    will be updated as appropriate on return
*/    
BESTATUS
be_eth_rq_create_async (
    PBE_FUNCTION_OBJECT    function_object, 
    PSA_SGL                sgl, 
    PBE_CQ_OBJECT          cq_object, 
    PBE_CQ_OBJECT          bcmc_cq_object,
    u32                    frag_size,
    u32                    max_frame_size,
    boolean                rss_queue,
    PBE_RXIF_OBJECT        rxif_object,     
    PBE_ETHRQ_OBJECT       eth_rq,
    BE_GENERIC_QUEUE_CONTEXT* queue_context,
    MCC_WRB_CQE_CALLBACK   callback,
    PVOID                  callback_context)
{
    BESTATUS           status = 0;
    MCC_WRB *wrb = NULL;
    IOCTL_ETH_RX_CREATE *ioctl = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    // MPU will set the 
    ASSERT (sgl);
    ASSERT (eth_rq);

    FUNCTION_ASSERT (function_object);
    CQ_ASSERT (cq_object);

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }
    
    eth_rq->magic            = BE_ETHRQ_MAGIC;
    eth_rq->parent_function  = function_object;
    eth_rq->cq_object        = cq_object;

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto Error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, ETH_RX_CREATE);

    if (max_frame_size > frag_size)  {
        ASSERT (sa_is_power2 (frag_size));
    }

    ioctl->params.request.num_pages = 2; // required length
    ioctl->params.request.cq_id = be_cq_get_id (cq_object);
    ioctl->params.request.interface_id = rxif_object->handle;
    ioctl->params.request.frag_size = sa_log2 (sa_ge_power2 (frag_size));
    ioctl->params.request.max_frame_size = max_frame_size;
    ioctl->params.request.rss_queue = rss_queue;

    if (bcmc_cq_object) {
        // Only pd 0 host & storage can receive broadcast/multicast.
        if (be_function_get_pd_number (function_object) != 0) {
            TRACE (DL_WARN, 
                   "bcmc_cq_object ignored for pd_number:%d", 
                   be_function_get_pd_number (function_object));
        }
        //ioctl->params.request.bcmc_cq_id = be_cq_get_id (bcmc_cq_object);
    } else {
        //ioctl->params.request.bcmc_cq_id = 0xFFFF;
    }

    // Create a page list for the IOCTL.
    be_sgl_to_pa_list (sgl, 
                       ioctl->params.request.pages, 
                       SA_NUMBER_OF (ioctl->params.request.pages));

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_internal_callback (function_object, 
                                                              wrb, 
                                                              generic_context, 
                                                              callback, 
                                                              callback_context, 
                                                              be_eth_rq_create_internal_callback, 
                                                              eth_rq, 
                                                              ioctl);
    if (status != BE_SUCCESS &&
        status != BE_PENDING)
    {
        TRACE (DL_ERR, "ioctl to map eth rxq frags failed.");
        goto Error;
    }

Error:

    be_unlock_wrb_post (function_object);
    return status;
}

#ifdef BECLIB_BE3_SUPPORT

BESTATUS
be_eth_rq_create_async_v1 (
    PBE_FUNCTION_OBJECT    function_object, 
    PSA_SGL                sgl,     
    PBE_RXIF_OBJECT        rxif_object,
    PBE_CQ_OBJECT          cq_object, 
    BE_ETH_RQ_PARAMETERS   *parameters,        
    PBE_ETHRQ_OBJECT       eth_rq,
    BE_GENERIC_QUEUE_CONTEXT* queue_context,
    MCC_WRB_CQE_CALLBACK   callback,
    PVOID                  callback_context)
{
    BESTATUS           status = 0;
    MCC_WRB *wrb = NULL;
    IOCTL_ETH_RX_CREATE_V1 *ioctl = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    // MPU will set the 
    ASSERT (sgl);
    ASSERT (eth_rq);

    FUNCTION_ASSERT (function_object);
    CQ_ASSERT (cq_object);

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }
    
    eth_rq->magic            = BE_ETHRQ_MAGIC;
    eth_rq->parent_function  = function_object;
    eth_rq->cq_object        = cq_object;

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto Error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, ETH_RX_CREATE_V1);

    if (parameters->max_frame_size > parameters->frag_size)  {
        ASSERT (sa_is_power2 (parameters->frag_size));
    }

    ioctl->header.request.version = 1;

    ioctl->params.request.num_pages            = 2; // required length
    ioctl->params.request.cq_id                = be_cq_get_id (cq_object);
    ioctl->params.request.interface_id         = rxif_object->handle;
    ioctl->params.request.frag_size            = sa_log2 (sa_ge_power2 (parameters->frag_size));
    ioctl->params.request.max_frame_size       = parameters->max_frame_size;
    ioctl->params.request.rss_queue            = parameters->rss_queue;
    ioctl->params.request.backfill             = parameters->backfill;
    ioctl->params.request.dca                  = parameters->dca;
    ioctl->params.request.fixed_split_location = parameters->fixed_split_location;
    ioctl->params.request.header_dma           = parameters->header_dma;
    ioctl->params.request.header_size          = parameters->header_size;
    ioctl->params.request.header_split         = parameters->header_split;
    ioctl->params.request.split_location       = parameters->split_location;    

    // Create a page list for the IOCTL.
    be_sgl_to_pa_list (sgl, 
                       (PHYS_ADDR*)ioctl->params.request.pages, 
                       SA_NUMBER_OF (ioctl->params.request.pages));

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_internal_callback (function_object, 
                                                              wrb, 
                                                              generic_context, 
                                                              callback, 
                                                              callback_context, 
                                                              be_eth_rq_create_internal_callback, 
                                                              eth_rq, 
                                                              ioctl);
    if (status != BE_SUCCESS &&
        status != BE_PENDING)
    {
        TRACE (DL_ERR, "ioctl to map eth rxq frags failed.");
        goto Error;
    }

Error:

    be_unlock_wrb_post (function_object);
    return status;
}

#endif

BESTATUS
be_eth_rq_create (
    PBE_FUNCTION_OBJECT    function_object, 
    PSA_SGL                sgl, 
    PBE_CQ_OBJECT          cq_object, 
    PBE_CQ_OBJECT          bcmc_cq_object,
    u32                    frag_size,
    u32                    max_frame_size,
    boolean                rss_queue,
    PBE_RXIF_OBJECT        rxif_object,     
    PBE_ETHRQ_OBJECT       eth_rq)
{
    return be_eth_rq_create_async (function_object,
                                   sgl, 
                                   cq_object,
                                   bcmc_cq_object,
                                   frag_size,
                                   max_frame_size,
                                   rss_queue,
                                   rxif_object,
                                   eth_rq,
                                   NULL, NULL, NULL);
}

/*!

@brief
    This routine returns the RID for an Ethernet receive queue     

@param    
    EthRq - Ethernet receive queue handle returned from EthRqCreate

@return   
    Returns BE_SUCCESS on success and an appropriate BESTATUS on failure.

@note    
    IRQL: <= DISPATCH_LEVEL
*/
u32
be_eth_rq_get_id(
    IN      PBE_ETHRQ_OBJECT        rq)
{
   ETHRQ_ASSERT (rq);
   return rq->rid;
}


/*
    Returns the CPU id for the given RQ.  Only valid for RSS RQs.
*/
u8
be_eth_rq_get_rss_cpu_index (
    IN  PBE_ETHRQ_OBJECT        rq)
{
   ETHRQ_ASSERT (rq);
   return rq->rss_cpu_index;
}



/*!

@brief
    This routine destroys an Ethernet receive queue     

@param    
    eth_rq - ethernet/toe default receive queue handle returned from eth_rq_create

@return   
    Returns BE_SUCCESS on success and an appropriate BESTATUS on failure.

@note    
    This function frees resourcs allocated by EthRqCreate.
    The erx::host_cqid (or host_stor_cqid) register and erx::ring_page registers
    will be updated as appropriate on return
    IRQL: < DISPATCH_LEVEL
*/

void be_eth_rq_destroy_internal_callback (PVOID context, BESTATUS status, MCC_WRB* wrb)
{
    PBE_ETHRQ_OBJECT eth_rq = (PBE_ETHRQ_OBJECT)context;
    ETHRQ_ASSERT (eth_rq);

    if (status != BE_SUCCESS && !eth_rq->parent_function->ue) {
        TRACE (DL_ERR, "Destroy eth rq failed in internal callback.\n");
    } 

    // Dereference any CQs associated with this queue.
    be_cq_object_dereference (eth_rq->cq_object);

    // Remove from the function object.
    _be_function_remove_eth_rq (eth_rq->parent_function, eth_rq);

    memset (eth_rq, 0, sizeof (*eth_rq));

    return;
}


BESTATUS
be_eth_rq_destroy_async (
    IN PBE_ETHRQ_OBJECT     eth_rq,
    IN MCC_WRB_CQE_CALLBACK callback,
    IN PVOID                callback_context)
{
    BESTATUS status = BE_SUCCESS;
    ETHRQ_ASSERT (eth_rq);

    // Send ioctl to destroy the RQ.
    status = be_function_ring_destroy_async (
                 eth_rq->parent_function,
                 eth_rq->rid,
                 IOCTL_RING_TYPE_ETH_RX,
                 callback,
                 callback_context,
                 be_eth_rq_destroy_internal_callback,
                 eth_rq);
    
    return status;
}

BESTATUS
be_eth_rq_destroy (
    IN      PBE_ETHRQ_OBJECT        eth_rq)
{
    return be_eth_rq_destroy_async (eth_rq, NULL, NULL);
}


//---------------------------------------------------------------------------------------
// Function: be_eth_rq_destroy_options
//
//   Destroys an ethernet receive ring with finer granularity options than the standard
//   be_eth_rq_destroy() API function. 
//
// eth_rq           - 
// flush            - Set to 1 to flush the ring, set to 0 to bypass the flush
// callback         - Callback function on completion
// callback_context - Callback context
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_rq_destroy_options (
   IN PBE_ETHRQ_OBJECT     eth_rq,
   IN boolean              flush,
   IN BE_GENERIC_QUEUE_CONTEXT* queue_context,
   IN MCC_WRB_CQE_CALLBACK callback,
   IN PVOID                callback_context)
{
    IOCTL_ETH_RX_DESTROY *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = BE_SUCCESS;
    PBE_FUNCTION_OBJECT function_object = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    ETHRQ_ASSERT (eth_rq);

    function_object = eth_rq->parent_function;
    FUNCTION_ASSERT (function_object);

    be_lock_wrb_post (function_object);

    TRACE (DL_INFO, "Destroy eth_rq ring id:%d, flush:%d", eth_rq->rid, flush);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }  

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, ETH_RX_DESTROY);

    ioctl->params.request.id = eth_rq->rid;
    ioctl->params.request.bypass_flush = ((0 == flush) ? 1 : 0);

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_internal_callback (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 be_eth_rq_destroy_internal_callback,
                 eth_rq,
                 ioctl);

    if (status != BE_SUCCESS &&
        status != BE_PENDING) {
        // TRACE (DL_ERR, "eth_rq ring destroy ioctl failed. id:%d, flush:%d", eth_rq->rid, flush);
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}
