/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        hwi.c

    Purpose:

        This file implements the Hardware Interface (hwi) for the ACIT library.
        Functions implemented in this library are specific to BladeEngine ASIC.
        

--*/


#include "acitpch.h"

ACIT_UEXACT32
hwi_get_ws_sizes(
    PACIT_UEXACT32          pcontroller_ws_size,
    PACIT_UEXACT32          ptarget_ws_size,
    PACIT_UEXACT32          plun_ws_size,
    PACIT_UEXACT32          pio_ws_size,
    PACIT_UEXACT32          ptmf_ws_size)
{
    *pcontroller_ws_size = sizeof(HWI_CONTROLLER_WS);
    *ptarget_ws_size = 0;
    *plun_ws_size = 0;
    *pio_ws_size = 0;
    *ptmf_ws_size = 0;

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_init_controller_ws(
    PACIT_CONTROLLER_WS         pcontroller)
{
    PHWI_CONTROLLER_WS          phwi_controller;

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    OSM_MEMSET(phwi_controller, 0, sizeof(HWI_CONTROLLER_WS));

    phwi_controller->pacit_ws = pcontroller;

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_init_mailbox_descriptor(
    PACIT_CONTROLLER_WS         pcontroller)
{
    PACIT_CONTROLLER_WS         pc = pcontroller;
    PACIT_PARAMETERS            p = &pc->params;
    PHWI_CONTROLLER_WS          ph = GET_HWI_CONTROLLER_WS(pc);

    ACIT_UEXACT32               mailbox_size;
    ACIT_UEXACT32               num_wrbh_desc, i;
    ACIT_UEXACT32               remaining_cxns, num_cxns;
    ACIT_UEXACT32               hwi_ctx_mem;

    const ACIT_UEXACT32 a1 = ACIT_ALIGN_1BYTE;
    const ACIT_UEXACT32 a4k = ACIT_ALIGN_4096BYTE;

    const ACIT_UEXACT32 cached = ACIT_MEM_CATEGORY_CACHED;
    const ACIT_UEXACT32 non_cached = ACIT_MEM_CATEGORY_NON_CACHED;

    // Context memory has connection specific and msix specific
    // additions
    hwi_ctx_mem = sizeof(HWI_CONTEXT_MEMORY) + 
                  sizeof(BE_CXN_QUEUE_OBJECT) * p->num_cxn +
                  sizeof(HWI_INT_QUEUE_OBJECTS) * p->num_cq;
       

    // ----------------------------------------------------
    // Pre-compute some of the intermediate sizes required.
    // ----------------------------------------------------
    mailbox_size = sizeof(MCC_WRB) + sizeof(MCC_CQ_ENTRY);
    ap_set_mem_const (pc, (PUEXACT8)"mailbox",    HWI_MEM_MAILBOX,          mailbox_size,               a4k, non_cached);   
    ap_set_mem_const (pc, (PUEXACT8)"headers",    HWI_MEM_HDRS,             0,                          a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"cq",         HWI_MEM_CQ,               0,                          a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"eq",         HWI_MEM_EQ,               0,                          a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"async_hdr",  HWI_MEM_ASYNC_HEADER_BUF, 0,                          a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"async_hdr_ring",HWI_MEM_ASYNC_HEADER_RING, 0,                      a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"async_hdr_h",HWI_MEM_ASYNC_HEADER_HANDLE,  0,                      a1, cached);
    ap_set_mem_const (pc, (PUEXACT8)"async_data", HWI_MEM_ASYNC_DATA_BUF,       0,                      a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"async_data_ring",HWI_MEM_ASYNC_DATA_RING,  0,                      a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"async_data_h",HWI_MEM_ASYNC_DATA_HANDLE,   0,                      a1, cached);
    ap_set_mem_const (pc, (PUEXACT8)"mcc",        HWI_MEM_MCC,              0,                          a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"mcc_cq",     HWI_MEM_MCC_CQ,           0,                          a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"context",    HWI_MEM_ADDN_CONTEXT,     hwi_ctx_mem,                a1, cached);
    ap_set_mem_const (pc, (PUEXACT8)"async_ctx",  HWI_MEM_ASYNC_PDU_CONTEXT,sizeof(HWI_ASYNC_PDU_CONTEXT), a1, cached);
    ap_set_mem_const (pc, (PUEXACT8)"sglh",       HWI_MEM_SGLH,             0,                          a1, cached);
    ap_set_mem_const (pc, (PUEXACT8)"mcc_ctx",    HWI_MEM_MCC_CONTEXT,      0,                          a1, non_cached);

    // SGE are 32 per I/O.
    ph->wrb_bytes_per_cxn = sa_round_up (sizeof (ISCSI_WRB) * p->num_cmd_per_cxn, BE_PAGE_SIZE);
    ph->sge_bytes_per_io = sizeof(ISCSI_SGE) * p->num_sge;

    ph->wrb_mem_index = pc->acit_var_desc_index.range.wrb;
    ap_set_mem_variable (pc, (PUEXACT8)"wrb", ph->wrb_mem_index, 0, 0, a4k, non_cached);

    // Allocate HWI_MEM_FRAG
    ph->frag_mem_index = pc->acit_var_desc_index.range.frag;
    ap_set_mem_variable (pc, (PUEXACT8)"sge", ph->frag_mem_index, 0, 0, a4k, non_cached);

    // Allocate HWI_MEM_WRBH
    // 256 WRBHs belonging to 8 cxns will go in 1 mem descriptor
    ph->wrbh_mem_index = pc->acit_var_desc_index.range.wrbh;
    num_wrbh_desc = sa_ceiling(p->num_cxn, 8);
    remaining_cxns = p->num_cxn;
    num_cxns = 0;

    for(i = 0; i < num_wrbh_desc; i++)
    {
        ACIT_UEXACT8 mem_name[8] = {"wrbh_"};
        mem_name[5]=(ACIT_UEXACT8)i+48;
        mem_name[6]=(ACIT_UEXACT8)'\0';

        num_cxns = (remaining_cxns > 8)? 8: num_cxns;
        ap_set_mem_variable (pc, mem_name, 
                             ph->wrbh_mem_index + i, 
                             0, 
                             0, a1, non_cached);
        remaining_cxns -= num_cxns;
    }

    ph->max_cxns_per_descriptor = 0;
    ph->max_sgl_per_descriptor  = 0;
    ph->max_wrbh_per_descriptor = 0;
    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_init_mem_descriptors(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PACIT_CONTROLLER_WS     pc = pcontroller;
    PACIT_PARAMETERS        p = &pc->params;
    PHWI_CONTROLLER_WS      ph = GET_HWI_CONTROLLER_WS(pc);

    ACIT_UEXACT32           num_cq_page_sz, num_eq_page_sz;
    ACIT_UEXACT32           num_async_pdu_buf_page_sz;
    ACIT_UEXACT32           num_mcc_page_sz, num_mcc_cq_page_sz;
    ACIT_UEXACT32           num_wrbh_desc, i, max_phys = ACIT_CFG_MAX_CONTIGUOUS_MEM;

    ACIT_UEXACT32           remaining_cxns, num_cxns, mailbox_size, max_mcc_wrbs;
    ACIT_UEXACT32           mem_default_pdu_ring_sz, mem_default_pdu_buffer_sz; 
    ACIT_UEXACT32           mem_default_pdu_handle_sz, async_pdu_context_sz;
    ACIT_UEXACT32           hwi_ctx_mem;

    const ACIT_UEXACT32 a1 = ACIT_ALIGN_1BYTE;
    const ACIT_UEXACT32 a4k = ACIT_ALIGN_4096BYTE;

    const ACIT_UEXACT32 cached = ACIT_MEM_CATEGORY_CACHED;
    const ACIT_UEXACT32 non_cached = ACIT_MEM_CATEGORY_NON_CACHED;

    // SGE are 32 per I/O.
    ph->wrb_bytes_per_cxn = sa_round_up (sizeof (ISCSI_WRB) * p->num_cmd_per_cxn, BE_PAGE_SIZE);
    ph->sge_bytes_per_io = sizeof(ISCSI_SGE) * p->num_sge;

    OSM_ASSERT(ph->wrb_bytes_per_cxn <= (sizeof(ISCSI_WRB) * BE_MAX_CMD_PER_CXN));

    // ----------------------------------------------------
    // Pre-compute some of the intermediate sizes required.
    // ----------------------------------------------------
    mailbox_size = sizeof(MCC_WRB) + sizeof(MCC_CQ_ENTRY);

    // number of completions is equal to the number of commands that can be pending
    num_cq_page_sz = p->num_cq_entries * sizeof (ISCSI_CQE);
    p->num_cq_pages = sa_ceiling(num_cq_page_sz, PAGE_SIZE);
    p->num_cq_entries = p->num_cq_pages * PAGE_SIZE/sizeof(ISCSI_CQE);

    num_eq_page_sz = p->num_eq_entries * sizeof (EQ_ENTRY);
    p->num_eq_pages = sa_ceiling(num_eq_page_sz, PAGE_SIZE);
    p->num_eq_entries = p->num_eq_pages * PAGE_SIZE/sizeof(EQ_ENTRY);

    num_async_pdu_buf_page_sz = p->num_async_pdus * p->default_pdu_buf_sz;
    p->num_async_pdu_buf_pages = sa_ceiling(num_async_pdu_buf_page_sz, PAGE_SIZE);
    OSM_ASSERT(num_async_pdu_buf_page_sz <= p->num_async_pdu_buf_pages * PAGE_SIZE);

    p->num_async_pdu_buf_sgl_pages = sa_ceiling(p->num_async_pdus * sizeof(PHYS_ADDR), PAGE_SIZE);

    // MCC ring is used by IOCTLs and invalidates, max can only be 8 pages
    max_mcc_wrbs = p->ioctls_per_ctrl;
    if(max_mcc_wrbs > MAX_MCC_WRBS)
        max_mcc_wrbs = MAX_MCC_WRBS;

    num_mcc_page_sz = (max_mcc_wrbs) * sizeof (MCC_WRB);
    p->num_mcc_pages = sa_ceiling(num_mcc_page_sz, PAGE_SIZE);

    if(p->num_mcc_pages > 8)
        p->num_mcc_pages = 8;

    num_mcc_cq_page_sz = p->ioctls_per_ctrl * sizeof (MCC_CQ_ENTRY);
    p->num_mcc_cq_pages = sa_ceiling(num_mcc_cq_page_sz, PAGE_SIZE);

    mem_default_pdu_buffer_sz = p->num_async_pdu_buf_pages*PAGE_SIZE;
    mem_default_pdu_ring_sz = p->num_async_pdu_buf_sgl_pages * PAGE_SIZE;
    mem_default_pdu_handle_sz = p->num_async_pdus * sizeof(ASYNC_PDU_HANDLE);

    async_pdu_context_sz = sizeof(HWI_ASYNC_PDU_CONTEXT) + 
                           ((p->num_async_pdus) * sizeof(HWI_ASYNC_ENTRY));

    // Context memory has connection specific and msix specific
    // additions
    hwi_ctx_mem = sizeof(HWI_CONTEXT_MEMORY) + 
                  sizeof(BE_CXN_QUEUE_OBJECT) * p->num_cxn +
                  sizeof(HWI_INT_QUEUE_OBJECTS) * p->num_cq;

    ap_set_mem_const (pc, (PUEXACT8)"mailbox",    HWI_MEM_MAILBOX,        mailbox_size,                        a4k, non_cached);   

    p->num_hdr_pages = sa_ceiling(p->num_cxn,32);
    ap_set_mem_const (pc, (PUEXACT8)"headers",    HWI_MEM_HDRS,           p->num_hdr_pages*PAGE_SIZE,         a4k, non_cached); // 32 headers per page

    ap_set_mem_const (pc, (PUEXACT8)"cq",         HWI_MEM_CQ,             p->num_cq_pages*PAGE_SIZE*p->num_cq, a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"eq",         HWI_MEM_EQ,             p->num_eq_pages*PAGE_SIZE*p->num_cq, a4k, non_cached);

    ap_set_mem_const (pc, (PUEXACT8)"async_hdr",  HWI_MEM_ASYNC_HEADER_BUF,mem_default_pdu_buffer_sz,          a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"async_hdr_ring",HWI_MEM_ASYNC_HEADER_RING, mem_default_pdu_ring_sz,       a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"async_hdr_h",HWI_MEM_ASYNC_HEADER_HANDLE, mem_default_pdu_handle_sz ,     a1,  cached);

    ap_set_mem_const (pc, (PUEXACT8)"def_buf",    HWI_MEM_ASYNC_DATA_BUF,mem_default_pdu_buffer_sz,            a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"def_buf_ring",HWI_MEM_ASYNC_DATA_RING, mem_default_pdu_ring_sz,           a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"def_buf_h",  HWI_MEM_ASYNC_DATA_HANDLE, mem_default_pdu_handle_sz ,       a1,  cached);

    ap_set_mem_const (pc, (PUEXACT8)"mcc",        HWI_MEM_MCC,            p->num_mcc_pages*PAGE_SIZE,         a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"mcc_cq",     HWI_MEM_MCC_CQ,         p->num_mcc_cq_pages*PAGE_SIZE,      a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"context",    HWI_MEM_ADDN_CONTEXT,   hwi_ctx_mem,                        a1, cached);
    ap_set_mem_const (pc, (PUEXACT8)"async_ctx",  HWI_MEM_ASYNC_PDU_CONTEXT,async_pdu_context_sz,             a1, cached);
    ap_set_mem_const (pc, (PUEXACT8)"sglh",       HWI_MEM_SGLH,           sizeof(SGL_HANDLE) * p->icds_per_controller, a1, cached);
    ap_set_mem_const (pc, (PUEXACT8)"mcc_ctx",    HWI_MEM_MCC_CONTEXT,    sizeof(BE_MCC_WRB_CONTEXT) * max_mcc_wrbs , a1, non_cached);

    OSM_ASSERT(sizeof(BE_MCC_WRB_CONTEXT) * p->ioctls_per_ctrl <= max_phys);

    // Allocate HWI_MEM_WRB
    ph->wrb_mem_index = pc->acit_var_desc_index.range.wrb;
    ap_set_mem_variable (pc, (PUEXACT8)"wrb", ph->wrb_mem_index, p->num_cxn, ph->wrb_bytes_per_cxn, a4k, non_cached);

    // Allocate HWI_MEM_FRAG
    ph->frag_mem_index = pc->acit_var_desc_index.range.frag;
    ap_set_mem_variable (pc, (PUEXACT8)"sge", ph->frag_mem_index, p->icds_per_controller, ph->sge_bytes_per_io, a4k, non_cached);

    // Allocate HWI_MEM_WRBH
    // 256 WRBHs belonging to 8 cxns will go in 1 mem descriptor
    ph->wrbh_mem_index = pc->acit_var_desc_index.range.wrbh;
    num_wrbh_desc = sa_ceiling(p->num_cxn, 8);
    remaining_cxns = p->num_cxn;
    num_cxns = p->num_cxn;

    for(i = 0; i < num_wrbh_desc; i++)
    {
        ACIT_UEXACT8 mem_name[8] = {"wrbh_"};
        mem_name[5]=(ACIT_UEXACT8)i+48;
        mem_name[6]=(ACIT_UEXACT8)'\0';

        num_cxns = (remaining_cxns > 8)? 8: num_cxns;
        ap_set_mem_variable (pc, mem_name, 
                             ph->wrbh_mem_index + i, 
                             num_cxns * p->num_cmd_per_cxn, 
                             sizeof(WRB_HANDLE), a1, non_cached);
        remaining_cxns -= num_cxns;
    }

    ph->max_cxns_per_descriptor = (max_phys / ph->wrb_bytes_per_cxn);
    ph->max_sgl_per_descriptor  = (max_phys / (p->num_sge * sizeof(ISCSI_SGE)));
    ph->max_wrbh_per_descriptor = (8 * p->num_cmd_per_cxn);

    return ACIT_STATUS_SUCCESS;
}

ACIT_U32
hwi_init_hiber_controller(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_VOID              osm_handle)
{
    PHWI_CONTROLLER_WS      phwi_controller;

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    phwi_controller->phwic = (PHWI_CONTEXT_MEMORY) pcontroller->
        mem_descriptors[HWI_MEM_ADDN_CONTEXT].virtual_address;

    if(hwi_init_beclib(pcontroller, osm_handle) != ACIT_STATUS_SUCCESS)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR , ("hwi_init_beclib failed\n"));
        return ACIT_STATUS_ERROR;
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_U32
hwi_init_controller(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_VOID              osm_handle)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    PHWI_CONTEXT_MEMORY     phwi_context;
    ACIT_UEXACT32           desc_index, i;

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    phwi_controller->pmcc_wrb_base = (PMCC_WRB) pcontroller->
        mem_descriptors[HWI_MEM_MCC].virtual_address;

    phwi_controller->phwic = (PHWI_CONTEXT_MEMORY) pcontroller->
        mem_descriptors[HWI_MEM_ADDN_CONTEXT].virtual_address;

    phwi_controller->phwic->pbe_mcc_context = (PBE_MCC_WRB_CONTEXT) pcontroller->
        mem_descriptors[HWI_MEM_MCC_CONTEXT].virtual_address;

    desc_index = phwi_controller->frag_mem_index;
    OSM_ASSERT(pcontroller->acit_var_desc_index.num.frag <= HWI_MEM_NUM_FRAG);
    for(i=0; i<pcontroller->acit_var_desc_index.num.frag; i++, desc_index++)
    {
        phwi_controller->sge_base[i] = \
                pcontroller->mem_descriptors[desc_index].virtual_address;
    }

    desc_index = phwi_controller->wrb_mem_index;
    OSM_ASSERT(pcontroller->acit_var_desc_index.num.wrb <= HWI_MEM_NUM_WRB);
    for(i=0; i<pcontroller->acit_var_desc_index.num.wrb; i++, desc_index++)
    {
        phwi_controller->wrb_base[i] = \
                pcontroller->mem_descriptors[desc_index].virtual_address;
    }

    //
    // Initialize the context memory
    //

    phwi_context = phwi_controller->phwic;

    // The WRB queues are at the end of the context
    phwi_context->pbe_wrbq = (PBE_CXN_QUEUE_OBJECT)(phwi_context + 1);

    // Following the WRB queues, the EQ/CQ objects are located
    phwi_context->phwi_queue = (PHWI_INT_QUEUE_OBJECTS)&phwi_context->pbe_wrbq[pcontroller->params.num_cxn].be_wrbq;

    // Check context setup
    {
        ACIT_UEXACT32 ctx_len;
        PUEXACT32 cva, cva_end;
        cva = pcontroller->mem_descriptors[HWI_MEM_ADDN_CONTEXT].virtual_address;
        ctx_len = pcontroller->mem_descriptors[HWI_MEM_ADDN_CONTEXT].size_in_bytes;

        cva_end = (PUEXACT32)&phwi_context->phwi_queue[pcontroller->params.num_cq].be_eq;

        OSM_ASSERT(((ACIT_UEXACT32)(cva_end - cva)) <= ctx_len);
    }


    hwi_init_wrb_handle(pcontroller);
    hwi_init_async_pdu_ctx(pcontroller);
    if(hwi_init_beclib(pcontroller, osm_handle) != ACIT_STATUS_SUCCESS)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR , ("hwi_init_beclib failed\n"));
        return ACIT_STATUS_ERROR;
    }

    // initialize the sgl handle now since we need hwi_init_beclib to get
    // the firmware config prior to initializing the sgl handle
    hwi_init_sgl_handle(pcontroller);

    // reserve a SGL for NOPOut.
    if(hwi_reserve_nop_sgl(pcontroller) != ACIT_STATUS_SUCCESS)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR , ("hwi_reserve_nop_sgl failed\n"));
        hwi_cleanup_beclib(pcontroller);
        return ACIT_STATUS_ERROR;
    }

    // Get PCIE Link info from CSR
    hwi_get_pci_link_info(pcontroller);

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
hwi_stop_controller(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PHWI_CONTROLLER_WS      phwc;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           isr, pid;

    // initiate cleanup of EP firmware
    mgmt_status = mgmt_epfw_cleanup(pcontroller, IOCTL_CONNECTION_CHUTE_0);
    OSM_ASSERT(mgmt_status == MGMT_STATUS_SUCCESS);

    // destroy all rings
    hwi_cleanup_beclib(pcontroller);

    // purge all EQ entries that may have been left out
    // This is to workaround a problem we've seen occasionally where
    // we exit from acit_stop_controller and we get an interrupt
    // with EQ entry bit set
    for(pid = 0; pid < pcontroller->params.num_cq; pid++)
    {
        hwi_purge_eq(pcontroller, pid);
    }

    // Read the ISR register to clear any pending interrupt bit
    isr = CsrRead(pcontroller, pcontroller->fw_config.isr_offset);
    
    // destroy SA device object
    phwc = GET_HWI_CONTROLLER_WS(pcontroller);
    sa_dev_destroy (&phwc->phwic->sa_dev);
}

ACIT_UEXACT32
hwi_register_arm_async_events(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PHWI_CONTROLLER_WS      phwc;
    PHWI_CONTEXT_MEMORY     phwi_context;
    BESTATUS                bestatus = 0;

    phwc = GET_HWI_CONTROLLER_WS(pcontroller);
    ASSERT (phwc);
    phwi_context = phwc->phwic;

    bestatus = be_mcc_add_async_event_callback(
                                &phwi_context->be_mcc,
                                (MCC_ASYNC_EVENT_CALLBACK)mgmt_async_event_handler,
                                pcontroller);

    if (bestatus != BE_SUCCESS)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Async Event registration failed!!\n"));
        return ACIT_STATUS_ERROR;
    }

    phwi_context->initialized.async_regn = 1; 
    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
hwi_check_health(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           log_error)
{
    PHWI_CONTROLLER_WS      phwic;
    PCICFG_UE_STATUS_LOW_CSR ue_lo;
    PCICFG_UE_STATUS_HI_CSR  ue_hi;

    PCICFG_ONLINE0_CSR      online0;
    PCICFG_ONLINE1_CSR      online1;

    ACIT_UEXACT32           ue_lo_mask, ue_hi_mask; 
    ACIT_UEXACT32           semaphore_offset = 0, fw_line = ACIT_U32_MINUS_ONE;

    ACIT_UEXACT32           ext_error[3];

    if(!pcontroller->eh_stats.ue_detected)
    {
        phwic = GET_HWI_CONTROLLER_WS(pcontroller);

        online0.dw = ACIT_PCICFG0_READ(pcontroller, online0);
        online1.dw = ACIT_PCICFG0_READ(pcontroller, online1);

        if((!online0.dw) || (!online1.dw))
        {
            ue_lo.dw = ACIT_PCICFG0_READ(pcontroller, ue_status_low);
            ue_hi.dw = ACIT_PCICFG0_READ(pcontroller, ue_status_hi);

            ue_lo_mask = ACIT_PCICFG0_READ(pcontroller, ue_status_low_mask);
            ue_hi_mask = ACIT_PCICFG0_READ(pcontroller, ue_status_hi_mask);

            ue_lo.dw = (ue_lo.dw & (~ue_lo_mask));
            ue_hi.dw = (ue_hi.dw & (~ue_hi_mask));

            if(ue_lo.dw || ue_hi.dw)
            {
//                online0.dw = 0;
//                online1.dw = 0;
//                online0.host1_online = 1;
//                ACIT_PCICFG0_WRITE(pcontroller, online0, online0.dw);
//                ACIT_PCICFG0_WRITE(pcontroller, online1, online1.dw);

                if(ue_lo.rxulp0_ue_status)
                    semaphore_offset = RXULP0_SEMAPHORE_OFFSET;
                else if(ue_lo.rxulp1_ue_status)
                    semaphore_offset = RXULP1_SEMAPHORE_OFFSET;
                else if(ue_lo.rxulp2_ue_status)
                    semaphore_offset = RXULP2_SEMAPHORE_OFFSET;
                else if(ue_lo.tpost_ue_status)
                    semaphore_offset = TPOST_SEMAPHORE_OFFSET;
                else if(ue_lo.tpre_ue_status)
                    semaphore_offset = TPRE_SEMAPHORE_OFFSET;
                else if(ue_lo.txulp0_ue_status)
                    semaphore_offset = TXULP0_SEMAPHORE_OFFSET;
                else if(ue_lo.txulp1_ue_status)
                    semaphore_offset = TXULP1_SEMAPHORE_OFFSET;
                else if(ue_lo.txulp2_ue_status)
                    semaphore_offset = TXULP2_SEMAPHORE_OFFSET;
                else if(ue_lo.mpu_ue_status)
                    semaphore_offset = MPU_SEMAPHORE_OFFSET;

                fw_line = CsrRead(pcontroller, semaphore_offset);

                OSM_DEBUG_PRINT(ACIT_DBGMSG_FATAL,(("UE Detected - UE Low 0x%x UE Hi 0x%x Line# 0x%x\n"),
                                    ue_lo.dw, ue_hi.dw, fw_line));

                ext_error[0] = ue_lo.dw;
                ext_error[1] = ue_hi.dw;
                ext_error[2] = fw_line;

                ap_log_osm_event(pcontroller,
                                 UNI_CFG_LOG_LEVEL_CRITICAL,
                                 UNI_SUBMODULE_ACIT_HWI,
                                 UNI_PHASE_DONT_CARE,
                                 3,
                                 ACIT_EC_HWI_UE,
                                 0, ACIT_NULL,
                                 &ext_error[0]);

                pcontroller->eh_stats.ue_low = ue_lo.dw;
                pcontroller->eh_stats.ue_hi = ue_hi.dw;
                pcontroller->eh_stats.ue_fw_line = fw_line;
                pcontroller->eh_stats.ue_detected = 1;

                OSM_ASSERT(0);
            }
        }
    }
}

//
// This routine dynamically configures the EQ timer based on
// the interrupt rate. This is run for every timer tick in ACIT
//

ACIT_VOID
hwi_check_interrupts(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           pid)
{
    ACIT_UEXACT32           reconfigure = 0;
    ACIT_SEXACT32           new_eq_delay = 0;
    ACIT_UEXACT32           io_load = 0; 
    ACIT_UEXACT32           load_factor = 0, target_load = 0;

    pcontroller->partition[pid].dic.interrupts_per_sec = \
                (pcontroller->partition[pid].dic.interrupts - 
                 pcontroller->partition[pid].dic.last_interrupt_count) *
                pcontroller->params.ticks_per_second;

    if(pcontroller->partition[pid].dic.interrupts_per_sec > 
       pcontroller->partition[pid].dic.max_interrupts_per_sec)
    {
        pcontroller->partition[pid].dic.max_interrupts_per_sec = 
                        pcontroller->partition[pid].dic.interrupts_per_sec;
    }

    // extrapolate to ios_per second
    pcontroller->partition[pid].dic.ios_per_sec *= pcontroller->params.ticks_per_second;

    if(pcontroller->partition[pid].dic.ios_per_sec > 
       pcontroller->partition[pid].dic.max_ios_per_sec)
    {
        pcontroller->partition[pid].dic.max_ios_per_sec = 
            pcontroller->partition[pid].dic.ios_per_sec;
    }

   {
        ACIT_UEXACT32 eq_run;

        eq_run = pcontroller->partition[pid].dic.eq_timer / pcontroller->params.dic.timer_increments;

        if(eq_run <= 31)
        {
            pcontroller->partition[pid].dic.eq_runs[eq_run]++;
        }
    }

    ap_determine_load(pcontroller, &io_load, &load_factor, &target_load, pid);

    pcontroller->partition[pid].dic.current_target_load = target_load;
    pcontroller->partition[pid].dic.current_load_factor = load_factor;

    if(!io_load)
    {
        // if there is no load, then the interrupt rate and timers must go back to default
        if((pcontroller->partition[pid].dic.interrupt_rate != pcontroller->params.dic.max_interrupt_rate) ||
           (pcontroller->partition[pid].dic.eq_timer != pcontroller->params.dic.min_timer_val))
        {
            pcontroller->partition[pid].dic.interrupt_rate = pcontroller->params.dic.max_interrupt_rate;
            new_eq_delay = pcontroller->params.dic.min_timer_val;
            reconfigure = 1;
        }
    }
    else
    {
        pcontroller->partition[pid].dic.interrupt_rate = pcontroller->params.dic.max_interrupt_rate -
                                            target_load * pcontroller->params.dic.interrupt_dec_steps;

        if(pcontroller->partition[pid].dic.interrupt_rate < pcontroller->params.dic.min_interrupt_rate)
        {
            pcontroller->partition[pid].dic.interrupt_rate = pcontroller->params.dic.min_interrupt_rate;
        }

        if((ACIT_SEXACT32)pcontroller->partition[pid].dic.interrupts_per_sec > 
            (pcontroller->partition[pid].dic.interrupt_rate + (ACIT_SEXACT32)pcontroller->params.dic.interrupt_resolution))
        {
            // is the system loaded at the same time with lots of IOs?
            // In this case modify eq timer thresholds at a faster rate and also
            // modify the expected interrupt rate

            new_eq_delay = pcontroller->partition[pid].dic.eq_timer + 
                            pcontroller->params.dic.timer_increments * load_factor;

            if(new_eq_delay > (ACIT_SEXACT32)pcontroller->params.dic.max_timer_val)
                new_eq_delay = (ACIT_SEXACT32)pcontroller->params.dic.max_timer_val;

            reconfigure = 1;
        }
        else if((ACIT_SEXACT32)pcontroller->partition[pid].dic.interrupts_per_sec < 
                (pcontroller->partition[pid].dic.interrupt_rate - (ACIT_SEXACT32)pcontroller->params.dic.interrupt_resolution))
        {
            new_eq_delay = pcontroller->partition[pid].dic.eq_timer - pcontroller->params.dic.timer_increments;

            if(new_eq_delay < (ACIT_SEXACT32)pcontroller->params.dic.min_timer_val)
                new_eq_delay = (ACIT_SEXACT32)pcontroller->params.dic.min_timer_val;

            reconfigure = 1;
        }
    }

    if(reconfigure && 
       (new_eq_delay != (ACIT_SEXACT32)(pcontroller->partition[pid].dic.eq_timer)))
    {
        PBE_FUNCTION_OBJECT pfo;
        ACIT_UEXACT32 delay_multiplier;
        OSM_ASSERT(new_eq_delay);

        pfo = hwi_get_be_pfo(pcontroller);
        delay_multiplier = be_eq_delay_to_multiplier(pfo, new_eq_delay);
        
        mgmt_modify_eq_delay(pcontroller, 
                             pid,
                             pcontroller->phwi_ws->partition[pid].eq.id,
                             delay_multiplier);

        pcontroller->partition[pid].dic.eq_timer = new_eq_delay;            
    }

    // make this the new interrupt rate
    pcontroller->partition[pid].dic.last_interrupt_count = pcontroller->partition[pid].dic.interrupts;

    pcontroller->partition[pid].dic.last_ios_per_sec = pcontroller->partition[pid].dic.ios_per_sec;
    pcontroller->partition[pid].dic.ios_per_sec = 0;

}   

#if 0

ACIT_VOID
hwi_check_interrupts(
    PACIT_CONTROLLER_WS     pcontroller)
{
    ACIT_UEXACT32           reconfigure = 0;
    ACIT_SEXACT32           new_eq_delay = 0;
    ACIT_UEXACT32           io_load = 0; 
    ACIT_UEXACT32           load_factor = 0, target_load = 0;

    pcontroller->dic.interrupts_per_sec = \
                (pcontroller->dic.interrupts - pcontroller->dic.last_interrupt_count) *
                pcontroller->params.ticks_per_second;

    if(pcontroller->dic.interrupts_per_sec > pcontroller->dic.max_interrupts_per_sec)
    {
        pcontroller->dic.max_interrupts_per_sec = pcontroller->dic.interrupts_per_sec;
    }

    // extrapolate to ios_per second
    pcontroller->dic.ios_per_sec *= pcontroller->params.ticks_per_second;

    if(pcontroller->dic.ios_per_sec > pcontroller->dic.max_ios_per_sec)
    {
        pcontroller->dic.max_ios_per_sec = pcontroller->dic.ios_per_sec;
    }

    {
        ACIT_UEXACT32 eq_run;

        eq_run = pcontroller->dic.eq_timer / pcontroller->params.dic.timer_increments;

        if(eq_run <= 31)
        {
            pcontroller->dic.eq_runs[eq_run]++;
        }
    }

    ap_determine_load(pcontroller, &io_load, &load_factor, &target_load);

    pcontroller->dic.current_target_load = target_load;
    pcontroller->dic.current_load_factor = load_factor;

    if(!io_load)
    {
        // default to max interrupt rate and min eq timer when there's no load
        if((pcontroller->dic.interrupt_rate != pcontroller->params.dic.max_interrupt_rate) ||
            (pcontroller->dic.eq_timer != pcontroller->params.dic.min_timer_val))
        {
            pcontroller->dic.interrupt_rate = pcontroller->params.dic.max_interrupt_rate;
            new_eq_delay = pcontroller->params.dic.min_timer_val;
            reconfigure = 1;
        }
    }
    else
    {
        pcontroller->dic.interrupt_rate = pcontroller->params.dic.max_interrupt_rate -
                                            target_load * pcontroller->params.dic.interrupt_dec_steps;

        if(pcontroller->dic.interrupt_rate < pcontroller->params.dic.min_interrupt_rate)
        {
            pcontroller->dic.interrupt_rate = pcontroller->params.dic.min_interrupt_rate;
        }

        if((ACIT_SEXACT32)pcontroller->dic.interrupts_per_sec > 
           (pcontroller->dic.interrupt_rate + pcontroller->params.dic.interrupt_resolution))
        {
            // is the system loaded at the same time with lots of IOs?
            // In this case modify eq timer thresholds at a faster rate and also
            // modify the expected interrupt rate

            new_eq_delay = pcontroller->dic.eq_timer + pcontroller->params.dic.timer_increments * load_factor;

            if(new_eq_delay > (ACIT_SEXACT32)pcontroller->params.dic.max_timer_val)
                new_eq_delay = (ACIT_SEXACT32)pcontroller->params.dic.max_timer_val;

            reconfigure = 1;
        }
        else if((ACIT_SEXACT32)pcontroller->dic.interrupts_per_sec <
                (pcontroller->dic.interrupt_rate - pcontroller->params.dic.interrupt_resolution))
        {
            new_eq_delay = pcontroller->dic.eq_timer - pcontroller->params.dic.timer_increments;

            if(new_eq_delay < (ACIT_SEXACT32)pcontroller->params.dic.min_timer_val)
                new_eq_delay = (ACIT_SEXACT32)pcontroller->params.dic.min_timer_val;

            reconfigure = 1;
        }
    }

    if(reconfigure && (new_eq_delay != pcontroller->dic.eq_timer))
    {
        PBE_FUNCTION_OBJECT pfo;
        ACIT_UEXACT32 delay_multiplier;
        OSM_ASSERT(new_eq_delay);

        pfo = hwi_get_be_pfo(pcontroller);
        delay_multiplier = be_eq_delay_to_multiplier(pfo, new_eq_delay);
        
        mgmt_modify_eq_delay(pcontroller,
                             0,
                             pcontroller->phwi_ws->eq.id,
                             delay_multiplier);

        pcontroller->dic.eq_timer = new_eq_delay;            
    }

    // make this the new interrupt rate
    pcontroller->dic.last_interrupt_count = pcontroller->dic.interrupts;

    pcontroller->dic.last_ios_per_sec = pcontroller->dic.ios_per_sec;
    pcontroller->dic.ios_per_sec = 0;

}   

#endif

ACIT_VOID
hwi_build_sa_sgl(
    IN  PACIT_CONTROLLER_WS pcontroller,
    IN  ACIT_UEXACT32       mem_descriptor_index,
    OUT PSA_SGL             sgl)
{
    PHWI_CONTROLLER_WS  phwi_controller;

    ASSERT (sgl);
    ASSERT (pcontroller);
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    ASSERT (phwi_controller);

    // Destroy any existing SGL.
    if (sa_sgl_get_base_va (sgl)) {
        sa_sgl_destroy_contiguous (sgl);
    }

    sa_sgl_create_contiguous (
        pcontroller->mem_descriptors[mem_descriptor_index].virtual_address,
        pcontroller->mem_descriptors[mem_descriptor_index].bus_address.u.a64.address,
        pcontroller->mem_descriptors[mem_descriptor_index].size_in_bytes,
        sgl);
}

ACIT_VOID
hwi_build_sa_sgl_by_offset (
    IN  PACIT_CONTROLLER_WS pcontroller,
    IN  ACIT_UEXACT32       mem_descriptor_index,
    IN  ACIT_UEXACT32       offset,
    IN  ACIT_UEXACT32       size,
    OUT PSA_SGL             sgl)
{
    PHWI_CONTROLLER_WS  phwi_controller;

    ASSERT (sgl);
    ASSERT (pcontroller);
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    ASSERT (phwi_controller);

    // Destroy any existing SGL.
    if (sa_sgl_get_base_va (sgl)) {
        sa_sgl_destroy_contiguous (sgl);
    }

    sa_sgl_create_contiguous (
        (PACIT_UEXACT8)pcontroller->mem_descriptors[mem_descriptor_index].virtual_address + offset,
        pcontroller->mem_descriptors[mem_descriptor_index].bus_address.u.a64.address + offset,
        size,
        sgl);
}

ACIT_VOID
hwi_cleanup_beclib(
    PACIT_CONTROLLER_WS     pcontroller)    
{
    PHWI_CONTROLLER_WS      phwc = NULL;
    PHWI_CONTEXT_MEMORY     phwi_context;
    BESTATUS bestatus = 0;

    ASSERT (pcontroller);
    phwc = GET_HWI_CONTROLLER_WS (pcontroller);
    ASSERT (phwc);
    phwi_context = phwc->phwic;
    OSM_ASSERT(phwi_context);

    bestatus = be_function_cleanup(&phwi_context->be_function);
    OSM_ASSERT(bestatus == BE_SUCCESS);

    return;
}


ACIT_UEXACT32
hwi_init_beclib(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_VOID              osm_handle)
{
    BESTATUS                bestatus;
    SA_DEV_BAR_LOCATIONS    bars[3];
    PBE_FUNCTION_OBJECT     pfo = NULL;
    SA_SGL                  sgl = {0}, *pemul = NULL;
    PACIT_PARAMETERS        p = &pcontroller->params;
    PHWI_CONTROLLER_WS      phwc;
    PHWI_CONTEXT_MEMORY     phwi_context;
    ACIT_U32 i = 0, desc_index = 0;
    ACIT_UEXACT32           def_pdu_ring_sz;
    ACIT_UEXACT32           timer_threshold, page_offset;
    ACIT_UEXACT32           num_sge = 0, bits = 0;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           err_status = 0;
    ACIT_UEXACT32           pid, num_eq_entries, num_cq_entries;
    PACIT_VOID              eq_va, cq_va;
    ACIT_UEXACT32           wrb_mem_index, offset, size;
	ACIT_UEXACT32			sgl_num_pg_to_post,num_pg_post;

#ifdef BE_EMULATION
    SA_SGL                  emul_sgl = {0};
#endif

    phwc = GET_HWI_CONTROLLER_WS(pcontroller);
    ASSERT (phwc);

    phwi_context = phwc->phwic;
    OSM_ASSERT(phwi_context);

    pfo = &phwi_context->be_function; // abbreviation

    bars[0].base_pa = pcontroller->bar0_pa.u.a64.address;
    bars[0].base_va = pcontroller->bar0_va;
    bars[0].length = sizeof (PCICFG0_CSRMAP);
    bars[0].mem_or_io_mapped = SA_MEM_MAPPED;
    bars[0].type = SA_BAR_TYPE_PCI;
    DBG_LOG_OP_PARAMETER(&pcontroller->vase_master_block,
                        "BAR0",
                        bars[0].base_va);

    bars[1].base_pa = pcontroller->bar1_pa.u.a64.address;
    bars[1].base_va = pcontroller->bar1_va;
    bars[1].length = sizeof (BLADE_ENGINE_CSRMAP);
    bars[1].mem_or_io_mapped = SA_MEM_MAPPED;
    bars[1].type = SA_BAR_TYPE_CSR;
    DBG_LOG_OP_PARAMETER(&pcontroller->vase_master_block,
                        "BAR1",
                        bars[1].base_va);

    bars[2].base_pa = pcontroller->bar2_pa.u.a64.address;
    bars[2].base_va = pcontroller->bar2_va;
    bars[2].length = sizeof (PROTECTION_DOMAIN_DBMAP);
    bars[2].mem_or_io_mapped = SA_MEM_MAPPED;
    bars[2].type = SA_BAR_TYPE_PD;
    DBG_LOG_OP_PARAMETER(&pcontroller->vase_master_block,
                        "BAR2",
                        bars[2].base_va);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("Creating SA Device..\n"));

    bestatus = (BESTATUS)sa_dev_create (bars, 3, osm_handle, &phwc->phwic->sa_dev);

    if(bestatus != BE_SUCCESS)
    {
        OSM_ASSERT(0);
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("sa_dev_create failed - %d\n", bestatus));
        err_status = HWI_ERR_SA_DEV_CREATE;

        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_CRITICAL,
                         UNI_SUBMODULE_ACIT_HWI,
                         UNI_PHASE_DONT_CARE,
                         1,
                         ACIT_EC_HWI_INIT_FAILED,
                         0, ACIT_NULL,
                         &err_status);

        return ACIT_STATUS_ERROR;
    }

#ifndef ACIT_NO_CHIP_RESET_FOR_CRASHDUMP
init_start:
#endif

    /*
     * BEClib initialization.
     */

    /*
     * First initialize the library...
     */
    bestatus = be_initialize_library ();
    if (bestatus != BE_SUCCESS)
    {
        TRACE (DL_ERR, "Initialize beclib library failed.");
        err_status = HWI_ERR_BECLIB_INIT;
        goto error;
    }

    phwi_context->initialized.lib_init = 1;

    /*
     * .. and let the lib do its stuff for creating its internal objects...
     */
    hwi_build_sa_sgl (pcontroller, HWI_MEM_MAILBOX, &sgl);

#ifdef BE_EMULATION
    hwi_build_sa_sgl (pcontroller, HWI_MEM_EMUL, &emul_sgl);
    pemul = &emul_sgl;
#endif

    bestatus = be_function_and_chip_create (
                        &phwc->phwic->sa_dev,
                        BE_FUNCTION_TYPE_ISCSI,
                        &sgl,
                        pemul,
                        pfo,
                        &phwi_context->be_chip); // object to initialize

    if (bestatus != BE_SUCCESS)
    {
        TRACE (DL_ERR, "function create failed.");
        err_status = HWI_ERR_FUN_CHIP_CREATE;
        goto error;
    }

    phwi_context->initialized.chip_function = 1;

#ifndef ACIT_NO_CHIP_RESET_FOR_CRASHDUMP
    if(pcontroller->state.init_type == ACIT_INIT_TYPE_CRASHDUMP &&
       (pcontroller->state.soft_reset == 0) &&
       (pcontroller->params.crash_dump.dump_rst == 1))
    {
        bestatus = be_pci_soft_reset(pfo);

        if (bestatus != BE_SUCCESS)
        {
            TRACE (DL_ERR, "soft reset failed.");
            err_status = HWI_ERR_SOFT_RESET;
            goto error;
        }

        pcontroller->state.soft_reset = 1;

        bestatus = be_function_and_chip_destroy (
                            pfo,
                            &phwi_context->be_chip); // object to destroy

        goto init_start;
    }
#endif

    if((pcontroller->state.init_type == ACIT_INIT_TYPE_CRASHDUMP) &&
        (pcontroller->params.crash_dump.dump_rst == 0))
    {
        //
        // Bug 18349: A soft reset has some issues on some platforms where the CLP configured
        // data is not on the flash. So a reset will wipe out this info and ARM fw will not
        // be able to establish session to boot target. So avoid the chip reset and go with
        // function level reset.
        // FLR will work only when there's no UE. So check for UE first and then issue FLR 
        // IOCTL.
        //
        // Note!!
        // There's still a way to trigger a chip reset. Parameter dump_rst must be configured 
        // in the registry/insmod time and must be passed to the dump driver
        //
        hwi_check_health(pcontroller, ACIT_FALSE);

        if(pcontroller->eh_stats.ue_detected == 0)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Issuing function level reset..\n"));
            mgmt_status = mgmt_issue_function_reset(pcontroller);

            if(mgmt_status != MGMT_STATUS_SUCCESS)
            {
                bestatus = BE_NOT_OK;
                err_status = HWI_ERR_FUNCTION_RESET;
                TRACE (DL_ERR, "Failed to issue Function Reset, quitting\n");
                goto error;
            }
            else
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Function level reset completed successfully\n"));
            }

        }
        else
        {
            // UE during dump?! Sorry folks, the show's over.
            bestatus = BE_NOT_OK;
            TRACE (DL_ERR, "UE detected by dump driver, quitting\n");
            goto error;
        }
    }

    /* 
     * If in the hibernation or crash dump mode then no need
     * to initialize any rings. Driver will use the mailbox
     * mode to send IO.
     */
    if(pcontroller->state.api_type == ACIT_OP_MODE_RESTRICTED)
        return ACIT_STATUS_SUCCESS;

    def_pdu_ring_sz = (p->num_async_pdus * sizeof(PHYS_ADDR));

    /*
     * Create SA rings for CQ and EQ
     */

    eq_va = (PACIT_UEXACT8)pcontroller->mem_descriptors[HWI_MEM_EQ].virtual_address;
    cq_va = (PACIT_UEXACT8)pcontroller->mem_descriptors[HWI_MEM_CQ].virtual_address;

    num_eq_entries = (pcontroller->mem_descriptors[HWI_MEM_EQ].size_in_bytes / sizeof (EQ_ENTRY))/
                            pcontroller->params.num_cq;

    num_cq_entries = (pcontroller->mem_descriptors[HWI_MEM_CQ].size_in_bytes / sizeof (ISCSI_CQE))/
                            pcontroller->params.num_cq;

    for(pid = 0; pid < pcontroller->params.num_cq; pid++)
    {
        sa_ring_create(&phwc->partition[pid].eq,
                num_eq_entries, // num entries
                sizeof (EQ_ENTRY),
                (PACIT_UEXACT8)eq_va + pid * num_eq_entries * sizeof(EQ_ENTRY));

        sa_ring_create(&phwc->partition[pid].cq,
                num_cq_entries, // num entries
                sizeof (ISCSI_CQE),
                (PACIT_UEXACT8)cq_va + pid * num_cq_entries * sizeof(ISCSI_CQE));
    }

    sa_ring_create(&phwc->default_pdu_hdr,
              pcontroller->mem_descriptors[HWI_MEM_ASYNC_HEADER_RING].size_in_bytes / sizeof (PHYS_ADDR), // num entries
              sizeof (PHYS_ADDR),
              pcontroller->mem_descriptors[HWI_MEM_ASYNC_HEADER_RING].virtual_address);

    sa_ring_create(&phwc->default_pdu_buf,
              pcontroller->mem_descriptors[HWI_MEM_ASYNC_DATA_RING].size_in_bytes / sizeof (PHYS_ADDR), // num entries
              sizeof (PHYS_ADDR),
              pcontroller->mem_descriptors[HWI_MEM_ASYNC_DATA_RING].virtual_address);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("EQ VA - 0x%p\n", pcontroller->mem_descriptors[HWI_MEM_EQ].virtual_address));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("CQ VA - 0x%p\n", pcontroller->mem_descriptors[HWI_MEM_CQ].virtual_address));

#ifdef BE_EMULATION
    // initialize for emulation
    if(emul_hwi_init(pcontroller) != ACIT_STATUS_SUCCESS)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR , ("Initialization failed for emulation."));
        goto error;
    }
#endif

    // config params is in usecs and the units need to be in 8 usec units
    timer_threshold = p->eq_timer / 8;
    if(timer_threshold > CEV_TIMER_MAX)
        timer_threshold = CEV_TIMER_MAX;
    /*
     * .. start with creation of EQ ring...
     */

    for(pid = 0; pid < pcontroller->params.num_cq; pid++)
    {

        hwi_build_sa_sgl_by_offset (pcontroller,
                                    HWI_MEM_EQ,
                                    (pid * p->num_eq_pages * PAGE_SIZE),
                                    p->num_eq_pages * PAGE_SIZE,
                                    &sgl);

        bestatus = be_eq_create (pfo,
                                &sgl,
                                sizeof (EQ_ENTRY),
                                p->num_eq_entries,
                                be_eq_delay_to_multiplier(pfo, timer_threshold), // timer delay in 8 usec units
                                &phwi_context->phwi_queue[pid].be_eq);

        if (bestatus != BE_SUCCESS)
        {
            TRACE (DL_ERR, "eq create failed.");
            goto error;
        }

        phwc->partition[pid].eq.id = be_eq_get_id(&phwi_context->phwi_queue[pid].be_eq);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
            ("EQ created with RID %d\n", phwc->partition[pid].eq.id));

        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("EQ Object : 0x%p\n", &phwi_context->phwi_queue[pid].be_eq));
    }

    phwi_context->initialized.eq = 1;


    /*
     * Check the firmware version to see if we support it
     */
#if defined (BE_EMULATION) || defined (FPGA_ARM)

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Emulation or FPGA build, skipping fw_config IOCTL\n"));

#else
    mgmt_status = mgmt_check_supported_fw(pcontroller);

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        bestatus = BE_NOT_OK;
        TRACE (DL_ERR, "Unsupported fw version\n");
        err_status = HWI_ERR_SUPPORTED_FW_CHECK;
        goto error;
    }

    mgmt_status = mgmt_get_fw_config(pcontroller);

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        bestatus = BE_NOT_OK;
        TRACE (DL_ERR, "Error getting fw config\n");
        err_status = HWI_ERR_FW_CONFIG_GET;
        goto error;
    }

#endif

    /*
     * Check if interrupt redirection is supported by this FW
     */

    mgmt_status = mgmt_get_addi_attribs(pcontroller);

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        //
        // Ideally the IOCTL should not have failed. Lets continue with
        // the initialization anyway..
        //

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Failed to query additional attribs!\n"));
        pcontroller->addi_attribs.interrupt_redirection = 0;
        OSM_ASSERT(0);
    }

    /*
     * .. then comes the MCC CQ ring...
     */
    hwi_build_sa_sgl (pcontroller, HWI_MEM_MCC_CQ, &sgl);
    bestatus = be_cq_create (pfo, 
                             &sgl, 
                             sa_sgl_get_byte_count (&sgl), // num entries
                             ACIT_FALSE,                   // not eventable
                             ACIT_TRUE,                    // force interrupt
                             BE_CQ_DMA_COALESCE_NONE,      // DMA coalescing
                             &phwi_context->phwi_queue[0].be_eq, // associated event queue object
                             &phwi_context->be_mcc_cq);

    if (bestatus != BE_SUCCESS)
    {
        TRACE (DL_ERR, "mcc cq create failed.");
        err_status = HWI_ERR_MCC_CQ_CREATE;
        goto error;
    }
    phwi_context->initialized.mcc_cq = 1;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
        ("MCC CQ created with RID %d\n", be_cq_get_id(&phwi_context->be_mcc_cq)));

    // disabled for emulation..

    /*
     * .. followed by the MCC ring, which also transitions the chip to MCC mode...
     */
    hwi_build_sa_sgl (pcontroller, HWI_MEM_MCC, &sgl);
    bestatus = be_mcc_ring_create (pfo, 
                                   &sgl, 
                                   sa_sgl_get_byte_count (&sgl), 
                                   phwi_context->pbe_mcc_context, 
                                   p->ioctls_per_ctrl,
                                   &phwi_context->be_mcc_cq, 
                                   &phwi_context->be_mcc);
    if (bestatus != BE_SUCCESS)
    {
        TRACE (DL_ERR, "mcc create failed.");
        err_status = HWI_ERR_MCC_RING_CREATE;
        goto error;
    }
    phwi_context->initialized.mcc = 1;   

    for(pid = 0; pid < pcontroller->params.num_cq; pid++)
    {
        hwi_build_sa_sgl_by_offset (pcontroller, 
                                    HWI_MEM_CQ, 
                                    (pid * p->num_cq_pages * PAGE_SIZE),
                                    p->num_cq_pages * PAGE_SIZE,
                                    &sgl);

        bestatus = be_cq_create (pfo, 
                                &sgl, 
                                sa_sgl_get_byte_count (&sgl), 
                                ACIT_FALSE,    // not eventable
                                ACIT_FALSE,    // honor EQ timers
                                BE_CQ_DMA_COALESCE_NONE, // DMA coalescing
                                &phwi_context->phwi_queue[pid].be_eq, // associated event queue object
                                &phwi_context->phwi_queue[pid].be_cq);

        if (bestatus != BE_SUCCESS)
        {
            TRACE (DL_ERR, "cq create failed.");
            err_status = HWI_ERR_ISCSI_CQ_CREATE;
            goto error;
        }

        phwc->partition[pid].cq.id = be_cq_get_id(&phwi_context->phwi_queue[pid].be_cq);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
            ("iSCSI CQ created with RID %d\n", phwc->partition[pid].cq.id));

        phwi_context->initialized.cq = 1;
    }

    /*
     * .. and rings for posting async PDU buffers...
     */

    hwi_build_sa_sgl (pcontroller, HWI_MEM_ASYNC_HEADER_RING, &sgl);
    bestatus = be_iscsi_defq_create (pfo, 
                                     &sgl, 
                                     def_pdu_ring_sz, // length
                                     p->default_pdu_buf_sz,
                                     &phwi_context->phwi_queue[0].be_cq, 
                                     &phwi_context->be_def_hdrq);
    if (bestatus != BE_SUCCESS)
    {
        TRACE (DL_ERR, "defq create failed.");
        err_status = HWI_ERR_DPDU_HDR_RING_CREATE;
        goto error;
    }
    phwi_context->initialized.def_hdrq = 1;
    phwc->default_pdu_hdr.id = be_iscsi_defq_get_id(&phwi_context->be_def_hdrq);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
        ("Default PDU Header Ring created with RID %d\n", phwc->default_pdu_hdr.id));

    /*
     * Now that the default PDU ring header has been created, let the EP know 
     * about it...
     */
    hwi_post_async_buffers(pcontroller, 1); // header

    hwi_build_sa_sgl (pcontroller, HWI_MEM_ASYNC_DATA_RING, &sgl);
    bestatus = be_iscsi_defq_create (pfo, 
                                     &sgl, 
                                     def_pdu_ring_sz, // length
                                     p->default_pdu_buf_sz,
                                     &phwi_context->phwi_queue[0].be_cq, 
                                     &phwi_context->be_def_bufq);
    if (bestatus != BE_SUCCESS)
    {
        TRACE (DL_ERR, "defq create failed.");
        err_status = HWI_ERR_DPDU_DATA_RING_CREATE;
        goto error;
    }
    phwi_context->initialized.def_bufq = 1;
    phwc->default_pdu_buf.id = be_iscsi_defq_get_id(&phwi_context->be_def_bufq);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
        ("Default PDU Data Ring created with RID %d\n", phwc->default_pdu_buf.id));



    /*
     * Now that the default PDU data ring has been created, let the EP know 
     * about it...
     */
    hwi_post_async_buffers(pcontroller, 0); // header

    {
        IOCTL_COMMON_ADD_TEMPLATE_HEADER_BUFFERS_RESPONSE response; 

		memset(&response,0,sizeof(response));

        hwi_build_sa_sgl (pcontroller, HWI_MEM_HDRS, &sgl);
        bestatus = be_function_post_template_header_memory (pfo, &sgl, &response);    
        if (bestatus != BE_SUCCESS)
        {
            TRACE (DL_ERR, "post templates failed.");
            err_status = HWI_ERR_TEMPLATE_HEADER_POST;
            goto error;
        }
        phwi_context->initialized.templates = 1;
    }

    /*
     * Set the frag_num bits. Technically its not needed for default driver that uses
     * 32 ISCSI_SGEs, but no harm in writing it here. This also gives us a chance to
     * support configurability for frag_num if needed.
     */
    OSM_ASSERT(p->num_sge);
    num_sge = p->num_sge;

    if(num_sge != 32)
    {
        while(num_sge != 0x1)
        {
            num_sge = (num_sge >> 1);
            bits++;
        }

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Using %d bits for frag_num in ISCSI_SGL_CRA\n", bits));

        mgmt_status = mgmt_set_fragnum_bits(pcontroller, bits);

        if(mgmt_status != MGMT_STATUS_SUCCESS)
        {
            bestatus = BE_NOT_OK;
            err_status = HWI_ERR_FRAGNUM_SET;
            TRACE (DL_ERR, "Cannot set fragnum bits\n");
            goto error;
        }
    }

    // post sgl pages. Note that the page offset indicates the starting range 
    // of icd

    desc_index = phwc->frag_mem_index;
    page_offset = (phwc->sge_bytes_per_io * pcontroller->fw_config.iscsi_icd_start)/BE_PAGE_SIZE;
	sgl_num_pg_to_post = (pcontroller->fw_config.iscsi_icd_count * phwc->sge_bytes_per_io)/BE_PAGE_SIZE;

    for(i=0; i<pcontroller->acit_var_desc_index.run_time.frag; i++, desc_index++)
    {
        hwi_build_sa_sgl (pcontroller, desc_index, &sgl);

		num_pg_post = pcontroller->mem_descriptors[desc_index].size_in_bytes / BE_PAGE_SIZE;
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Posting SGLPage: offset %d, num_pages %d\n",
                page_offset,
                pcontroller->mem_descriptors[desc_index].size_in_bytes / BE_PAGE_SIZE));

		/*
		 * ACIT allocates memory even before the fw_config is called. Post pages
		 * only for the FW_CONFIG_ICD_COUNT.
		 */
	
		if(sgl_num_pg_to_post < num_pg_post) 
			num_pg_post = sgl_num_pg_to_post;
		else
			sgl_num_pg_to_post -= num_pg_post;

        bestatus = be_iscsi_post_sgl_pages (
                        pfo,
                        &sgl,
                        page_offset,
                        num_pg_post);

        page_offset += num_pg_post;

        OSM_ASSERT(bestatus == BE_SUCCESS);

        if (bestatus != BE_SUCCESS)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR , ("post sgl failed."));
            err_status = HWI_ERR_SGL_PAGES_POST;
            goto error;
        }
    }

    phwi_context->initialized.sgl_pages = 1;

    /*
     * create all the WRB ring objects.
     */

    for (i=0; i<p->num_cxn; i++) {

        wrb_mem_index = 0;
        offset = 0;
        size = 0;

        hwi_get_wrb_mem(phwc, i, &wrb_mem_index, &offset, &size);

        hwi_build_sa_sgl_by_offset (pcontroller, 
                                    wrb_mem_index,
                                    offset,
                                    size,
                                    &sgl);

        bestatus = be_iscsi_wrb_queue_create (pfo,
                                &sgl,
                                sa_sgl_get_byte_count (&sgl),
                                &phwi_context->pbe_wrbq[i].be_wrbq);

        if (bestatus != BE_SUCCESS)
        {
            TRACE (DL_ERR, "wrbq create failed.");
            err_status = HWI_ERR_WRB_QUEUE_CREATE;
            goto error;
        }

        /*
         * This is a table which maps the Connection Resource Index
         * and the CID used by the hardware. CRI is a resource shared
         * with iSCSI and is used so that iSCSI gets a consistent 0
         * based number for managing connection resources.
         */
        phwc->wrb_context[i].cid = \
           (ACIT_UEXACT16)be_iscsi_wrbq_get_cid(&phwi_context->pbe_wrbq[i].be_wrbq);

        if(i == 0)
        {
            OSM_ASSERT(pcontroller->fw_config.iscsi_cid_start == phwc->wrb_context[i].cid);

            // The first connection will be the cid_offset for this
            // function
            //pcontroller->fw_config.iscsi_cid_start = phwc->wrb_context[i].cid;

            //OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,
            //   ("CID 0x%x will be used as iscsi_cid_start..\n", i, 
            //                                        phwc->wrb_context[i].cid));
        }

        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
           ("iSCSI WRB Ring created for CRI %d with CID 0x%x\n", i, 
                                                phwc->wrb_context[i].cid));

        phwi_context->initialized.num_wrbq = i + 1;

    }
    
error:

    if (bestatus != BE_SUCCESS) {
        TRACE (DL_ERR, "beclib init failed");

        if(phwi_context->initialized.chip_function)
        {
            hwi_cleanup_beclib (pcontroller);
        }

        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_CRITICAL,
                         UNI_SUBMODULE_ACIT_HWI,
                         UNI_PHASE_DONT_CARE,
                         1,
                         ACIT_EC_HWI_INIT_FAILED,
                         0, ACIT_NULL,
                         &err_status);

        return ACIT_STATUS_ERROR;
    }
#if 0
    // inform the chip about the number of msi-x interrupts we need
    // Fixme!!! Do not call this for non-MSIx mode
    if(pcontroller->params.num_cq)
    {
        bestatus = be_cev_modify_msi_messages(pfo, 
                                              pcontroller->params.num_cq,
                                              ACIT_NULL, ACIT_NULL, ACIT_NULL);

        if(bestatus != BE_SUCCESS)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("ModifyMSIMessages failed msix %d status 0x%x\n", 
                                    pcontroller->params.num_cq, bestatus));
            OSM_ASSERT(0);
        }
    }
#endif
    /*
     * .. DONE! now log some important address in debug memory
     */
    DBG_LOG_OP_PARAMETER(&pcontroller->vase_master_block,"psgl_handle_base",phwc->psgl_handle_base);
//    DBG_LOG_OP_PARAMETER(&pcontroller->vase_master_block,"pasync_buff_base",phwc->pasync_buff_base);
    DBG_LOG_OP_PARAMETER(&pcontroller->vase_master_block,"wrb_base",phwc->wrb_base[0]);
    DBG_LOG_OP_PARAMETER(&pcontroller->vase_master_block,"pmcc_wrb_base",phwc->pmcc_wrb_base);
    DBG_LOG_OP_PARAMETER(&pcontroller->vase_master_block,"CQ Base",
                    (ACIT_UEXACT64)pcontroller->mem_descriptors[HWI_MEM_CQ].virtual_address);
    DBG_LOG_OP_PARAMETER(&pcontroller->vase_master_block,"EQ Base",
                    (ACIT_UEXACT64)pcontroller->mem_descriptors[HWI_MEM_EQ].virtual_address);
    DBG_LOG_OP_PARAMETER(&pcontroller->vase_master_block,"MCC CQ Base",
                    (ACIT_UEXACT64)pcontroller->mem_descriptors[HWI_MEM_MCC_CQ].virtual_address);


    return ACIT_STATUS_SUCCESS;
}

ACIT_U8
hwi_disable_irq(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PBE_FUNCTION_OBJECT     pfo;

    if(pcontroller->state.api_type != ACIT_OP_MODE_RESTRICTED)
    {
        pfo = hwi_get_be_pfo(pcontroller);
        OSM_ASSERT(pfo);
        be_function_disable_interrupts(pfo);
    }
    else
    {
        // not handled currently. This will be enabled once we determine
        // how the crashdump mode will be handled
        OSM_ASSERT(0);
    }

    pcontroller->state.interrupt_enabled = ACIT_FALSE;
    return ACIT_TRUE;
}

ACIT_U8
hwi_enable_irq(
    PACIT_CONTROLLER_WS     pcontroller)

{
    PCICFG_MEMBAR_CTRL_INT_CTRL_CSR ctrl;   
    CEV_ISR0_CSR       isr;
    ACIT_UEXACT32       pid = 0;

    ctrl.dw = ACIT_PCICFG0_READ (pcontroller, membar_ctrl_intr_ctrl);

    isr.dw = CsrRead(pcontroller, pcontroller->fw_config.isr_offset);

    if (!ctrl.hostintr) 
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("HostIntr not set, setting now..\n"));

        ctrl.hostintr = 1;        
        ACIT_PCICFG0_WRITE(pcontroller, membar_ctrl_intr_ctrl, ctrl.dw);        
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("HostIntr set, only consuming pending eq..\n"));
    }

    for(pid = 0; pid < pcontroller->params.num_cq; pid++)
    {
        //
        // Bug 9732: If the EQ is not processed and rearmed, when resuming from
        // local boot hibernation, we stop receiving interrupts. Its also important
        // to have the correct value for num_processed. The beclib function that
        // enables interrupt always writes a value of 0 for num_processed. Therefore
        // we write our own enable interrupt routine
        //
        hwi_process_eq(pcontroller, pid);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Rearming EQ - id 0x%x num_processed %d\n",
            pcontroller->phwi_ws->partition[pid].eq.id, pcontroller->partition[pid].dic.num_eq_processed));

        // Rearm the EQ since we may have missed the last interrupt while
        // the host interrupts were disabled.
        hwi_ring_eq_db (pcontroller, 
                        pcontroller->phwi_ws->partition[pid].eq.id, 
                        0, // clear interrupt
                        pcontroller->partition[pid].dic.num_eq_processed, 
                        1, 
                        1);

        pcontroller->partition[pid].dic.num_eq_processed = 0;
    }

    pcontroller->state.interrupt_enabled = ACIT_TRUE;
    return ACIT_TRUE;
}

ACIT_VOID
hwi_process_mcc_cq(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    be_mcc_process_cq(&phwi_controller->phwic->be_mcc, ACIT_FALSE);   // no-rearm
}

ACIT_VOID
hwi_process_cq (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           message_id)
{            
    PHWI_CONTROLLER_WS      phwi_controller;
    SA_RING                 *ring;
    PISCSI_CQE              cqe;
    PI_SOL_CQE              sol;
    PI_T_DMSG_CQE           dmsg;
    ACIT_UEXACT32           num_processed = 0; // num processed this time
    ACIT_UEXACT32           count = 0;

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    DBG_LOG_FUNC_ENTER(&pcontroller->vase_master_block,
                        FUNC_hwi_process_cq,
                        VASE_MODULE_ACIT,
                        0,
                        0,
                        0);

    ring = &phwi_controller->partition[message_id].cq;

    cqe = (PISCSI_CQE) sa_ring_current (ring); // CQ entry to process
    ACIT_UPDATE_CQ_INT_COUNT(pcontroller, message_id);

    while (cqe->valid && (count < pcontroller->params.max_cqe_per_isr)) 
    {
        ACIT_UPDATE_CQ_PROCESSED_COUNT(pcontroller, message_id);

        if (num_processed >= 32) 
        {
            // Occassionally free space in CQ for the hardware to write
            // more entries without enabling the interrupt yet.
            hwi_ring_cq_db (pcontroller, ring->id, num_processed, 0, 0);
            num_processed = 0;
        }     

        switch(cqe->code)
        {
        case SOL_CMD_COMPLETE:

            hwi_complete_cmd(phwi_controller, message_id, (PI_SOL_CQE)cqe);
            ACIT_UPDATE_ISCSI_CMD_INT_COUNT(pcontroller, message_id);

            break;

        case DRIVERMSG_NOTIFY:

            dmsg = (PI_T_DMSG_CQE)cqe;
            // Free the WRB index in the chain until this WRB.  
            hwi_complete_drvr_msgs(phwi_controller, dmsg->cid, dmsg->wrb_index);
            ACIT_UPDATE_DRIVER_MSG_INT_COUNT(pcontroller, message_id);

            break;

        case UNSOL_HDR_NOTIFY:
        case UNSOL_DATA_NOTIFY:

            hwi_process_default_pdu_ring(pcontroller, (PI_T_DPDU_CQE)cqe);
            ACIT_UPDATE_DFT_PDU_INT_COUNT(pcontroller, message_id);

            break;

        case CXN_INVALIDATE_INDEX_NOTIFY:

            // FixmeBE2: Not implemented!!
            OSM_ASSERT(0);

            break;

        case CMD_INVALIDATED_NOTIFY:
        case CXN_INVALIDATE_NOTIFY:

            // These can be ignored
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Ignoring CQ Error notification for cmd/cxn invalidate\n"));
            break;

        case SOL_CMD_KILLED_DATA_DIGEST_ERR:
        case CMD_KILLED_INVALID_STATSN_RCVD:
        case CMD_KILLED_INVALID_R2T_RCVD:
        case CMD_CXN_KILLED_LUN_INVALID:
        case CMD_CXN_KILLED_ICD_INVALID:
        case CMD_CXN_KILLED_ITT_INVALID:
        case CMD_CXN_KILLED_SEQ_OUTOFORDER:
        case CMD_CXN_KILLED_INVALID_DATASN_RCVD:

            // Though some of the above code say CMD_CXN, it only affects the command in 
            // initiator mode. Target mode affects the connection, hence the _CXN..

            sol = (PI_SOL_CQE)cqe;

            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("CQ Error notification for cmd.. code 0x%x cid 0x%x\n",
                                                cqe->code, sol->cid));

            iscsi_update_statistics(pcontroller, sol->cid, cqe->code);

            hwi_complete_cmd_error(phwi_controller, 
                                    message_id,
                                    (ACIT_UEXACT16)sol->cid, 
                                    (ACIT_UEXACT16)sol->wrb_index, 
                                    (ACIT_UEXACT16)cqe->code);

            pcontroller->eh_stats.cq_errors[cqe->code]++;

            break;

        case UNSOL_DATA_DIGEST_ERROR_NOTIFY:

            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Digest error on def pdu ring, dropping..\n"));

            // Increment default PDU producer index and post it back
            iscsi_update_statistics(pcontroller, cqe->cid, cqe->code);

            //
            // Bug 24334: This CQE does not consume a DPDU buffer and hence we need
            // not post the DPDU back. Mark the connection so the next DPDU on the
            // data ring need to be discarded because its got a digest error
            //
            iscsi_process_unsol_data_digest_err_notify(pcontroller, cqe->cid);
            
            pcontroller->eh_stats.cq_errors[cqe->code]++;

            break;

        case CXN_KILLED_PDU_SIZE_EXCEEDS_DSL:
        case CXN_KILLED_BURST_LEN_MISMATCH:
        case CXN_KILLED_AHS_RCVD:
        case CXN_KILLED_HDR_DIGEST_ERR:
        case CXN_KILLED_UNKNOWN_HDR:
        case CXN_KILLED_STALE_ITT_TTT_RCVD:
        case CXN_KILLED_INVALID_ITT_TTT_RCVD:
        case CXN_KILLED_FIN_RCVD:
        case CXN_KILLED_BAD_UNSOL_PDU_RCVD:
        case CXN_KILLED_BAD_WRB_INDEX_ERROR:
        case CXN_KILLED_OVER_RUN_RESIDUAL:
        case CXN_KILLED_UNDER_RUN_RESIDUAL:
        case CXN_KILLED_CMND_DATA_NOT_ON_SAME_CONN:

            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("CQ Error 0x%x, resetting CID 0x%x...\n", cqe->code, cqe->cid));
            iscsi_update_statistics(pcontroller, cqe->cid, cqe->code);
            iscsi_rst_send_notify(pcontroller, cqe->cid);
            pcontroller->eh_stats.cq_errors[cqe->code]++;

            break;

        case CXN_KILLED_RST_SENT:        
        case CXN_KILLED_RST_RCVD:
        case CXN_KILLED_TCP_PROT_ERR:

            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("CQ Error 0x%x, reset received/sent on CID 0x%x...\n", cqe->code, cqe->cid));
            iscsi_update_statistics(pcontroller, cqe->cid, cqe->code);
            iscsi_rst_recv_notify(pcontroller, cqe->cid);
            pcontroller->eh_stats.cq_errors[cqe->code]++;

            break;

        default:
            OSM_ASSERT(0);
            pcontroller->partition[message_id].interrupt_stats.unknown_cqes++;
            break;

        }

        // Mark this CQ_ENTRY as invalid
        cqe->valid = FALSE;

        // Circularly increment the consumer index, and move to next CQ_ENTRY
        // to process.
        cqe = (PISCSI_CQE) sa_ring_next (ring);
        
        num_processed++;
        count++;
    }

    // Update on-chip ring index.
    if (num_processed > 0) 
    {
        hwi_ring_cq_db (pcontroller, ring->id, num_processed, 1, 0);
    }

    if(count > 64)
        pcontroller->partition[message_id].interrupt_stats.cqe_gt64++;
    else if (count > 32)
        pcontroller->partition[message_id].interrupt_stats.cqe_gt32++;
    else if (count > 16)
        pcontroller->partition[message_id].interrupt_stats.cqe_gt16++;
    else
        pcontroller->partition[message_id].interrupt_stats.cqe_lt16++;

    DBG_LOG_FUNC_EXIT(&pcontroller->vase_master_block,
                        FUNC_hwi_process_cq,
                        VASE_MODULE_ACIT,
                        0,
                        0,
                        0);
    return;
}

ACIT_VOID
hwi_purge_eq (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           message_id)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    SA_RING *ring;
    EQ_ENTRY *eqe; 

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    ring = &phwi_controller->partition[message_id].eq;
    eqe = (EQ_ENTRY*) sa_ring_current (ring); 

    //
    // Zero out all valid entries
    //

    while (eqe->Valid) {

        //
        // Mark this EQ_ENTRY as invalid
        //
        eqe->Valid = 0;

        //
        // Circularly increment the consumer index, and move to next EQ_ENTRY
        // to process.
        //
        eqe = (EQ_ENTRY*) sa_ring_next (ring);
    }
}


ACIT_VOID
hwi_process_eq (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           message_id)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    SA_RING *ring;
    EQ_ENTRY *eqe; 

    DBG_LOG_FUNC_ENTER(&pcontroller->vase_master_block,
                        FUNC_hwi_process_eq,
                        VASE_MODULE_ACIT,
                        0,
                        0,
                        0);

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    ring = &phwi_controller->partition[message_id].eq;

    eqe = (EQ_ENTRY*) sa_ring_current (ring); 

    ACIT_UPDATE_INT_COUNT(pcontroller, message_id);

    //
    // Process the Event Queue
    //

    DBG_LOG_FUNC_EXEC(&pcontroller->vase_master_block,
                        FUNC_hwi_process_eq,
                        VASE_MODULE_ACIT,
                        0,
                        0,
                        0);

    while (eqe->Valid) {
        switch (eqe->MajorCode) {
            case EQ_MAJOR_CODE_COMPLETION:
                //
                // Determine which CQ to process.
                //
                if (eqe->ResourceID == phwi_controller->partition[message_id].cq.id) {
                    
                    hwi_process_cq (pcontroller, message_id);
                    ACIT_UPDATE_EQ_MAJ_COMP_INT_COUNT(pcontroller, message_id);

                } 
                else
                {
                    be_mcc_process_cq(&phwi_controller->phwic->be_mcc,
                                      ACIT_TRUE);   // rearm
                }
                break;

            default:
                OSM_DEBUG_PRINT (ACIT_DBGMSG_ERROR, ("Unexpected EQ_ENTRY MajorCode: 0x%x\n", eqe->MajorCode));
                pcontroller->partition[message_id].interrupt_stats.unknown_eqes++;
                OSM_ASSERT (0);
        }

        //
        // Mark this EQ_ENTRY as invalid
        //
        eqe->Valid = 0;

        //
        // Circularly increment the consumer index, and move to next EQ_ENTRY
        // to process.
        //
        eqe = (EQ_ENTRY*) sa_ring_next (ring);

        pcontroller->partition[message_id].dic.num_eq_processed++;
    }

    // assert on no EQ entries found
    if(pcontroller->state.interrupt_enabled)
    {
        if(pcontroller->partition[message_id].dic.num_eq_processed == 0)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("No valid EQ entries on message_id %d\n",
                                            message_id));
            // OSM_ASSERT(0);
        }
    }

    DBG_LOG_FUNC_EXIT(&pcontroller->vase_master_block,
                        FUNC_hwi_process_eq,
                        VASE_MODULE_ACIT,
                        0,
                        0,
                        0);
}


ACIT_U32
hwi_process_irq(
    PACIT_CONTROLLER_WS     pcontroller)
{
    CEV_ISR0_CSR       isr;

    DBG_LOG_FUNC_ENTER(&pcontroller->vase_master_block,
                        FUNC_hwi_process_irq,
                        VASE_MODULE_ACIT,
                        pcontroller->counters.pending_ios,
                        0,
                        0);

    //
    // Read interrupt status register to determine source of interrupt.
    // This is a read clear register so we must process all posted interrupts
    // before exiting the interrupt handler.
    // 
    isr.dw = CsrRead(pcontroller, pcontroller->fw_config.isr_offset);

    DBG_LOG_FUNC_EXEC(&pcontroller->vase_master_block,
                        FUNC_hwi_process_irq,
                        VASE_MODULE_ACIT,
                        pcontroller->counters.pending_ios,
                        isr.dw,
                        0);

    if (!isr.dw)
    {
        if(pcontroller->state.interrupt_enabled)
        {
            //
            // This is not our interrupt.
            // 
            pcontroller->partition[0].interrupt_stats.unknown_ints++;
            DBG_LOG_FUNC_EXIT(&pcontroller->vase_master_block,
                                FUNC_hwi_process_irq,
                                VASE_MODULE_ACIT,
                                pcontroller->counters.pending_ios,
                                0,
                                0);

            DBG_LOG_STATISTICS_INCREMENT(&pcontroller->vase_master_block,
                                        InvalidInterrupts);
        }

        return ACIT_STATUS_UNKNOWN_INTERRUPT;      
    }


    hwi_process_eq (pcontroller, 0);

    DBG_LOG_FUNC_EXIT(&pcontroller->vase_master_block,
                        FUNC_hwi_process_irq,
                        VASE_MODULE_ACIT,
                        pcontroller->counters.pending_ios,
                        0,
                        0);
    DBG_LOG_STATISTICS_INCREMENT(&pcontroller->vase_master_block,
                                ValidInterrupts);

    return ACIT_STATUS_SUCCESS;
}

ACIT_U32
hwi_process_msi_irq(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           message_id)
{

    DBG_LOG_FUNC_ENTER(&pcontroller->vase_master_block,
                        FUNC_hwi_process_irq,
                        VASE_MODULE_ACIT,
                        pcontroller->counters.pending_ios,
                        0,
                        0);

    hwi_process_eq (pcontroller, message_id);

    DBG_LOG_FUNC_EXIT(&pcontroller->vase_master_block,
                        FUNC_hwi_process_irq,
                        VASE_MODULE_ACIT,
                        pcontroller->counters.pending_ios,
                        0,
                        0);
    DBG_LOG_STATISTICS_INCREMENT(&pcontroller->vase_master_block,
                                ValidInterrupts);

    return ACIT_STATUS_SUCCESS;
}


ACIT_STATUS
hwi_alloc_eh_resources(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           cri,
    PWRB_HANDLE             *pwrb_handle,
    PSGL_HANDLE             *psgl_handle)
{
    PWRB_HANDLE             pwrb_h = ACIT_NULL;
    PSGL_HANDLE             psgl_h = ACIT_NULL;

    OSM_ASSERT(phwi_controller);
    OSM_ASSERT(pwrb_handle);
    OSM_ASSERT(psgl_handle);

    pwrb_h = hwi_alloc_wrb_handle(phwi_controller, cri);

    if(pwrb_h)
    {
        psgl_h = hwi_alloc_eh_sgl_handle(phwi_controller);
        if(psgl_h) 
        {
            *pwrb_handle = pwrb_h;
            *psgl_handle = psgl_h;
            return ACIT_STATUS_SUCCESS;
        }
        hwi_free_wrb_handle(phwi_controller, cri, pwrb_h);
    }

    OSM_ASSERT(0);
    return ACIT_STATUS_INSUFFICIENT_RESOURCE;
}

ACIT_VOID
hwi_free_eh_resources(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           cri,
    PWRB_HANDLE             pwrb_handle,
    PSGL_HANDLE             psgl_handle)
{
    hwi_free_wrb_handle(phwi_controller, 
                        cri, 
                        pwrb_handle);

    pwrb_handle->type = HWH_TYPE_INVALID;
    pwrb_handle->pio_handle = ACIT_NULL;

    hwi_free_eh_sgl_handle(phwi_controller, psgl_handle);
}


ACIT_STATUS
hwi_alloc_io_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    pio_handle->pwrb_handle = hwi_alloc_wrb_handle(phwi_controller, 
                                                   pio_handle->cri);
    if(pio_handle->pwrb_handle)
    {
        pio_handle->pwrb_handle->type = HWH_TYPE_IO;
        pio_handle->pwrb_handle->pio_handle = pio_handle;

        OSM_ASSERT(pio_handle->psgl_handle);
        pio_handle->itt = (((ACIT_UEXACT32)(pio_handle->pwrb_handle->wrb_index) << 16) |
                            (ACIT_UEXACT32)(pio_handle->psgl_handle->sgl_index));

        return ACIT_STATUS_SUCCESS;
    }

    return ACIT_STATUS_INSUFFICIENT_RESOURCE;
}

ACIT_VOID
hwi_free_io_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    hwi_free_wrb_handle(phwi_controller, 
                        pio_handle->cri, 
                        pio_handle->pwrb_handle);

    pio_handle->pwrb_handle->type = HWH_TYPE_INVALID;
    pio_handle->pwrb_handle->pio_handle = ACIT_NULL;

    pio_handle->pwrb_handle = ACIT_NULL;
}

ACIT_STATUS
hwi_alloc_logout_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LOGOUT_HANDLE     plogout_handle)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    hwi_alloc_eh_resources(phwi_controller, 
                           plogout_handle->cri,
                           &plogout_handle->pwrb_handle,
                           &plogout_handle->psgl_handle);

    plogout_handle->pwrb_handle->type = HWH_TYPE_LOGOUT;
    plogout_handle->pwrb_handle->plogout_handle = plogout_handle;

    plogout_handle->itt = (((ACIT_UEXACT32)(plogout_handle->pwrb_handle->wrb_index) << 16) |
                            (ACIT_UEXACT32)(plogout_handle->psgl_handle->sgl_index));

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
hwi_free_logout_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LOGOUT_HANDLE     plogout_handle)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    hwi_free_eh_resources(phwi_controller,
                          plogout_handle->cri,
                          plogout_handle->pwrb_handle,
                          plogout_handle->psgl_handle);

    plogout_handle->pwrb_handle = ACIT_NULL;
    plogout_handle->psgl_handle = ACIT_NULL;
}


ACIT_STATUS
hwi_alloc_tmf_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    hwi_alloc_eh_resources(phwi_controller, 
                           ptmf_handle->cri,
                           &ptmf_handle->pwrb_handle,
                           &ptmf_handle->psgl_handle);

    ptmf_handle->pwrb_handle->type = HWH_TYPE_TMF;
    ptmf_handle->pwrb_handle->ptmf_handle = ptmf_handle;

    ptmf_handle->itt = (((ACIT_UEXACT32)(ptmf_handle->pwrb_handle->wrb_index) << 16) |
                         (ACIT_UEXACT32)(ptmf_handle->psgl_handle->sgl_index));

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
hwi_free_tmf_resources(
    PACIT_CONTROLLER_WS         pcontroller,
    PACIT_TMF_HANDLE            ptmf_handle)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    hwi_free_eh_resources(phwi_controller,
                         ptmf_handle->cri,
                         ptmf_handle->pwrb_handle,
                         ptmf_handle->psgl_handle);

    ptmf_handle->pwrb_handle = ACIT_NULL;
    ptmf_handle->psgl_handle = ACIT_NULL;
}

ACIT_UEXACT32
hwi_alloc_nop_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_NOP_HANDLE        pnop_handle,
    ACIT_UEXACT32           cri,
    PACIT_NOP_HANDLE        poptional_nop_handle)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    ACIT_UEXACT32           status;

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    //
    // Only if optional handle is present, then allocate
    // an SGL. Otherwise, ITT will be -1 and no SGL will
    // be required since it will complete in drvr_message
    // path
    //
    if(poptional_nop_handle)
    {
        status = hwi_alloc_eh_resources(phwi_controller, 
                                        cri,
                                        &pnop_handle->pwrb_handle,
                                        &pnop_handle->psgl_handle);

        if(status == ACIT_STATUS_SUCCESS)
        {
            pnop_handle->pwrb_handle->type = HWH_TYPE_NOP;
            pnop_handle->pwrb_handle->pnop_handle = pnop_handle;

            pnop_handle->itt = (((ACIT_UEXACT32)(pnop_handle->pwrb_handle->wrb_index) << 16) |
                                 (ACIT_UEXACT32)(pnop_handle->psgl_handle->sgl_index));
        }
        else
        {
            return status;
        }
    }
    else
    {
        pnop_handle->psgl_handle = ACIT_NULL;
        pnop_handle->pwrb_handle = hwi_alloc_wrb_handle(phwi_controller, cri);
        OSM_ASSERT(pnop_handle->pwrb_handle);
        if(pnop_handle->pwrb_handle)
        {
            OSM_ASSERT(phwi_controller->nop_icd);
            pnop_handle->pwrb_handle->type = HWH_TYPE_NOP;
            pnop_handle->pwrb_handle->pnop_handle = pnop_handle;
            pnop_handle->itt = (((ACIT_UEXACT32)(pnop_handle->pwrb_handle->wrb_index) << 16) |
                                 (ACIT_UEXACT32)(phwi_controller->nop_icd));
        }
        else
        {
            return ACIT_STATUS_INSUFFICIENT_RESOURCE;
        }
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
hwi_free_nop_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_NOP_HANDLE        pnop_handle)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    PWRB_HANDLE             pwrb_h;

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    pwrb_h = pnop_handle->pwrb_handle;

    // to handle those cases where a drvr message has arrived yet
    LIST_REMOVE_ELEMENT(&pwrb_h->link);
    LIST_INIT_ELEMENT(&pwrb_h->link);

    if(pnop_handle->psgl_handle)
    {
        hwi_free_eh_resources(phwi_controller,
                              pnop_handle->cri,
                              pnop_handle->pwrb_handle,
                              pnop_handle->psgl_handle);
    }
    else
    {
        hwi_free_wrb_handle(phwi_controller, 
                            pnop_handle->cri, 
                            pnop_handle->pwrb_handle);

        pnop_handle->pwrb_handle->type = HWH_TYPE_INVALID;
        pnop_handle->pwrb_handle->pio_handle = ACIT_NULL;
    }

    pnop_handle->pwrb_handle = ACIT_NULL;
    pnop_handle->psgl_handle = ACIT_NULL;
}

PACIT_IO_HANDLE
hwi_io_handle_from_itt(
    PACIT_CONTROLLER_WS     pcontroller, 
    ACIT_UEXACT32           cri, 
    ACIT_UEXACT32           itt)
{
    PHWI_CONTROLLER_WS      phwi;
    PWRB_HANDLE             pwrb_handle;
    ACIT_UEXACT16           wrb_index = 0, sgl_index = 0;

    phwi = GET_HWI_CONTROLLER_WS(pcontroller);

    wrb_index = (ACIT_UEXACT16)((itt & 0x00ff0000) >> 16);
    sgl_index = (ACIT_UEXACT16)(itt & 0x00000fff);

    if((wrb_index < pcontroller->params.num_cmd_per_cxn) &&
        ((sgl_index >= pcontroller->fw_config.iscsi_icd_start) &&
        (sgl_index < (pcontroller->fw_config.iscsi_icd_start + pcontroller->params.icds_per_controller))))
    {
        OSM_ASSERT(phwi->wrb_context[cri].cid == CID_FROM_CRI(pcontroller, cri));

        pwrb_handle = (PWRB_HANDLE)((PACIT_UEXACT8)phwi->wrb_context[cri].\
                        pwrb_handle_base + wrb_index * sizeof(WRB_HANDLE));

        if(pwrb_handle->type == HWH_TYPE_IO)
        {
            return pwrb_handle->pio_handle;
        }
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("ITT 0x%x - WRB 0x%x SGL 0x%x not found cri = %x\n", 
                                        itt, wrb_index, sgl_index, cri));
    OSM_ASSERT(0);
    return ACIT_NULL;
}


PBE_FUNCTION_OBJECT
hwi_get_be_pfo(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PHWI_CONTROLLER_WS      phwc;
    PHWI_CONTEXT_MEMORY     phwi_context;
    PBE_FUNCTION_OBJECT     pfo;

    phwc = GET_HWI_CONTROLLER_WS(pcontroller);
    ASSERT (phwc);
    phwi_context = phwc->phwic;
    OSM_ASSERT(phwi_context);

    pfo = &phwi_context->be_function;

    return pfo;
}

ACIT_UEXACT32
hwi_fill_mcc_wrb_sgl(
    PACIT_CONTROLLER_WS     pcontroller,
    PMCC_WRB                pwrb,
    PACIT_VOID              ppayload,
    ACIT_UEXACT32           payload_len,
    PACIT_VOID              presponse_hdr,
    ACIT_UEXACT32           response_hdr_len,
    PACIT_VOID              posm_reference)
{
    ACIT_BUS_ADDRESS        pa;
    ACIT_SEXACT32           length, offset, index;

    if(payload_len)
    {
        length = payload_len;
        offset = 0;
        pwrb->sge_count = 0;

        while(length > 0)
        {
            if(pwrb->sge_count > BE_MAX_SGE_IN_MCC_WRB)
            {
                return ACIT_STATUS_ERROR;
            }

            pa = OSM_GET_BUS_ADDRESS(pcontroller->pcontroller_key,
                                     posm_reference,
                                     (PACIT_UEXACT8)ppayload + offset,
                                     (PACIT_UEXACT32)&length);

            index = pwrb->sge_count + 1;

            OSM_ASSERT(length && pa.u.a64.address);

            ACIT_PUT_LE32(pwrb->payload.sgl[index].pa_lo,
                          pa.u.a32.address_lo);
            ACIT_PUT_LE32(pwrb->payload.sgl[index].pa_hi,
                          pa.u.a32.address_hi);

            if(length <= (ACIT_SEXACT32)payload_len)
            {
                ACIT_PUT_LE32(pwrb->payload.sgl[index].length,
                              length);
            }
            else
            {
                ACIT_PUT_LE32(pwrb->payload.sgl[index].length,
                              payload_len);
            }

            offset += length;
            length = payload_len - offset;
            pwrb->sge_count++;

            if(length <= 0)
                break;
        }
    }

    if(response_hdr_len)
    {
        length = response_hdr_len;
        pa = OSM_GET_BUS_ADDRESS(pcontroller->pcontroller_key,
                                    posm_reference,
                                    (PACIT_UEXACT8)presponse_hdr,
                                    (PACIT_UEXACT32)&length);

        OSM_ASSERT(length && pa.u.a64.address);

        ACIT_PUT_LE32(pwrb->payload.sgl[0].pa_lo,
                        pa.u.a32.address_lo);
        ACIT_PUT_LE32(pwrb->payload.sgl[0].pa_hi,
                        pa.u.a32.address_hi);

        if(length <= (ACIT_SEXACT32)response_hdr_len)
        {
            ACIT_PUT_LE32(pwrb->payload.sgl[0].length,
                          length);
        }
        else
        {
            ACIT_PUT_LE32(pwrb->payload.sgl[0].length,
                          response_hdr_len);
        }

        pwrb->sge_count++;
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_STATUS
hwi_fill_non_embedded_wrb(
    PACIT_CONTROLLER_WS     pcontroller,
    PMCC_WRB                pwrb,
    PACIT_VOID              ppayload,
    ACIT_UEXACT32           payload_len,
    PACIT_VOID              presponse_hdr,
    ACIT_UEXACT32           response_hdr_len,
    PACIT_VOID              posm_reference)
{
    ACIT_STATUS             status;

    /*
     * For non embedded WRBs, the payload length should also contain
     * the IOCTL header length
     */
    ACIT_PUT_LE32(pwrb->payload_length, payload_len + response_hdr_len);

    /*
     * Construct a SGL for the payload and fill them into
     * MCC WRB
     */

    pwrb->embedded = 0;

    status = hwi_fill_mcc_wrb_sgl(pcontroller,
                                    pwrb,
                                    ppayload,
                                    payload_len,
                                    presponse_hdr,
                                    response_hdr_len,
                                    posm_reference);

    return status;
}
ACIT_STATUS
hwi_fill_embedded_wrb(
    PACIT_CONTROLLER_WS     pcontroller,
    PMCC_WRB                pwrb,
    PACIT_VOID              ppayload,
    ACIT_UEXACT32           payload_len)
{

    /*
     * Directly copy the payload into the MCC WRB. This is an embedded
     * MCC WRB since there's no response header. The CQ entry is good
     * enough to return the status
     */

    pwrb->embedded = 1;

    return ACIT_STATUS_SUCCESS;
}

#if 0
// STORPORT_DUPLEX not implemented

ACIT_UEXACT32
hwi_check_wrb_index(
    PACIT_CONTROLLER_WS     pcontroller, 
    ACIT_UEXACT16           cri, 
    ACIT_UEXACT16           wrb_index)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    PISCSI_CQE              pcqe;
    SA_RING                 *ring;
    ACIT_U32                entries, found = 0;
    ACIT_UEXACT32           cid = 0;
 
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    ring = &phwi_controller->cq;
    pcqe = (PISCSI_CQE) ring->va;
    cid = CID_FROM_CRI(pcontroller, cri);

    for(entries = 0; entries < ring->num; entries++)
    {
        if(pcqe->valid)
        {
            // compare CID and WRB index
            if(cid == pcqe->cid)
            {
                PI_SOL_CQE psol = (PI_SOL_CQE)pcqe;

                if((psol->code == SOL_CMD_COMPLETE) &&
                  (wrb_index == psol->wrb_index))
                {
                    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("Found at VA 0x%p - CQ entry:", pcqe));
                    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("0x%x 0x%x 0x%x 0x%x\n", pcqe->dw[0], pcqe->dw[1], pcqe->dw[2], pcqe->dw[3]));
                    found = 1;
                }
            }
        }
        pcqe++;
    }

    return found;
}

ACIT_UEXACT32
hwi_check_pending_handle(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cid,
    ACIT_UEXACT32           wrb_index)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    PWRB_HANDLE             pwrb_handle;
    ACIT_UEXACT32           found;

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    pwrb_handle = (PWRB_HANDLE)(phwi_controller->wrb_context[CRI_FROM_CID(pcontroller, cid)].pwrb_handle_base) +
                                wrb_index;

    if(pwrb_handle->type != HWH_TYPE_INVALID)
        found = 1;
    else
        found = 0;

    return found;
}

ACIT_VOID
hwi_check_stale_cq(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    PISCSI_CQE              pcqe;
    SA_RING                 *ring;
    ACIT_UEXACT32           entries, found = 0, stales = 0;
    PI_SOL_CQE              psol;

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    ring = &phwi_controller->cq;
    pcqe = (PISCSI_CQE) ring->va;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Scanning CQEs for stale entries\n"));

    for(entries = 0; entries < ring->num; entries++)
    {
        if(pcqe->valid && pcqe->code == SOL_CMD_COMPLETE)
        {
            psol = (PI_SOL_CQE)pcqe;

            found = hwi_check_pending_handle(pcontroller, 
                                             psol->cid, 
                                             psol->wrb_index);

            if(!found)
            {
                stales++;
                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("!!!!!Entry at VA 0x%p stale!! - CQ entry:\n", pcqe));
                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("0x%x 0x%x 0x%x 0x%x\n", pcqe->dw[0], pcqe->dw[1], pcqe->dw[2], pcqe->dw[3]));

            }
        }
        pcqe++;
    }

    if(!stales)
	{
       	 OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("No stale entries found in CQ\n"));
	}
    else
	{
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("%d stale entries found in CQ\n", stales));
        OSM_ASSERT(0);
	}
}

ACIT_VOID
hwi_validate_pending_ios(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget_to_search,
    PACIT_LUN_WS            plun_to_search)
{
    ACIT_UEXACT32           target, found = 0;
    PACIT_TARGET_WS         ptarget;
    PACIT_DC_LIST           flink, pio_link;
    PACIT_IO_HANDLE         pio_handle;
    PACIT_TARGET_ENTRY      ptarget_entry = ACIT_NULL;
    PACIT_IO_HANDLE         pioh_oldest_t, pioh_oldest_c = ACIT_NULL;
    ACIT_UEXACT32           num_ios_t = 0, num_ios_c = 0, isread;
    ACIT_UEXACT32           num_r_ios_t = 0, num_r_ios_c = 0, num_w_ios_t = 0, num_w_ios_c = 0;

    //
    // First scan the pending list for each target and compare
    // the CQ to see if there is an entry
    //
    ptarget_entry = &pcontroller->ptarget_table->target_entry[0];

    for(target = 0; target < pcontroller->ptarget_table->occupied_entries; target++)
    {
        ptarget = GET_TARGET_WS_SAFE(ptarget_entry[target].ptarget_key);

        if(!ptarget)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Ignoring deleted or not present target..\n"));
            continue;
        }

        if(ptarget_to_search && (ptarget_to_search != ptarget))
            continue;

        pioh_oldest_t = ACIT_NULL;
        num_ios_t = 0;
        num_r_ios_t = 0;
        num_w_ios_t = 0;

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Scanning target %d ws 0x%p...\n", target, ptarget));

        flink = ptarget->io_handle_busy_list.next;
        pio_link = ptarget->io_handle_busy_list.next;

        while(flink != &ptarget->io_handle_busy_list)
        {
            pio_handle = STRUCT_BASE(ACIT_IO_HANDLE,
                                     target_link,
                                     pio_link);

            OSM_ASSERT(pio_handle);

            flink = flink->next;
            pio_link = pio_link->next;

            if(plun_to_search && (plun_to_search != pio_handle->plun))
                continue;

            if(pio_handle->pio->pio_request->flags.inbound_data)
            {
                isread = 1;
                num_r_ios_t++;
            }
            else
            {
                isread = 0;
                num_w_ios_t++;
            }

            found = hwi_check_wrb_index(pcontroller, 
                                        pio_handle->cri, 
                                        (pio_handle->itt >> 16) & 0xffff);

            if(!found)
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,(("Missing CEV on IO Handle 0x%p Cid 0x%x CmdSN 0x%x ITT 0x%x Read=%d TS 0x%x\n"), 
                                                            (PACIT_UEXACT32)pio_handle, 
                                                            CID_FROM_CRI(pcontroller, pio_handle->cri),
                                                            pio_handle->cmd_sn,
                                                            pio_handle->itt,
                                                            isread,
                                                            pio_handle->starting_tick_count));
            }

            if(!pioh_oldest_t)
                pioh_oldest_t = pio_handle;
            else 
                if(pio_handle->starting_tick_count < pioh_oldest_t->starting_tick_count)
                    pioh_oldest_t = pio_handle;

            num_ios_t++;
        }

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Summary for target %d\n", target));
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Pending IOs %d Reads %d Writes %d\n", num_ios_t, num_r_ios_t, num_w_ios_t));
        if(num_ios_t)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Oldest IOH 0x%p TS 0x%x\n", pioh_oldest_t, pioh_oldest_t->starting_tick_count));
        }

        if(num_w_ios_t > 32)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("!!DRIVER ERROR - %d WRITES OUTSTANDING!!\n",num_w_ios_t));
        }

        if(!pioh_oldest_c)
            pioh_oldest_c = pioh_oldest_t;
        else 
            if((pioh_oldest_t) && 
               (pioh_oldest_t->starting_tick_count < pioh_oldest_c->starting_tick_count))
                    pioh_oldest_c = pioh_oldest_t;

        num_ios_c += num_ios_t;
        num_r_ios_c += num_r_ios_t;
        num_w_ios_c += num_w_ios_t;
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Summary for controller\n"));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Pending IOs %d Reads %d Writes %d\n", num_ios_c, num_r_ios_c, num_w_ios_c));
    if(num_ios_c)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, (("Oldest IOH 0x%p TS 0x%x Current TS 0x%x\n"), 
                                            pioh_oldest_c, 
                                            pioh_oldest_c->starting_tick_count,
                                            pcontroller->counters.current_tick_count));
    }

    //
    // Then do the reverse. Go through the CQ first and try and 
    // account for each entry in controller's list. Not sure how useful 
    // it is, but may catch some stale CQ entries
    //
    hwi_check_stale_cq(pcontroller);

}
#endif

ACIT_VOID
hwi_get_pci_link_info(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PCICFG_PCIE_LINK_STATUS_CSR pcie_link_status;
    PCICFG_PCIE_LINK_CAP_CSR pcie_link_cap;

    pcie_link_status.dw = ACIT_PCICFG0_READ(pcontroller, pcie_link_status);
    pcie_link_cap.dw = ACIT_PCICFG0_READ(pcontroller, pcie_link_cap);

    pcontroller->pcie_info.current_link_width = (ACIT_UEXACT8)pcie_link_status.NegLinkWidth;
    pcontroller->pcie_info.max_link_width = (ACIT_UEXACT8)pcie_link_cap.MaxLinkWidth;

    pcontroller->pcie_info.current_link_speed = (ACIT_UEXACT8)pcie_link_status.LinkSpeed;
    pcontroller->pcie_info.max_link_speed =  (ACIT_UEXACT8)pcie_link_cap.MaxLinkSpeed;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, (("LinkWidth - Current/Max=%d/%d LinkSpeed - Current/Max=%d/%d\n"), 
                                            pcontroller->pcie_info.current_link_width, 
                                            pcontroller->pcie_info.max_link_width,
                                            pcontroller->pcie_info.current_link_speed,
                                            pcontroller->pcie_info.max_link_speed));

    return;
}

