/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*!
@file
   mpu_ll.c

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/drivers/beclib/mpu_ll.c,v 1.36 2006/04/17 01:35:06 menz Exp $
*/
#include "pch.h"

//============================================================================
//                            F I L E   S C O P E
//============================================================================

static i32 be_mpu_post_timeout = 60; /* 60 secs */
static i32 be_mpu_poll_timeout = 120;

BESTATUS
be_function_check_hardware_ue (
    BE_FUNCTION_OBJECT *function_object);

void
_be_mcc_free_wrb_buffers (
    PBE_MCC_OBJECT      mcc,
    PBE_MCC_WRB_CONTEXT wrb_context);

void
be_function_disable_ioctls (
    IN PBE_FUNCTION_OBJECT function_object);

/*!

@brief
    This routine converts from a BladeEngine Mcc wrb completion status to
    a standard NT status;    
    
@param    
    MccStatus   - BE Mcc WRB completion status to convert

@return   
    An NT status
                                                         
@note
    IRQL < DISPATCH_LEVEL    

*/
STATIC
BESTATUS
be_mcc_wrb_status_to_bestatus(
    IN u32 mcc_status )
{
    // TODO -- Fix this if we want better status codes.
    switch ( mcc_status ) {
        case MGMT_STATUS_SUCCESS:                    return  SA_SUCCESS;        
        default:
            return BE_NOT_OK;
    }
}

/*!
@brief
    This routine waits for a previously posted mailbox WRB to be completed.
    Specifically it waits for the mailbox to say that it's ready to accept more data
    by setting the LSB of the mailbox pd register to 1.    
    
@param    
    pcontroller      - The function object to post this data to

@return   

@note
    IRQL < DISPATCH_LEVEL    
*/
STATIC
BESTATUS
be_mcc_mailbox_wait(
    IN      PBE_FUNCTION_OBJECT              function_object)
{
    MPU_MAILBOX_DB  mailbox_db;
    u32 i = 0;
    u32 tl;
    u32 mbox_corruption_DBG = 0, mbox_corruption_toggle_DBG = 0;    // for debug/workaround of 8673
    BESTATUS status = BE_SUCCESS;

    const u32 us_per_loop = 500;                             
    const u32 print_frequency_loops = 1000000 / us_per_loop; // 1 second print intervals
    u32 max_loops;
    u32 loops = 0;

    // be_mpu_poll_timeout is in seconds.
    // The print frequency loops is the number of loops per 1 second.
    max_loops = be_mpu_poll_timeout * print_frequency_loops; 

    FUNCTION_ASSERT (function_object);

    if (function_object->emulate) {
        // No waiting for mailbox in emulated mode.
        return BE_SUCCESS;
    }

    /* mailbox_db.dw = PD_READ( function_object, mpu_mailbox_db );*/
    BeSetDword (&mailbox_db, PD_READ(function_object, mpu_mailbox_db));

    tl = sa_trace_set_level (sa_trace_get_level () & ~DL_HW);

    while (((mailbox_db.ready == FALSE) || ((*(u32 *)&mailbox_db) == 0xFFFFFFFF)) &&
           loops++ < max_loops)
    {   
        // Async UE detected.
        if (function_object->ue) {
            status = BE_STATUS_UE_DETECTED;
            break;
        }

        if ((*(u32 *)&mailbox_db) == 0xFFFFFFFF)
        {   // keep counting how many successive times in a row we see the problem
            mbox_corruption_DBG++;
        }
        else if (mbox_corruption_DBG)
        {   // keep counting how many times we toggle between corrupt and non-corrupt
            mbox_corruption_toggle_DBG++;
        }

        if ((++i % print_frequency_loops) == 0 ) {

            TRACE (DL_WARN, 
                   "Waiting for mailbox ready for %dk polls (corruption_cnt = %d, toggle_cnt = %d)", 
                   i/1000, 
                   mbox_corruption_DBG, 
                   mbox_corruption_toggle_DBG);

            // Actively check for a UE.
            //if (be_function_detect_ue (function_object)) {
            //    break;
            //}
        }

        sa_dev_stall (function_object->sa_dev, us_per_loop); 

        BeSetDword (&mailbox_db, PD_READ(function_object, mpu_mailbox_db ));
    }

    //
    // Check if a timeout occurred.
    //
    if (loops >= max_loops) {        

        TRACE (DL_ERR, "Timeout waiting for mailbox. function:%d",                
               function_object->pci_function_number);

        be_function_disable_ioctls (function_object);

        //if (function_object->links.errobj) {
        //    be_ue_generate_software_ue (function_object->links.errobj, UE_SOFTWARE_MAILBOX_TIMEOUT);
        //}
        
        status = BE_STATUS_MAILBOX_TIMEOUT;
    }

    sa_trace_set_level (tl);

    return status;

}


/*!

@brief
    This routine tells the MCC mailbox that there is data to processed
    in the mailbox. It does this by setting the physical address for the
    mailbox location and clearing the LSB.  This routine returns immediately
    and does not wait for the WRB to be processed.
    
    
@param    
    pcontroller      - The function object to post this data to

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
STATIC
BESTATUS
be_mcc_mailbox_notify (
    IN PBE_FUNCTION_OBJECT function_object)
{        
    MPU_MAILBOX_DB mailbox_db;
    BESTATUS status;

    FUNCTION_ASSERT (function_object);    
    ASSERT (function_object->mailbox.pa);
    ASSERT (function_object->mailbox.va);

    // If emulated, do not ring the mailbox
    if (function_object->emulate) {
        TRACE (DL_WARN, "MPU disabled. Skipping mailbox notify.");
        return BE_SUCCESS;
    }

    if (function_object->ue) {
        // No waiting for mailbox after a UE.
        return BE_STATUS_UE_DETECTED;
    }

    function_object->stats.mailbox_wrbs++;

    // form the higher bits in the address
    // init
/*    mailbox_db.dw = 0; */
    BeSetDword(&mailbox_db, 0);

    mailbox_db.hi = 1;
    mailbox_db.ready = 0;
    SA_C_ASSERT (BE_BIT_SIZE (MPU_MAILBOX_DB, address) == 30);
    mailbox_db.address = (u32)sa_upper_bits64 (function_object->mailbox.pa, 30);

    // Wait for the MPU to be ready
    status = be_mcc_mailbox_wait (function_object);
    if (status != BE_SUCCESS) {
        return status;
    }

    // Ring doorbell 1st time
    PD_WRITE (function_object, mpu_mailbox_db, mailbox_db);

    // Wait for 1st write to be acknowledged.
    be_mcc_mailbox_wait( function_object );

    // form the lower bits in the address
    mailbox_db.hi = 0;
    mailbox_db.ready = 0;
    mailbox_db.address = (u32)sa_bit_range64 (function_object->mailbox.pa, 4, 30);

    // Ring doorbell 2nd time
    PD_WRITE (function_object, mpu_mailbox_db, mailbox_db);

    return BE_SUCCESS;
}


/*!
@brief
    This routine tells the MCC mailbox that there is data to processed
    in the mailbox. It does this by setting the physical address for the
    mailbox location and clearing the LSB.  This routine spins until the
    MPU writes a 1 into the LSB indicating that the data has been received
    and is ready to be processed.

@param    
    pcontroller      - The function object to post this data to
@return   
@note
    IRQL < DISPATCH_LEVEL    
*/
STATIC
BESTATUS
be_mcc_mailbox_notify_and_wait(
    IN      PBE_FUNCTION_OBJECT              function_object)
{
    BESTATUS status;
    
    if (function_object->ue) {
        // No waiting for mailbox after a UE.
        return BE_STATUS_UE_DETECTED;
    }

    //
    // Notify it
    //
    status = be_mcc_mailbox_notify( function_object );
    if (status != BE_SUCCESS) {
        return status;
    }

    //
    // Now wait for completion of WRB
    //
    status = be_mcc_mailbox_wait( function_object );
    return status;
}


void
be_mcc_process_cqe (PBE_FUNCTION_OBJECT function_object,
                    PMCC_CQ_ENTRY cqe)
{
    
    PBE_MCC_WRB_CONTEXT wrb_context = NULL;
    
    FUNCTION_ASSERT (function_object);
    ASSERT (cqe);

    // A command completed.  Commands complete out-of-order.  Determine which
    // command completed from the TAG.
    SA_C_ASSERT (sizeof (cqe->mcc_tag) >= sizeof (u64));
    wrb_context = (PBE_MCC_WRB_CONTEXT) SA_U64_TO_PTR ((u64)cqe->mcc_tag[0] | ((u64)cqe->mcc_tag[1] << 32));

    ASSERT (wrb_context);    
    
    if (!wrb_context) {
        TRACE (DL_ERR, "NULL WRB context during completion.");
        // TODO - ASSERT UE
        return;
    }

    // Try to detect a corrupted tag, at least for debug builds.
    ASSERT ((wrb_context == &function_object->mailbox.default_context)
             ||
            (wrb_context >= function_object->links.mcc->wrb_context.base &&
             wrb_context < function_object->links.mcc->wrb_context.base + function_object->links.mcc->wrb_context.num));
    
    if (FALSE == 
        ((wrb_context == &function_object->mailbox.default_context)
          ||
         (wrb_context >= function_object->links.mcc->wrb_context.base &&
          wrb_context < function_object->links.mcc->wrb_context.base + function_object->links.mcc->wrb_context.num))
       )
    {
        TRACE (DL_ERR, "Illegal WRB context during completion.");
        // TODO - ASSERT UE
        return;
    }
    
    #ifdef SA_DEBUG
        // Assert that at least one WRB was consumed since we posted this one.
        // If this assert fires, it indicates that a WRB was completed before it was consumed.
        if (wrb_context->ring_wrb && !function_object->ue) {
            ASSERT (C_GT (function_object->stats.consumed_wrbs, wrb_context->consumed_count));
        }
    #endif

    // Perform a response copy if requested.
    // Only copy data if the IOCTL is successful.
    if (cqe->completion_status == MGMT_STATUS_SUCCESS && 
        wrb_context->copy.length > 0) {
        ASSERT (wrb_context->wrb);
        ASSERT (wrb_context->wrb->embedded);
        ASSERT (wrb_context->copy.va);
        ASSERT (wrb_context->copy.ioctl_offset + wrb_context->copy.length <= SA_SIZEOF_FIELD (MCC_WRB, payload));
        sa_memcpy (wrb_context->copy.va, 
                   (u8*)&wrb_context->wrb->payload + wrb_context->copy.ioctl_offset, 
                   wrb_context->copy.length);
    }


    #if defined(SA_DEBUG)
        TRACE (DL_IOCTL,        
                "Complete IOCTL. emul:? embed:%d sub:0x%x op:0x%x ctx:%p (%s/%s) status:%s(0x%x)",                
                wrb_context->wrb && wrb_context->wrb->embedded,
                wrb_context->subsystem,                
                wrb_context->opcode,
                wrb_context,
                ioctl_subsystems_get_name (wrb_context->subsystem), 
                be_query_opcode_name (wrb_context->subsystem, wrb_context->opcode),
                mgmt_error_codes_get_name (cqe->completion_status),
                cqe->completion_status       
                );
    #endif


    // internal callback
    if (wrb_context->internal_callback) {
        wrb_context->internal_callback (wrb_context->internal_callback_context, 
                                        be_mcc_wrb_status_to_bestatus (cqe->completion_status), 
                                        wrb_context->wrb);
    }


    // callback
    if (wrb_context->callback) {
        wrb_context->callback (wrb_context->callback_context, 
                               be_mcc_wrb_status_to_bestatus (cqe->completion_status), 
                               wrb_context->wrb);
    }

    // Free the context structure
    _be_mcc_free_wrb_context (function_object, wrb_context);
}

boolean
be_drive_mcc_wrb_queue ( 
    IN      PBE_MCC_OBJECT mcc)
{
    PBE_FUNCTION_OBJECT function_object = NULL;
    BESTATUS status = BE_PENDING;
    PBE_GENERIC_QUEUE_CONTEXT queue_context;
    PSA_LIST_ENTRY entry;
    MCC_WRB *wrb;
    MCC_WRB *queue_wrb;
    u32 length;    

    MCC_ASSERT (mcc);

    function_object = mcc->parent_function;   

    do {
        
        be_lock_wrb_post (function_object);    

        if (mcc->driving_backlog) {
            be_unlock_wrb_post_no_redrive (function_object);
            return FALSE;
        }

        function_object->pend_queue_driving = 0;

        // Acquire the flag to limit 1 thread to redrive posts.
        mcc->driving_backlog = 1;

        while (!sa_is_list_empty (&mcc->backlog)) {

            wrb = _be_mpu_peek_ring_wrb (mcc, TRUE); // Driving the queue
            if (!wrb) {
                break; // No space in the ring yet.
            }

            // Get the next queued entry to process.
            entry = sa_remove_head_list (&mcc->backlog);
            ASSERT (entry);
            queue_context = SA_CONTAINING_RECORD (entry, BE_GENERIC_QUEUE_CONTEXT, context.list);
            function_object->links.mcc->backlog_length--;

            // Compute the required length of the WRB.  Since the queue element may be smaller than
            // the complete WRB, copy only the required number of bytes.
            queue_wrb = (MCC_WRB*)&queue_context->wrb_header;
            if (queue_wrb->embedded) {
                length = sizeof (BE_MCC_WRB_HEADER) + queue_wrb->payload_length;
            } else {
                ASSERT (queue_wrb->sge_count == 1); // Right now we only support 1 fragment.
                length = sizeof (BE_MCC_WRB_HEADER) + queue_wrb->sge_count * sizeof (MCC_SGE);
            }

            // Truncate the length based on the size of the queue element.  Some elements that have output parameters
            // can be smaller than the payload_length field would indicate.  We really only need to copy the request 
            // parameters, not the response.
            ASSERT (queue_context->context.bytes >= SA_FIELD_OFFSET (BE_GENERIC_QUEUE_CONTEXT, wrb_header));
            length = MIN (length, queue_context->context.bytes - SA_FIELD_OFFSET (BE_GENERIC_QUEUE_CONTEXT, wrb_header));         
                    
            // Copy the queue element WRB into the ring.
            sa_memcpy (wrb, &queue_context->wrb_header, length);
            
            // Post the wrb.  This should not fail assuming we have enough context structs.
            status = be_function_post_mcc_wrb_complete (
                        function_object,
                        wrb,
                        NULL, // Do not queue
                        queue_context->context.callback,
                        queue_context->context.callback_context,
                        queue_context->context.internal_callback,
                        queue_context->context.internal_callback_context,
                        queue_context->context.optional_ioctl_va,
                        queue_context->context.copy);        

            if (status != BE_PENDING) {

                // Synchronous completion. Since it was queued, we will invoke the callback.
                // To the user, this is an asynchronous request.
                be_unlock_wrb_post_no_redrive (function_object);
                
                ASSERT (queue_context->context.callback);

                queue_context->context.callback (
                    queue_context->context.callback_context,
                    status,
                    NULL);
                
                be_lock_wrb_post (function_object);

            } 
        }

        function_object->stats.queue_length = function_object->links.mcc->backlog_length;

        // Free the flag to limit 1 thread to redrive posts.
        mcc->driving_backlog = 0;

        be_unlock_wrb_post_no_redrive (function_object);

        // Redrive again in the rare case that everything is completed by another
        // CQ thread before we release the driving_backlog flag.  This indicates that the 
        // CQ thread did not get a chance to drive the backlog, so we should continue looping.
        // (Note that status == SUCCESS will indicate that we are polling and everything will
        // complete in order, so there is no need to loop.  And status == failure indicates
        // another major problem, so we bail out to avoid spinning forever.
    } while (status == BE_PENDING && 
             function_object->pend_queue_driving &&
             sa_ring_num_pending (&mcc->sq.ring) == 0 && 
             !sa_is_list_empty (&mcc->backlog));

    return FALSE;
}

// This function asserts that the WRB was consumed in order.
#ifdef SA_DEBUG
u32
be_mcc_wrb_consumed_in_order (
    PBE_MCC_OBJECT mcc,
    PMCC_CQ_ENTRY cqe)
{
    PBE_FUNCTION_OBJECT function_object = mcc->parent_function;
    PBE_MCC_WRB_CONTEXT wrb_context = NULL;
    u32 wrb_index;
    u32 wrb_consumed_in_order;    
    
    FUNCTION_ASSERT (function_object);
    ASSERT (cqe);   

    // A command completed.  Commands complete out-of-order.  Determine which
    // command completed from the TAG.
    SA_C_ASSERT (sizeof (cqe->mcc_tag) >= sizeof (u64));
    wrb_context = (PBE_MCC_WRB_CONTEXT) SA_U64_TO_PTR ((u64)cqe->mcc_tag[0] | ((u64)cqe->mcc_tag[1] << 32));
    ASSERT (wrb_context);    

    // Try to detect a corrupted tag, at least for debug builds.
    ASSERT ((wrb_context == &function_object->mailbox.default_context)
             ||
            (wrb_context >= function_object->links.mcc->wrb_context.base &&
             wrb_context < function_object->links.mcc->wrb_context.base + function_object->links.mcc->wrb_context.num));

    wrb_index = (u32)(((SA_PTR_TO_U64(wrb_context->ring_wrb) - SA_PTR_TO_U64 (mcc->sq.ring.va))) / sizeof (MCC_WRB));
    ASSERT (wrb_index < sa_ring_num (&mcc->sq.ring));

    wrb_consumed_in_order = (u32)(wrb_index == mcc->consumed_index);    
    mcc->consumed_index = sa_addc (mcc->consumed_index, 1, sa_ring_num (&mcc->sq.ring));

    return wrb_consumed_in_order;
}
#endif


BESTATUS
be_mcc_process_cq ( 
    IN      PBE_MCC_OBJECT mcc,
    IN      boolean        rearm)
{
    PBE_FUNCTION_OBJECT function_object = NULL;
    PMCC_CQ_ENTRY cqe;
    CQ_DB db;
    PSA_RING cq_ring = &mcc->cq.ring;
    PSA_RING mcc_ring = &mcc->sq.ring;
    u32 num_processed = 0;
    u32 consumed = 0;

    function_object = mcc->parent_function;
    FUNCTION_ASSERT (function_object);

    be_lock_cq_process (function_object);

    function_object->stats.processed_cq++;
    
    // Verify that only one thread is processing the CQ at once.  We cannot
    // hold the lock while processing the CQ due to the callbacks into the OS.
    // Therefore, this flag is used to control it.  If any of the threads want to 
    // rearm the CQ, we need to honor that.
    if (mcc->processing != 0) {        
        mcc->rearm = mcc->rearm || rearm;
        goto Error;
    } else {
        mcc->processing = 1; // lock processing for this thread.
        mcc->rearm = rearm;  // set our rearm setting
    }

    be_unlock_cq_process (function_object);


    cqe = sa_ring_current (cq_ring);

    while (cqe->valid) {

        function_object->stats.cq_entries++;

        if (num_processed >= 8) { 
            // coalesce doorbells, but free space in cq ring while processing. 
            // clear
/*            db.dw = 0;      */
            BeSetDword(&db, 0);
            db.qid          = cq_ring->id;
            db.rearm        = FALSE;    // Rearm at the end!
            db.event        = FALSE;
            db.num_popped   = num_processed ;
            num_processed = 0;

            PD_WRITE( function_object, cq_db, db );
        }
        

        if (cqe->async_event) {
            // This is an asynchronous event.                        
            ASYNC_EVENT_TRAILER *async_trailer = (ASYNC_EVENT_TRAILER *)
                ((u8*)cqe + sizeof (MCC_CQ_ENTRY) - sizeof(ASYNC_EVENT_TRAILER));

            // The async_event bit must be in the same location in both structs.
            SA_C_ASSERT (BE_WORD_OFFSET (MCC_CQ_ENTRY, async_event) * 32 + BE_BIT_OFFSET (MCC_CQ_ENTRY, async_event) == 
                         BE_BIT_OFFSET (ASYNC_EVENT_TRAILER, async_event) + 8*(sizeof (MCC_CQ_ENTRY) - sizeof(ASYNC_EVENT_TRAILER)));
            
            ASSERT (async_trailer->async_event == 1);
            ASSERT (async_trailer->valid == 1);

            function_object->stats.async_events++;
            
            // Call the async event handler if it is installed.
            if (mcc->async_callback) {
                TRACE (DL_INFO, "Async event. Type:0x%x Code:0x%x", async_trailer->event_type, async_trailer->event_code);
                mcc->async_callback (mcc->async_context, (u32)async_trailer->event_code, (PVOID)cqe);
            } else {
                TRACE (DL_WARN, "No async callback for event. Type:0x%x Code:0x%x", async_trailer->event_type, async_trailer->event_code);
                function_object->stats.ignored_async_events++;
            }            

        } else {
            // This is a completion entry.
            if (cqe->consumed) {
                // A command on the MCC ring was consumed. Update the consumer index.
                // These occur in order.
                ASSERT (be_mcc_wrb_consumed_in_order (mcc, cqe));
                consumed++;                
                function_object->stats.consumed_wrbs++;
            }

            if (cqe->completed) {
                // A command completed.  Use tag to determine which command.                
                be_mcc_process_cqe (function_object, cqe);
                function_object->stats.completed_wrbs++;
            }                        
        }

        // Reset the CQE
        cqe->valid  = FALSE;
        num_processed++;

        TRACE (DL_IOCTL, " update the cidx now cidx=%x", cq_ring->cidx );
        // Update our tracking for the CQ ring.
        cqe = sa_ring_next (cq_ring);

    }


    TRACE (DL_INFO, "num_processed:0x%x, and consumed:0x%x", num_processed, consumed);
    // Grab the CQ lock to synchronize the "rearm" setting for the doorbell, and
    // for clearing the "processing" flag.
    be_lock_cq_process (function_object);    

    // Rearm the cq.  This is done based on the global mcc->rearm flag which combines
    // the rearm parameter from the current call to process_cq and any other threads
    // that tried to process the CQ while this one was active.  This handles the situation
    // where a sync. ioctl was processing the CQ while the interrupt/dpc tries to process it.
    // The sync process gets to continue -- but it is now responsible for the rearming.
    if (num_processed > 0 || mcc->rearm == TRUE) {
        // clear
/*      db.dw = 0; */
        BeSetDword(&db, 0);
        db.qid          = cq_ring->id;
        db.rearm        = mcc->rearm;
        db.event        = FALSE;
        db.num_popped   = num_processed;

        PD_WRITE( function_object, cq_db, db );
    }

    // Update the consumer index after ringing the CQ doorbell.
    // We don't want another thread to post more WRBs before we have CQ space available.
    sa_ring_consume_multiple (mcc_ring, consumed);

    // Clear the processing flag.
    mcc->processing = 0;

Error:

    be_unlock_cq_process (function_object);

    // Use the local variable to detect if the current thread holds the WRB post lock.
    // If rearm is FALSE, this is either a synchronous command, or the upper layer driver is polling
    // from a thread.  We do not drive the queue from that context since the driver may hold the 
    // wrb post lock already.
    if (rearm) {        
        be_drive_mcc_wrb_queue (mcc);        
    } else {        
        function_object->pend_queue_driving = num_processed > 0 ? 1 : 0;
    }

    return BE_SUCCESS;
}

//============================================================================
//                  P U B L I C  R O U T I N E S
//============================================================================

/*!

@brief
    This routine creates an MCC object.  This object contains an MCC send queue
    and a CQ private to the MCC.

@param    
    pcontroller      - Handle to a function object                     

@param        
    EqObject            - EQ object that will be used to dispatch this MCC
    
@param        
    ppMccObject         - Pointer to an internal Mcc Object returned.                    

@return   
    BE_SUCCESS if successfull,, otherwise a useful error code is returned.
    
@note
    IRQL < DISPATCH_LEVEL    

*/

BESTATUS 
be_mcc_ring_create_ext (
    IN PBE_FUNCTION_OBJECT function_object,
    IN PSA_SGL             sgl,
    IN u32                 length,
    IN PBE_MCC_WRB_CONTEXT context_array,
    IN u32                 num_context_entries,
    IN PBE_CQ_OBJECT       cq,
    IN PBE_MCC_OBJECT      mcc,
    IN u32				   flags)
{
    BESTATUS                status = 0;
    IOCTL_COMMON_MCC_CREATE_EXT* ioctl = NULL;
    IOCTL_COMMON_MCC_CREATE* old_ioctl = NULL;
    MCC_WRB    *wrb = NULL;    
    u32 num_entries_encoded;
    PVOID va = NULL;
    u32 i;
    
    if (length < sizeof (MCC_WRB) * 2) {
        TRACE (DL_ERR, "Invalid MCC ring length:%d", length);
        return BE_NOT_OK;
    }

    // Reduce the actual ring size to be less than the number of context entries.  This
    // ensures that we run out of ring WRBs first so the queueing works correctly.  We never
    // queue based on context structs.
    if (num_context_entries + 1 < length / sizeof (MCC_WRB) - 1) {
        
        u32 max_length = (num_context_entries + 2) * sizeof (MCC_WRB);
        
        length = sa_gt_power2 (max_length) / 2;
        ASSERT (length <= max_length);

        TRACE (DL_WARN, 
            "MCC ring length reduced based on context entries. length:%d wrbs:%d context_entries:%d", 
            length, 
            length / sizeof (MCC_WRB),
            num_context_entries);
    }


    be_lock_wrb_post (function_object);


    num_entries_encoded = be_ring_length_to_encoding (length, sizeof (MCC_WRB));

    // Init MCC object.
    SA_ZERO_MEM (mcc);
    mcc->magic           = BE_MCC_MAGIC;
    mcc->parent_function = function_object;
    mcc->cq_object       = cq;

    sa_initialize_list_head (&mcc->backlog);

    // Force zeroing of CQ.
    sa_zero_mem (cq->va, cq->num_entries * 16);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        ASSERT (wrb);
        TRACE (DL_ERR, "No free MCC WRBs in create EQ.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    if (flags)
    {
        // Prepares an embedded ioctl, including request/response sizes.
        ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_MCC_CREATE_EXT);

        ioctl->params.request.num_pages = sa_ceiling (length, SA_PAGE_SIZE);
        ASSERT (ioctl->params.request.num_pages <= sa_sgl_get_page_count (sgl));
        
        //
        // Program MCC ring context
        //
        ioctl->params.request.context.pdid				= be_function_get_pd_number (function_object);
        ioctl->params.request.context.fid				= be_function_get_function_number(function_object);
        ioctl->params.request.context.valid				= TRUE;
        ioctl->params.request.context.ring_size			= num_entries_encoded;
        ioctl->params.request.context.cq_id             = be_cq_get_id (cq);
        ioctl->params.request.async_event_bitmap[0]		= flags; 

        // Create a page list for the IOCTL.
        be_sgl_to_pa_list (sgl, ioctl->params.request.pages, SA_NUMBER_OF (ioctl->params.request.pages));

        // Post the Ioctl
        status = be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, ioctl);
        if (status != BE_SUCCESS)
        {
            TRACE (DL_ERR, "COMMON_MCC_CREATE_EXT to create CQ failed.");
            be_unlock_wrb_post (function_object);
            status = be_mcc_ring_create (function_object,
                                         sgl,
                                         length,
                                         context_array,
                                         num_context_entries,
                                         cq,
                                         mcc);
            be_lock_wrb_post (function_object);
            goto error;
        }
    }
    else
    {
        // Prepares an embedded old_ioctl, including request/response sizes.
        old_ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_MCC_CREATE);

        old_ioctl->params.request.num_pages = sa_ceiling (length, SA_PAGE_SIZE);
        ASSERT (old_ioctl->params.request.num_pages <= sa_sgl_get_page_count (sgl));
        //
        // Program MCC ring context
        //
        old_ioctl->params.request.context.pdid				= be_function_get_pd_number (function_object);
        old_ioctl->params.request.context.fid				= be_function_get_function_number(function_object);
        old_ioctl->params.request.context.valid				= TRUE;
        old_ioctl->params.request.context.ring_size			= num_entries_encoded;
        old_ioctl->params.request.context.cq_id             = be_cq_get_id (cq);

        // Create a page list for the IOCTL.
        be_sgl_to_pa_list (sgl, old_ioctl->params.request.pages, SA_NUMBER_OF (old_ioctl->params.request.pages));

        // Post the Ioctl
        status = be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, old_ioctl);
        if (status != BE_SUCCESS)
        {
            TRACE (DL_ERR, "COMMON_MCC_CREATE to create CQ failed.");
            goto error;
        }
    }

    mcc->wrb_context.base = context_array;
    mcc->wrb_context.num = num_context_entries;
    sa_initialize_list_head (&mcc->wrb_context.list_head);
    sa_zero_mem (context_array, sizeof(BE_MCC_WRB_CONTEXT) * num_context_entries);
    for (i=0; i<mcc->wrb_context.num; i++) {
        sa_insert_tail_list (&mcc->wrb_context.list_head, &context_array[i].next);
    }

    // 
    // Create an SA_RING for tracking WRB hw ring
    //
    va = sa_sgl_get_base_va (sgl);
    ASSERT (va);
    sa_ring_create (&mcc->sq.ring, 
                    length/sizeof (MCC_WRB), 
                    sizeof (MCC_WRB), 
                    va);

    // save the butler id!
    if (flags)
        mcc->sq.ring.id = ioctl->params.response.id;
    else
        mcc->sq.ring.id = old_ioctl->params.response.id;
    
    //
    // Init a SA_RING for tracking the MCC CQ.
    //
    ASSERT (cq->va);
    sa_ring_create (&mcc->cq.ring, cq->num_entries, 16, cq->va);
    mcc->cq.ring.id = be_cq_get_id (cq);

    // Initialize debug index.
    mcc->consumed_index = 0;

    be_cq_object_reference (cq);
    _be_function_add_mcc (function_object, mcc);

    TRACE (DL_INFO, 
        "MCC ring created. id:%d bytes:%d cq_id:%d cq_entries:%d num_context:%d",
        flags? ioctl->params.response.id: old_ioctl->params.response.id,
        length,
        be_cq_get_id (cq),
        cq->num_entries,
        num_context_entries);

error:
    be_unlock_wrb_post (function_object);
    return status;
}

BESTATUS 
be_mcc_ring_create (
    IN PBE_FUNCTION_OBJECT function_object,
    IN PSA_SGL             sgl,
    IN u32                 length,
    IN PBE_MCC_WRB_CONTEXT context_array,
    IN u32                 num_context_entries,
    IN PBE_CQ_OBJECT       cq,
    IN PBE_MCC_OBJECT      mcc)
{
    return (be_mcc_ring_create_ext (function_object,
                                    sgl,
                                    length,
                                    context_array,
                                    num_context_entries,
                                    cq,
                                    mcc,
                                    0));
}


u32
be_mcc_get_id (PBE_MCC_OBJECT mcc) {
    MCC_ASSERT (mcc);
    return mcc->sq.ring.id;
}

/*!

@brief
    This routine destroys an MCC send queue    

@param        
    MccObject         - Internal Mcc Object to be destroyed.                    

@return   
    BE_SUCCESS if successfull,, otherwise a useful error code is returned.
    
@note
    IRQL < DISPATCH_LEVEL   
    
    The caller of this routine must ensure that no other WRB may be posted
    until this routine returns. 

*/
BESTATUS 
be_mcc_ring_destroy (
    IN PBE_MCC_OBJECT mcc)
{
    BESTATUS status = 0;

    MCC_ASSERT( mcc );  

    ASSERT (mcc->processing == 0);  

    // Remove the ring from the function object.
    // This transitions back to mailbox mode.
    _be_function_remove_mcc (mcc->parent_function, mcc);

    // Send ioctl to destroy the queue.  (Using the mailbox.)
    status = be_function_ring_destroy (mcc->parent_function, mcc->sq.ring.id, IOCTL_RING_TYPE_MCC);

    // Release the SQ reference to the CQ
    be_cq_object_dereference ( mcc->cq_object );

    return status;
}


STATIC 
void
mcc_wrb_sync_callback (
    IN      PVOID               context, 
    IN      BESTATUS            completion_status, 
    IN      PMCC_WRB       wrb)
{
    PBE_MCC_WRB_CONTEXT  wrb_context = (PBE_MCC_WRB_CONTEXT)context;

    SA_NOT_USED (wrb);

    ASSERT (wrb_context);

    *wrb_context->users_final_status = completion_status;
}


/*!

@brief
    This routine posts a command to the MCC send queue    

@param        
    MccObject       - Internal Mcc Object to be destroyed.                    
    
@param        
    wrb             - wrb to post.
    
    
@param    
    CompletionCallback  - Address of a callback routine to invoke once the WRB
                          is completed.

@param    
    CompletionCallbackContext - Opaque context to be passed during the call to
                                the CompletionCallback.        

@return   
    BE_SUCCESS if successfull,, otherwise a useful error code is returned.
    
@note
    IRQL < DISPATCH_LEVEL if CompletionCallback is not NULL
    IRQL <=DISPATCH_LEVEL if CompletionCallback is  NULL
    
    If this routine is called with CompletionCallback != NULL the
    call is considered to be asynchronous and will return as soon
    as the WRB is posted to the MCC with BE_PENDING.
    
    If CompletionCallback is NULL, then this routine will not return until
    a completion for this MCC command has been processed.                                                                     
    If called at DISPATCH_LEVEL the CompletionCallback must be NULL.
    
    This routine should only be called if the MPU has been boostraped past
    mailbox mode.
    

*/
BESTATUS
_be_mpu_post_wrb_ring (
    IN PBE_MCC_OBJECT      mcc,
    IN PMCC_WRB       wrb,
    IN PBE_MCC_WRB_CONTEXT wrb_context)
{
    
    PMCC_WRB ring_wrb = NULL;
    BESTATUS status = BE_PENDING;
    volatile BESTATUS final_status = BE_PENDING;
    MCC_WRB_CQE_CALLBACK callback = NULL;
    MCC_DB mcc_db;    

    MCC_ASSERT (mcc);
    ASSERT (sa_ring_num_empty (&mcc->sq.ring) > 0);
    
    // Input wrb is most likely the next wrb in the ring, since the client
    // can peek at the address.
    ring_wrb = sa_ring_producer_ptr (&mcc->sq.ring);
    if (wrb != ring_wrb) {
        // If not equal, copy it into the ring.
        sa_memcpy (ring_wrb, wrb, sizeof (MCC_WRB));
    }

    TRACE (DL_INFO, "WRB : %x and Ring_WRB : %x.", wrb, ring_wrb);
    TRACE (DL_INFO, "address of WRB : %x and address of Ring_WRB : %x.", &wrb, &ring_wrb);
    // Minimal verification of posted WRB.
    ASSERT (wrb->embedded || wrb->sge_count > 0);       

    SA_DBG_ONLY (
        wrb_context->ring_wrb = ring_wrb;
    )
    
    if (ring_wrb->embedded) {
        // embedded commands will have the response within the WRB.
        wrb_context->wrb = ring_wrb;
    } else {
        // non-embedded commands will not have the response within the WRB, and
        // they may complete out-of-order.  The WRB will not be valid to inspect
        // during the completion.
        wrb_context->wrb = NULL;
    }

    callback = wrb_context->callback;
    
    if (callback == NULL) {        
        // Assign our internal callback if this is a synchronous call.
        wrb_context->callback = mcc_wrb_sync_callback;
        wrb_context->callback_context = wrb_context;
        wrb_context->users_final_status = &final_status;

        mcc->parent_function->stats.synchronous_wrbs++;
    }

    // Increment producer index
    sa_ring_produce (&mcc->sq.ring);

    // initialize
    BeSetDword (&mcc_db, 0);

    mcc_db.rid       = mcc->sq.ring.id;
    mcc_db.numPosted = 1;
    
    mcc->parent_function->stats.posted_wrbs++;

    PD_WRITE (mcc->parent_function, mpu_mcc_db, mcc_db); // Ring the doorbell

    if (callback == NULL) {

        i32 polls = 0; 
        u64 start;

        start = sa_get_system_time ();
        
        // Wait until this command completes, polling the CQ.
        do {
            
            TRACE (DL_INFO, "IOCTL submitted in the poll mode.");

            // Async UE detection.
            if (mcc->parent_function->ue) {
                
                TRACE (DL_ERR, "Abort MCC CQ poll after UE.");
                
                final_status = BE_STATUS_UE_DETECTED;

                break;
            }

            // Do not rearm CQ in this context.
            be_mcc_process_cq (mcc, FALSE);

            if (final_status == BE_PENDING) {

                if ((++polls & 0x7FFFF) == 0) 
                {
                    TRACE (DL_WARN, "Warning : polling MCC CQ for %d ms.", polls/1000);

                    if (be_function_check_hardware_ue (mcc->parent_function) != BE_SUCCESS) {
                        
                        be_function_disable_ioctls (mcc->parent_function);

                        final_status = BE_STATUS_UE_DETECTED;

                        break;
                    }

                    // Each poll is 1us, the be_mpu_poll_timeout is in seconds.
                    if ( (polls/1000) >= (be_mpu_poll_timeout * 1000)) {

                        TRACE (DL_ERR, "MCC CQ poll timeout after %d ms", 
                               (u32)(sa_diff_system_time (start, sa_get_system_time ()) / 1000));

                        be_function_disable_ioctls (mcc->parent_function);

                        //if (mcc->parent_function->links.errobj) {
                        //    be_ue_generate_software_ue (mcc->parent_function->links.errobj, UE_SOFTWARE_CQ_TIMEOUT);
                        //}

                        final_status = BE_STATUS_POLLING_CQ_TIMEOUT;

                        break;
                    }
                }

                sa_dev_stall (mcc->parent_function->sa_dev, 1);  // microsecond delay
            }

        } while (final_status == BE_PENDING);

        status = final_status;
    }

    return status;
}




PMCC_WRB
_be_mpu_peek_ring_wrb (
    IN PBE_MCC_OBJECT mcc,
    boolean driving_queue)
{
    MCC_ASSERT (mcc);

    // If we have queued items, do not allow a post to bypass the queue.
    if (!driving_queue && !sa_is_list_empty (&mcc->backlog)) {
        return NULL;
    }

    if (sa_ring_num_empty (&mcc->sq.ring) <= 0) {
        return NULL;
    }

    return (PMCC_WRB)sa_ring_producer_ptr (&mcc->sq.ring);
}




BESTATUS
be_mpu_init_mailbox (    
    IN      PBE_FUNCTION_OBJECT  function_object,
    IN      PSA_SGL              mailbox)
{
    BESTATUS status;

    FUNCTION_ASSERT (function_object);
    ASSERT (mailbox);    
    
//    function_object->mailbox.sgl    = mailbox;
    function_object->mailbox.va     = sa_sgl_get_base_va (mailbox);
    function_object->mailbox.pa     = sa_sgl_get_page (mailbox, 0) + sa_sgl_get_offset (mailbox);
    function_object->mailbox.length = MIN (SA_PAGE_SIZE - sa_sgl_get_offset (mailbox), 
                                           sa_sgl_get_byte_count (mailbox));

    ASSERT ((SA_PTR_TO_U32 (function_object->mailbox.va) & 0xf) == 0); // need 16 byte alignment
    ASSERT ((SA_PTR_TO_U32 (function_object->mailbox.pa) & 0xf) == 0); // need 16 byte alignment

    //
    // Issue the WRB to set MPU endianness
    //
    {
        u8 *endian_check = (u8*)&function_object->mailbox.va->wrb;
        *endian_check++ = 0xFF; // byte 0
        *endian_check++ = 0x12; // byte 1
        *endian_check++ = 0x34; // byte 2
        *endian_check++ = 0xFF; // byte 3
        *endian_check++ = 0xFF; // byte 4
        *endian_check++ = 0x56; // byte 5
        *endian_check++ = 0x78; // byte 6
        *endian_check++ = 0xFF; // byte 7
    }

    status = be_mcc_mailbox_notify_and_wait (function_object);
    return status;
}


BESTATUS
be_mpu_uninit_mailbox (    
    IN      PBE_FUNCTION_OBJECT  function_object)
{
    BESTATUS status;

    FUNCTION_ASSERT (function_object);
    ASSERT (function_object->mailbox.va);    
    ASSERT ((SA_PTR_TO_U32 (function_object->mailbox.va) & 0xf) == 0); // need 16 byte alignment
    ASSERT ((SA_PTR_TO_U32 (function_object->mailbox.pa) & 0xf) == 0); // need 16 byte alignment

    //
    // Issue the WRB to indicate mailbox teardown
    //
    {
        u8 *mbox_unint = (u8*)&function_object->mailbox.va->wrb;
        *mbox_unint++ = 0xFF; // byte 0
        *mbox_unint++ = 0xAA; // byte 1
        *mbox_unint++ = 0xBB; // byte 2
        *mbox_unint++ = 0xFF; // byte 3
        *mbox_unint++ = 0xFF; // byte 4
        *mbox_unint++ = 0xCC; // byte 5
        *mbox_unint++ = 0xDD; // byte 6
        *mbox_unint++ = 0xFF; // byte 7
    }

    status = be_mcc_mailbox_notify_and_wait (function_object);
    return status;
}


void
be_mpu_cleanup_mailbox (PBE_FUNCTION_OBJECT function_object) {

    FUNCTION_ASSERT (function_object);
    //function_object->mailbox.sgl = NULL;
    function_object->mailbox.va = NULL;
    function_object->mailbox.pa = 0ULL;
    function_object->mailbox.length = 0UL;
}


/*!
@brief
    This routine posts a command to the MCC mailbox.    
@param        
    FuncObj         - Function Object to post the WRB on behalf of.
@param        
    wrb             - wrb to post.
@param    
    CompletionCallback  - Address of a callback routine to invoke once the WRB
                          is completed.
@param    
    CompletionCallbackContext - Opaque context to be passed during the call to
                                the CompletionCallback.        
@return   
    BE_SUCCESS if successfull,, otherwise a useful error code is returned.
@note
    IRQL <=DISPATCH_LEVEL if CompletionCallback is  NULL
    
    This routine will block until a completion for this MCC command has been
     processed. If called at DISPATCH_LEVEL the CompletionCallback must be NULL.
    
    This routine should only be called if the MPU has not been boostraped past
    mailbox mode.
*/
BESTATUS
_be_mpu_post_wrb_mailbox (
    IN PBE_FUNCTION_OBJECT function_object,
    IN PMCC_WRB            wrb,
    IN PBE_MCC_WRB_CONTEXT wrb_context)
{
    PMCC_MAILBOX mailbox = NULL;
    BESTATUS status;
    
    FUNCTION_ASSERT (function_object);
    ASSERT (function_object->links.mcc == NULL);

    mailbox = function_object->mailbox.va;
    ASSERT (mailbox);

    if ((PMCC_WRB)&mailbox->wrb != wrb) {
        SA_ZERO_MEM (mailbox);
        sa_memcpy (&mailbox->wrb, wrb, sizeof (MCC_WRB));  // copy wrb into mailbox
    }

    // The callback can inspect the final WRB to get output parameters.
    wrb_context->wrb = &mailbox->wrb;
    
    status = be_mcc_mailbox_notify_and_wait (function_object);

    if (status != BE_SUCCESS)
    {
        // Create a fake "failed" CQ entry.
        mailbox->cq.completion_status = ENUM_MGMT_STATUS_FAILED;        
        mailbox->cq.extended_status = MGMT_ADDI_STATUS_NO_STATUS;        
        mailbox->cq.mcc_tag[0] = (u32) SA_PTR_TO_U64 (wrb_context);         // low
        mailbox->cq.mcc_tag[1] = (u32)(SA_PTR_TO_U64 (wrb_context) >> 32);  // high
        mailbox->cq.consumed = 1;
        mailbox->cq.completed = 1;
        mailbox->cq.async_event = 0;
        mailbox->cq.valid = 1;
    }

    // A command completed.  Use tag to determine which command.
    // Process this even in the face of a failure.
    be_mcc_process_cqe (function_object, &mailbox->cq);

    return be_mcc_wrb_status_to_bestatus (mailbox->cq.completion_status);
}




PBE_MCC_WRB_CONTEXT
_be_mcc_allocate_wrb_context (PBE_FUNCTION_OBJECT function_object)
{
    PBE_MCC_WRB_CONTEXT context = NULL;
    SA_IRQ irq;

    FUNCTION_ASSERT (function_object);

    sa_acquire_spinlock (&function_object->mcc_context_lock, &irq);

    if (!function_object->mailbox.default_context_allocated && 
        (!function_object->links.mcc || sa_list_is_empty (&function_object->links.mcc->wrb_context.list_head)))
    {
        // Use the single default context that we always have allocated, only
        // when all other MCC context items are used.  This gives maximum debugability
        // for tracing thru the history of commands.
        function_object->mailbox.default_context_allocated = TRUE;
        context = &function_object->mailbox.default_context;
    } else if (function_object->links.mcc) {
        // Get a context from the free list. If any are available.
        if (!sa_is_list_empty (&function_object->links.mcc->wrb_context.list_head)) {
            PSA_LIST_ENTRY list_entry = sa_remove_head_list (&function_object->links.mcc->wrb_context.list_head);
            context = SA_CONTAINING_RECORD (list_entry, BE_MCC_WRB_CONTEXT, next);
        }
    }

    if (context) {
        SA_ZERO_MEM (context);        

        //
        // Add to list of in-progress WRBs.
        //
        function_object->PendingWRBs++;
        sa_list_insert_tail (&function_object->InProgressList, &context->next);
        ASSERT (sa_list_count (&function_object->InProgressList) == function_object->PendingWRBs);
    }

    sa_release_spinlock (&function_object->mcc_context_lock, &irq);

    return context;
}

void
_be_mcc_free_wrb_context (
    PBE_FUNCTION_OBJECT function_object,
    PBE_MCC_WRB_CONTEXT context)
{
    SA_IRQ irq;

    FUNCTION_ASSERT (function_object);
    ASSERT (context);

    sa_acquire_spinlock (&function_object->mcc_context_lock, &irq);

    //
    // Remove from the in-progress list.
    //
    ASSERT (function_object->PendingWRBs > 0);
    function_object->PendingWRBs--;
    sa_list_remove_entry (&context->next);

    ASSERT (sa_list_count (&function_object->InProgressList) == function_object->PendingWRBs);

    if (context == &function_object->mailbox.default_context) {
        // Free the default context.
        ASSERT (function_object->mailbox.default_context_allocated);
        function_object->mailbox.default_context_allocated = FALSE;
    } else {
        // Add to free list.
        ASSERT (function_object->links.mcc);
        sa_insert_tail_list (&function_object->links.mcc->wrb_context.list_head, &context->next);
    }

    sa_release_spinlock (&function_object->mcc_context_lock, &irq);
}

void
be_function_abort_in_progress_wrbs (
    IN PBE_FUNCTION_OBJECT       function_object)
{
    SA_IRQ irq;
    SA_LIST_ENTRY *entry;
    BE_MCC_WRB_CONTEXT *context;
    BE_GENERIC_QUEUE_CONTEXT *queue_context;
    PBE_MCC_OBJECT mcc;

    FUNCTION_ASSERT (function_object);

    //
    // Assume that UE must be set prior to aborting so further requests 
    // are not queued.
    //
    ASSERT (function_object->ue);
    
    sa_acquire_spinlock (&function_object->mcc_context_lock, &irq);

    //
    // Abort requests on the MCC ring (or the 1 request posted via the doorbell).
    //
    while (!sa_list_is_empty (&function_object->InProgressList))
    {
        MCC_CQ_ENTRY cqe;

        entry = sa_list_head (&function_object->InProgressList);
        
        context = SA_CONTAINING_RECORD (entry, BE_MCC_WRB_CONTEXT, next);

        sa_release_spinlock (&function_object->mcc_context_lock, &irq);
        be_lock_cq_process (function_object);

        memset (&cqe, 0, sizeof (cqe));

        // Create a fake CQ entry with error status.
        cqe.completion_status = ENUM_MGMT_STATUS_FAILED;
        cqe.extended_status = MGMT_ADDI_STATUS_NO_STATUS;        
        cqe.mcc_tag[0] = (u32) SA_PTR_TO_U64 (context);         // low
        cqe.mcc_tag[1] = (u32)(SA_PTR_TO_U64 (context) >> 32);  // high
        cqe.consumed = 1;                        
        cqe.completed = 1;                       
        cqe.async_event = 0;                     
        cqe.valid = 1;        
        
        #ifdef SA_DEBUG
        TRACE (DL_WARN, 
              "Abort posted MCC WRB BE_MCC_WRB_CONTEXT:%p wrb:%p",
               context, 
               SA_DBG_PTR (context->ring_wrb)
               );
        #endif

        // Process this entry with error status.
        be_mcc_process_cqe (function_object, &cqe);

        be_unlock_cq_process (function_object);
        sa_acquire_spinlock (&function_object->mcc_context_lock, &irq);
    }

    sa_release_spinlock (&function_object->mcc_context_lock, &irq);

    mcc = function_object->links.mcc;

    // 
    // Exit if the MCC ring isn't created yet.
    //
    if (!mcc) {
        return;
    }

    //
    // Abort the queued requests that haven't been posted to the chip yet.
    //
    be_lock_wrb_post (function_object);    
    
    ASSERT (mcc->driving_backlog == 0);

    while (!sa_is_list_empty (&mcc->backlog)) {

        // Get the next queued entry to process.
        entry = sa_remove_head_list (&mcc->backlog);
        ASSERT (entry);
        queue_context = SA_CONTAINING_RECORD (entry, BE_GENERIC_QUEUE_CONTEXT, context.list);
        mcc->backlog_length--;

        // Synchronous completion. Since it was queued, we will invoke the callback.
        // To the user, this is an asynchronous request.
        be_unlock_wrb_post_no_redrive (function_object);
        
        TRACE (DL_WARN, "Abort queued MCC WRB BE_GENERIC_QUEUE_CONTEXT:%p", queue_context);

        // internal callback optional
        if (queue_context->context.internal_callback) {
            queue_context->context.internal_callback (
                queue_context->context.internal_callback_context, 
                ENUM_MGMT_STATUS_FAILED, 
                NULL);
        }

        // External callback required for queued items.
        ASSERT (queue_context->context.callback);
        queue_context->context.callback (
            queue_context->context.callback_context,
            ENUM_MGMT_STATUS_FAILED, // generic error status
            NULL);
        
        be_lock_wrb_post (function_object);

    }

    //
    // Clear the ring since everything was emptied out.  
    // This allows cleanup IOCTLs to get a WRB before they fail due to the UE.
    //
    sa_ring_clear (&mcc->sq.ring);

    be_unlock_wrb_post_no_redrive (function_object);
}



BESTATUS
be_mcc_add_async_event_callback (
   IN BE_MCC_OBJECT*           mcc_object,
   IN MCC_ASYNC_EVENT_CALLBACK callback,
   IN PVOID                    callback_context
)
{
    MCC_ASSERT (mcc_object);

    // Lock against anyone trying to change the callback/context pointers while being used.
    be_lock_cq_process (mcc_object->parent_function);

    // Assign the async callback.
    mcc_object->async_context = callback_context;
    mcc_object->async_callback = callback;

    be_unlock_cq_process (mcc_object->parent_function);

    return BE_SUCCESS;
}

#define MPU_EP_CONTROL 0
#define MPU_EP_SEMAPHORE 0xac
#define MPU_SCRATCHPAD0_DB 0x180

//---------------------------------------------------------------------------------------
// Function: be_wait_for_POST_complete
//
//   Waits until the BladeEngine POST completes (either in error or success). 
//
// function_object - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_wait_for_POST_complete (
   IN PBE_FUNCTION_OBJECT function_object)
{
    MGMT_HBA_POST_STATUS_STRUCT postStatus;
    BESTATUS status;
    boolean is_asic;
    const u32 us_per_loop = 1000;                           // 1000us
    const u32 print_frequency_loops = 1000000 / us_per_loop;// 1 second print intervals
    u32 max_loops;
    u32 loops = 0;

    // Determine maximum loop count. This should be based on FPGA vs. ASIC.
    // Asumming tpre.rev_id < 10 covers up to the max number of ASICs we may spin.
    if (BeCsrBarAvailable(function_object->sa_dev))
    {
        is_asic = (BeCsrRead (function_object->sa_dev, 0x910, "") < 10);
    }
    else
    {
        is_asic = 1;
    }

    if (is_asic) {
        max_loops = 60 * print_frequency_loops;             // 60 second wait max
    } else {
        max_loops = 180 * print_frequency_loops;            // 180 second wait max
    }

    // Wait for arm fw indicating it is done or a fatal error happened. 
    // Note: POST can take some time to complete depending on configuration
    // settings (consider ARM attempts to acquire an IP address over DHCP!!!).
    // 
    TRACE(DL_NOTE, "Wait for management firmware POST completion...");

    do {        
       
        if (BeCsrBarAvailable(function_object->sa_dev))
        {
            BeSetDword(&postStatus, BeCsrRead(function_object->sa_dev, MPU_EP_SEMAPHORE, NULL));
        }
        else
        {
            BeSetDword(&postStatus, BePdRead(function_object->sa_dev, MPU_SCRATCHPAD0_DB, NULL));
        }

        if (0 == (loops % print_frequency_loops)) { 
            // Print current status
            TRACE (DL_INFO, 
                   "POST status = 0x%x (stage = 0x%x)", 
                   (* (u32 *)&postStatus), 
                   postStatus.stage);
        }

        sa_dev_stall (function_object->sa_dev, us_per_loop); 

    } while ( (postStatus.error != 1) && 
              (postStatus.stage != POST_STAGE_ARMFW_READY) && 
              (++loops < max_loops) );

    if (postStatus.error == 1) {
        TRACE (DL_ERR, "POST error! Status = 0x%x (stage = 0x%x)", (*(u32*)&postStatus), postStatus.stage);
        status = BE_STATUS_POST | postStatus.stage;
    } else if (postStatus.stage != POST_STAGE_ARMFW_READY) {
        TRACE(DL_ERR, "POST time-out! Status = 0x%x (stage = 0x%x)", 
                (*(u32 *)&postStatus), postStatus.stage);
        status = BE_STATUS_POST | postStatus.stage;
    } else {
        status = BE_SUCCESS;
    }
    
    return status;
}



//---------------------------------------------------------------------------------------
// Function: be_kickoff_and_wait_for_POST
//
//   Interacts with the BladeEngine management processor to initiate POST, and
//   subsequently waits until POST completes (either in error or success). The caller must
//   acquire the reset semaphore before initiating POST to prevent multiple drivers
//   interacting with the management processor. Once POST is complete the caller must
//   release the reset semaphore. Callers who only want to wait for POST complete may call
//   be_wait_for_POST_complete. 
//
// function_object - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_kickoff_and_wait_for_POST (
   IN PBE_FUNCTION_OBJECT function_object)
{
    MGMT_HBA_POST_STATUS_STRUCT postStatus;
    BESTATUS status;

    const u32 us_per_loop = 1000;                           // 1000us
    const u32 print_frequency_loops = 1000000 / us_per_loop;// 1 second print intervals
    const u32 max_loops = 5 * print_frequency_loops;        // 5 second wait max
    u32 loops = 0;

    // Wait for arm fw awaiting host ready or a fatal error happened.
    TRACE(DL_WARN, "Wait for BladeEngine ready to POST");

    do {

        if (BeCsrBarAvailable(function_object->sa_dev))
        {
            BeSetDword(&postStatus, BeCsrRead(function_object->sa_dev, MPU_EP_SEMAPHORE, NULL));
        }
        else
        {
            BeSetDword(&postStatus, BePdRead(function_object->sa_dev, MPU_SCRATCHPAD0_DB, NULL));
        }

        if (0 == (loops % print_frequency_loops)) { 
            // Print current status
            TRACE (DL_WARN, "POST status = 0x%x (stage = 0x%x)", (*(u32*)&postStatus), postStatus.stage); 
        }

        sa_dev_stall (function_object->sa_dev, us_per_loop); 

    } while ( (postStatus.error != 1) && 
              (postStatus.stage < POST_STAGE_AWAITING_HOST_RDY) &&
              (++loops < max_loops) );

    if (postStatus.error == 1) {
/*      TRACE(DL_ERR, "Pre-POST error! Status = 0x%x (stage = 0x%x)", postStatus.dw, postStatus.stage);*/
        TRACE(DL_ERR, "Pre-POST error! Status = 0x%x (stage = 0x%x)", 
            (*(u32 *)&postStatus), postStatus.stage);
    
        status = BE_NOT_OK;
    } else if (postStatus.stage == POST_STAGE_AWAITING_HOST_RDY && BeCsrBarAvailable(function_object->sa_dev)) {
        // Give arm fw host ready indication.
        //CSR_WRITE_CONST(function_object, mpu.ep.ep_semaphore, POST_STAGE_HOST_RDY);

        BeCsrWrite(function_object->sa_dev, MPU_EP_SEMAPHORE, POST_STAGE_HOST_RDY, NULL);

        // Wait for POST to complete
        status = be_wait_for_POST_complete(function_object);
    } else {
        // Either a timeout waiting for host ready signal or POST has 
        // moved ahead without requiring a host ready signal. 
        // Might as well give POST a chance to complete (or timeout again).
        status = be_wait_for_POST_complete(function_object);
    }

    return status;
}

//---------------------------------------------------------------------------------------
// Function: be_pci_soft_reset_nowait
//
//   This function is called to issue a BladeEngine soft reset.
//   This routine DOES NOT WAIT for BladeEngine POST to complete. 
//
// function_object - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_pci_soft_reset_nowait(
    PBE_FUNCTION_OBJECT function_object)
{
 
    PCICFG_SOFT_RESET_CSR pciSoftReset;
    PCICFG_ONLINE0_CSR pciOnline0;
    PCICFG_ONLINE1_CSR pciOnline1;
    BESTATUS status = BE_SUCCESS;
    u32 i;
    
    TRACE(DL_WARN, "PCI reset...");

    // Issue soft reset #1 to get BladeEngine into a known state.
/*  pciSoftReset.dw = PCICFG_READ(function_object, soft_reset);*/
    BeSetDword(&pciSoftReset, PCICFG_READ(function_object, soft_reset));
    
    pciSoftReset.softreset = 1;
    PCICFG0_WRITE(function_object, soft_reset, pciSoftReset);

    // wait til soft reset is deasserted - hardware
    // deasserts after some time.
    i = 0;
    while (pciSoftReset.softreset && (i++ < 1024) ) {
        sa_dev_stall (function_object->sa_dev, 50); 
/*      pciSoftReset.dw = PCICFG_READ(function_object, soft_reset);*/
        BeSetDword(&pciSoftReset, PCICFG_READ(function_object, soft_reset));
        TRACE (DL_WARN, "Stall...");
    };
    if (pciSoftReset.softreset != 0) {
        TRACE(DL_ERR, "Soft-reset #1 did not deassert as expected.");
        status = BE_NOT_OK;
        goto Error_label;
    }

    TRACE (DL_ERR, "Masking");

    // Mask everything 
    PCICFG0_WRITE_CONST(function_object, ue_status_low_mask, 0xFFFFFFFF);
    PCICFG0_WRITE_CONST(function_object, ue_status_hi_mask, 0xFFFFFFFF);

    // Set everything offline except MPU IRAM (it is offline with the soft-reset, but
    // soft-reset does not reset the PCICFG registers!)
/*
    pciOnline0.dw = 0;
    pciOnline1.dw = 0;
*/
    BeSetDword( &pciOnline0, 0);
    BeSetDword( &pciOnline1, 0);
    pciOnline1.mpu_iram_online = 1;
    PCICFG0_WRITE(function_object, online0, pciOnline0);
    PCICFG0_WRITE(function_object, online1, pciOnline1);

    sa_dev_stall (function_object->sa_dev, 50000);  // delay 5ms

    TRACE (DL_WARN, "Reset2...");

    // Issue soft reset #2.
    pciSoftReset.softreset = 1;
    PCICFG0_WRITE(function_object, soft_reset, pciSoftReset);

    // wait til soft reset is deasserted - hardware
    // deasserts after some time.
    i = 0;
    while (pciSoftReset.softreset && (i++ < 1024) ) {
        sa_dev_stall (function_object->sa_dev, 50); 
/*      pciSoftReset.dw = PCICFG_READ(function_object, soft_reset);*/
        BeSetDword( &pciSoftReset, PCICFG_READ(function_object, soft_reset));

        TRACE (DL_WARN, "Stall2...");
    };
    if (pciSoftReset.softreset != 0) {
        TRACE(DL_ERR, "Soft-reset #2 did not deassert as expected.");
        status = BE_NOT_OK;
        goto Error_label;
    }

    sa_dev_stall (function_object->sa_dev, 50000);  // delay 5ms

    TRACE (DL_WARN, "MPU...");

    // Take MPU out of reset.
    ASSERT(BeCsrBarAvailable(function_object->sa_dev));
    if (BeCsrBarAvailable(function_object->sa_dev))
        BeCsrWrite(function_object->sa_dev, MPU_EP_CONTROL, 0, NULL);

Error_label:

    return status;
}



//---------------------------------------------------------------------------------------
// Function: be_pci_soft_reset
//
//   This function is called to issue a BladeEngine soft reset. Callers should acquire the
//   soft reset semaphore before calling this function. Additionaly, callers should ensure
//   they cannot be pre-empted while the routine executes. Upon completion of this
//   routine, callers should release the reset semaphore. This routine implicitly waits
//   for BladeEngine POST to complete. 
//
// function_object - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_pci_soft_reset(
    PBE_FUNCTION_OBJECT function_object)
{
    BESTATUS status = BE_SUCCESS;

    ASSERT (BeCsrBarAvailable (function_object->sa_dev));

    status = be_pci_soft_reset_nowait(function_object);

    if (BE_SUCCESS != status) goto Error_label;
    
    // Kickoff BE POST and wait for completion
    status = be_kickoff_and_wait_for_POST(function_object);

Error_label:
    return status;
}


//---------------------------------------------------------------------------------------
// Function: be_acquire_reset_semaphore
//
//   Acquires the reset semaphore. If success is returned, the caller owns the reset
//   semaphore. Otherwise the caller does not own the reset semaphore. Callers must
//   acquire the reset semaphore before inducing a BladeEngine runtime reset. Callers must
//   also release the reset semaphore once they are done with a soft reset. Release of the
//   reset semaphore is accomplished with be_release_reset_semaphore. 
//
// function_object - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_acquire_reset_semaphore (
   IN PBE_FUNCTION_OBJECT function_object)
{
    ASSERT(FALSE == function_object->own_semaphore);
    if (FALSE == function_object->own_semaphore) {
        // BUG -- This is BE2 only.  BE3 register changed to VPD data.
        if (0 == PCICFG_READ(function_object, semaphore)) {
            // This driver now owns the resource
            function_object->own_semaphore = TRUE;
            return BE_SUCCESS;
        } else {
            // Some other driver owns the resource
            return BE_NOT_OK;
        }
    }

    return BE_SUCCESS;
}

//---------------------------------------------------------------------------------------
// Function: be_release_reset_semaphore
//
//   Releases the reset semaphore. Callers must release the reset semaphore once they are
//   done with a soft reset. 
//
// function_object    - 
// return return_type - 
//---------------------------------------------------------------------------------------
void
be_release_reset_semaphore (
   IN PBE_FUNCTION_OBJECT function_object)
{
    ASSERT(TRUE == function_object->own_semaphore);

    if (TRUE == function_object->own_semaphore) {
        PCICFG0_WRITE_CONST (function_object, semaphore, 0);
        function_object->own_semaphore = FALSE;
    }

    return;
}


//---------------------------------------------------------------------------------------
// Function: be_pci_reset_required
//
//   This private function is called to detect if a host entity is required to issue a
//   PCI soft reset and subsequently drive BladeEngine POST. Scenarios where this is
//   required:
//   1) BIOS-less configuration
//   2) Hot-swap/plug/power-on
//
// function_object - 
// return   TRUE if a reset is required, FALSE otherwise
//---------------------------------------------------------------------------------------
boolean
be_pci_reset_required(
   IN PBE_FUNCTION_OBJECT function_object)
{
    MGMT_HBA_POST_STATUS_STRUCT postStatus;
    boolean doReset = FALSE;

    // Read the management processor semaphore register to retrieve the POST status
/*  postStatus.dw = BeCsrRead(function_object->sa_dev, MPU_EP_SEMAPHORE, NULL);*/
    if (BeCsrBarAvailable(function_object->sa_dev))
    {
        BeSetDword(&postStatus, BeCsrRead(function_object->sa_dev, MPU_EP_SEMAPHORE, NULL));
    }
    else
    {
        BeSetDword(&postStatus, BePdRead(function_object->sa_dev, MPU_SCRATCHPAD0_DB, NULL));
    }
    
    if (postStatus.stage <= POST_STAGE_AWAITING_HOST_RDY) {
        // If BladeEngine is waiting for host ready indication, we want to
        // do a PCI reset.
        doReset = TRUE;
    }

    return doReset;
}


//---------------------------------------------------------------------------------------
// Function: be_drive_POST
//
//   This function is called to drive BladeEngine POST. The caller should ensure they
//   cannot be pre-empted while this routine executes. 
//
// function_object - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_drive_POST (
   IN PBE_FUNCTION_OBJECT function_object)
{
    BESTATUS status;

    if (FALSE != be_pci_reset_required(function_object)) {
        // PCI reset is needed. Acquire reset semaphore.
        if (BeCsrBarAvailable(function_object->sa_dev) && BE_SUCCESS == be_acquire_reset_semaphore(function_object)) {
            // Issue PCI reset (implicitly starts and waits for POST to complete).
            status = be_pci_soft_reset(function_object);

            // Release the reset semaphore.
            be_release_reset_semaphore(function_object);

        } else {
            // Someone else owns the reset semaphore, just wait for POST to complete.
            status = be_wait_for_POST_complete(function_object);
        }
    } else {
        // No PCI reset is needed, start POST.
        status = be_kickoff_and_wait_for_POST(function_object);
    }

    return status;
}


#ifdef ENABLE_MANAGEMENT_FIRMWARE_DOWNLOAD

// Include the header generated from the ARMfw binary. This file needs to
// be in the include path somewhere. It contains the DDR data to download 
// (the runtime management firmware).
#include "skyhawk_cfg00_fw_ddr_data.h"

// FIXME TODO @@@@ these CSR offsets and the MPU_DDR_ADDR_CSR definitions are
// put here because the register declarations and definitions are cloaked in
// the normal srcgen header output. It'd be nice to get these out of the 
// source, but I'm not sure we want to expose that interface to the world...
// For now, this is hacked in here since it's expected the download method 
// is for FPGA bringup purposes. --Vahid
#define BE_DDR_DATA_CSR 0x0188
#define BE_DDR_ADDR_CSR 0x018C
#define _DDR_BUSY_TIMEOUT 0x1000

typedef struct _MPU_DDR_ADDR_CSR {
    union {
        struct { 
            u32 rsvd0 : 2;                     // DWORD 0
            u32 data_ram_addr : 26;            // DWORD 0
            u32 def : 1;                       // DWORD 0
            u32 reg : 1;                       // DWORD 0
            u32 write : 1;                     // DWORD 0
            u32 busy : 1;                      // DWORD 0
        } SG_PACK ; // unnamed struct
        u32 dw; // dword union
    }; // unnamed union
} SG_PACK MPU_DDR_ADDR_CSR, *PMPU_DDR_ADDR_CSR;

unsigned int
_be_ddr_busy (
    IN PBE_FUNCTION_OBJECT function_object );

BESTATUS
_be_write_ddr_single (
    IN PBE_FUNCTION_OBJECT function_object,
    IN u32 dword_addr,
    IN u32 dword_data );


//---------------------------------------------------------------------------------------
// Function: _be_ddr_busy
//
//   This function is private to beclib and used for management firmware download to DDR. 
//   It retuns the state of the DDR interface.
//
// function_object - 
// return - 1 if interface is busy, 0 otherwise.
//---------------------------------------------------------------------------------------
unsigned int
_be_ddr_busy (
    IN PBE_FUNCTION_OBJECT function_object )
{
    MPU_DDR_ADDR_CSR ddr_addr_csr;
    ddr_addr_csr.dw = BeCsrRead(function_object->sa_dev, BE_DDR_ADDR_CSR, NULL);
    return (ddr_addr_csr.busy);
}

//---------------------------------------------------------------------------------------
// Function: _be_read_ddr_single
//
//   This function is private to beclib and used for management firmware download to DDR. 
//   It reads a single dword of data from DDR.
//
// function_object - 
// dword_addr - dword (4B) DDR address
// return - BE_STATUS_BUSY if DDR interface is busy, BE_NOT_OK if the write did not 
//  complete or succeed, BE_SUCCESS (0) if successful.
//---------------------------------------------------------------------------------------
BESTATUS
_be_read_ddr_single (
    IN PBE_FUNCTION_OBJECT function_object,
    IN u32 dword_addr )
{
    MPU_DDR_ADDR_CSR ddr_addr_csr;
    u32 wait_count = 0;

    // Assume interface is not busy to begin with

    // Initiate read command
    ddr_addr_csr.dw = 0;
    ddr_addr_csr.data_ram_addr = dword_addr;
    ddr_addr_csr.write = 0;
    BeCsrWrite(function_object->sa_dev, BE_DDR_ADDR_CSR, ddr_addr_csr.dw, NULL);

    // Wait for the DDR read to complete (DDR busy will deassert)
    while (_be_ddr_busy(function_object)) {
        if (++wait_count > _DDR_BUSY_TIMEOUT) {
            TRACE(DL_ERR, "DDR interface busy, failed completing read @ dword address 0x%x", dword_addr);
            return BE_STATUS_BUSY; 
        }
        sa_dev_stall (function_object->sa_dev, 1); 
    }

    return BeCsrRead(function_object->sa_dev, BE_DDR_DATA_CSR, NULL);
}

//---------------------------------------------------------------------------------------
// Function: _be_write_ddr_single
//
//   This function is private to beclib and used for management firmware download to DDR. 
//   It writes a single dword of data to DDR.
//
// function_object - 
// dword_addr - dword (4B) DDR address
// dword_data - dword (4B) DDR data
// return - BE_STATUS_BUSY if DDR interface is busy, BE_NOT_OK if the write did not 
//  complete or succeed, BE_SUCCESS (0) if successful.
//---------------------------------------------------------------------------------------
BESTATUS
_be_write_ddr_single (
    IN PBE_FUNCTION_OBJECT function_object,
    IN u32 dword_addr,
    IN u32 dword_data )
{
    MPU_DDR_ADDR_CSR ddr_addr_csr;
    u32 wait_count = 0;

    // Assume interface is not busy to begin with

    // Set data to write
    BeCsrWrite(function_object->sa_dev, BE_DDR_DATA_CSR, dword_data, NULL);

    // Initiate write command
    ddr_addr_csr.dw = 0;
    ddr_addr_csr.data_ram_addr = dword_addr;
    ddr_addr_csr.write = 1;
    BeCsrWrite(function_object->sa_dev, BE_DDR_ADDR_CSR, ddr_addr_csr.dw, NULL);

    // Wait for the DDR write to complete (DDR busy will deassert)
    while (_be_ddr_busy(function_object)) {
        if (++wait_count > _DDR_BUSY_TIMEOUT) {
            TRACE(DL_ERR, "DDR interface busy, failed completing write @ dword address 0x%x", dword_addr);
            return BE_STATUS_BUSY;
        }
        sa_dev_stall (function_object->sa_dev, 1); 
    }

    return BE_SUCCESS;
}
#endif  // ENABLE_MANAGEMENT_FIRMWARE_DOWNLOAD


//---------------------------------------------------------------------------------------
// Function: be_download_management_firmware
//
//   This function is called to download management firmware to DDR. The caller should
//   ensure they cannot be pre-empted while this routine executes. 
//
// function_object - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_download_management_firmware (
   IN PBE_FUNCTION_OBJECT function_object )
{
#ifdef ENABLE_MANAGEMENT_FIRMWARE_DOWNLOAD
    ASSERT (BeCsrBarAvailable(function_object->sa_dev));
    MGMT_HBA_POST_STATUS_STRUCT post_status;
    u32 count;
    u32 dword_addr;
    boolean is_asic;
    u32 max_loops;
    const u32 us_per_loop = 1000;
    const u32 print_frequency_loops = 1000000 / us_per_loop;

    // Asumming tpre.rev_id < 10 covers up to the max number of ASICs we may spin.
    is_asic = (BeCsrRead (function_object->sa_dev, 0x910, "") < 10);

    //
    // Wait until boot firmware initialization is complete. This is
    // the last stage in the boot firmware process before it will either
    // load management firmware from flash (if a boot script is present),
    // or halt at the RedBoot prompt (if no boot script is present or 
    // a user aborts the boot script execution countdown timer via serial
    // port). 
    // 
    // It takes ~30 seconds on the FPGA and ~5 seconds on ASIC for the 
    // boot firmware to complete initialization. The post code will be set 
    // to POST_STAGE_REDBOOT_INIT_DONE at that point. If a boot script is 
    // present, a 5 second countdown timer will execute before the script
    // is executed. Once a boot script begins execution, the post code will
    // change. Thus this code should wait until it observes post code 
    // POST_STAGE_REDBOOT_INIT_DONE (up to 30 seconds), and then wait another
    // 5 seconds once observed to ensure a boot script is not executed. 
    //

    // Determine maximum loop count. This should be based on FPGA vs. ASIC.
    // Wait for a finite amount of time until POST_STAGE_REDBOOT_INIT_DONE is observed, 
    // or if any POST code that typically happens afterwards is observed.
    TRACE(DL_WARN, "Wait until Boot firmware completes initialization prior to management firmware download attempt.");
    max_loops = (is_asic) ? (5 * print_frequency_loops) : (30 * print_frequency_loops);
    count = 1;
    do {
        sa_dev_stall (function_object->sa_dev, us_per_loop); 
        post_status.dw = BeCsrRead(function_object->sa_dev, MPU_EP_SEMAPHORE, NULL);
        if (0 == (count % print_frequency_loops)) { 
            // Print current status
            TRACE(DL_WARN, "POST status = 0x%x (stage = 0x%x)", post_status.dw, post_status.stage);
        }
    } while ( (++count < max_loops) &&
              (post_status.error != 1) && 
              (post_status.stage != POST_STAGE_REDBOOT_INIT_DONE)   &&
              (post_status.stage != POST_STAGE_POWER_ON_RESET)      &&
              (post_status.stage != POST_STAGE_ARMFW_READY)         &&
              (post_status.stage != POST_STAGE_ARMFW_UE)            &&
              (post_status.stage != POST_STAGE_FW_IMAGE_LOAD_START) &&
              (post_status.stage != POST_STAGE_FW_IMAGE_LOAD_DONE)  &&
              (post_status.stage != POST_STAGE_FW_DOWNLOAD_START)   &&
              (post_status.stage != POST_STAGE_FW_DOWNLOAD_DONE)    &&
              (post_status.stage != POST_STAGE_ARMFW_START)         &&
              (post_status.stage != POST_STAGE_ARMFW_READY) );

    if (post_status.stage != POST_STAGE_REDBOOT_INIT_DONE)
    {
        // Assume no download necessary
        TRACE(DL_WARN, "Boot firmware initialization not detected (status = 0x%08x), management firmware download aborted.", post_status.dw);
        return BE_SUCCESS;
    }

    // Wait for a finite amount of time unless post code changes from 
    // POST_STAGE_REDBOOT_INIT_DONE. This is at most a 5 second countdown timer
    // once RedBoot reaches this code before it executes a boot script.
    TRACE(DL_WARN, "10 second countdown to be sure Boot firmware is idling...");
    max_loops = 10 * print_frequency_loops;
    count = 0;
    do {
        sa_dev_stall (function_object->sa_dev, us_per_loop);
        post_status.dw = BeCsrRead(function_object->sa_dev, MPU_EP_SEMAPHORE, NULL);
    } while ( (++count != max_loops) &&
              (post_status.error != 1) && 
              (post_status.stage == POST_STAGE_REDBOOT_INIT_DONE) );

    if (post_status.stage != POST_STAGE_REDBOOT_INIT_DONE)
    {
        // Assume no download necessary
        TRACE(DL_WARN, "Boot firmware not idling (status = 0x%08x), management firmware download skipped.", post_status.dw);
        return BE_SUCCESS;
    }

    TRACE(DL_ALWAYS, "Begin management firmware download.");
    
    // Check to ensure the DDR interface is not busy
    if (_be_ddr_busy(function_object)) {
        TRACE(DL_ERR, "DDR interface busy, failed initiating management firmware download");
        return BE_STATUS_BUSY;
    }

    // Tell boot firmware we are beginning firmware download to DDR
    post_status.stage = POST_STAGE_FW_DOWNLOAD_START;
    BeCsrWrite(function_object->sa_dev, MPU_EP_SEMAPHORE, post_status.dw, NULL);

    // Write data at requested address
    dword_addr = ARMFW_DDR_DATA_BASE_ADDRESS >> 2;
    count = 0;
    for (count = 0; count < ARMFW_DDR_DATA_DWORDS; count++) {
        if (_be_write_ddr_single(function_object, 
                dword_addr + count, 
                armfw_ddr_data[count])) {
            return BE_STATUS_BUSY;
        }
    }

#if 0   // Enable for read-verify of DDR download
    // Read data at requested address
    dword_addr = ARMFW_DDR_DATA_BASE_ADDRESS >> 2;
    count = 0;
    for (count = 0; count < ARMFW_DDR_DATA_DWORDS; count++) {
        u32 got_data = _be_read_ddr_single(function_object, (dword_addr + count));
        if ( got_data != armfw_ddr_data[count] ) {
            TRACE(DL_ERR, "Management firmware download read verify failed at DDR addr %d, got 0x%08x, exp 0x%08x.",(dword_addr + count), got_data, armfw_ddr_data[count]);
            return BE_NOT_OK;
        }
    }
#endif

    // Tell boot firmware the firmware download to DDR is complete
    post_status.stage = POST_STAGE_FW_DOWNLOAD_DONE;
    BeCsrWrite(function_object->sa_dev, MPU_EP_SEMAPHORE, post_status.dw, NULL);
    TRACE(DL_ALWAYS, "Management firmware download complete.");
#else
    TRACE(DL_NOTE, "Management firmware download disabled.");
#endif
    return BE_SUCCESS;
}

BESTATUS
be_mpu_set_poll_timeout (
   IN u32                  timeout
)
{
    be_mpu_poll_timeout = timeout;
    return BE_SUCCESS;
}

BESTATUS
be_mpu_set_post_timeout (
   IN u32                  timeout
)
{
    be_mpu_post_timeout = timeout;
    return BE_SUCCESS;
}

