/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*!

@file
    pdudefault.cpp

@brief
    Provides support for creating and managing iSCSI default PDU queues.

@author jflores

@notes
    jflores   8/8/2004    initial
    $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/drivers/beclib/pdudefault_ll.c,v 1.21 2006/04/09 23:32:54 menz Exp $

*/
#include "pch.h"


//============================================================================
//                  P U B L I C  R O U T I N E S
//============================================================================

u32
be_iscsi_defq_get_id (PBE_DEFAULT_PDU_QUEUE_OBJECT   defq) {

    DEFAULT_PDU_QUEUE_ASSERT (defq);
    return defq->rid;
}

/*!

@brief
    This routine creates an iSCSI default PDU ring.          

@param
    FunctionObject      - Handle to a function object                           
@param
    SendQueueVa         - Base VA for a the iSCSI default PDU ring
@param
    RingSizePowerOfTwo  - Number of entries in the ring (power of two). 64K = 0 
@param
    EntrySize           - Encoded size of each PDU entry
@param
    CqObject            - Handle to a previously created CQ
@param
    ppIscsiDefQ         - Pointer to an opqaue handle where an iSCSI default PDU 
                          object is returned.                          

@return
    BE_SUCCESS if successfull,, otherwise a useful error code is returned.

@note
    IRQL: < DISPATCH_LEVEL
    This function allocates a BE_ISCSI_CONNECTION_OBJECT object.

*/       
BESTATUS 
be_iscsi_defq_create (
    IN  PBE_FUNCTION_OBJECT           function_object,
    IN  PSA_SGL                       sgl,
    IN  u32                           length,      // bytes in total ring
    IN  u32                           entry_size,  // bytes per entry
    IN  PBE_CQ_OBJECT                 cq_object,
    OUT PBE_DEFAULT_PDU_QUEUE_OBJECT  defq)
{
    BESTATUS status = 0;
    MCC_WRB *wrb = NULL;
    IOCTL_COMMON_ISCSI_DEFQ_CREATE *ioctl = NULL;
    u32 entry_size_encoding = 0, ring_count_encoding = 0;

    ASSERT (sgl);
    ASSERT (defq);

    be_lock_wrb_post (function_object);

    // Encode the values
    entry_size_encoding = sa_log2 (entry_size);
    ASSERT (entry_size_encoding >= 6 && entry_size_encoding <= 13);
    entry_size_encoding -= 6;

    ring_count_encoding = be_ring_length_to_encoding (length, sizeof (u64));

    TRACE (DL_INFO, 
           "iscsi defq. Entry:%d bytes (0x%x encoding) Length:%d bytes (0x%x encoding)", 
           entry_size, 
           entry_size_encoding, 
           length, 
           ring_count_encoding);
           
    FUNCTION_ASSERT (function_object);
    
    sa_zero_mem (defq, sizeof (*defq));

    defq->magic           = BE_DEFAULT_PDU_QUEUE_MAGIC;
    defq->parent_function = function_object;
    defq->rid             = (u32)0xFFFFFFFF;        
    defq->cq_object       = cq_object;
    defq->sgl             = sgl;

    //
    // map the ring into the UT
    //
    TRACE (DL_INFO, "Length:%d PageCount:%d", length, sa_sgl_get_page_count (sgl));    
    //status = be_ut_map_ring (function_object, defq->rid, CRA_DFT_TYPE, sgl, length, &defq->ut_mapping);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        ASSERT (wrb);
        TRACE (DL_ERR, "No free MCC WRBs in create EQ.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_ISCSI_DEFQ_CREATE);

    ioctl->params.request.num_pages = (UEXACT16)SA_PAGES_SPANNED (sa_sgl_get_offset (sgl), length);

    // populate chip context data structure
    ioctl->params.request.context.rx_pdid               = be_function_get_pd_number (function_object);
    ioctl->params.request.context.rx_pdid_valid         = TRUE;
    ioctl->params.request.context.ring_size             = ring_count_encoding;
    ioctl->params.request.context.pci_func_id           = be_function_get_function_number (function_object);
    ioctl->params.request.context.default_buffer_size   = entry_size;
    ioctl->params.request.context.cq_id_recv            = be_cq_get_id(cq_object);

    // Create a page list for the IOCTL.
    be_sgl_to_pa_list (sgl, ioctl->params.request.pages, SA_NUMBER_OF (ioctl->params.request.pages));

    // Post the Ioctl
    status = be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, ioctl);
    if (status != BE_SUCCESS) {
        TRACE (DL_ERR, "ioctl to create iscsi default queue failed.");
        goto error;
    }

    defq->rid = ioctl->params.response.id; // this is ID used to ring doorbell.
    
    be_cq_object_reference (cq_object);

    _be_function_add_default_pdu_queue (function_object, defq);

error:
    be_unlock_wrb_post (function_object);
    return  status;
}


#if (BE_CONFIG == 7)

/*!

@brief
    This routine creates an FCoE default PDU ring.          

@param
    FunctionObject      - Handle to a function object                           
@param
    SendQueueVa         - Base VA for a the FCoE default PDU ring
@param
    RingSizePowerOfTwo  - Number of entries in the ring (power of two). 64K = 0 
@param
    EntrySize           - Encoded size of each PDU entry
@param
    CqObject            - Handle to a previously created CQ
@param
    ppDefQ              - Pointer to an opqaue handle where an FCoE default PDU 
                          object is returned.                          

@return
    BE_SUCCESS if successfull,, otherwise a useful error code is returned.

@note
    IRQL: < DISPATCH_LEVEL

*/       
BESTATUS 
be_fcoe_defq_create (
    IN  PBE_FUNCTION_OBJECT           function_object,
    IN  PSA_SGL                       sgl,
    IN  u32                           length,      // bytes in total ring
    IN  u32                           entry_size,  // bytes per entry
    IN  PBE_CQ_OBJECT                 cq_object,
    OUT PBE_DEFAULT_PDU_QUEUE_OBJECT  defq)
{
    BESTATUS status = 0;
    MCC_WRB *wrb = NULL;
    IOCTL_COMMON_FCOE_DEFQ_CREATE *ioctl = NULL;
    u32 entry_size_encoding = 0, ring_count_encoding = 0;

    ASSERT (sgl);
    ASSERT (defq);

    be_lock_wrb_post (function_object);

    // Encode the values
    entry_size_encoding = sa_log2 (entry_size);
    ASSERT (entry_size_encoding >= 6 && entry_size_encoding <= 13);
    entry_size_encoding -= 6;

    ring_count_encoding = be_ring_length_to_encoding (length, sizeof (u64));

    TRACE (DL_INFO, 
           "fcoe defq. Entry:%d bytes (0x%x encoding) Length:%d bytes (0x%x encoding)", 
           entry_size, 
           entry_size_encoding, 
           length, 
           ring_count_encoding);
           
    FUNCTION_ASSERT (function_object);
    
    sa_zero_mem (defq, sizeof (*defq));

    defq->magic           = BE_DEFAULT_PDU_QUEUE_MAGIC;
    defq->parent_function = function_object;
    defq->rid             = (u32)0xFFFFFFFF;        
    defq->cq_object       = cq_object;
    defq->sgl             = sgl;

    //
    // map the ring into the UT
    //
    TRACE (DL_INFO, "Length:%d PageCount:%d", length, sa_sgl_get_page_count (sgl));    
    //status = be_ut_map_ring (function_object, defq->rid, CRA_DFT_TYPE, sgl, length, &defq->ut_mapping);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        ASSERT (wrb);
        TRACE (DL_ERR, "No free MCC WRBs in create EQ.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_FCOE_DEFQ_CREATE);

    ioctl->params.request.num_pages = SA_PAGES_SPANNED (sa_sgl_get_offset (sgl), length);

    // populate chip context data structure
    ioctl->params.request.context.rx_pdid               = be_function_get_pd_number (function_object);
    ioctl->params.request.context.rx_pdid_valid         = TRUE;
    ioctl->params.request.context.ring_size             = ring_count_encoding;
    ioctl->params.request.context.pci_func_id           = be_function_get_function_number (function_object);
    ioctl->params.request.context.default_buffer_size   = entry_size;
    ioctl->params.request.context.cq_id_recv            = be_cq_get_id(cq_object);

    // Create a page list for the IOCTL.
    be_sgl_to_pa_list (sgl, ioctl->params.request.pages, SA_NUMBER_OF (ioctl->params.request.pages));

    // Post the Ioctl
    status = be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, ioctl);
    if (status != BE_SUCCESS) {
        TRACE (DL_ERR, "ioctl to create iscsi default queue failed.");
        goto error;
    }

    defq->rid = ioctl->params.response.id; // this is ID used to ring doorbell.
    
    be_cq_object_reference (cq_object);

    _be_function_add_default_pdu_queue (function_object, defq);

error:
    be_unlock_wrb_post (function_object);
    return  status;
}

#endif

/*!

@brief
    This routine destroys an iSCSI default PDU queue     

@param    
    IscsiDefQ - iSCSI default PDU queue handle returned from PduDefaultCreate

@return   
    Returns BE_SUCCESS on success and an appropriate BESTATUS on failure.

@note
    This function frees resourcs allocated by PduDefaultCreate.
    IRQL: < DISPATCH_LEVEL

*/

BESTATUS 
be_iscsi_defq_destroy (
    IN     PBE_DEFAULT_PDU_QUEUE_OBJECT defq )
{
    BESTATUS status = BE_SUCCESS;

    DEFAULT_PDU_QUEUE_ASSERT (defq);

    // Send ioctl to destroy the queue.
    status = be_function_ring_destroy (defq->parent_function, defq->rid, IOCTL_RING_TYPE_ISCSI_DEFQ);
    ASSERT (status == 0);

    // Dereference any associated cq.
    be_cq_object_dereference (defq->cq_object);

    // Remove from the function object
    _be_function_remove_default_pdu_queue (defq->parent_function, defq);

    // Zero software tracking object.
    sa_zero_mem (defq, sizeof (*defq));

    return status;
}



