/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!   

@file
    $RCSfile: sa_crc.h,v $

@brief
    Header file defining the CRC32 exports.
    

CVS Header Information:
    $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/runtime/sa_crc.h,v 1.6 2006/03/10 19:16:12 menz Exp $
*/

#ifndef	__sa_crc_h__
#define	__sa_crc_h__

#include "system/sa_system.h"
SA_CPP_HEADER


#define SA_CRC_ISCSI_POLYNOMIAL (0x1EDC6F41)
#define SA_CRC_ETHER_POLYNOMIAL (0x04c11db7)


// -------------------------------------------------------
// Ethernet CRC functions.
// -------------------------------------------------------

// Returns big-endian CRC for data.
u32 sa_crc_ethernet (u8 *buffer, u32 length);

// Returns TRUE is given crc matches data.
boolean sa_crc_ethernet_validate (u8 *data, u32 len, u32 crc_bigEndian);

// Lower level functions.
// Starts CRC calculation, continues, and completes crc calculation
// in multiple steps.  Start & update return intermediate crc result.
// End return big-endian CRC.
u32 sa_crc_start_ethernet (u8 *buffer, u32 length);
u32 sa_crc_update_ethernet (u32 crc, u8 *buffer, u32 length);
u32 sa_crc_end_ethernet (u32 crc, u8 *buffer, u32 length);

// -------------------------------------------------------
// iSCSI CRC functions.
// -------------------------------------------------------

// Returns big-endian CRC for data.
u32 sa_crc_iscsi (u8 *buffer, u32 length);

// Returns TRUE is given crc matches data.
boolean sa_crc_iscsi_validate (u8 *data, u32 len, u32 crc_bigEndian);

// Lower level functions.
// Starts, continues, and ends crc calculation.
// Start & update return intermediate crc result.
// End return big-endian CRC.
u32 sa_crc_start_iscsi (u8 *buffer, u32 length);
u32 sa_crc_update_iscsi (u32 crc, u8 *buffer, u32 length);
u32 sa_crc_end_iscsi (u32 crc, u8 *buffer, u32 length);

// -------------------------------------------------------
// Generic crc32 functions.
// -------------------------------------------------------

// Computes a CRC without pre-computing the table.  Any 
// polynomial may be used.  
// Polynomial LSB is x^0 term.  MSB is x^31 term.  Term x^32 is implied.
u32 sa_crc_slow (u8 *buf, u32 len, u32 polynomial);

// Creates a table of 256 u32's for a given polynomial.
// Polynomial LSB is x^0 term.  MSB is x^31 term.  Term x^32 is implied.
void sa_create_crc_table (u32 polynomial, u32 *crc_table);

// Adds the given data to the crc calculation, returing intermediate crc32 result.
// Crc32 generally should start with 0xFFFFFFFF and the final crc result is
// the ~ the returned value.  crc_table must be precomputed with
// the sa_create_crc_table function.
u32 sa_crc_update (u8 *buffer, u32 length, u32 crc32, u32 *crc_table);

// Adds the given data to the crc calculation, returing intermediate crc32 result.
// Copys data from *from to *to, also. crc_table must be precomputed with
// the sa_create_crc_table function.
u32 sa_crc_update_copy (u8 *to, u8 *from, u32 length, u32 crc32, u32 *crc_table);

// -------------------------------------------------------
// Test
// -------------------------------------------------------
#if !defined(SA_OMIT_TEST_CODE)
    // Tests CRC algorithm with known cases.
    void sa_crc32_test (void);
#endif


SA_CPP_TRAILER
#endif // __sa_crc_h__

