/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*!

@file
   sa_mring.h

@brief
  
  The file defines a software ring data structure that contains
  multiple markers dividing non-overlapping regions.  
  It is a marker-ring (mring) for short.  
  
  The degenerate case of the marker ring contains 2 markers: the
  consumer and producer.  This case is handled by the sa_ring type.

  The multi-marker case assumes the ring has N non-overlapping
  regions.  The user defines an enumeration of N markers, where
  the ultimate producer is value 0 and the ultimate consumer is
  value N-1.  These should be typedef'd/#define'd for ease of
  use.  
  
  For example:
      #define MARKER_PRODUCER (0)
      #define MARKER_MIDDLE   (1)
      #define MARKER_CONSUMER (2)

  In this example, 2 regions exist.  From MARKER_CONSUMER to MARKER_MIDDLE,
  and from MARKER_MIDDLE to MARKER_PRODUCER.  These regions do not
  overlap.  

  The available entries in each region of the ring are queried with 
  the sa_mring_num_pending function.  This function can also be
  used with non-adjacent regions, for example, to query the number
  from the MARKER_PRODUCER all the way to MARKER_CONSUMER.  Several
  function that cmark and pmark parameters, corresponding to the
  consumer and producer markers to use in the operation.  The only
  requirement is cmark > pmark.
  
  Each marker is moved with the sa_mring_increment_marker/decrement_marker 
  function.  (It is rare to use the decrement function, since rings
  by nature increment circularly.)

  The advantage of the mring data structure is the synchronization
  requirements.  Each index may be synchronized independent of the 
  other indices.
*/
#ifndef __sa_mring_h__
#define __sa_mring_h__

#include "system/sa_system.h"
#include "runtime/sa_int_math.h"

SA_CPP_HEADER

//
// Software ring structure with N indices separating the N-1 regions
// of the ring.  Index 0 is the ultimate producer, index N-1 is
// the ultimate consumer.  The user may specify N up to 8 indices.
//
typedef struct _SA_MRING {
    u32 num;      // number of elements in queue
    u32 itemSize; // size in bytes of one object

    #if defined(SA_DEBUG)
        u32 power2;   // boolean -- true if num is a power of 2.
    #endif

    PVOID  va; // The virtual address of the ring.  This should be last
               // to allow 32 & 64 bit debugger extensions to work.
    
    u32 num_marker;
    u32 rsvd;
    u32 idx[8]; // Fixed max number of indices for the ring.
    
} SA_MRING, *PSA_MRING;


STATIC
INLINE void sa_mring_create (PSA_MRING ring, 
                             u32 num_marker,
                             u32 num_elements, 
                             u32 itemSize, 
                             PVOID va)
{
    ASSERT (ring);
    ASSERT (num_marker > 0);

    sa_zero_mem (ring, sizeof (SA_MRING));
    
    // Assert that the number of markers doesn't exceed the max.
    ASSERT (num_marker < SA_NUMBER_OF (ring->idx));

    ring->num = num_elements;
    ring->itemSize = itemSize;
    ring->va = va;
    ring->num_marker = num_marker;

    #if defined(SA_DEBUG)
        ring->power2 = sa_is_power2 (num_elements);
    #endif
}

STATIC INLINE
u32 sa_mring_num (PSA_MRING ring) {
    ASSERT (ring);
    return ring->num;
}

STATIC INLINE
PVOID sa_mring_va (PSA_MRING ring) {
    ASSERT (ring);
    return ring->va;
}

STATIC INLINE
void sa_mring_clear (PSA_MRING ring) {
    u32 i;
    ASSERT (ring);
    for (i=0; i<SA_NUMBER_OF(ring->idx); i++) {
        ring->idx[i] = 0;
    }
}

// Returns number items pending between the producer (pdix) and
// the consumer (cdix).  pdix must be less than
// cdix, following the convention that 0 is the ultimate producer
// and N-1 is the ultimate consumer.
STATIC INLINE u32 sa_mring_num_pending (PSA_MRING ring, u32 cmark, u32 pmark) {
    ASSERT (ring);
    if (ring->num == 0) { return 0; }
    ASSERT (pmark < cmark);
    ASSERT (cmark < ring->num_marker);
    ASSERT (sa_subc (ring->idx[pmark], ring->idx[cmark], ring->num) < ring->num);

    return sa_subc (ring->idx[pmark], ring->idx[cmark], ring->num);
}
STATIC INLINE u32 sa_mring_power2_num_pending (PSA_MRING ring, u32 cmark, u32 pmark) {
    ASSERT (ring);
    ASSERT (ring->power2);
    ASSERT (pmark < cmark);
    ASSERT (cmark < ring->num_marker);
    ASSERT (sa_subc_power2 (ring->idx[pmark], ring->idx[cmark], ring->num) < ring->num);

    return sa_subc_power2 (ring->idx[pmark], ring->idx[cmark], ring->num);
}

// Same as sa_mring_num_pending, but casts the IDX variables as volatile to 
// force the compiler to reread the variable from memory instead of optimizing
// away the variable read due to a cached version of the variable.
STATIC INLINE u32 sa_mring_volatile_num_pending (PSA_MRING ring, u32 cmark, u32 pmark) {
    u32 cidx;
    u32 pidx;

    ASSERT (ring);
    if (ring->num == 0) { return 0; }
    ASSERT (pmark < cmark);
    ASSERT (cmark < ring->num_marker);

    cidx = *((volatile u32*)&ring->idx[cmark]);
    pidx = *((volatile u32*)&ring->idx[pmark]);

    ASSERT (sa_subc (pidx, cidx, ring->num) < ring->num);
    return sa_subc (pidx, cidx, ring->num);
}
STATIC INLINE u32 sa_mring_power2_volatile_num_pending (PSA_MRING ring, u32 cmark, u32 pmark) {
    u32 cidx;
    u32 pidx;

    ASSERT (ring);
    ASSERT (ring->power2);
    ASSERT (pmark < cmark);
    ASSERT (cmark < ring->num_marker);

    cidx = *((volatile u32*)&ring->idx[cmark]);
    pidx = *((volatile u32*)&ring->idx[pmark]);

    ASSERT (sa_subc_power2 (pidx, cidx, ring->num) < ring->num);
    return sa_subc_power2 (pidx, cidx, ring->num);
}


// Returns TRUE if entire queue is full.  Equivalent to the number pending
// between the ultimate producer and consumer indices == num - 1.
STATIC INLINE boolean sa_mring_is_full (PSA_MRING ring) {
    ASSERT (ring);
    return (boolean)(sa_mring_num_pending (ring, ring->num_marker-1, 0) + 1 == ring->num);
}
STATIC INLINE boolean sa_mring_power2_is_full (PSA_MRING ring) {
    ASSERT (ring);
    return (boolean)(sa_mring_power2_num_pending (ring, ring->num_marker-1, 0) + 1 == ring->num);
}

// Returns TRUE if the queue is completely empty.
STATIC INLINE boolean sa_mring_is_empty (PSA_MRING ring) {
    ASSERT (ring);
    return (boolean)(sa_mring_num_pending (ring, ring->num_marker-1, 0) == 0);
}
STATIC INLINE boolean sa_mring_power2_is_empty (PSA_MRING ring) {
    ASSERT (ring);
    return (boolean)(sa_mring_power2_num_pending (ring, ring->num_marker-1, 0) == 0);
}


// Increments the given index.  
STATIC INLINE void sa_mring_increment_marker (PSA_MRING ring, u32 mark, u32 count) {
    ASSERT (ring);
    ASSERT (mark < ring->num_marker);

    if (mark == 0) {
        // We are ultimately producing a new item on the list. Verify space.
        ASSERT (sa_mring_num_pending (ring, ring->num_marker-1, mark) + count < ring->num);
    } else {
        // Verify something is pending between 2 adjacent indices to consume.
        ASSERT (sa_mring_num_pending (ring, mark, mark-1) >= count);
    }

    ring->idx[mark] = sa_addc (ring->idx[mark], count, ring->num);    
}
STATIC INLINE void sa_mring_power2_increment_marker (PSA_MRING ring, u32 mark, u32 count) {
    ASSERT (ring);
    ASSERT (ring->power2);

    ASSERT (mark < ring->num_marker);

    if (mark == 0) {
        // We are ultimately producing a new item on the list. Verify space.
        ASSERT (sa_mring_power2_num_pending (ring, ring->num_marker-1, mark) + count < ring->num);
    } else {
        // Verify something is pending between 2 adjacent indices to consume.
        ASSERT (sa_mring_power2_num_pending (ring, mark, mark-1) >= count);
    }

    ring->idx[mark] = sa_addc_power2 (ring->idx[mark], count, ring->num);    
}



// Decrements the given index.  
STATIC INLINE void sa_mring_decrement_marker (PSA_MRING ring, u32 mark, u32 count) {
    ASSERT (ring);
    ASSERT (mark < ring->num_marker);

    if (mark == ring->num_marker-1) {
        // We are ultimately adding a new item on the list. Verify space.
        ASSERT (sa_mring_num_pending (ring, ring->num_marker-1, 0) + count < ring->num);
    } else {
        // Verify something is pending between 2 adjacent indices to consume.
        ASSERT (sa_mring_num_pending (ring, mark+1, mark) >= count);
    }

    ring->idx[mark] = sa_subc (ring->idx[mark], count, ring->num);    
}
STATIC INLINE void sa_mring_power2_decrement_marker (PSA_MRING ring, u32 mark, u32 count) {
    ASSERT (ring);
    ASSERT (ring->power2);

    ASSERT (mark < ring->num_marker);

    if (mark == ring->num_marker-1) {
        // We are ultimately adding a new item on the list. Verify space.
        ASSERT (sa_mring_power2_num_pending (ring, ring->num_marker-1, 0) + count < ring->num);
    } else {
        // Verify something is pending between 2 adjacent indices to consume.
        ASSERT (sa_mring_power2_num_pending (ring, mark+1, mark) >= count);
    }

    ring->idx[mark] = sa_subc_power2 (ring->idx[mark], count, ring->num);    
}



// Returns a pointer to the given item at any point in the ring.
// It does NOT asserts that the item index is valid.
STATIC INLINE PVOID sa_mring_item (PSA_MRING ring, u32 ring_index) {
    ASSERT (ring);
    // The index must fit in the ring buffer.
    ASSERT (ring_index < ring->num);
    return (u8*)ring->va + ring_index * ring->itemSize;
}
#define sa_mring_power2_item sa_mring_item 

// Returns a pointer to the given item at any point in the ring.
// It asserts that the item index is valid.
STATIC INLINE PVOID sa_mring_valid_item (PSA_MRING ring, u32 ring_index) {
    ASSERT (ring);

    // The index must fit in the ring buffer.
    ASSERT (ring_index < ring->num);

    // The index must fall in a valid portion of the ring.
    if (ring->idx[0] >= ring->idx[ring->num_marker-1]) {
        ASSERT (ring_index >= ring->idx[ring->num_marker-1] &&
                ring_index < ring->idx[0]);
    } else {
        
        ASSERT (ring_index < ring->idx[0] || 
                ring_index >= ring->idx[ring->num_marker-1]);

    }

    return sa_mring_item (ring, ring_index);
}
#define sa_mring_power2_valid_item sa_mring_valid_item 



// Returns a pointer to the given item at any point in the ring.
// It asserts that the item index is valid.
STATIC INLINE PVOID sa_mring_marker_item (PSA_MRING ring, u32 mark) {
    ASSERT (ring);
    ASSERT (mark < ring->num_marker);
    return sa_mring_item (ring, ring->idx[mark]);
}
#define sa_mring_power2_marker_item    sa_mring_marker_item



// Query any of the indices.
STATIC INLINE u32 sa_mring_marker_index (PSA_MRING ring, u32 mark) {
    ASSERT (ring);
    ASSERT (mark < ring->num_marker);
    return ring->idx[mark];
}
#define sa_mring_power2_marker_index  sa_mring_marker_index



// Consume/produce 1 item and return pointer to next item if the ring 
// is not empty.  Returns NULL if the last item was consumed/produced.
STATIC INLINE PVOID sa_mring_increment_next_ptr (PSA_MRING ring, u32 mark) {
    ASSERT (ring);
    ASSERT (mark < ring->num_marker);

    sa_mring_increment_marker (ring, mark, 1);

    if (mark == 0) {
        // Ultimate producer.  Verify ring space.
        if (sa_mring_num_pending (ring, ring->num_marker, 0) + 1 == ring->num) {
            return NULL;
        } 
    } else {
        // One of the consumers.  Verify pending item.
        if (sa_mring_num_pending (ring, mark, mark-1) == 0) {
            return NULL;
        } 
    }
    
    // Next item is available.
    return sa_mring_marker_item (ring, mark);
}
STATIC INLINE PVOID sa_mring_power2_increment_next_ptr (PSA_MRING ring, u32 mark) {
    ASSERT (ring);
    ASSERT (ring->power2);
    ASSERT (mark < ring->num_marker);

    sa_mring_power2_increment_marker (ring, mark, 1);

    if (mark == 0) {
        // Ultimate producer.  Verify ring space.
        if (sa_mring_power2_num_pending (ring, ring->num_marker, 0) + 1 == ring->num) {
            return NULL;
        } 
    } else {
        // One of the consumers.  Verify pending item.
        if (sa_mring_power2_num_pending (ring, mark, mark-1) == 0) {
            return NULL;
        } 
    }
    
    // Next item is available.
    return sa_mring_power2_marker_item (ring, mark);
}



// Special cases for a ring containing pointers.

// Produce 1 item that must be a pointer. Copy the pointer into the ring.
// This is a special case of marker 0, the ultimate producer.
STATIC INLINE void sa_mring_produce_ptr_item (PSA_MRING ring, PVOID ptrToProduce) {
    PVOID *va;
    ASSERT (ring);
    ASSERT (ring->itemSize == sizeof (PVOID)); // Ring must contain pointers
    ASSERT (ring->va);
    
    va = (PVOID*)sa_mring_marker_item (ring, 0);
    *va = ptrToProduce;
    sa_mring_increment_marker (ring, 0, 1); // Inc producer index by 1.
}
STATIC INLINE void sa_mring_power2_produce_ptr_item (PSA_MRING ring, PVOID ptrToProduce) {
    PVOID *va;
    ASSERT (ring);
    ASSERT (ring->itemSize == sizeof (PVOID)); // Ring must contain pointers
    ASSERT (ring->va);
    
    va = (PVOID*)sa_mring_power2_marker_item (ring, 0);
    *va = ptrToProduce;
    sa_mring_power2_increment_marker (ring, 0, 1);  // Inc producer index by 1.
}

// Produce 1 item that must be a pointer. Copy the pointer into the ring.
// This is a special case of marker 0, the ultimate producer.
// Uses a memory barrier between writing the item and the index -- useful if an 
// async process may be polling the index.
STATIC INLINE void sa_mring_produce_ptr_item_mb (PSA_MRING ring, PVOID ptrToProduce) {
    PVOID *va;
    ASSERT (ring);
    ASSERT (ring->itemSize == sizeof (PVOID)); // Ring must contain pointers
    ASSERT (ring->va);
    
    va = (PVOID*)sa_mring_marker_item (ring, 0);
    *va = ptrToProduce;
    
    // Force memory write to the ring to preceed the memory write
    // to the index variable.
    sa_memory_barrier (); // WARNING -- Not defined on all platforms!

    sa_mring_increment_marker (ring, 0, 1); // Inc producer index by 1.
}
STATIC INLINE void sa_mring_power2_produce_ptr_item_mb (PSA_MRING ring, PVOID ptrToProduce) {
    PVOID *va;
    ASSERT (ring);
    ASSERT (ring->itemSize == sizeof (PVOID)); // Ring must contain pointers
    ASSERT (ring->va);
    
    va = (PVOID*)sa_mring_power2_marker_item (ring, 0);
    *va = ptrToProduce;

    // Force memory write to the ring to preceed the memory write
    // to the index variable.
    sa_memory_barrier (); // WARNING -- Not defined on all platforms!

    sa_mring_power2_increment_marker (ring, 0, 1);  // Inc producer index by 1.
}


// Consume 1 pointer item.  Return the consumed pointer.  This may be
// any marker except the ultimate producer (0).
STATIC INLINE PVOID sa_mring_consume_ptr_item (PSA_MRING ring, u32 mark) {
    PVOID item;

    ASSERT (ring);
    ASSERT (ring->itemSize == sizeof (PVOID)); // Ring must contain pointers
    ASSERT (ring->va);
    ASSERT (mark > 0);
    ASSERT (mark < ring->num_marker);
    
    item = *((PVOID*)sa_mring_marker_item (ring, mark));
    sa_mring_increment_marker (ring, mark, 1);
    return item;
}
STATIC INLINE PVOID sa_mring_power2_consume_ptr_item (PSA_MRING ring, u32 mark) {
    PVOID item;

    ASSERT (ring);
    ASSERT (ring->power2);
    ASSERT (ring->itemSize == sizeof (PVOID)); // Ring must contain pointers
    ASSERT (ring->va);
    ASSERT (mark > 0);
    ASSERT (mark < ring->num_marker);
    
    item = *((PVOID*)sa_mring_power2_marker_item (ring, mark));
    sa_mring_power2_increment_marker (ring, mark, 1);
    return item;
}

// Return the pointer item at the given marker.
STATIC INLINE PVOID sa_mring_marker_ptr_item (PSA_MRING ring, u32 mark) {
    PVOID item;

    ASSERT (ring);
    ASSERT (ring->itemSize == sizeof (PVOID)); // Ring must contain pointers
    ASSERT (ring->va);
    ASSERT (mark < ring->num_marker);
    
    item = *((PVOID*)sa_mring_marker_item (ring, mark));
    return item;
}
STATIC INLINE PVOID sa_mring_power2_marker_ptr_item (PSA_MRING ring, u32 mark) {
    PVOID item;

    ASSERT (ring);
    ASSERT (ring->power2);
    ASSERT (ring->itemSize == sizeof (PVOID)); // Ring must contain pointers
    ASSERT (ring->va);
    ASSERT (mark < ring->num_marker);
    
    item = *((PVOID*)sa_mring_power2_marker_item (ring, mark));
    return item;
}

SA_CPP_TRAILER
#endif
