/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   spring.h

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/runtime/sa_ring.h,v 1.9 2006/01/25 21:59:41 menz Exp $
*/
#ifndef __sa_ring_h__
#define __sa_ring_h__

#include "system/sa_system.h"
#include "runtime/sa_int_math.h"

SA_CPP_HEADER

//
// This structure stores information about a ring shared between hardware
// and software.  Each ring is allocated by the driver in the uncached 
// extension and mapped into BladeEngine's unified table.  BEKLIB manages
// creation and destruction of the rings in hardware.  It returns the 
// corresponding HANDLE to the ring.
//
typedef struct _SA_RING {
    u32 pages;    // queue size in pages
    u32 id;       // queue id assigned by beklib
    u32 num;      // number of elements in queue
    u32 cidx;     // consumer index
    u32 pidx;     // producer index -- not used by most rings
    u32 itemSize; // size in bytes of one object

    union {
    #if defined(SA_DEBUG)
        u32 power2;   // boolean -- true if num is a power of 2.
    #endif
        u8 pad[4]; // use union so sizeof(SA_RING) is the same between fre and chk driver.
    };

    PVOID  va; // The virtual address of the ring.  This should be last
               // to allow 32 & 64 bit debugger extensions to work.
    
} SA_RING, *PSA_RING;

#define SA_RING_FOREACH_ITEM(_pRing_,_pItem_,_type_,_code_)                                      \
    {                                                                                            \
        u32 ring_num_pending = sa_ring_num_pending ((_pRing_));                                  \
        u32 ring_index;                                                                          \
        for (ring_index = 0; ring_index < ring_num_pending; ring_index++) {                      \
            _pItem_ = (_type_)((u8*)(_pRing_)->va +                                              \
                      (_pRing_)->itemSize * (((_pRing_)->cidx + ring_index) % (_pRing_)->num));  \
            ASSERT (sizeof (*(_pItem_)) == (_pRing_)->itemSize);                              \
            _code_                                                                               \
        }                                                                                        \
    }                                                                                            \

STATIC
INLINE void sa_ring_create (PSA_RING ring, 
                            u32 num, 
                            u32 itemSize, 
                            PVOID va)
{
    ASSERT (ring);
    sa_zero_mem (ring, sizeof (SA_RING));
    ring->num = num;
    ring->pages = sa_ceiling (num * itemSize, SA_PAGE_SIZE);
    ring->itemSize = itemSize;
    ring->va = va;
    #if defined(SA_DEBUG)
    ring->power2 = sa_is_power2 (num);
    #endif
}

STATIC INLINE 
void sa_ring_set_id (PSA_RING ring, u32 id) {
    ASSERT (ring);
    ring->id = id;
}

STATIC INLINE
u32 sa_ring_num (PSA_RING ring) {
    ASSERT (ring);
    return ring->num;
}

STATIC INLINE
PVOID sa_ring_va (PSA_RING ring) {
    ASSERT (ring);
    return ring->va;
}

STATIC INLINE
void sa_ring_clear (PSA_RING ring) {
    ASSERT (ring);
    ring->pidx = ring->cidx = 0;
}

// -----------------------------------------------------------------------------
// Interface for 2 index rings. i.e. consumer/producer rings
// -----------------------------------------------------------------------------

// Returns TRUE if ring is empty
STATIC INLINE boolean sa_ring_is_empty (PSA_RING ring) {
    ASSERT (ring);
    return (ring->pidx == ring->cidx);
}

// Define the power2 version equivalent
#define sa_ring_power2_is_empty sa_ring_is_empty 

// Returns number items pending on ring.
STATIC INLINE u32 sa_ring_num_pending (PSA_RING ring) {
    ASSERT (ring);
    if (ring->num == 0) { return 0; }
    return sa_subc (ring->pidx, ring->cidx, ring->num);
}

// Returns number items pending on ring.
STATIC INLINE u32 sa_ring_volatile_num_pending (PSA_RING ring) {
    u32 cidx;
    u32 pidx;

    ASSERT (ring);    
    if (ring->num == 0) { return 0; }
    
    cidx = *((volatile u32*)&ring->cidx);
    pidx = *((volatile u32*)&ring->pidx);

    return sa_subc (pidx, cidx, ring->num);
}


// Returns number items pending on ring.
STATIC INLINE u32 sa_ring_power2_num_pending (PSA_RING ring) {
    ASSERT (ring);
    ASSERT (ring->power2);
    return sa_subc_power2 (ring->pidx, ring->cidx, ring->num);
}

// Returns number items pending on ring.
// Casts producer and consumer as volatile.
STATIC INLINE u32 sa_ring_power2_volatile_num_pending (PSA_RING ring) {
    u32 cidx;
    u32 pidx;

    ASSERT (ring);
    ASSERT (ring->power2);
    
    cidx = *((volatile u32*)&ring->cidx);
    pidx = *((volatile u32*)&ring->pidx);

    return sa_subc_power2 (pidx, cidx, ring->num);
}


// Returns number items free on ring.
STATIC INLINE u32 sa_ring_num_empty (PSA_RING ring) {
    ASSERT (ring);
    return ring->num - 1 - sa_ring_num_pending (ring);
}

// Returns number items free on ring.
STATIC INLINE u32 sa_ring_power2_num_empty (PSA_RING ring) {
    ASSERT (ring);
    ASSERT (ring->power2);
    return ring->num - 1 - sa_ring_power2_num_pending (ring);
}


// Returns TRUE if ring is full
STATIC INLINE boolean sa_ring_is_full (PSA_RING ring) {
    ASSERT (ring);
    return sa_ring_num_pending (ring) == (ring->num-1);
}

// Returns TRUE if ring is full
STATIC INLINE boolean sa_ring_power2_is_full (PSA_RING ring) {
    ASSERT (ring);
    ASSERT (ring->power2);
    return sa_ring_power2_num_pending (ring) == (ring->num-1);
}


// Consume 1 item
STATIC INLINE void sa_ring_consume (PSA_RING ring) {
    ASSERT (ring);
    ASSERT (!sa_ring_is_empty (ring));
    ring->cidx = sa_addc (ring->cidx, 1, ring->num);    
}

// Consume 1 item
STATIC INLINE void sa_ring_power2_consume (PSA_RING ring) {
    ASSERT (ring);
    ASSERT (ring->power2);
    ASSERT (!sa_ring_is_empty (ring));
    ring->cidx = sa_addc_power2 (ring->cidx, 1, ring->num);    
}


// Produce 1 item
STATIC INLINE void sa_ring_produce (PSA_RING ring) {
    ASSERT (ring);
    ASSERT (!sa_ring_is_full (ring));
    ring->pidx = sa_addc (ring->pidx, 1, ring->num);
}

// Produce 1 item
STATIC INLINE void sa_ring_power2_produce (PSA_RING ring) {
    ASSERT (ring);
    ASSERT (ring->power2);
    ASSERT (!sa_ring_is_full (ring));
    ring->pidx = sa_addc_power2 (ring->pidx, 1, ring->num);
}


// Consume count items
STATIC INLINE void sa_ring_consume_multiple (PSA_RING ring, u32 count) {
    ASSERT (ring);
    ASSERT (sa_ring_num_pending (ring) >= count);
    ring->cidx = sa_addc (ring->cidx, count, ring->num);
}

// Consume count items
STATIC INLINE void sa_ring_power2_consume_multiple (PSA_RING ring, u32 count) {
    ASSERT (ring);
    ASSERT (ring->power2);
    ASSERT (sa_ring_num_pending (ring) >= count);
    ring->cidx = sa_addc_power2 (ring->cidx, count, ring->num);
}


// Produce count items
STATIC INLINE void sa_ring_produce_multiple (PSA_RING ring, u32 count) {
    ASSERT (ring);
    ASSERT (sa_ring_num_empty (ring) >= count);
    ring->pidx = sa_addc (ring->pidx, count, ring->num);
}

// Produce count items
STATIC INLINE void sa_ring_power2_produce_multiple (PSA_RING ring, u32 count) {
    ASSERT (ring);
    ASSERT (ring->power2);
    ASSERT (sa_ring_num_empty (ring) >= count);
    ring->pidx = sa_addc_power2 (ring->pidx, count, ring->num);
}


// Reduce the producer index by count items
STATIC INLINE void sa_ring_unproduce_multiple (PSA_RING ring, u32 count) {
    ASSERT (ring);
    ASSERT (sa_ring_num_pending (ring) >= count);
    ring->pidx = sa_subc (ring->pidx, count, ring->num);
}

// Reduce the producer index by count items
STATIC INLINE void sa_ring_power2_unproduce_multiple (PSA_RING ring, u32 count) {
    ASSERT (ring);
    ASSERT (ring->power2);
    ASSERT (sa_ring_num_pending (ring) >= count);
    ring->pidx = sa_subc_power2 (ring->pidx, count, ring->num);
}


STATIC INLINE PVOID sa_ring_item (PSA_RING ring, u32 index) {
    ASSERT (ring);
    ASSERT (index < ring->num);
    ASSERT (ring->itemSize > 0);
    return (u8*)ring->va + index * ring->itemSize;
}

#define sa_ring_power2_item sa_ring_item 


// Ptr to consume item
STATIC INLINE PVOID sa_ring_consumer_ptr (PSA_RING ring) {
    ASSERT (ring);
    ASSERT (!sa_ring_is_empty (ring));
    return sa_ring_item (ring, ring->cidx);
}

#define sa_ring_power2_consumer_ptr sa_ring_consumer_ptr 


// Ptr to produce item
STATIC INLINE PVOID sa_ring_producer_ptr (PSA_RING ring) {
    ASSERT (ring);
    ASSERT (!sa_ring_is_full (ring));
    return sa_ring_item (ring, ring->pidx);
}

#define sa_ring_power2_producer_ptr sa_ring_producer_ptr


// Index to consumer
STATIC INLINE u32 sa_ring_consumer_index (PSA_RING ring) {
    ASSERT (ring);
    return ring->cidx;
}

#define sa_ring_power2_consumer_index  sa_ring_consumer_index

// Ptr to produce item
STATIC INLINE u32 sa_ring_producer_index (PSA_RING ring) {
    ASSERT (ring);
    return ring->pidx;
}

#define sa_ring_power2_producer_index  sa_ring_producer_index

// Consume 1 item and return pointer to next item if the ring 
// is not empty.  Returns NULL if the last item was consumed.
STATIC INLINE PVOID sa_ring_consume_next_ptr (PSA_RING ring) {
    ASSERT (ring);
    sa_ring_consume (ring);
    if (sa_ring_is_empty (ring)) {
        return NULL;
    } else {
        return sa_ring_consumer_ptr (ring);
    }
}

// Consume 1 item and return pointer to next item if the ring 
// is not empty.  Returns NULL if the last item was consumed.
STATIC INLINE PVOID sa_ring_power2_consume_next_ptr (PSA_RING ring) {
    ASSERT (ring);
    ASSERT (ring->power2);
    sa_ring_power2_consume (ring);
    if (sa_ring_power2_is_empty (ring)) {
        return NULL;
    } else {
        return sa_ring_power2_consumer_ptr (ring);
    }
}



// Produce 1 item.  Returns pointer to next free item to produce
// if the ring is not full.  If the ring is full, it returns NULL.
STATIC INLINE PVOID sa_ring_produce_next_ptr (PSA_RING ring) {
    ASSERT (ring);
    sa_ring_produce (ring);
    if (sa_ring_is_full (ring)) {
        return NULL;
    } else {
        return sa_ring_producer_ptr (ring);
    }
}

// Produce 1 item.  Returns pointer to next free item to produce
// if the ring is not full.  If the ring is full, it returns NULL.
STATIC INLINE PVOID sa_ring_power2_produce_next_ptr (PSA_RING ring) {
    ASSERT (ring);
    ASSERT (ring->power2);
    sa_ring_power2_produce (ring);
    if (sa_ring_power2_is_full (ring)) {
        return NULL;
    } else {
        return sa_ring_power2_producer_ptr (ring);
    }
}


// Produce 1 item (copying it into ring)
STATIC INLINE void sa_ring_produce_item (PSA_RING ring, PVOID item) {
    ASSERT (ring);
    ASSERT (ring->itemSize > 0);
    ASSERT (ring->va);
    ASSERT (item);
    
    sa_memcpy (sa_ring_producer_ptr (ring), item, ring->itemSize);
    sa_ring_produce (ring);
}

// Produce 1 item (copying it into ring)
STATIC INLINE void sa_ring_power2_produce_item (PSA_RING ring, PVOID item) {
    ASSERT (ring);
    ASSERT (ring->power2);
    ASSERT (ring->itemSize > 0);
    ASSERT (ring->va);
    ASSERT (item);
    
    sa_memcpy (sa_ring_power2_producer_ptr (ring), item, ring->itemSize);
    sa_ring_power2_produce (ring);
}


// Consume 1 item (copying it out of ring)
STATIC INLINE void sa_ring_consume_item (PSA_RING ring, PVOID item) {
    ASSERT (ring);
    ASSERT (ring->itemSize > 0);
    ASSERT (ring->va);
    ASSERT (item);
    
    sa_memcpy (item, sa_ring_consumer_ptr (ring), ring->itemSize);
    sa_ring_consume (ring);
}

// Consume 1 item (copying it out of ring)
STATIC INLINE void sa_ring_power2_consume_item (PSA_RING ring, PVOID item) {
    ASSERT (ring);
    ASSERT (ring->power2);
    ASSERT (ring->itemSize > 0);
    ASSERT (ring->va);
    ASSERT (item);
    
    sa_memcpy (item, sa_ring_power2_consumer_ptr (ring), ring->itemSize);
    sa_ring_power2_consume (ring);
}

// Produce 1 item that must be a pointer. Copy the pointer into the ring.
STATIC INLINE void sa_ring_atomic_produce_ptr_item (PSA_RING ring, PVOID ptrToProduce) {
    PVOID *va;
    u32 pIdx;
    ASSERT (ring);
    ASSERT (ring->itemSize == sizeof (PVOID)); // Ring must contain pointers
    ASSERT (ring->va);
    
    pIdx = *((volatile u32*)&ring->pidx);
    
    do {
        ASSERT (sa_ring_num_empty (ring) > 0);

        va = (PVOID*)((u8*)ring->va + sizeof (PVOID)*pIdx);

        pIdx = sa_addc (pIdx, 1, ring->num);

    } while (sa_atomic_compare_exchange_ptr (va, ptrToProduce, NULL) != NULL);

    ring->pidx = pIdx;
}

STATIC INLINE PVOID sa_ring_atomic_consume_ptr_item (PSA_RING ring) {
    PVOID *va;
    u32 cIdx;
    PVOID ptr;
    ASSERT (ring);
    ASSERT (ring->itemSize == sizeof (PVOID)); // Ring must contain pointers
    ASSERT (ring->va);
    
    cIdx = *((volatile u32*)&ring->cidx);
    va = (PVOID*)((u8*)ring->va + sizeof (PVOID) * cIdx);
    ptr = *va;    
    cIdx = sa_addc (cIdx, 1, ring->num);

    if (sa_atomic_compare_exchange_ptr (va, NULL, ptr) == ptr) {
        ring->cidx = cIdx;
    } else {
        ptr = NULL;
    }

    return ptr;
}

// Produce 1 item that must be a pointer. Copy the pointer into the ring.
STATIC INLINE void sa_ring_produce_ptr_item (PSA_RING ring, PVOID ptrToProduce) {
    PVOID *va;
    ASSERT (ring);
    ASSERT (ring->itemSize == sizeof (PVOID)); // Ring must contain pointers
    ASSERT (ring->va);
    
    va = (PVOID*)sa_ring_producer_ptr (ring);
    *va = ptrToProduce;
    sa_ring_produce (ring);
}

// Produce 1 item that must be a pointer. Copy the pointer into the ring.
// Uses a memory barrier between writing the item and the index -- useful if an 
// async process may be polling the index.
STATIC INLINE void sa_ring_produce_ptr_item_mb (PSA_RING ring, PVOID ptrToProduce) {
    PVOID *va;
    ASSERT (ring);
    ASSERT (ring->itemSize == sizeof (PVOID)); // Ring must contain pointers
    ASSERT (ring->va);
    
    va = (PVOID*)sa_ring_producer_ptr (ring);
    *va = ptrToProduce;

    // Force memory write to the ring to preceed the memory write
    // to the index variable.
    sa_memory_barrier (); // WARNING -- Not defined on all platforms!

    sa_ring_produce (ring);
}

// Produce 1 item that must be a pointer. Copy the pointer into the ring.
// Uses a memory barrier between writing the item and the index -- useful if an 
// async process may be polling the index.
STATIC INLINE void sa_ring_power2_produce_ptr_item_mb (PSA_RING ring, PVOID ptrToProduce) {
    PVOID *va;
    ASSERT (ring);
    ASSERT (ring->itemSize == sizeof (PVOID)); // Ring must contain pointers
    ASSERT (ring->va);
    
    va = (PVOID*)sa_ring_power2_producer_ptr (ring);
    *va = ptrToProduce;

    // Force memory write to the ring to preceed the memory write
    // to the index variable.
    sa_memory_barrier (); // WARNING -- Not defined on all platforms!

    sa_ring_power2_produce (ring);
}

// Produce 1 item that must be a pointer. Copy the pointer into the ring.
STATIC INLINE void sa_ring_power2_produce_ptr_item (PSA_RING ring, PVOID ptrToProduce) {
    PVOID *va;
    ASSERT (ring);
    ASSERT (ring->itemSize == sizeof (PVOID)); // Ring must contain pointers
    ASSERT (ring->va);
    
    va = (PVOID*)sa_ring_power2_producer_ptr (ring);
    *va = ptrToProduce;
    sa_ring_power2_produce (ring);
}


// Consume 1 pointer item.  Return the consumed pointer.
STATIC INLINE PVOID sa_ring_consume_ptr_item (PSA_RING ring) {
    PVOID item;

    ASSERT (ring);
    ASSERT (ring->itemSize == sizeof (PVOID)); // Ring must contain pointers
    ASSERT (ring->va);
    
    item = *((PVOID*)sa_ring_consumer_ptr (ring));
    sa_ring_consume (ring);
    return item;
}

// Consume 1 pointer item.  Return the consumed pointer.
STATIC INLINE PVOID sa_ring_power2_consume_ptr_item (PSA_RING ring) {
    PVOID item;

    ASSERT (ring);
    ASSERT (ring->power2);
    ASSERT (ring->itemSize == sizeof (PVOID)); // Ring must contain pointers
    ASSERT (ring->va);
    
    item = *((PVOID*)sa_ring_power2_consumer_ptr (ring));
    sa_ring_power2_consume (ring);
    return item;
}

// -----------------------------------------------------------------------------
// Interface for 1 index rings.
// -----------------------------------------------------------------------------

// Returns the index of the current location in the ring.
// This is used for rings with 1 index.
STATIC INLINE u32 sa_ring_current_index (PSA_RING ring) {
    ASSERT (ring);
    ASSERT (ring->pidx == 0); // not used

    return ring->cidx;
}

#define sa_ring_power2_current_index  sa_ring_current_index

// Increments the index of the ring.
// Returns the index after it is updated.
// This is used for rings with 1 index.
STATIC INLINE u32 sa_ring_next_index (PSA_RING ring) {
    ASSERT (ring);
    ASSERT (ring->num > 0);
    ASSERT (ring->pidx == 0); // not used

    ring->cidx = sa_addc (ring->cidx, 1, ring->num);
    return ring->cidx;
}

// Increment index for rings with only 1 index.
// This is used for rings with 1 index.
STATIC INLINE u32 sa_ring_power2_next_index (PSA_RING ring) {
    ASSERT (ring);
    ASSERT (ring->power2);
    ASSERT (ring->num > 0);
    ASSERT (ring->pidx == 0); // not used

    ring->cidx = sa_addc_power2 (ring->cidx, 1, ring->num);
    return ring->cidx;
}

// Returns a pointer to the current location in the ring.
// This is used for rings with 1 index.
STATIC INLINE PVOID sa_ring_current (PSA_RING ring) {
    ASSERT (ring);
    ASSERT (ring->pidx == 0); // not used

    return sa_ring_item (ring, ring->cidx);
}

#define sa_ring_power2_current sa_ring_current

// Increment index for rings with only 1 index.
// This is used for rings with 1 index.
STATIC INLINE PVOID sa_ring_next (PSA_RING ring) {
    ASSERT (ring);
    ASSERT (ring->num > 0);
    ASSERT (ring->pidx == 0); // not used

    ring->cidx = sa_addc (ring->cidx, 1, ring->num);
    return sa_ring_current (ring);
}

// Increment index for rings with only 1 index.
// This is used for rings with 1 index.
STATIC INLINE PVOID sa_ring_power2_next (PSA_RING ring) {
    ASSERT (ring);
    ASSERT (ring->power2);
    ASSERT (ring->num > 0);
    ASSERT (ring->pidx == 0); // not used

    ring->cidx = sa_addc_power2 (ring->cidx, 1, ring->num);
    return sa_ring_power2_current (ring);
}

// Add the given item to the current location in ring, and increment pointer.
// This is used for rings with 1 index.
STATIC INLINE void sa_ring_push (PSA_RING ring, PVOID item) {
    ASSERT (ring);
    ASSERT (item);
    ASSERT (ring->pidx == 0); // not used
    
    sa_memcpy (sa_ring_current (ring), item, ring->itemSize);
    sa_ring_next (ring);
}

// Add the given item to the current location in ring, and increment pointer.
// This is used for rings with 1 index.
STATIC INLINE void sa_ring_power2_push (PSA_RING ring, PVOID item) {
    ASSERT (ring);
    ASSERT (item);
    ASSERT (ring->pidx == 0); // not used
    
    sa_memcpy (sa_ring_power2_current (ring), item, ring->itemSize);
    sa_ring_power2_next (ring);
}


// Add the given item to the current location in ring, and increment pointer.
// This is used for rings with 1 index.
STATIC INLINE void sa_ring_pop (PSA_RING ring, PVOID item) {
    ASSERT (ring);
    ASSERT (item);
    ASSERT (ring->pidx == 0); // not used

    sa_memcpy (item, sa_ring_current (ring), ring->itemSize);
    sa_ring_next (ring);
}

// Add the given item to the current location in ring, and increment pointer.
// This is used for rings with 1 index.
STATIC INLINE void sa_ring_power2_pop (PSA_RING ring, PVOID item) {
    ASSERT (ring);
    ASSERT (item);
    ASSERT (ring->pidx == 0); // not used

    sa_memcpy (item, sa_ring_power2_current (ring), ring->itemSize);
    sa_ring_power2_next (ring);
}


SA_CPP_TRAILER
#endif
