/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
#ifndef __sa_toeplitz_hash_h__
#define __sa_toeplitz_hash_h__

#include "runtime/sa_rand.h"

// Compute toeplitz hash using iterative algorithm.
// Assumes key_len >= data_len.
u32
sa_toeplitz_hash(
    u8 *key,
    u32 key_len,
    u8* data,
    u32 data_len);


// Struct for caching a precomputed toeplitz table.
typedef struct _SA_TOEPLITZ_TABLE {
    u32 bytes;
    u32 key_len;
    u32 *result;
} SA_TOEPLITZ_TABLE, *PSA_TOEPLITZ_TABLE;

// Create a precomputed table for efficiently computing hashes with short keys.
SA_STATUS
sa_toeplitz_alloc_table (
    u8* key,      
    u32 key_len,                // Minimum of key length or bytes of data to hash.
    SA_TOEPLITZ_TABLE *table    // The caller provides a SA_TOEPLITZ_TABLE struct that is filled in.
    );

// Free the precomputed table.
void
sa_toeplitz_free_table (
    SA_TOEPLITZ_TABLE *table);

// Compute a Toeplitz hash using a precomputed table.
u32
sa_toeplitz_hash_with_table (
    u8 *data,
    u32 len,
    SA_TOEPLITZ_TABLE *table);


// ------------  RSS using TOEPLITZ hashes ------------------------------

// Big-Endian inputs. Directly from network packet.
u32
sa_toeplitz_rss_hash (
    u8* key,        
    u32 key_length, // Requires 16 bytes
    u32 ip_src,
    u32 ip_dst,
    u16 tcp_src,
    u16 tcp_dst);

// Big-Endian inputs. Directly from network packet.
u32
sa_toeplitz_with_table_rss_hash (
    SA_TOEPLITZ_TABLE *table,
    u32 ip_src,
    u32 ip_dst,
    u16 tcp_src,
    u16 tcp_dst);

// Big-Endian inputs, directly from network packet.
// Uses iterative algorithm.
u32
sa_toeplitz_ipv6_rss_hash (
    u8* key,
    u32 key_length,
    SA_IPV6_ADDRESS *ip_src,
    SA_IPV6_ADDRESS *ip_dst,
    u16 tcp_src,
    u16 tcp_dst);

// Big-Endian inputs, directly from network packet.
// Uses precomputed table.
u32
sa_toeplitz_with_table_ipv6_rss_hash (
    SA_TOEPLITZ_TABLE *table,
    SA_IPV6_ADDRESS *ip_src,
    SA_IPV6_ADDRESS *ip_dst,
    u16 tcp_src,
    u16 tcp_dst);

#if !defined(SA_OMIT_TEST_CODE)
    // Perform a functional and performance test on hash computation.
    void
    sa_toeplitz_test (void);
#endif

#endif // __sa_toeplitz_hash_h__


