/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_output.h

@brief
   
CVS Header Information:
   $Header: $
*/
#ifndef __sa_output_h__
#define __sa_output_h__

#include "sa_types.h"
#include "sa_std.h"

SA_CPP_HEADER

#define SA_OUTPUT_ASSERT(_pSaOutput)   \
        ASSERT (_pSaOutput);           \
        ASSERT ((_pSaOutput)->print);  \

struct _SA_OUTPUT;
typedef void (SA_STDCALL *SA_OUTPUT_PRINT) (struct _SA_OUTPUT *out, const c8* str);   // Required function
typedef void (SA_STDCALL *SA_OUTPUT_FLUSH) (struct _SA_OUTPUT *out);                  // Optional
typedef void (SA_STDCALL *SA_OUTPUT_VPRINTF) (struct _SA_OUTPUT *out, const c8* format, va_list args); // Optional

typedef struct _SA_OUTPUT {
    SA_OUTPUT_PRINT   print;   // required
    SA_OUTPUT_FLUSH   flush;   // optional, may be NULL
    SA_OUTPUT_VPRINTF vprintf; // optional, may be NULL
    PVOID sa_rsvd[2];          // reserved for future functions -- future SA versions may modify
    PVOID rsvd[2];             // output specific parameters that are used for
                               // specific output needs, such as a FILE* handle.                       
} SA_OUTPUT, *PSA_OUTPUT;

// Index for each default output location in array below.
#define SA_OUTPUT_NULL  (0)
#define SA_OUTPUT_TRACE (1)
#define SA_OUTPUT_DBG   (2)
#define SA_OUTPUT_OUT   (3)
#define SA_OUTPUT_ERR   (4)
#define SA_OUTPUT_MAX   (5) // Number of std outputs locations

// Globals defined based on the environment.  The user program
// may modify these variables to change the location of debug, trace,
// or stdout/err prints.  
extern SA_OUTPUT *sa_output_std[SA_OUTPUT_MAX];

// These may be passed for any SA_OUTPUT* parameter.  The default locations
// may be modified using the sa_output_redirect function.
#define sa_stdtrace (sa_output_std[SA_OUTPUT_TRACE])
#define sa_stdnull  (sa_output_std[SA_OUTPUT_NULL])
#define sa_stddbg   (sa_output_std[SA_OUTPUT_DBG])
#define sa_stdout   (sa_output_std[SA_OUTPUT_OUT])
#define sa_stderr   (sa_output_std[SA_OUTPUT_ERR])

// Modifies the SA_OUTPUT object used for one of the 5 defaults.
// e.g. sa_output_redirect (SA_OUTPUT_ERR, sa_stdout); // Moves all stderr output to stdout.
// std_output_to_redirect is one of the indices defined above.
// out may be any SA_OUTPUT object.
void sa_output_std_redirect (u32 sa_output_index, SA_OUTPUT *out);

// Code that wishes to print text should use these functions.  Generally, it 
// should either use one of the global SA_OUTPUT variables or accept an
// SA_OUTPUT parameter so the calling routine can choose the output location.
// This allows all utility-type functions (like dumping a trace log) 
// to be very general purpose.  They can be used in the debugger, console app, 
// windows app, or debugger extension.  All functions are valid for all SA_OUTPUT
// structs.
void sa_output_print (SA_OUTPUT *out, const c8* str);
void sa_output_flush (SA_OUTPUT *out);
void sa_output_vprintf (SA_OUTPUT *output, const c8* format, va_list args);
void sa_output_printf (SA_OUTPUT *output, const c8* format, ...);
void sa_output_putc (SA_OUTPUT *output, c8 character);

// Closes an output struct. It is still safe to pass the struct to an output 
// function, but all output is discarded after it is closed.
void sa_output_close (SA_OUTPUT *out);

// Tees the outputs one & two.  The resulting SA_OUTPUT struct pipes the same output to both 
// destinations.  All three SA_OUTPUT structs must remain allocated for the duration of
// using the tee.
void sa_output_tee_create (IN SA_OUTPUT *one, IN SA_OUTPUT *two, OUT SA_OUTPUT *out);


void sa_output_string_create (SA_OUTPUT *se_out, PVOID log, u32 log_size);

// These functions automatically use the sa_stdout global variable.
void sa_putchar (const c8 c);
void sa_printf (const c8* format, ...);
void sa_vprintf (const c8* format, va_list args);

#if defined(SA_USER)
    // Creates an SA_OUTPUT object for an open file.  The handle must be valid.
    // SA_OUTPUT must point to an allocated struct.
    void sa_output_cfile_create (FILE* handle, SA_OUTPUT *out);
#endif

SA_CPP_TRAILER

#endif // __sa_output_h__

