/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
#include "sa_types.h"
#include "sa_defines.h"
#include "sa_assert.h"
#include "sa_trace.h"
#include "sa_std.h"

#if defined(SA_USER) && defined(SA_CONSOLE)

    boolean 
    sa_assert_handler (const c8 *module_name, 
                       const c8 *expression, 
                       const c8 *file, 
                       const c8 *function, 
                       u32 line)
    {
        #define MAX_IGNORE_ALL (16)
        static struct { const c8 *file; u32 line; } db[MAX_IGNORE_ALL];
        static u32 ignored = 0;
        u32 i;
        i32 c = 0;

        for (i=0; i<ignored; i++) {
            if (file == db[i].file && line == db[i].line)
                return 0; // ignore all
        }
        
        // Print to debugger
        sa_trace_vargs (DL_ERR, module_name, "ASSERT FAILURE: %s", expression);
        sa_trace_vargs(DL_ERR, module_name, "Function %s() in %s line %d", function, file, line);

        // ... and to console.
        sa_printf ("ASSERT FAILURE: %s in %s\n", expression, module_name);
        sa_printf ("Function %s() in %s line %u\n", function, file, line);
        
        while ( (c != 'a') &&
                (c != 'b') && 
                (c != 'i') && 
                (c != 'g') ) {
            printf ("(a)bort, (b)reak, (i)gnore, i(g)nore-all? ");
            c = SA_GETC ();
            printf ("\n");

            if (c == EOF) {
                // stdin was closed. just crash
                sa_crash ();
            }
        }

        if (c == 'a') {
            exit (-1);
        } else if (c == 'i') {
            return 0;
        } else if (c == 'g') {
            if (ignored < MAX_IGNORE_ALL) {
                db[ignored].file = file;
                db[ignored++].line = line;
            }
            return 0;
        } else {
            return 1;
        }
    }

#else

    boolean 
    sa_assert_handler (const c8 *module, 
                       const c8 *expression, 
                       const c8 *file, 
                       const c8 *function, 
                       u32 line)
    {
        sa_trace_vargs (DL_ERR, module, "ASSERT FAILURE: %s", expression);
        sa_trace_vargs (DL_ERR, module, "Function %s() in %s line %d", function, file, line);
        return TRUE; // always break
    }

#endif
  

    //
    // Code to get user mode stack trace and overwrite the breakpoint with a NOP.
    //
    //STATIC INLINE
    //void sa_assert_win_user ()
    //{
    //    CONTEXT         CurrentContext;
    //    unsigned c8*  RetAddr;
    //    DWORD_PTR*      pRetAddr;

    //    memset( &CurrentContext, 0, sizeof(CurrentContext) );
    //    CurrentContext.ContextFlags = CONTEXT_FULL;
    //    GetThreadContext( GetCurrentThread(), &CurrentContext );

    //    printf(".\nStack trace:\n");
    //    DumpStack( &CurrentContext );      
    //    printf(".\n.\n");

    //    RetAddr = (unsigned c8*)_ReturnAddress();
    //    pRetAddr = (DWORD_PTR*)_AddressOfReturnAddress();
    //    if ( *RetAddr == 0xCC )
    //    {
    //        c8 c = 0;

    //        
    //        while (c != 'A' && c != 'I' && c != 'Z' && c != 'D') {
    //            printf( "(A)bort (B)reak (I)gnore (Z)ap (D)ump?  " );
    //            c = (c8)getc8();
    //            fflush(stdin);
    //            c = _toupper(c);
    //            
    //            if ( c == 'A' )
    //            {
    //                exit(0xBADDC0DE);
    //            }
    //            else if ( c == 'I' )
    //            {
    //                (*pRetAddr)++;
    //            }
    //            else if ( c == 'Z' )
    //            {
    //                DWORD OldProtect, Dummy;
    //                if ( VirtualProtect( RetAddr, 1, PAGE_EXECUTE_READWRITE, &OldProtect ) )
    //                {
    //                    *RetAddr = (unsigned c8)0x90;
    //                    VirtualProtect( RetAddr, 1, OldProtect, &Dummy );
    //                }
    //                else
    //                {
    //                    printf("Couldn't zap\n" );
    //                }
    //            }    
    //            else if ( c == 'D' )
    //            {
    //                BeBugCheck();
    //                (*pRetAddr)++;
    //            }
    //            else if ( c == 'B' )
    //            {
    //                break;
    //            }
    //        }        
    //    }
    //    else if (*RetAddr != 0x90)
    //    {
    //        // Tell jose
    //        __debugbreak(); 
    //    }
    //}

