/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!   

@file
    $RCSfile: sa_crc.c,v $

@brief
    This implements the iSCSI protocol using both the software TCP stack and
    the offload engine.

@notes
    This code was derived from the UNH Sample version 1.5.03.

CVS Header Information:
    $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/runtime/sa_crc.c,v 1.8 2006/03/13 19:42:46 menz Exp $
*/
#include "sa_crc.h"
#include "sa_endian.h"
#include "sa_trace.h"
#include "sa_assert.h"
#include "sa_int_math.h"


// Polynomial LSB is x^0 term.  MSB is x^31 term.  It x^32 is implied.
// Outputs the crc_table[256].
void sa_create_crc_table (u32 polynomial, u32 *crc_table)
{
    u32 c;
    u32 n, k;
    u32 reverse_poly = sa_bit_reverse (polynomial);

    for (n = 0; n < 256; n++)
    {
        c = n;

        for (k = 0; k < 8; k++)
            c = c & 1 ? reverse_poly ^ (c >> 1) : c >> 1;

        crc_table[n] = c;
    }
}

// Updates next byte of calculation.
STATIC
FORCEINLINE
u32 sa_crc32c_byte (u32 crc, u8 byte, u32 *table)
{
    return (crc >> 8) ^ table[(crc ^ byte) & 0xFF];
}

// Adds more bytes to the calculation
u32 sa_crc_update (u8 *buffer, u32 length, u32 crc32, u32 *crc_table)
{
	u32 i;

    for (i = 0; i < length ; i++)
		crc32 = sa_crc32c_byte (crc32, buffer[i], crc_table);

	return crc32;
}

u32 sa_crc_update_copy (u8 *to, u8 *from, u32 length, u32 crc32, u32 *crc_table)
{
	u32 i;
	u32 *_to = (u32 *)to;
	u32 *_from = (u32 *)from;
	
	for (i = 0; i < length; i+=4) {
		_to[i] = _from[i];

        crc32 = sa_crc32c_byte (crc32, from[i], crc_table);
        crc32 = sa_crc32c_byte (crc32, from[i+1], crc_table);
        crc32 = sa_crc32c_byte (crc32, from[i+2], crc_table);
        crc32 = sa_crc32c_byte (crc32, from[i+3], crc_table);
	}

	return crc32;
}

u32 sa_crc_slow (u8 *buf, u32 len, u32 polynomial)
{    
    u32 crc;
    u8 byte;
    u32 b;
    u32 reverse_poly = sa_bit_reverse (polynomial);
    
    crc = 0xffffffff;
    while( len-- > 0 )
    {
        byte = *buf++;
        for( b = 0; b < 8; b++ )
        {
            if( (byte & 0x1) ^ (crc & 0x1) )
            {
                crc >>= 1;
                crc ^= reverse_poly; // reverse_poly is 0xedb88320 for ethernet
            }
            else
            {
                crc >>= 1;
            }
            byte >>= 1;
        }
    } 
    
    return( ~ crc );
}

// ------------------------------------------------------------------
//  Ethernet
// ------------------------------------------------------------------
STATIC boolean sa_ether_table_exists = FALSE;
STATIC u32 sa_ether_crc_table[256] = {0};

u32 sa_crc_update_ethernet (u32 crc, u8 *buffer, u32 length)
{
    ASSERT (sa_ether_table_exists);
    return sa_crc_update (buffer, length, crc, sa_ether_crc_table);
}


u32 sa_crc_start_ethernet (u8 *buffer, u32 length)
{
    // Create the table first time.
    if (!sa_ether_table_exists) {
        sa_create_crc_table (SA_CRC_ETHER_POLYNOMIAL, sa_ether_crc_table);
        sa_ether_table_exists = TRUE;
    }

    // update the crc with the current data set.
    return sa_crc_update_ethernet ((u32)0xFFFFFFFF, buffer, length);
}

// Return big-endian CRC
u32 sa_crc_end_ethernet (u32 crc, u8 *buffer, u32 length)
{
    crc = sa_crc_update_ethernet (crc, buffer, length);
    crc = ~crc;    
    return crc;  // big-endian
}

// Return big-endian CRC.
u32 sa_crc_ethernet (u8 *buffer, u32 length) {
    u32 crc = sa_crc_start_ethernet (buffer, length);
    return sa_crc_end_ethernet (crc, NULL, 0);
}

boolean
sa_crc_ethernet_validate (u8 *data, 
                          u32 len, 
                          u32 crc_bigEndian) 
{    
    u32 crc_calc;
    boolean valid;

    crc_calc = sa_crc_ethernet (data, len);
    valid = (boolean)(crc_bigEndian == crc_calc);

    if (!valid) {
        TRACE (DL_ERR, 
               "Ethernet CRC. Received %x, expected %x\n", 
               crc_bigEndian, 
               crc_calc);
    }
    return valid;
}

// ------------------------------------------------------------------
//  ISCSI
// ------------------------------------------------------------------
STATIC boolean sa_iscsi_table_exists = FALSE;
STATIC u32 sa_iscsi_crc_table[256] = {0};

u32 sa_crc_update_iscsi (u32 crc, u8 *buffer, u32 length)
{
    ASSERT (sa_iscsi_table_exists);
    return sa_crc_update (buffer, length, crc, sa_iscsi_crc_table);
}

u32 sa_crc_start_iscsi (u8 *buffer, u32 length)
{
    // Create the table first time.
    if (!sa_iscsi_table_exists) {
        sa_create_crc_table (SA_CRC_ISCSI_POLYNOMIAL, sa_iscsi_crc_table);
        sa_iscsi_table_exists = TRUE;
    }

    // update the crc with the current data set.
    return sa_crc_update_iscsi ((u32)0xFFFFFFFF, buffer, length);
}

// Return big-endian CRC
u32 sa_crc_end_iscsi (u32 crc, u8 *buffer, u32 length)
{
    crc = sa_crc_update_iscsi (crc, buffer, length);
    crc = ~crc;    
    return crc;  // big-endian
}

// Return big-endian CRC.
u32 sa_crc_iscsi (u8 *buffer, u32 length) {
    u32 crc = sa_crc_start_iscsi (buffer, length);
    return sa_crc_end_iscsi (crc, NULL, 0);
}

boolean
sa_crc_iscsi_validate (u8 *data, 
                       u32 len, 
                       u32 crc_bigEndian) 
{    
    u32 crc_calc;
    boolean valid;

    crc_calc = sa_crc_iscsi (data, len);
    valid = (boolean)(crc_bigEndian == crc_calc);

    if (!valid) {
        TRACE (DL_ERR, 
               "Digest error. Received %x, expected %x\n", 
               crc_bigEndian, 
               crc_calc);
    }
    return valid;
}


// ------------------------------------------------------------------
//  Test Code
// ------------------------------------------------------------------

#if !defined(SA_OMIT_TEST_CODE)

    void sa_crc32_perf_test (u32 *x, u32 *y)
    {
        #define SA_CRC32_TEST_SIZE (4096)     // size of buffer to crc
        #define SA_CRC32_TEST_ITERATIONS (64) // # of times to compute CRC to get avg time

        u8 *buf = sa_malloc (SA_CRC32_TEST_SIZE, 0, SA_TAG("crc."));
        u32 i;
        u64 slow, fast;

        SA_RELEASE_ASSERT (buf);

        for (i = 0; i < SA_CRC32_TEST_SIZE; i++) {
            buf[i] = (u8)i;
        }

        // Using slow method.
        slow = sa_timestamp ();
        for (i = 0; i < SA_CRC32_TEST_ITERATIONS; i++) {
            *x += sa_crc_slow (buf, SA_CRC32_TEST_SIZE, SA_CRC_ISCSI_POLYNOMIAL);
        }
        slow = sa_timestamp () - slow;

        // Using fast (table) method
        fast = sa_timestamp ();
        for (i = 0; i < SA_CRC32_TEST_ITERATIONS; i++) {
            *y += sa_crc_iscsi (buf, SA_CRC32_TEST_SIZE);
        }
        fast = sa_timestamp () - fast;

        // Report the avg timing results.
        sa_trace (DL_INFO, 
                "CRC Calculation. Slow ticks: %d Fast Ticks: %d", 
                (u32)slow/SA_CRC32_TEST_ITERATIONS, 
                (u32)fast/SA_CRC32_TEST_ITERATIONS
                //(u32) (fast * 100 / slow)
                );

        sa_free (buf);
    }


    // Tests CRC algorithm with examples in iSCSI spec.
    void sa_crc32_test ()
    {
        u8 data[32];
        u32 crc, crc2;
        u8 i;
        
        // All 0's
        sa_memset (data, 0, sizeof (data));
        crc = sa_crc_iscsi (data, sizeof (data));     
        ASSERT (crc == 0x8a9136aa); // big-endian result
        crc2 = sa_crc_slow (data, sizeof (data), SA_CRC_ISCSI_POLYNOMIAL);
        ASSERT (crc == crc2);

        // All 1's
        for (i = 0; i<32; i++) data[i] = 0xFF;
        crc = sa_crc_iscsi (data, sizeof (data));     
        ASSERT (crc == 0x62a8ab43); // big-endian result
        crc2 = sa_crc_slow (data, sizeof (data), SA_CRC_ISCSI_POLYNOMIAL);
        ASSERT (crc == crc2);

        // Incrementing
        for (i = 0; i<32; i++) data[i] = i;    
        crc = sa_crc_iscsi (data, sizeof (data));     
        ASSERT (crc == 0x46dd794e); // big-endian result
        crc2 = sa_crc_slow (data, sizeof (data), SA_CRC_ISCSI_POLYNOMIAL);
        ASSERT (crc == crc2);

        // Decrementing
        for (i = 0; i<32; i++) data[i] = 31 - i;
        crc = sa_crc_iscsi (data, sizeof (data));     
        ASSERT (crc == 0x113fdb5c); // big-endian result
        crc2 = sa_crc_slow (data, sizeof (data), SA_CRC_ISCSI_POLYNOMIAL);
        ASSERT (crc == crc2);

        // Verify the CRC table first few guys.
        ASSERT (sa_iscsi_crc_table[0] == 0x00000000);
        ASSERT (sa_iscsi_crc_table[1] == 0xF26B8303);
        ASSERT (sa_iscsi_crc_table[2] == 0xE13B70F7);
        ASSERT (sa_iscsi_crc_table[3] == 0x1350F3F4);


        sa_crc32_perf_test (&crc, &crc2);

        // -----------------------------------------------
        // Ethernet test
        // -----------------------------------------------

        // All 0's
        sa_memset (data, 0, sizeof (data));
        crc = sa_crc_ethernet (data, sizeof (data));
        crc2 = sa_crc_slow (data, sizeof (data), SA_CRC_ETHER_POLYNOMIAL);
        ASSERT (crc == 0x190a55ad); // big-endian result
        ASSERT (crc == crc2);

        // All 1's
        for (i = 0; i<32; i++) data[i] = 0xFF;
        crc = sa_crc_ethernet (data, sizeof (data)); 
        crc2 = sa_crc_slow (data, sizeof (data), SA_CRC_ETHER_POLYNOMIAL);
        ASSERT (crc == 0xff6cab0b); // big-endian result
        ASSERT (crc == crc2);

        // Incrementing
        for (i = 0; i<32; i++) data[i] = i;    
        crc = sa_crc_ethernet (data, sizeof (data)); 
        crc2 = sa_crc_slow (data, sizeof (data), SA_CRC_ETHER_POLYNOMIAL);
        ASSERT (crc == 0x91267e8a); // big-endian result
        ASSERT (crc == crc2);

        // Decrementing
        for (i = 0; i<32; i++) data[i] = 31 - i;
        crc = sa_crc_ethernet (data, sizeof (data)); 
        crc2 = sa_crc_slow (data, sizeof (data), SA_CRC_ETHER_POLYNOMIAL);
        ASSERT (crc == 0x9ab0ef72); // big-endian result
        ASSERT (crc == crc2);

        // Verify the CRC table first few guys.
        ASSERT (sa_ether_crc_table[0] == 0x00000000);
        ASSERT (sa_ether_crc_table[1] == 0x77073096);
        ASSERT (sa_ether_crc_table[2] == 0xee0e612c);
        ASSERT (sa_ether_crc_table[3] == 0x990951ba);
        
    }

#endif

