/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*!

@file 
    $RCSfile: sa_object_allocator.c,v $
    
@brief 
    Implements generic index and object allocators that use a 
    free list to efficiently allocate/deallocate items.  Supports
    both multi-thread safe, and non-safe methods.
    
@notes

CVS Header Information:
    $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/runtime/sa_object_allocator.c,v 1.4 2005/08/05 17:29:08 menz Exp $
*/

#include "sa_object_allocator.h"
#include "sa_mem.h"


PSA_OBJECT_ALLOCATOR
sa_objalloc_allocate (u32 num,
                      u32 object_size,
                      u32 alloc_flags,
                      const c8* optional_name)
{
    SA_STATUS status = SA_STATUS_SUCCESS;

    // Allocate the object itself.
    PSA_OBJECT_ALLOCATOR oa = sa_malloc (sizeof (SA_OBJECT_ALLOCATOR),
                                          alloc_flags,
                                          SA_TAG ("oa.."));

    // Allocate object array.
    PVOID objects = sa_malloc (object_size * num,
                                alloc_flags,
                                SA_TAG("oaob"));

    PSA_INDEX_ALLOCATOR ia = sa_indexalloc_allocate (num, alloc_flags, optional_name);

    if (oa == NULL || objects == NULL || ia == NULL) {
        status = SA_STATUS_NOT_OK;
    } else {
        status = sa_objalloc_create (oa, num, object_size, objects, ia);
    }

    if (status != SA_STATUS_SUCCESS) {
        // sa_free handles null pointers.
        sa_free (oa);
        sa_free (objects);
        sa_indexalloc_free (ia);
        return NULL;
    }

    return oa;
}


void
sa_objalloc_free (PSA_OBJECT_ALLOCATOR oa)
{
    sa_objalloc_destroy (oa);

    sa_indexalloc_free (oa->ia);
    sa_free (oa->object_array);    
    sa_free (oa);
}



/*!
@brief 
    Initializes an object allocator.

@param 
    oa - pointer to object allocator to intialize
@param 
    entries - pointer to array of LIST_ENTRY objects 
@param
    num - number of LIST_ENTRY objects in array. 
@param
    optionalDbgVector - optional array of booleans used only during
        debug builds for tracking double frees.  This vector must
        have "num" elements.
                      
@return
    None
*/
SA_STATUS
sa_objalloc_create (PSA_OBJECT_ALLOCATOR oa, 
                    u32 num, 
                    u32 objectSize, 
                    PVOID objectArray, 
                    PSA_INDEX_ALLOCATOR ia) 
{
    ASSERT (oa);
    ASSERT (objectArray);
    ASSERT (objectSize > 0);
    
    sa_zero_mem (oa, sizeof (SA_OBJECT_ALLOCATOR));

    SA_NOT_USED (num);

    oa->object_size = objectSize;
    oa->object_array = objectArray;
    oa->ia = ia;

    //
    // Initialize the index allocator object
    //
    return SA_STATUS_SUCCESS;
}

/*!
@brief
    Destroys a index allocator object.  It is safe to
    call this function even if the oaInit fails.
@param
    oa - Pointer to index allocator object.  
@return
    none
*/
void
sa_objalloc_destroy (PSA_OBJECT_ALLOCATOR oa) {
    ASSERT (oa);
    SA_NOT_USED (oa);

    // nop
    
    return;
}


/*!
@brief 
    Allocates an index value, removing it from the free list.
    If no values are available, IA_BAD_INDEX is returned.
@param 
    oa - pointer to index allocator                      
@return
    allocated index on success, IA_BAD_INDEX on failure
*/
PVOID
sa_objalloc_allocate_object (PSA_OBJECT_ALLOCATOR oa) 
{
    u32 index;

    ASSERT (oa);

    index = sa_indexalloc_allocate_index (oa->ia);

    if (index == SA_IA_BAD_INDEX) {
        return NULL;
    }

    return (u8*)oa->object_array + (oa->object_size * index);
}

/*
@brief
    Thread safe allocator.
*/
PVOID
sa_objalloc_interlocked_allocate_object (PSA_OBJECT_ALLOCATOR oa) {
    //
    // TBD
    //
    SA_NOT_USED (oa);

    return NULL;
}

/*!
@brief 
    Frees an object
@param 
    oa - pointer to index allocator
@param
    object - pointer to object to free
@return
    none
*/
void
sa_objalloc_free_object (PSA_OBJECT_ALLOCATOR oa, PVOID object) {
    u32 index = 0;
    
    ASSERT (oa);
    ASSERT (oa->object_size > 0);
    ASSERT (object >= oa->object_array &&
            object < (PVOID)((u8*)oa->object_array + oa->ia->num * oa->object_size));

    //
    // Compute index for this object
    //
    index = sa_objalloc_index (oa, object);

    //
    // Free the index
    //
    sa_indexalloc_free_index (oa->ia, index);
}

/*
@brief
    Thread safe allocator.
*/
void
sa_objalloc_interlocked_free_object (PSA_OBJECT_ALLOCATOR oa, PVOID object) {
    //
    // TBD
    //
    SA_NOT_USED (oa);
    SA_NOT_USED (object);
    return;
}

/*
@brief
    Returns the index of the given object.
*/
u32 
sa_objalloc_index (PSA_OBJECT_ALLOCATOR oa, PVOID object)
{
    u32 index;

    ASSERT (oa);
    ASSERT (object);

    //
    // Compute index for this object
    //
    index = (u32)
                (
                    ((u8*)object - (u8*)oa->object_array) / 
                    (oa->object_size)
                );

    ASSERT (index < oa->ia->num);

    return index;
}

