/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_output.c

@brief
   
CVS Header Information:
   $Header: $
*/

#include "sa_types.h"
#include "sa_defines.h"
#include "sa_output.h"
#include "sa_assert.h"
#include "sa_mem.h"

void SA_STDCALL sa_output_null_print (SA_OUTPUT *out, const c8* outstr) { 
    SA_NOT_USED (out);
    SA_NOT_USED (outstr); 
}

void SA_STDCALL sa_output_null_vprintf (SA_OUTPUT *out, const c8* format, va_list args) { 
    SA_NOT_USED (out);
    SA_NOT_USED (format); 
    SA_NOT_USED (args);
}

void SA_STDCALL sa_output_null_flush (SA_OUTPUT *out) {
    SA_NOT_USED (out);
}

#if defined(SA_KERNEL) && defined(SA_MICROSOFT)
    
    void SA_STDCALL sa_output_debugger_print (SA_OUTPUT *out, const c8* outstr) { 
        #if defined(SA_DEBUG) || defined(SA_ENABLE_RELEASE_TRACES)
            #if ( defined(NDIS60_MINIPORT) || defined(NDIS61_MINIPORT) || defined(NDIS620_MINIPORT))
                DbgPrintEx(DPFLTR_IHVNETWORK_ID, DPFLTR_ERROR_LEVEL, "%s", outstr); 
            #else
            //  Argument '1': using a read-only string as a writable string argument. 
            // This will attempt to write into static read-only memory and cause random crashes
            #pragma prefast (disable : 6298) 
                DbgPrint ("%s", outstr); // Incorrect Microsoft prototype... needs const 1st arg. 
            #pragma prefast (default : 6298) 
            #endif

        #endif
    }

#elif defined(SA_LINUXK)

    void SA_STDCALL sa_output_debugger_print (SA_OUTPUT *out, const c8* outstr) { printk (KERN_ERR "%s", outstr); }

#elif defined(SA_WIN32U)

    void SA_STDCALL sa_output_debugger_print (SA_OUTPUT *out, const c8* outstr) { OutputDebugString (outstr); }

#endif

#if defined(SA_USER)

    #if defined (SA_CONSOLE)
            
        void SA_STDCALL sa_output_stdout_print (SA_OUTPUT *out, const c8* outstr) { printf ("%s", outstr); }
        void SA_STDCALL sa_output_stdout_vprintf (SA_OUTPUT *out, const c8* format, va_list args) { vfprintf (stdout, format, args); }
        void SA_STDCALL sa_output_stdout_flush (SA_OUTPUT *out) { fflush (stdout); }

        void SA_STDCALL sa_output_stderr_print (SA_OUTPUT *out, const c8* outstr) { fprintf (stderr, "%s", outstr); }
        void SA_STDCALL sa_output_stderr_vprintf (SA_OUTPUT *out, const c8* format, va_list args) { vfprintf (stderr, format, args); }
        void SA_STDCALL sa_output_stderr_flush (SA_OUTPUT *out) { fflush (stderr); }

    #endif

    void SA_STDCALL sa_output_cfile_vprintf (SA_OUTPUT *out, const c8* format, va_list args) { 
        vfprintf (out->rsvd[0], format, args); }
    
    void SA_STDCALL sa_output_cfile_print (SA_OUTPUT *out, const c8* outstr) { 
        fprintf (out->rsvd[0], "%s", outstr); }
    
    void SA_STDCALL sa_output_cfile_flush (SA_OUTPUT *out) { 
        fflush (out->rsvd[0]); }

    void sa_output_cfile_create (FILE* handle, SA_OUTPUT *out) {
        out->print = sa_output_cfile_print;
        out->flush = sa_output_cfile_flush;
        out->vprintf = sa_output_cfile_vprintf;
        out->rsvd[0] = handle;
        out->rsvd[1] = NULL;
    }

#endif

// Initializers for the rsvd fields.
#define SA_OUTPUT_RSVD_FIELDS    {NULL, NULL}, {NULL, NULL}

SA_OUTPUT sa_output_stdnull  = {sa_output_null_print, sa_output_null_flush, sa_output_null_vprintf, SA_OUTPUT_RSVD_FIELDS};

#if defined(SA_KERNEL)
    
    SA_OUTPUT sa_output_stddbg   = {sa_output_debugger_print, NULL, NULL, SA_OUTPUT_RSVD_FIELDS};
    SA_OUTPUT sa_output_stdtrace = {sa_output_debugger_print, NULL, NULL, SA_OUTPUT_RSVD_FIELDS};
    SA_OUTPUT sa_output_stdout   = {sa_output_debugger_print, NULL, NULL, SA_OUTPUT_RSVD_FIELDS};
    SA_OUTPUT sa_output_stderr   = {sa_output_debugger_print, NULL, NULL, SA_OUTPUT_RSVD_FIELDS};

#elif defined(SA_WIN32U)

        SA_OUTPUT sa_output_stddbg   = {sa_output_debugger_print, NULL,                   NULL,                     SA_OUTPUT_RSVD_FIELDS};

        #if defined(SA_CONSOLE)
            SA_OUTPUT sa_output_stdout   = {sa_output_stdout_print,   sa_output_stdout_flush, sa_output_stdout_vprintf, SA_OUTPUT_RSVD_FIELDS};
            SA_OUTPUT sa_output_stdtrace = {sa_output_stdout_print,   sa_output_stdout_flush, sa_output_stdout_vprintf, SA_OUTPUT_RSVD_FIELDS};
            SA_OUTPUT sa_output_stderr   = {sa_output_stderr_print,   sa_output_stderr_flush, sa_output_stderr_vprintf, SA_OUTPUT_RSVD_FIELDS};
        #else        
            SA_OUTPUT sa_output_stdtrace = {sa_output_debugger_print, NULL,                   NULL,                     SA_OUTPUT_RSVD_FIELDS};
            SA_OUTPUT sa_output_stdout   = {sa_output_debugger_print, NULL,                   NULL,                     SA_OUTPUT_RSVD_FIELDS};
            SA_OUTPUT sa_output_stderr   = {sa_output_debugger_print, NULL,                   NULL,                     SA_OUTPUT_RSVD_FIELDS};
        #endif

#elif defined(SA_LINUXU) || defined(SA_DOS)
    
    // TODO -- Is there a debugger output for user mode linux?
    SA_OUTPUT sa_output_stddbg   = {sa_output_stdout_print, sa_output_stdout_flush, sa_output_stdout_vprintf, SA_OUTPUT_RSVD_FIELDS};
    SA_OUTPUT sa_output_stdout   = {sa_output_stdout_print, sa_output_stdout_flush, sa_output_stdout_vprintf, SA_OUTPUT_RSVD_FIELDS};
    SA_OUTPUT sa_output_stderr   = {sa_output_stderr_print, sa_output_stderr_flush, sa_output_stderr_vprintf, SA_OUTPUT_RSVD_FIELDS};
    SA_OUTPUT sa_output_stdtrace = {sa_output_stdout_print, sa_output_stdout_flush, sa_output_stdout_vprintf, SA_OUTPUT_RSVD_FIELDS};

#endif

// Define the array of pointers that the user may modify to change
// default output behavior.
SA_OUTPUT *sa_output_std[5] = {&sa_output_stdnull,
                               &sa_output_stdtrace,
                               &sa_output_stddbg,
                               &sa_output_stdout,
                               &sa_output_stderr};

void sa_output_print (SA_OUTPUT *output, const c8* outstr)
{
    SA_OUTPUT_ASSERT (output);
    output->print (output, outstr); // required function
}

void sa_output_flush (SA_OUTPUT *output) {
    SA_OUTPUT_ASSERT (output);
    if (output->flush) {
        output->flush (output);
    }
}

void sa_output_vprintf (SA_OUTPUT *output, const c8* format, va_list args)
{
    SA_OUTPUT_ASSERT (output);

    if (output->vprintf) {
        
        // If the method support a vprintf semantic, use it directly.
        output->vprintf (output, format, args);

    } else {
        // First try the easy method of printing to a local buffer.
        c8 buf[256];
        u32 len;

        len = SA_VSNPRINTF (buf, sizeof (buf), format, args);
        buf[sizeof(buf)-1] = 0;

        if (len < sizeof (buf)) {
            // good, just print the buf
            output->print (output, buf);
        } else {
            // revert to slow method of printing each arg separately
            // TODO -- Do it right...
            output->print (output, buf);
        }
    }
}


void sa_output_printf (SA_OUTPUT *output, const c8* format, ...)
{
    va_list args;
    va_start (args, format);

    SA_OUTPUT_ASSERT (output);

    if (output->vprintf) {
        
        // If the method support a vprintf semantic, use it directly.
        output->vprintf (output, format, args);

    } else {
        // First try the easy method of printing to a local buffer.
        c8 buf[256];
        u32 len;

        len = SA_VSNPRINTF (buf, sizeof (buf), format, args);
        buf[sizeof(buf)-1] = 0;

        if (len < sizeof (buf)) {
            // good, just print the buf
            output->print (output, buf);
        } else {
            // revert to slow method of printing each arg separately
            // TODO -- Do it right... this truncates the string...
            output->print (output, buf);
        }
    }

    va_end (args);
}

void sa_output_putc (SA_OUTPUT *output, c8 character)
{
    c8 c[2];
    c[0] = character;
    c[1] = 0; // terminate
    SA_OUTPUT_ASSERT (output);
    output->print (output, c); // required function
}


void SA_STDCALL sa_output_tee_print (SA_OUTPUT *out, const c8* outstr) {
    SA_OUTPUT_ASSERT (out);
    sa_output_print ((SA_OUTPUT*)(out->rsvd[0]), outstr);
    sa_output_print ((SA_OUTPUT*)(out->rsvd[1]), outstr);
}

void SA_STDCALL sa_output_tee_flush (SA_OUTPUT *out) {
    SA_OUTPUT_ASSERT (out);
    sa_output_flush ((SA_OUTPUT*)(out->rsvd[0]));
    sa_output_flush ((SA_OUTPUT*)(out->rsvd[1]));
}

void SA_STDCALL sa_output_tee_vprintf (SA_OUTPUT *out, const c8* format, va_list args) {
    SA_OUTPUT_ASSERT (out);
    sa_output_vprintf ((SA_OUTPUT*)(out->rsvd[0]), format, args);
    sa_output_vprintf ((SA_OUTPUT*)(out->rsvd[1]), format, args);
}

void sa_output_tee_create (SA_OUTPUT *one, SA_OUTPUT *two, SA_OUTPUT *out)
{
    SA_OUTPUT_ASSERT (one);
    SA_OUTPUT_ASSERT (two);
    ASSERT (out);
    ASSERT (out != one); // Try to detect any infinite recursive loops
    ASSERT (out != two); // Try to detect any infinite recursive loops
    out->print = sa_output_tee_print;
    out->flush = sa_output_tee_flush;
    out->vprintf = sa_output_tee_vprintf;
    out->rsvd[0] = one;
    out->rsvd[1] = two;
}

void sa_putchar (const c8 c)
{
    sa_output_putc (sa_stdout, c);
}

void sa_printf (const c8* format, ...) 
{
    va_list args;
    va_start (args, format);
    sa_output_vprintf (sa_stdout, format, args);
    va_end (args);
}

void sa_vprintf (const c8* format, va_list args)
{
    sa_output_vprintf (sa_stdout, format, args);
}

void sa_output_close (SA_OUTPUT *out) {
    ASSERT (out);
    *out = sa_output_stdnull;    
}

void sa_output_std_redirect (u32 sa_output_index, SA_OUTPUT *out)
{   
    ASSERT (sa_output_index < SA_NUMBER_OF (sa_output_std));
    SA_OUTPUT_ASSERT (out);
    sa_output_std[sa_output_index] = out;
}

//
// SA String Output
//

void SA_STDCALL sa_output_string_print (SA_OUTPUT *out, const c8* outstr) {

    u8* buf = (u8*)out->rsvd[0];
    u32 buflen = SA_PTR_TO_U32(out->rsvd[1]);

    if (buf && buflen > 1) {

        u32 slen = (u32)strlen (outstr);
    
        slen = MIN (slen, buflen-1);
    
        // Copy next string
        memcpy (buf, outstr, slen); 
    
        // terminate
        buf[slen] = 0;

        buf += slen;
        buflen -= slen;

        out->rsvd[0] = (PVOID)buf;
        out->rsvd[1] = SA_U32_TO_PTR (buflen);        
    }
}  

void SA_STDCALL sa_output_string_flush (SA_OUTPUT *out) {    
    // nop
}

void sa_output_string_create (SA_OUTPUT *out, PVOID str, u32 str_size) {
    out->print = sa_output_string_print;
    out->flush = sa_output_string_flush;
    out->vprintf = NULL;
    out->rsvd[0] = str;
    out->rsvd[1] = SA_U32_TO_PTR(str_size);        
}


