/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

//#define SA_MEM_DEBUG
#include "sa_runtime.h" // verify all header files
#define SA_ENDIAN_PERF_ITERATIONS (8*1024*1024)

#if !defined(SA_OMIT_TEST_CODE)

    //
    void
    sa_system_time_test (void) {
        u32 us_times[] = {1, 10, 100, 1000, 10000, 20000, 50000};
        u64 start, sleep_time, spin_time;
        u32 i;

        sa_sleep (1);
        sa_spin (1);

        for (i=0; i<SA_NUMBER_OF(us_times); i++) {
            start = sa_get_system_time ();
            sa_usleep (us_times[i]);
            sleep_time = sa_diff_system_time (start, sa_get_system_time ());

            start = sa_get_system_time ();
            sa_uspin (us_times[i]);
            spin_time = sa_diff_system_time (start, sa_get_system_time ());

            sa_trace (DL_INFO, 
                      "Interval request_interval_us:%6d sleep:%6d spin:%6d\n", 
                      us_times[i], 
                      (u32)sleep_time, 
                      (u32)spin_time);    
        }
    }

    // z, y, & z are arbitrary output values to avoid compiler optimizations
    // removing my test loops.
    void
    sa_addc_test (u32 *x, u32 *y, u32 *z, u32 max, u32 maxPow2) 
    {
        u64 slow;
        u32 i;        
        u32 a;
        u32 mask = maxPow2 - 1;

        *z = *y = *x = 0;

        a = 0;
        slow = sa_timestamp ();
        for (i=0; i<SA_ENDIAN_PERF_ITERATIONS; i++) {
            a = sa_addc_mod (a, i & mask, maxPow2); // use the result to avoid optimizations removing this instruction
            *x += a;
        }
        slow = sa_timestamp () - slow;
        sa_trace (DL_INFO, "addc_mod avg ticks/loop: %d", (u32)slow/SA_ENDIAN_PERF_ITERATIONS);


        a = 0;
        slow = sa_timestamp ();
        for (i=0; i<SA_ENDIAN_PERF_ITERATIONS; i++) {
            a = sa_addc_if (a, i & mask, maxPow2); // use the result to avoid optimizations removing this instruction
            *y += a;
        }
        slow = sa_timestamp () - slow;
        sa_trace (DL_INFO, "addc_if avg ticks/loop: %d", (u32)slow/SA_ENDIAN_PERF_ITERATIONS);


        a = 0;
        slow = sa_timestamp ();
        for (i=0; i<SA_ENDIAN_PERF_ITERATIONS; i++) {
            a = sa_addc_power2 (a, i & mask, maxPow2); // use the result to avoid optimizations removing this instruction
            *z += a;
        }
        slow = sa_timestamp () - slow;
        sa_trace (DL_INFO, "addc_power2 avg ticks/loop: %d", (u32)slow/SA_ENDIAN_PERF_ITERATIONS);

        SA_RELEASE_ASSERT (*x == *z);
        SA_RELEASE_ASSERT (*x == *y);
        

    }

    void
    sa_endian_perf16_test (u64 *x, u64 *y)
    {
        u64 fast, slow;
        u32 i;

        *x = 0; *y = 0;

        slow = sa_timestamp ();
        for (i=0; i<SA_ENDIAN_PERF_ITERATIONS; i++) {
            *x += SA_SWAP16 ((u16)i); // use the result to avoid optimizations removing this instruction
        }
        slow = sa_timestamp () - slow;
        sa_trace (DL_INFO, "u16 swap C-code avg ticks/loop: %d", (u32)slow/SA_ENDIAN_PERF_ITERATIONS);

        
        fast = sa_timestamp ();
        for (i=0; i<SA_ENDIAN_PERF_ITERATIONS; i++) {
            *y += sa_asm_swap16 ((u16)i); // use the result to avoid optimizations removing this instruction
        }
        fast = sa_timestamp () - fast;
        sa_trace (DL_INFO, "u16 swap ASM-code avg ticks/loop: %d", (u32)fast/SA_ENDIAN_PERF_ITERATIONS);
        
        //sa_trace (DL_INFO, "Fast u16 takes %d percent as much time as slow.  (lower is better.)", (u32)(fast * 100 / slow));
        SA_RELEASE_ASSERT (*x == *y);
    }

    void
    sa_endian_perf32_test (u64 *x, u64 *y)
    {
        u64 fast, slow;
        u32 i;
        u32 swap;

        *x = 0; 
        *y = 0;

        slow = sa_timestamp ();
        for (i=0; i<SA_ENDIAN_PERF_ITERATIONS; i++) {
            swap = i; // temp variable used to avoid optimization problem on amd64
            *x += (u32)SA_SWAP32 (swap);
        }
        slow = sa_timestamp () - slow;
        sa_trace (DL_INFO, "u32 swap C-code avg ticks/loop: %d", (u32)slow/SA_ENDIAN_PERF_ITERATIONS, slow);

        
        fast = sa_timestamp ();
        for (i=0; i<SA_ENDIAN_PERF_ITERATIONS; i++) {
            swap = i;
            *y += (u32)sa_asm_swap32 (swap);
        }
        fast = sa_timestamp () - fast;
        sa_trace (DL_INFO, "u32 swap ASM-code avg ticks/loop: %d", (u32)fast/SA_ENDIAN_PERF_ITERATIONS, fast);
        
        //sa_trace (DL_INFO, "Fast u32 takes %d as much time as slow.  (lower is better.)", (u32)(fast * 100 / slow));
        SA_RELEASE_ASSERT (*x == *y);
    }

    void
    sa_endian_perf64_test (u64 *x, u64 *y)
    {
        u64 fast, slow;
        u64 i;

        SA_C_ASSERT (sizeof (u64) == 8);

        *x = 0; *y = 0;

        slow = sa_timestamp ();
        for (i=0; i<SA_ENDIAN_PERF_ITERATIONS; i++) {
            *x += SA_SWAP64 (i);
        }
        slow = sa_timestamp () - slow;
        sa_trace (DL_INFO, "u64 swap C-code avg ticks/loop: %d", (u32)slow/SA_ENDIAN_PERF_ITERATIONS);
        
        fast = sa_timestamp ();
        for (i=0; i<SA_ENDIAN_PERF_ITERATIONS; i++) {
            *y += sa_asm_swap64 (i);
        }
        fast = sa_timestamp () - fast;
        sa_trace (DL_INFO, "u64 swap ASM-code avg ticks/loop: %d", (u32)fast/SA_ENDIAN_PERF_ITERATIONS);
        
        //sa_trace (DL_INFO, "Fast u64 takes %d percent as much time as slow.  (lower is better.)", (u32)(fast * 100 / slow));
        SA_RELEASE_ASSERT (*x == *y);
    }

    void
    sa_endian_test ()
    {    
        u64 x, y;
        u64 a, b;
        u64 u, v;

        SA_C_ASSERT (sizeof (u8) == 1);
        SA_C_ASSERT (sizeof (i8) == 1);
        SA_C_ASSERT (sizeof (u16) == 2);
        SA_C_ASSERT (sizeof (i16) == 2);
        SA_C_ASSERT (sizeof (u32) == 4);
        SA_C_ASSERT (sizeof (i32) == 4);
        SA_C_ASSERT (sizeof (u64) == 8);
        SA_C_ASSERT (sizeof (i64) == 8);

        SA_C_ASSERT (0x1234 == SA_SWAP16 (0x3412));
        SA_RELEASE_ASSERT (0x1234 == sa_asm_swap16 (0x3412));
        SA_C_ASSERT (0x12345678 == SA_SWAP32 (0x78563412));
        SA_RELEASE_ASSERT (0x12345678 == sa_asm_swap32 (0x78563412));
        SA_C_ASSERT (0x09efcdab12345678ULL == SA_SWAP64 (0x78563412abcdef09ULL));
        SA_RELEASE_ASSERT (0x09efcdab12345678ULL == sa_asm_swap64 (0x78563412abcdef09ULL));
        
        sa_endian_perf16_test (&a, &b);
        sa_endian_perf32_test (&x, &y);
        sa_endian_perf64_test (&u, &v);
    }

    void
    sa_trace_test (void)
    {
        u32 i;
        u8  xx[256];

        for (i=0; i<sizeof(xx); i++) xx[i] = (u8)i;

        sa_trace (DL_INFO, "default, u8:");
        sa_trace_buffer_options (DL_INFO, xx, sizeof (xx), 1, SA_TRACE_FLAG_DEFAULTS);
        sa_trace (DL_INFO, "default, u16:");
        sa_trace_buffer_options (DL_INFO, xx, sizeof (xx), 2, SA_TRACE_FLAG_DEFAULTS);
        sa_trace (DL_INFO, "default:");
        sa_trace_buffer (DL_INFO, xx, sizeof (xx)); // easy, standard function
        sa_trace (DL_INFO, "default, u64:");
        sa_trace_buffer_options (DL_INFO, xx, sizeof (xx), 8, SA_TRACE_FLAG_DEFAULTS);
        sa_trace (DL_INFO, "NO ascii, NO offset, u8:");
        sa_trace_buffer_options (DL_INFO, xx, sizeof (xx), 1, SA_TRACE_FLAG_NO_ASCII | SA_TRACE_FLAG_NO_OFFSET);
        sa_trace (DL_INFO, "NO ascii, swap, u16:");
        sa_trace_buffer_options (DL_INFO, xx, sizeof (xx), 2, SA_TRACE_FLAG_NO_ASCII | SA_TRACE_FLAG_ENDIAN_SWAP);
        sa_trace (DL_INFO, "NO offset, little-endian, u32:");
        sa_trace_buffer_options (DL_INFO, xx, sizeof (xx), 4, SA_TRACE_FLAG_NO_OFFSET | SA_TRACE_FLAG_LITTLE_ENDIAN);
        sa_trace (DL_INFO, "big-endian, u32:");
        sa_trace_buffer_options (DL_INFO, xx, sizeof (xx), 4, SA_TRACE_FLAG_BIG_ENDIAN);
        sa_trace (DL_INFO, "big-endian, u64:");
        sa_trace_buffer_options (DL_INFO, xx, sizeof (xx), 8, SA_TRACE_FLAG_BIG_ENDIAN);
    }

    void
    sa_int_test (u32 *x)
    {
        #define SA_POWER2_ITER (1024*1024)
        volatile u32 i, j;
        u64 slow;
        
        u32 offset;
        c8 results[256];

        // inputs.
        const u32 in[]               = {0,          1, 2, 3, 4, 5, 6, 7, 8,  9,  10, 11, 12, 13, 14, 15, 16, 0x7FFFFFFF, 0x40000000, 0xFFFFFFFF, 0x80000000};
        const u32 required_bit_out[] = {0,          1, 2, 2, 3, 3, 3, 3, 4,  4,  4,  4,  4,  4,  4,  4,  5,  31,         31,         32,         32};
        const u32 power2_out[]       = {0,          1, 1, 0, 1, 0, 0, 0, 1,  0,  0,  0,  0,  0,  0,  0,  1,  0,          1,          0,          1};
        const u32 log2_out[]         = {0xFFFFFFFF, 0, 1, 1, 2, 2, 2, 2, 3,  3,  3,  3,  3,  3,  3,  3,  4,  30,         30,         31,         31};
        const u32 gt_power2_out[]    = {1,          2, 4, 4, 8, 8, 8, 8, 16, 16, 16, 16, 16, 16, 16, 16, 32, 0x80000000, 0x80000000, 0xFFFFFFFF, 0xFFFFFFFF};
        const u32 ge_power2_out[]    = {1,          1, 2, 4, 4, 8, 8, 8, 8,  16, 16, 16, 16, 16, 16, 16, 16, 0x80000000, 0x40000000, 0xFFFFFFFF, 0x80000000};

        // outputs
        u32 out[SA_NUMBER_OF(in)]; // result values

        SA_C_ASSERT (SA_NUMBER_OF (in) == SA_NUMBER_OF(required_bit_out));
        SA_C_ASSERT (SA_NUMBER_OF (in) == SA_NUMBER_OF (power2_out));

    /*
    sar_w32 info  sa_num_bits_set_u8_loop                 . ticks/iter small:  27 medium:  41 big:  86  result:0x7088
    sar_w32 info  sa_num_bits_set_u8_add                  . ticks/iter small:  29 medium:  29 big:  29  result:0x7088
    sar_w32 info  sa_num_bits_set_u8_byte_table           . ticks/iter small:  23 medium:  22 big:  22  result:0x7088
    sar_w32 info  sa_num_bits_set_u8_nibble_table         . ticks/iter small:  26 medium:  25 big:  25  result:0x7088
    sar_w32 info  sa_num_bits_set_u32_loop                . ticks/iter small:  91 medium: 169 big: 355  result:0x1c031
    sar_w32 info  sa_num_bits_set_u32_add                 . ticks/iter small:  84 medium:  74 big:  74  result:0x1c031
    sar_w32 info  sa_num_bits_set_u32_byte_table          . ticks/iter small:  56 medium:  55 big:  55  result:0x1c031
    sar_w32 info  sa_num_bits_set_u32_nibble_table        . ticks/iter small:  41 medium:  40 big:  40  result:0x1c031
    sar_w32 info  sa_num_bits_set_u64_loop                . ticks/iter small: 245 medium: 481 big: 961  result:0x38071
    sar_w32 info  sa_num_bits_set_u64_add                 . ticks/iter small: 142 medium: 140 big: 140  result:0x38071
    sar_w32 info  sa_num_bits_set_u64_byte_table          . ticks/iter small: 112 medium: 108 big: 109  result:0x38071
    sar_w32 info  sa_num_bits_set_u64_nibble_table        . ticks/iter small:  72 medium:  71 big:  71  result:0x38071
    */

        {
            u64 times[3], mask[3];
            const u32 kNumInputs = 4096;
            u64 *inputs = (u64*)sa_malloc (kNumInputs * sizeof (u64), SA_MALLOC_NORMAL, SA_TAG ("inpu"));
            sa_rand_seed ((u32)sa_rdtsc (), SA_PTR_TO_U32 (out), (u32)(sa_rdtsc () >> 24));
            for (j=0; j<kNumInputs; j++) {
                
                u32 r = sa_rand ();
                u64 s = (u64)sa_rand () << 32 | r;
                u8 t = (u8)r;
                
                inputs[j] = s; // random for later            

                // Verify correctness of algorithms - init any static tables.
                SA_RELEASE_ASSERT (sa_num_bits_set_u32_count (r)        == sa_num_bits_set_u32_loop (r));
                SA_RELEASE_ASSERT (sa_num_bits_set_u32_fast_count (r)   == sa_num_bits_set_u32_loop (r));
                SA_RELEASE_ASSERT (sa_num_bits_set_u32_nifty (r)        == sa_num_bits_set_u32_loop (r));
                SA_RELEASE_ASSERT (sa_num_bits_set_u32_mit (r)          == sa_num_bits_set_u32_loop (r));
                SA_RELEASE_ASSERT (sa_num_bits_set_u32_add (r)          == sa_num_bits_set_u32_loop (r));
                SA_RELEASE_ASSERT (sa_num_bits_set_u32_byte_table (r)   == sa_num_bits_set_u32_loop (r));
                SA_RELEASE_ASSERT (sa_num_bits_set_u32_nibble_table (r) == sa_num_bits_set_u32_loop (r));

                SA_RELEASE_ASSERT (sa_num_bits_set_u64_count (s)        == sa_num_bits_set_u64_loop (s));
                SA_RELEASE_ASSERT (sa_num_bits_set_u64_fast_count (s)   == sa_num_bits_set_u64_loop (s));
                SA_RELEASE_ASSERT (sa_num_bits_set_u64_nifty (s)        == sa_num_bits_set_u64_loop (s));
                SA_RELEASE_ASSERT (sa_num_bits_set_u64_add (s)          == sa_num_bits_set_u64_loop (s));
                SA_RELEASE_ASSERT (sa_num_bits_set_u64_byte_table (s)   == sa_num_bits_set_u64_loop (s));
                SA_RELEASE_ASSERT (sa_num_bits_set_u64_nibble_table (s) == sa_num_bits_set_u64_loop (s));

                SA_RELEASE_ASSERT (sa_num_bits_set_u8_add (t)          == sa_num_bits_set_u8_loop (t));
                SA_RELEASE_ASSERT (sa_num_bits_set_u8_byte_table (t)   == sa_num_bits_set_u8_loop (t));
                SA_RELEASE_ASSERT (sa_num_bits_set_u8_nibble_table (t) == sa_num_bits_set_u8_loop (t));
            }

            #define SA_TEST_NUM_BITS(_func_, _type_)                                                   \
                mask[0] = (1ULL << (sizeof (_type_) * 2ULL)) - 1;                                      \
                mask[1] = (1ULL << (sizeof (_type_) * 4ULL)) - 1;                                      \
                mask[2] = 0xFFFFFFFFFFFFFFFFULL;                                                       \
                *x = 0;                                                                                \
                /*sa_trace (DL_ALWAYS, "Masks:0x%I64x 0x%I64x 0x%I64x", mask[0], mask[1], mask[2]);*/  \
                for (j=0; j<SA_NUMBER_OF (times); j++) {                                               \
                    times[j] = sa_timestamp ();                                                        \
                    for (i=0; i<kNumInputs; i++) {                                                     \
                        *x += _func_ ((_type_)(inputs[i] & mask[j]));                                  \
                    }                                                                                  \
                    times[j] = (sa_timestamp () - times[j]) / kNumInputs;                              \
                }                                                                                      \
                sa_trace (DL_INFO, "%-40s. ticks/iter small:%4d medium:%4d big:%4d  result:0x%x",      \
                        #_func_, (u32)times[0], (u32)times[1], (u32)times[2], *x);                   \


            SA_TEST_NUM_BITS (sa_num_bits_set_u8_loop,         u8);
            SA_TEST_NUM_BITS (sa_num_bits_set_u8_add,          u8);
            SA_TEST_NUM_BITS (sa_num_bits_set_u8_byte_table,   u8);
            SA_TEST_NUM_BITS (sa_num_bits_set_u8_nibble_table, u8);

            SA_TEST_NUM_BITS (sa_num_bits_set_u32_loop,         u32);
            SA_TEST_NUM_BITS (sa_num_bits_set_u32_count,        u32);
            SA_TEST_NUM_BITS (sa_num_bits_set_u32_fast_count,   u32);
            SA_TEST_NUM_BITS (sa_num_bits_set_u32_nifty,        u32);
            SA_TEST_NUM_BITS (sa_num_bits_set_u32_mit,          u32);
            SA_TEST_NUM_BITS (sa_num_bits_set_u32_add,          u32);
            SA_TEST_NUM_BITS (sa_num_bits_set_u32_byte_table,   u32);
            SA_TEST_NUM_BITS (sa_num_bits_set_u32_nibble_table, u32);

            SA_TEST_NUM_BITS (sa_num_bits_set_u64_loop,         u64);
            SA_TEST_NUM_BITS (sa_num_bits_set_u64_count,        u64);
            SA_TEST_NUM_BITS (sa_num_bits_set_u64_fast_count,   u64);
            SA_TEST_NUM_BITS (sa_num_bits_set_u64_nifty,        u64);
            SA_TEST_NUM_BITS (sa_num_bits_set_u64_mit,          u64);
            SA_TEST_NUM_BITS (sa_num_bits_set_u64_add,          u64);
            SA_TEST_NUM_BITS (sa_num_bits_set_u64_byte_table,   u64);
            SA_TEST_NUM_BITS (sa_num_bits_set_u64_nibble_table, u64);

            sa_free (inputs);
        }

        #define SA_PRINT_RESULTS(_func_)                                                                \
            offset = 0;                                                                                 \
            for(i=0; i<SA_NUMBER_OF(in); i++) {                                                         \
                out[i] = _func_ (in[i]);                                                                \
                ASSERT (offset < sizeof (results) - 32);                                                \
                offset += SA_SNPRINTF (&results[offset], sizeof(results)-1, " %u->%u", in[i], out[i]);  \
            }                                                                                           \
            sa_trace (DL_INFO, "%s results:", #_func_);                                                 \
            sa_trace (DL_INFO, "  %s", results);                                                        \



        // For curiosity, look at the relative times for various input numbers
        // to perform the binary search for the required number of bits.  I would
        // think that some inputs are much faster, since the search accidentally
        // guesses then immediately.
        for (j=0; j<32; j++) {
            slow = sa_timestamp ();
            for (i=0; i<SA_POWER2_ITER; i++) {
                *x += sa_required_bits (1<<j);
            }
            slow = sa_timestamp () - slow;
            sa_trace (DL_INFO, "%d ticks/iter for required_bits of 0x%x", (u32)(slow / SA_POWER2_ITER), (u32)(1<<j));
        }

        SA_PRINT_RESULTS (sa_required_bits64);
        SA_RELEASE_ASSERT (sa_memcmp (out, required_bit_out, sizeof (required_bit_out)) == 0);

        SA_PRINT_RESULTS (sa_required_bits);
        SA_RELEASE_ASSERT (sa_memcmp (out, required_bit_out, sizeof (required_bit_out)) == 0);

        SA_PRINT_RESULTS (sa_is_power2);
        SA_RELEASE_ASSERT (sa_memcmp (out, power2_out, sizeof (power2_out)) == 0);

        SA_PRINT_RESULTS (sa_is_power2_binary);
        SA_RELEASE_ASSERT (sa_memcmp (out, power2_out, sizeof (power2_out)) == 0);

        SA_PRINT_RESULTS (sa_log2);
        SA_RELEASE_ASSERT (sa_memcmp (out, log2_out, sizeof (log2_out)) == 0);

        SA_PRINT_RESULTS (sa_gt_power2);
        SA_RELEASE_ASSERT (sa_memcmp (out, gt_power2_out, sizeof (gt_power2_out)) == 0);

        SA_PRINT_RESULTS (sa_ge_power2);
        SA_RELEASE_ASSERT (sa_memcmp (out, ge_power2_out, sizeof (ge_power2_out)) == 0);
    }

    void
    sa_pages_spanned_test (void)
    {
        //SA_C_ASSERT (SA_PAGE_SIZE == 4096);
        //SA_C_ASSERT ((1<<SA_PAGE_SHIFT) == 4096);

        SA_C_ASSERT (SA_PAGES_SPANNED ((PVOID)0, 4096) == 1);
        //SA_C_ASSERT (SA_PAGES_SPANNED ((PVOID)1, 4096) == 2);
        //SA_C_ASSERT (SA_PAGES_SPANNED ((PVOID)4095, 4096) == 2);
        SA_C_ASSERT (SA_PAGES_SPANNED ((PVOID)4096, 4096) == 1);
        SA_C_ASSERT (SA_PAGES_SPANNED ((PVOID)4097, 4096) == 2);

    }

    void
    sa_mem_test (void)
    {
        u32 size = 8 * 1024 + 16;
        u32 i;
        u32 *va;


        // -- Physical Memory Allocation Test ---


        // Allocate physical memory
        PSA_SGL sgl = sa_physalloc (size, ~0ULL);

        SA_RELEASE_ASSERT (sa_sgl_get_byte_count (sgl) >= size);

        // Map, access the memory, and unmap it.
        va = sa_sgl_map (sgl, TRUE);
        SA_RELEASE_ASSERT (va);
        for (i=0; i<size/sizeof(u32); i++) { va[i] = i; }
        sa_sgl_unmap (sgl);

        // Map again and verify the values still exist.
        va = sa_sgl_map (sgl, TRUE);
        SA_RELEASE_ASSERT (va);
        for (i=0; i<size/sizeof(u32); i++) { SA_RELEASE_ASSERT (va[i] == i); }
        sa_sgl_unmap ( sgl);

        // free memory
        sa_physfree (sgl);



        // -- Virtual Memory Allocation Test ---


        // Allocate virtual buffer.
        va = sa_malloc (size, SA_MALLOC_NORMAL, 0);
        SA_RELEASE_ASSERT (va);

        // Allocate SGL for it.
        sgl = sa_sgl_allocate_for_va (va, size);
        SA_RELEASE_ASSERT (sgl);
        SA_RELEASE_ASSERT (sa_sgl_get_base_va (sgl) == va);
        sa_mem_debug_dump  ();

        sa_sgl_free (sgl);
        sa_free (va);
    }

    void
    sa_list_test (void)
    {        
        SA_LIST_ENTRY lentries[4];
        SA_LIST_ENTRY head, head2;
        SA_QUEUE_ENTRY qentries[4];
        SA_QUEUE_HEADER q, q2;
        SA_STACK_ENTRY sentries[4];
        SA_STACK_HEADER stack;

        sa_list_initialize (&head);
        sa_list_initialize (&head2);
        sa_list_insert_head (&head, &lentries[0]);
        ASSERT (sa_list_count (&head) == 1);
        sa_list_insert_head (&head, &lentries[1]);
        ASSERT (sa_list_count (&head) == 2);
        sa_list_insert_head (&head, &lentries[2]);
        ASSERT (sa_list_count (&head) == 3);
        sa_list_insert_head (&head, &lentries[3]);
        ASSERT (sa_list_count (&head) == 4);
        ASSERT (sa_list_head (&head) == &lentries[3]);
        ASSERT (sa_list_tail (&head) == &lentries[0]);
        sa_list_remove_entry (&lentries[2]);
        ASSERT (sa_list_count (&head) == 3);
        sa_list_remove_entry (&lentries[1]);
        ASSERT (sa_list_count (&head) == 2);
        sa_list_remove_entry (&lentries[3]);
        ASSERT (sa_list_count (&head) == 1);
        sa_list_remove_entry (&lentries[0]);
        ASSERT (sa_list_count (&head) == 0);
        ASSERT (sa_list_is_empty (&head));
        sa_list_insert_tail (&head, &lentries[0]);
        ASSERT (sa_list_count (&head) == 1);
        sa_list_insert_tail (&head, &lentries[1]);
        ASSERT (sa_list_count (&head) == 2);
        sa_list_insert_tail (&head, &lentries[2]);
        ASSERT (sa_list_count (&head) == 3);
        sa_list_insert_tail (&head, &lentries[3]);
        ASSERT (sa_list_count (&head) == 4);
        ASSERT (sa_list_remove_head (&head) == &lentries[0]);
        ASSERT (sa_list_remove_tail (&head) == &lentries[3]);
        ASSERT (sa_list_remove_tail (&head) == &lentries[2]);
        ASSERT (sa_list_remove_head (&head) == &lentries[1]);
        ASSERT (sa_list_is_empty (&head));
        sa_list_insert_tail (&head, &lentries[0]);
        ASSERT (sa_list_count (&head) == 1);
        sa_list_insert_head (&head, &lentries[1]);
        ASSERT (sa_list_count (&head) == 2);
        sa_list_insert_tail (&head2, &lentries[2]);
        ASSERT (sa_list_count (&head2) == 1);
        sa_list_insert_head (&head2, &lentries[3]);
        ASSERT (sa_list_count (&head2) == 2);
        sa_list_prepend (&head, &head2);
        ASSERT (sa_list_count (&head) == 4);
        ASSERT (sa_list_count (&head2) == 0);
        ASSERT (sa_list_head (&head) == &lentries[3]);
        ASSERT (sa_list_tail (&head) == &lentries[0]);
        sa_list_move (&head2, &head);
        ASSERT (sa_list_count (&head) == 0);
        ASSERT (sa_list_count (&head2) == 4);
        ASSERT (sa_list_head (&head2) == &lentries[3]);
        ASSERT (sa_list_tail (&head2) == &lentries[0]);

        sa_queue_initialize (&q);
        sa_queue_push (&q, &qentries[0]);
        sa_queue_push (&q, &qentries[1]);
        ASSERT (sa_queue_count (&q) == 2);
        sa_queue_move (&q2, &q);
        ASSERT (sa_queue_count (&q2) == 2);
        ASSERT (sa_queue_count (&q) == 0);
        sa_queue_move (&q, &q2);
        ASSERT (&qentries[0] == sa_queue_pop (&q));
        ASSERT (sa_queue_count (&q) == 1);
        sa_queue_unpop (&q, &qentries[2]);
        ASSERT (sa_queue_count (&q) == 2);
        ASSERT (&qentries[2] == sa_queue_pop (&q));
        ASSERT (sa_queue_count (&q) == 1);
        ASSERT (&qentries[1] == sa_queue_pop (&q));
        ASSERT (sa_queue_count (&q) == 0);

        sa_stack_initialize (&stack);
        sa_stack_atomic_push (&stack, &sentries[0]);
        sa_stack_push (&stack, &sentries[1]);
        sa_stack_atomic_push (&stack, &sentries[2]);
        ASSERT (sa_stack_count (&stack) == 3);
        ASSERT (&sentries[2] == sa_stack_pop (&stack));
        ASSERT (sa_stack_count (&stack) == 2);
        ASSERT (&sentries[1] == sa_stack_atomic_pop (&stack));
        ASSERT (sa_stack_count (&stack) == 1);
        ASSERT (&sentries[0] == sa_stack_atomic_pop (&stack));
        ASSERT (sa_stack_count (&stack) == 0);
    }

    void
    sa_sync_perf_test (u32 *x, SA_SPINLOCK *l, PVOID *z)
    {
        u32 i;
        u64 start;
        SA_IRQ irql;
        PVOID newZ = NULL;

        #define SA_SYNC_PERF_LOOPS (1000000)

        start = sa_timestamp ();
        for (i=0; i<SA_SYNC_PERF_LOOPS; i++) {
            (*((volatile u32*)x))++;
        }
        start = sa_timestamp () - start;

        sa_trace (DL_INFO, 
                "++. ticks/add:%d\n", 
                (u32)(start / SA_SYNC_PERF_LOOPS));    


        start = sa_timestamp ();
        for (i=0; i<SA_SYNC_PERF_LOOPS; i++) {
            sa_atomic_increment (x);
        }
        start = sa_timestamp () - start;

        sa_trace (DL_INFO, 
                "atomic_inc. ticks/add:%d\n", 
                (u32)(start / SA_SYNC_PERF_LOOPS));    


        start = sa_timestamp ();
        for (i=0; i<SA_SYNC_PERF_LOOPS; i++) {
            sa_atomic_add (1, x);    
        }
        start = sa_timestamp () - start;

        sa_trace (DL_INFO, 
                "atomic_add. ticks/add:%d\n",                 
                (u32)(start / SA_SYNC_PERF_LOOPS));    

        start = sa_timestamp ();
        for (i=0; i<SA_SYNC_PERF_LOOPS; i++) {
            sa_acquire_spinlock (l, &irql);
            *x += 1;
            sa_release_spinlock (l, &irql);
        }
        start = sa_timestamp () - start;

        sa_trace (DL_INFO, 
                "spinlock. ticks/add:%d\n", 
                (u32)(start / SA_SYNC_PERF_LOOPS));   


        start = sa_timestamp ();
        for (i=0; i<SA_SYNC_PERF_LOOPS; i++) {
            newZ = SA_U64_TO_PTR (SA_PTR_TO_U64 (newZ) + 1);
            sa_atomic_compare_exchange_ptr (z, newZ, *z);
        }
        start = sa_timestamp () - start;

        sa_trace (DL_INFO, 
                "compare_exchange. ticks/add:%d\n", 
                (u32)(start / SA_SYNC_PERF_LOOPS));   

    }

    #if (defined(SA_LINUXU) && !defined(SA_SIM)) || defined(SA_WIN32U)

        void SA_CDECL 
        sa_thread_test_routine (
            SA_THREAD *thread,
            PVOID context)
        {
            PSA_EVENT event = (PSA_EVENT)context;

            sa_trace (DL_INFO, "Test Thread Started.");
            sa_wait_for_single_object (event, SA_INFINITE);
            sa_trace (DL_INFO, "Signaled! Sleeping 2 seconds before resetting event.");
            sa_sleep (1000 * 2000);
            sa_event_set (event);
            sa_trace (DL_INFO, "Sleeping 2 seconds before exiting thread.");
            sa_sleep (1000 * 2000);
            sa_trace (DL_INFO, "Test thread exiting.");
            sa_thread_exit (thread, 123);
        }

        void
        sa_thread_test ()
        {
            SA_THREAD thread;
            u32 code = 0;
            SA_STATUS status;
            SA_EVENT event;

            sa_event_init (&event, FALSE);

            sa_trace (DL_INFO, "Thread test start.");
            status = sa_thread_create (
                sa_thread_test_routine,
                &event, // SA_U32_TO_PTR (0x1234), // context
                NULL, // parameters
                &thread);
            ASSERT (status == SA_SUCCESS);
    
            sa_trace (DL_INFO, "Sleep 2 seconds before setting event.");
            sa_sleep (1000 * 2000);
            sa_event_set (&event);
            sa_wait_for_single_object (&event, SA_INFINITE);

            sa_trace (DL_INFO, "Signaled! Call sa_thread_join.");
            status = sa_thread_join (
                &thread,
                &code);
            ASSERT (status == SA_SUCCESS);
            ASSERT (code == 123);

            sa_trace (DL_INFO, "Thread test complete.");
        }

    #else

        // not supported in other environments
        #define sa_thread_test()

    #endif


    void
    sa_runtime_test (void)
    {
        u32 x = sa_tag ("nope");
        u32 y = 0;
        u32 z = 0;
        u32 a = 0;
        SA_SPINLOCK lock;
        PVOID v = NULL;
        
        SA_RELEASE_ASSERT (x == 0x65706F6E);

        sa_toeplitz_test ();

        // Do some performance tests.
        sa_init_spinlock (&lock);
        sa_sync_perf_test (&a, &lock, &v);
        
        // Execute individual test functions.    
        sa_log_test ();
        sa_tlog_test ();
        sa_trace_test ();
        sa_ringalloc_test ();
        sa_indexalloc_test ();
        sa_print_bit_tables ();

        sa_crc32_test ();
        sa_chap_test ();
        sa_endian_test ();
        sa_addc_test (&x, &y, &z, 1234567, 1024*1024);
        sa_addc_test (&x, &y, &z, 0xFFFFFFFF, 2048);
        sa_int_test (&x);   
        sa_pages_spanned_test ();
        sa_mem_test ();
        sa_list_test ();
        sa_system_time_test ();        

        SA_TRACE_NN (DL_INFO, "This should be on");
        SA_TRACE_NN (DL_INFO, " one (1) line.\n");
        SA_TRACE_NN (DL_INFO, "And this should be\non 2 lines.\n");    
        SA_TRACE_NN (DL_INFO, "This should be a line.");
        SA_TRACE (DL_INFO, "And so should this.\n\n");
        SA_TRACE (DL_INFO, "And this.");

        sa_thread_test ();
        sa_mem_debug_dump  ();
    }

#endif // SA_OMIT_TEST_CODE


