/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_sync.h

@brief

    Synchronization objects.
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/system/linuxk/sa_sync_linuxk.c,v 1.8 2006/03/10 20:22:48 menz Exp $
*/

#include <asm/atomic.h>
#include <linux/version.h>
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0))
#include <linux/interrupt.h>
#endif

#include "sa_types.h"
#include "sa_sync.h"
#include "sa_assert.h"

// Returns resulting value.
u32 sa_atomic_decrement (u32* value)
{
    SA_C_ASSERT (sizeof(u32) == sizeof (atomic_t));
    atomic_dec ((atomic_t*)value);
    return *value;
}

// Returns resulting value.
u32 sa_atomic_increment(u32* value )
{
    SA_C_ASSERT (sizeof(u32) == sizeof (atomic_t));
    atomic_inc ((atomic_t*)value);
    return *value;
}

// Atomically does this: *v += i;
void sa_atomic_add (u32 i, u32 *v) {
    SA_C_ASSERT (sizeof(u32) == sizeof (atomic_t));
    atomic_add (i, (atomic_t*)v);
}

// Atomically does this: *v -= i;
void sa_atomic_sub (u32 i, u32 *v) {
    SA_C_ASSERT (sizeof(u32) == sizeof (atomic_t));
    atomic_sub (i, (atomic_t*)v);
}


boolean sa_can_block( void ) 
{ 
    // TODO - check this stuff somehow, if possible.
    // You can block if:
    // 1. You are in user context. 
    // 2. You do not own any spinlocks. 
    // 3. You have interrupts enabled 

    SA_BREAKPOINT_MSG ("Not Implemented.");
    return FALSE; 
}


boolean
sa_is_passive ()
{
    SA_BREAKPOINT_MSG ("Not Implemented.");
    return TRUE;
}

boolean
sa_is_dispatch () 
{
    SA_BREAKPOINT_MSG ("Not Implemented.");
    return FALSE;
}

boolean 
sa_is_not_dirql ()
{
    SA_BREAKPOINT_MSG ("Not Implemented.");
    return TRUE;
}

#if !defined(CONFIG_X86_CMPXCHG) && !defined(cmpxchg)
PVOID
cmpxchg (
	PVOID *ptr,
	PVOID val,
	PVOID newval)
{
	PVOID		curval;
	unsigned long	flags;

	local_irq_save(flags);

	curval = *ptr;

	if (curval == val)
		*ptr = newval;

	local_irq_restore(flags);

	return curval;
}
#endif


PVOID
sa_atomic_compare_exchange_ptr (
    OUT PVOID *destination, 
    IN  PVOID exchange,
    IN  PVOID compare)
{
    // TODO - verify supported on platforms we need.  64-bit platform support?

    // Returns initial value of (*destination).
    // Exchange only occurs if (*destination)==compare
    return (PVOID)cmpxchg (destination, 
                    compare,   // old
                    exchange); // new
}

/*!
@brief
    Initializes a spinlock
@param    
    SpinLock           - Pointer to a spinlock
@return
@note
*/
void 
sa_init_spinlock_real( 
    PSA_REAL_SPINLOCK spinlock )
{
    spin_lock_init (spinlock);
}

/*
    FIX for Bug 7841
    In iSCSI Driver the IRQs are always disabled
    before any BECLIB call
    But in NIC Driver, the IRQs are never disabled 
    before any BECLIB call
    Using this difference for different definitions
    of sa_acquire_spinlock_real() and 
    sa_release_spinlock_real()
*/

/*!

@brief
    Acquires a spinlock
    
@param    
    SpinLock        - Pointer to a spinlock
    
@param    
    Irq             - Recieves the current IRQ level
    
@return

@note

*/
void 
sa_acquire_spinlock_real( 
    IN      PSA_REAL_SPINLOCK         spinlock, 
    OUT     PSA_IRQ              irq )
{
#if ((LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)) || (defined (__VMKERNEL_MODULE__)))
    if (irqs_disabled()) {
        // iSCSI Call
        spin_lock_irqsave(spinlock, *irq);
    }
    else {
        // NIC call
        spin_lock_bh(spinlock);
    }
#else
        spin_lock_irqsave(spinlock, *irq);
#endif
}

void 
sa_acquire_spinlock_noirq_real( 
    IN      PSA_REAL_SPINLOCK         spinlock)
{
    spin_lock (spinlock);
}

/*!

@brief
    Releases a spinlock
    
@param    
    SpinLock        - Pointer to a spinlock
    
@param    
    Irq             - The IRQ level when the spinlock was acquired
    
@return

@note

*/
void 
sa_release_spinlock_real( 
    PSA_REAL_SPINLOCK         spinlock, 
    PSA_IRQ              irq )
{
#if ((LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)) || (defined (__VMKERNEL_MODULE__)))
    if (irqs_disabled()) {
        // iSCSI Call
        spin_unlock_irqrestore(spinlock, *irq);
    }
    else {
        // NIC call
        spin_unlock_bh(spinlock);
    }
#else
        spin_unlock_irqrestore(spinlock, *irq);
#endif
}

void 
sa_release_spinlock_noirq_real( 
    PSA_REAL_SPINLOCK spinlock)
{
    spin_unlock(spinlock);
}


boolean
sa_is_spin_lock_held_real (PSA_REAL_SPINLOCK lock)
{    
    return spin_is_locked (lock);
}

boolean
sa_is_spinlock_not_held_real (PSA_REAL_SPINLOCK lock)
{
    return !(spin_is_locked (lock));
}


/*!

@brief
    Initializes a FastMutex 
    
@param    
    FastMutex           - Pointer to a FastMutex
    
@return
   

@note

*/
void 
sa_initialize_fast_mutex(
    PSA_FAST_MUTEX       fast_mutex )
{
     sema_init (fast_mutex, 1);
}

/*!

@brief
    Acquires a FastMutex
    
@param    
    FastMutex        - Pointer to a FastMutex
    
@return

@note
    Current thread must be in a blockable state

*/
void 
sa_acquire_fast_mutex( 
    PSA_FAST_MUTEX       fast_mutex )
{
    down (fast_mutex);       
}


/*!

@brief
    Releases a FastMutex
    
@param    
    FastMutex        - Pointer to a FastMutex
    
@return

@note

*/
void 
sa_release_fast_mutex( 
    PSA_FAST_MUTEX       fast_mutex )
{
    up (fast_mutex);
}



//
// You can skip reading this section
//
// These routines are really only used in the MPU code. If the MPU is not
// being used (as will be the case for device bringup) these functions need
// not be implemented correctly but must have a body in order to link.
// Implementing as SA_BREAKPOINT_MSG("Not implemented"); is fine.
//
// The semi-high probability that all the MPU code may need to be rewritten to
// accomidate the ARM characteristics may render all these functions unnecessary.
//


//
// Initializes a watiable event. The event can either 1) wake all threads blocked
// on it when signaled, or 2) wake all threads
// 
// Replaces KeInitializeEvent
//
void
sa_event_init (
    PSA_EVENT  event,
    boolean    wake_all )
{
    sema_init (event, 1);
}

// 
// This function signals the event unblocking waiting threads.  The number of
// threads unblocked depends on the WakeAll argument that was passed during
// sa_EventInit()
//
// Replaces KeSetEvent
//
void
sa_event_set (
    PSA_EVENT event)
{
    up (event);
}

// 
// This function signals the event unblocking waiting threads.  The number of
// threads unblocked depends on the WakeAll argument that was passed during
// sa_EventInit()
//
// A timeout of SA_INFINITE means wait forever.  A timeout of 0 means check to see
// if the object is signaled, and then return but don't block.
//
// Replaces KeWaitForSingleObject
//
// Return values include
//      STATUS_SUCCESS - Object was signaled
//      STATUS_ALERTED - The wait was completed because of an alert/signal to the thread. 
//      STATUS_TIMEOUT - The object was not signaled
//
SA_STATUS
sa_wait_for_single_object (
    PVOID  object,
    u32    timeout_in_milliseconds )
{    
    // BUG - handle the timeout 
    ASSERT (timeout_in_milliseconds == SA_INFINITE);

    if (down_interruptible (object) == EINTR) {
        return SA_STATUS_ALERTED;
    } else {
        return SA_SUCCESS;
    }
}



