/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
#include "sa_types.h"
#include "sa_trace.h"
#include "sa_endian.h"
#include "sa_assert.h"
#include "sa_std.h"
#include "sa_output.h"

// All tracing if only defined for release builds, unless explicitly required by SA makefile/sources
//#if defined(SA_DEBUG) || defined(_SA_TRACE_COMPILE)

    // Current debug level bitmask
    STATIC u32 sa_debug_level = 0xFFFFFFFF;

    #if defined(SA_LINUX_USER) || defined(SA_DOS) || defined(SA_WIN_USER)
    STATIC FILE* g_trace_file = NULL;
    #endif

    // Strings corresponding to debug levels
    STATIC const c8* sa_dbg_level_strings[33] =
    {
    "err",   // 0x00000001
    "warn",  // 0x00000002
    "note",  // 0x00000004
    "info",  // 0x00000008

    "verb",  // 0x00000010
    "entry", // 0x00000020
    "exit",  // 0x00000040
    "??",    // 0x00000080
       
    NULL,    // 0x00000100
    NULL,    // 0x00000200
    NULL,    // 0x00000400
    NULL,    // 0x00000800
       
    NULL,    // 0x00001000
    NULL,    // 0x00002000
    NULL,    // 0x00004000
    NULL,    // 0x00008000
       
    NULL,    // 0x00010000
    NULL,    // 0x00020000
    NULL,    // 0x00040000
    NULL,    // 0x00080000
       
    NULL,    // 0x00100000
    NULL,    // 0x00200000
    NULL,    // 0x00400000
    NULL,    // 0x00800000
       
    NULL,    // 0x01000000
    NULL,    // 0x02000000 
    NULL,    // 0x04000000
    NULL,    // 0x08000000
       
    NULL,    // 0x10000000
    NULL,    // 0x20000000
    NULL,    // 0x40000000
    NULL,    // 0x80000000

    " ** " // 0
    };


    STATIC u32 sa_trace_column_number = 0;

    // Maps a debug level to an index for the sa_dbg_level_strings array.
    STATIC INLINE u32 sa_debug_level_to_index (u32 level) {
        u32 index  = 0;
        u32 i;

        if (level == DL_ALWAYS)
            return SA_NUMBER_OF (sa_dbg_level_strings) - 1;
            
        for ( i=0; i<32; ++i )
        {
            if ( (level & (1<<i))  )
            {
                index  = i;
                break;
            }
        }

        ASSERT (index < SA_NUMBER_OF (sa_dbg_level_strings)-1);

        return index;
    }

    // Sets an entry in the sa_dbg_level_strings array.
    void sa_trace_set_debug_level_string (u32 level, const c8* str)
    {
        u32 index = sa_debug_level_to_index (level);
        sa_dbg_level_strings[index] = str;
    }

    const c8* sa_trace_get_debug_level_string (u32 level)
    {
        u32 index = sa_debug_level_to_index (level);    
        return sa_dbg_level_strings[index];
    }

    u32 sa_trace_set_level (u32 level ) 
    { 
        u32 existing = sa_debug_level;
        sa_debug_level = level; 
        return existing;
    }


    u32 sa_trace_get_level ()            
    { 
        return sa_debug_level; 
    }

    SA_STATUS sa_trace_start_file (const c8* filename)
    {
        SA_NOT_USED (filename) ;

        #if  defined(SA_LINUX_USER) || defined(SA_DOS)
        {
            FILE *out;
            out = fopen (filename, "wt");
            if (!out) {
                SA_TRACE (DL_ERR, "Failed to open trace file: %s", filename);
                return SA_STATUS_INVALID_PARAMETER;
            }

            g_trace_file = out;
            return SA_STATUS_SUCCESS;
        }
        #elif defined(SA_WIN_USER)
            #if defined(SA_TARGET_WIN2008) 
            {
                FILE *out ;
                errno_t err ;
                err = fopen_s(&out, filename, "wt") ;
                if( err != 0 ) {
                    SA_TRACE (DL_ERR, "Failed to open trace file: %s", filename) ;
                    return SA_STATUS_INVALID_PARAMETER ;
                }

                g_trace_file = out ;
                return SA_STATUS_SUCCESS ;
            }
            #else
            {
                FILE *out;
                out = fopen (filename, "wt");
                if (!out) {
                    SA_TRACE (DL_ERR, "Failed to open trace file: %s", filename);
                    return SA_STATUS_INVALID_PARAMETER;
                }

                g_trace_file = out;
                return SA_STATUS_SUCCESS;
            }
            #endif
        #endif

        return SA_STATUS_NOT_SUPPORTED;
    }


    void sa_trace_end_file ()
    {
        #if defined(SA_LINUX_USER) || defined(SA_DOS) || defined(SA_WIN_USER)
            if (g_trace_file) {
                FILE *out = g_trace_file;
                g_trace_file = NULL;

                // BUG - Synchronization problem here if multi-threads writing to file!

                fclose (out);
            }
        #endif
    }


    //
    // Prints the given message using debug level "level".  This adds a prefix
    // to the debug print with the module name and debug level string.
    // The _SA_MODULE_NAME macro should be defined outside of this function based 
    // on the current module being compiled.
    //
    void sa_trace_vlist ( u32 level, const c8* module_name, const c8* format, va_list args )
    {
        c8 out_buff[256];
        u32 len, prelen;
        const c8* dbg_str = sa_trace_get_debug_level_string (level);

        // Check if current print is masked.
        if ((level != DL_ALWAYS) && 
            ((sa_trace_get_level () & level) == 0)) {
            return;
        }

        if (sa_trace_column_number > 0) {
            sa_output_print (sa_stdtrace, "\n");
            sa_trace_column_number = 0;
        }        

        prelen = SA_SNPRINTF (out_buff, sizeof (out_buff) - 1, "%-7s %-6s", module_name, (dbg_str ? dbg_str : "?!?"));
        prelen = MIN (prelen, sizeof (out_buff) - 1);

        len = SA_VSNPRINTF (&out_buff[prelen], sizeof (out_buff) - prelen - sizeof ("\n"), format, args) ;
        len = MIN (len, sizeof (out_buff) - prelen - sizeof ("\n"));

        if (out_buff[prelen + len - 1] != '\n') {
            out_buff[prelen + len] = '\n';
            out_buff[prelen + len + 1] = 0;
        }

        sa_output_print (sa_stdtrace, out_buff);
        sa_output_flush (sa_stdtrace);
    }


    void sa_trace_vlist_no_newline ( u32 level, const c8* module_name, const c8* format, va_list args )
    {
        c8 line_prefix[32];
        c8 out_buff[256];
        i32 line_start;
        i32 len, prelen;
        i32 i;
        const c8* dbg_str = sa_trace_get_debug_level_string (level);       

        // Check if current print is masked.
        if ((level != DL_ALWAYS) && ((sa_trace_get_level () & level) == 0)) {
            return;
        }

        // prefix
        prelen = SA_SNPRINTF (line_prefix, sizeof (line_prefix) - 1, "%-7s %-6s", module_name, (dbg_str ? dbg_str : "?!?"));
        line_prefix[sizeof(line_prefix)-1] = 0;

        // string
        len = SA_VSNPRINTF (out_buff, sizeof (out_buff), format, args) ;
        len = MIN (len, sizeof (out_buff));

        // Find any newlines and print the prefix appropriately.
        line_start = 0;
        for (i=0; i<len; i++) {

            if (out_buff[i] == '\n') {
                c8 next = out_buff[i+1];

                if (sa_trace_column_number == 0) {
                    sa_output_print (sa_stdtrace, line_prefix);
                }

                out_buff[i+1] = 0; // terminate
                sa_output_print (sa_stdtrace, &out_buff[line_start]);
                out_buff[i+1] = next; // restore

                line_start = i+1;
                sa_trace_column_number = 0;
            }
        }

        // Print the last partial line.
        if (line_start != len) {
            if (sa_trace_column_number == 0) {
                sa_output_print (sa_stdtrace, line_prefix);
            }
            sa_output_print (sa_stdtrace, out_buff);
            sa_trace_column_number++;
        }
    }

    void sa_trace_vargs_no_newline ( u32 level, const c8* module_name, const c8* format, ... )
    {    
        va_list args;
        va_start (args, format) ;
        sa_trace_vlist_no_newline (level, module_name, format, args); 
        va_end (args);
    }

    void sa_trace_vargs ( u32 level, const c8* module_name, const c8* format, ... )
    {    
        va_list args;
        va_start (args, format) ;
        sa_trace_vlist (level, module_name, format, args); 
        va_end (args);
    }


    // Returns # c8's printed.
    u32 sa_print_hex (c8* line, u32 width, boolean endianSwap, u8* data)
    {
        c8 *to = line;

        if (endianSwap) {
            switch (width) {
            // special case to back bytes together w/o spaces
            case 0: 
                return sprintf (to, "%02x", (u32)SA_GET8 (data));
                break;
            case 1:
                return sprintf (to, "%02x ", (u32)SA_GET8 (data));
                break;
            case 2:     
                return sprintf (to, "%04x ", (u32)sa_swap16 (SA_GET16 (data)));
                break;
            case 4:
                return sprintf (to, "%08x ", sa_swap32 (SA_GET32 (data)));
                break;
            case 8:
                #if defined(SA_MICROSOFT)
                    return sprintf (to, "%016I64x ", sa_swap64 (SA_GET64 (data)));
                #else
                    return sprintf (to, "%016llx ", sa_swap64 (SA_GET64 (data)));
                #endif
                
                break;
            default:
                SA_BREAKPOINT_MSG ("Illegal width in sa_print_hex");
            }
        } else {
            switch (width) {
            // special case to back bytes together w/o spaces
            case 0: 
                return sprintf (to, "%02x", (u32)SA_GET8 (data));
                break;
            case 1:
                return sprintf (to, "%02x ", (u32)SA_GET8 (data));
                break;
            case 2:     
                return sprintf (to, "%04x ", (u32)SA_GET16 (data));
                break;
            case 4:
                return sprintf (to, "%08x ", SA_GET32 (data));
                break;
            case 8:
                #if defined(SA_MICROSOFT)
                    return sprintf (to, "%016I64x ", SA_GET64 (data));
                #else
                    return sprintf (to, "%016llx ", SA_GET64 (data));
                #endif
                break;
            default:
                SA_BREAKPOINT_MSG ("Illegal width in sa_print_hex");
            }
        }
        return 0;
    }

    typedef struct _SA_TRACE_OUTPUT {
        SA_OUTPUT base; // must be first
        u32 magic; 
        u32 level;
        const c8* module_name;
    } SA_TRACE_OUTPUT, *PSA_TRACE_OUTPUT;

    #define SA_TRACE_OUTPUT_MAGIC (0xabcd1589)

    void SA_STDCALL sa_trace_print_internal (SA_OUTPUT *output, const c8* str) {
        SA_TRACE_OUTPUT *out = (SA_TRACE_OUTPUT*)output;
        ASSERT (out->magic == SA_TRACE_OUTPUT_MAGIC);
        sa_trace_vargs (out->level, out->module_name, "%s", str);
    }

    void sa_trace_buffer_internal (
        u32 level, 
        const c8* module_name, 
        u8* data,              // Array of data
        u32 size,              // Length of data in bytes to print
        u32 width,             // Width of hex numbers, 1, 2 or 4 bytes
        u32 flags)
    {
        SA_TRACE_OUTPUT out;
        out.base.print = sa_trace_print_internal;
        out.base.flush = NULL;
        out.base.vprintf = NULL;
        out.magic = SA_TRACE_OUTPUT_MAGIC;
        out.level = level;
        out.module_name = module_name;        

        sa_buffer_print ((SA_OUTPUT*)&out, data, size, width, flags);
    }

    void sa_buffer_print (
        SA_OUTPUT *out,        
        u8* data,              // Array of data
        u32 size,              // Length of data in bytes to print
        u32 width,             // Width of hex numbers, 1, 2 or 4 bytes
        u32 flags)
    {
        u32 lines, i, j;
        u32 bytesWritten = 0;
        u8 *startOfLineData;
        u32 startOfLineBytesWritten;
        u32 c8sPrinted;
        const u32 kBytesPerLine = 16; // probably shouldn't change this...  
        c8 line[120];
        boolean endianSwap = ((flags & SA_TRACE_FLAG_ENDIAN_SWAP)) ||
                            (sa_is_host_little_endian () && (flags & SA_TRACE_FLAG_BIG_ENDIAN)) ||
                            (sa_is_host_big_endian () && (flags & SA_TRACE_FLAG_LITTLE_ENDIAN));
        
        lines = size / kBytesPerLine;
        if (size % kBytesPerLine > 0) lines++;
        
        for (i = 0; i < lines; i++) {
            startOfLineData = data;
            startOfLineBytesWritten = bytesWritten;
            c8sPrinted = 0;

            if (!(flags & SA_TRACE_FLAG_NO_OFFSET)) {
                c8sPrinted += sprintf (&line[c8sPrinted], "[%03x] ", bytesWritten);
            }

            // Write full "width" length values
            for (j = 0; j < kBytesPerLine && size - bytesWritten >= width; j+=width) {
                if (j == kBytesPerLine/2) c8sPrinted += sprintf (&line[c8sPrinted], "  "); // add space in middle
                c8sPrinted += sa_print_hex (&line[c8sPrinted], width, endianSwap, data);
                bytesWritten += width;
                data += width;
            }

            // Write any remaining byte values
            for (; j < kBytesPerLine && size - bytesWritten >= 1; j+=1) {
                if (j == kBytesPerLine/2) c8sPrinted += sprintf (&line[c8sPrinted], "  "); // add space in middle
                c8sPrinted += sa_print_hex (&line[c8sPrinted], 0, endianSwap, data);
                bytesWritten += 1;
                data += 1;
            }

            if (!(flags & SA_TRACE_FLAG_NO_ASCII)) {
                // Move to the right...
                while (c8sPrinted < kBytesPerLine*2 + kBytesPerLine/width + 9) {
                    c8sPrinted += sprintf (&line[c8sPrinted], " ");
                }

                // Now print the ASCII representation...
                data = startOfLineData;
                bytesWritten = startOfLineBytesWritten;
                for (j = 0; j < kBytesPerLine && size - bytesWritten >= 1; j+=1) {
                    if (j == kBytesPerLine/2) c8sPrinted += sprintf (&line[c8sPrinted], " "); // add space in middle
                    if (*data >= 32 && *data <= 126) {
                        c8sPrinted += sprintf (&line[c8sPrinted], "%c", *data);
                    } else {
                        c8sPrinted += sprintf (&line[c8sPrinted], ".");
                    }
                    bytesWritten++;
                    data++;
                }
            }

            c8sPrinted += sprintf (&line[c8sPrinted], "\n");
            sa_output_print (out, line);
            
            c8sPrinted = 0;
            line[0] = 0; // clear the line
        }
    }

// #endif

