/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
#include "pch.h"
#include "ue_tiny_log.h"

// Doorbell space offsets
#define BE_UEREGISTER_MPU_SCRATCHPAD0_DB   0x180 // mpu_scratch0

// CSR space offsets - These are cloaked so offsets are coded directly.
#define BE_UEREGISTER_MPU_EP_SEMAPHORE     0xAC  // mpu.ep.ep_semaphore
#define BE_UEREGISTER_SEMAPHORE            0xDC8 // cev.host_semaphore1
#define BE_UEREGISTER_SCRATCHPAD           0xDE8 // cev.host_scratchpad1

// PCI Config Offsets
#define BE_UEREGISTER_PCICFG0_UESTATUS_LOW SA_FIELD_OFFSET (PCICFG_PF0_CSRMAP, ue_status_low)      // 0xA0
#define BE_UEREGISTER_PCICFG0_UESTATUS_HI  SA_FIELD_OFFSET (PCICFG_PF0_CSRMAP, ue_status_hi)       // 0xA4
#define BE_UEREGISTER_PCICFG0_UEMASK_LOW   SA_FIELD_OFFSET (PCICFG_PF0_CSRMAP, ue_status_low_mask) // 0xA8
#define BE_UEREGISTER_PCICFG0_UEMASK_HI    SA_FIELD_OFFSET (PCICFG_PF0_CSRMAP, ue_status_hi_mask)  // 0xAC
#define BE_UEREGISTER_SOFTRESET            SA_FIELD_OFFSET (PCICFG_PF0_CSRMAP, soft_reset)         // 0x5C

// Chip versions
#define BE_VERSION_UNKNOWN (0x0)
#define BE_VERSION_BE2 (0x20000)
#define BE_VERSION_BE3 (0x30000)

// Units for the waits
//#define BE_SLAVE_WAIT_UNITS  (10000000) // 10 seconds per 1 wait tick
//#define BE_MASTER_WAIT_UNITS (5000000)  // 5 seconds per 1 wait tick

// Enumeration for recovery state.
typedef enum {

    RSTATE_IDLE            = 0,
    RSTATE_START           = 1,
    RSTATE_MASTER_TIMEOUT  = 2,
    RSTATE_SLAVE_TIMEOUT   = 3,
    RSTATE_WAIT_FAT        = 4,
    RSTATE_RESET           = 5, 
    RSTATE_VERIFY_RECOVERY = 6,
    RSTATE_SUCCESS         = 7,
    RSTATE_ABORTED         = 8,
    RSTATE_AFTER_RESET_TIMEOUT = 9,
    
} RSTATE;


BESTATUS
be_ue_be3_reset_sequence (
    BE_ERROR_OBJECT *errobj);

void
be_ue_state_transition (
    BE_ERROR_OBJECT *errobj,
    u32 state);


//The following inline definitions can accomodate CSR space or PCICFG space 
//scratchpad and semaphore registers. BE3 uses CSR space registers; The plan
//is for the next generation chip to have a pcicfg space register to be able to
//work in sigle BAR mode.

u32 be_ue_read_scratchpad (
    BE_ERROR_OBJECT *errobj) 
{
    u32 val = 0xFFFFFFFF; // Sets the disable bits.

    switch (errobj->chip_version)
    {
        case BE_VERSION_BE3:
            val = BeCsrRead(errobj->sa_dev, BE_UEREGISTER_SCRATCHPAD, NULL);
            break;
        default:
            BREAKPOINT ();
            break;
    }

    return val;
}

void 
be_ue_write_scratchpad (BE_ERROR_OBJECT *errobj, u32 value)
{
    sa_tlog_trace64 (errobj->tlog, UE_TLOG_SCRATCHPAD_WRITE, value);

    switch (errobj->chip_version)
    {
        case BE_VERSION_BE3:
            BeCsrWrite (errobj->sa_dev, BE_UEREGISTER_SCRATCHPAD, value, NULL);
            break;
        default:
            BREAKPOINT ();
            break;
    }
}

void
_be_ue_force_release_semaphore (BE_ERROR_OBJECT *errobj)
{
    // Log our nasty behavior.
    sa_tlog_trace (errobj->tlog, UE_TLOG_STEAL_SEMAPHORE);

    switch (errobj->chip_version)
    {
    case BE_VERSION_BE3:
        // Write a 0 to release the semaphore.
        BeCsrWrite (errobj->sa_dev, BE_UEREGISTER_SEMAPHORE, 0, NULL);
        break;
    default:
        BREAKPOINT ();
        break;
    }
}

void
be_ue_release_semaphore (BE_ERROR_OBJECT *errobj)
{
    u32 val = 0;

    SA_ASSERT_DISPATCH ();

    switch (errobj->chip_version)
    {
    case BE_VERSION_BE3:
        val = BeCsrRead (errobj->sa_dev, BE_UEREGISTER_SEMAPHORE, NULL);
        
        //
        // Ensure that the function# calling release owns the semaphore
        //
        if (val == (errobj->pci_function_number + 1))
        {
            // Write a 0 to release the semaphore.
            BeCsrWrite (errobj->sa_dev, BE_UEREGISTER_SEMAPHORE, 0, NULL);
        } else {
            // Someone stole the semaphore!
            sa_tlog_trace64 (errobj->tlog, UE_TLOG_SEMAPHORE_STOLEN, val);
        }
        break;
    default:
        BREAKPOINT ();
        break;
    }

    sa_release_spinlock (&errobj->semaphore_lock, &errobj->irq);
}

BESTATUS
be_ue_try_to_acquire_semaphore (BE_ERROR_OBJECT *errobj)
{
    BESTATUS status;
    u32 readVal, writeVal;

    //
    // Acquire a spinlock to raise the CPU to DISPATCH (i.e. disable interrupts)
    //
    sa_acquire_spinlock (&errobj->semaphore_lock, &errobj->irq);

    SA_ASSERT_DISPATCH ();

    switch (errobj->chip_version)
    {
    case BE_VERSION_BE3:
        {
            writeVal = errobj->pci_function_number + 1;                         

            // Write to the semaphore reg & read it back to see
            // if write succeeded; if yes, the semaphore is ours
            // else it was already taken.                            
            BeCsrWrite (errobj->sa_dev, BE_UEREGISTER_SEMAPHORE, writeVal, NULL);
            readVal = BeCsrRead (errobj->sa_dev, BE_UEREGISTER_SEMAPHORE, NULL);
        
            if (readVal == writeVal) {
                status = BE_SUCCESS;
            } else {
                sa_release_spinlock (&errobj->semaphore_lock, &errobj->irq);

                // TRACE (DL_WARN, "Semaphore busy read:%d write:%d", readVal, writeVal);

                status = BE_STATUS_SEMAPHORE_BUSY;
            }
        }
        break;
    default:
        sa_release_spinlock (&errobj->semaphore_lock, &errobj->irq);
        BREAKPOINT ();
        status = BE_NOT_OK;
        break;
    }

    return status;
}


BESTATUS
be_ue_acquire_semaphore (
    BE_ERROR_OBJECT *errobj)
{
    BESTATUS status;
    i32 timeout;
    i32 stall_duration = 20; // microseconds
    u32 loops = 0;    

    do {

        timeout = errobj->semaphore_timeout; // timeout in microseconds

        status = be_ue_try_to_acquire_semaphore (errobj);
        while (status != BE_SUCCESS && timeout > 0) {

            sa_dev_stall (errobj->sa_dev, (u32)stall_duration); // stall 1 us
            
            timeout -= stall_duration;

            status = be_ue_try_to_acquire_semaphore (errobj);

            // Progessively reduce the stall to increase the fairness of
            // the semaphore acquisition.  We need to avoid starving one
            // function access to the semaphore.
            if (stall_duration > 1) {
                stall_duration -= 1;
            }
        }

        //
        // If overthrowing the master, we also are unscrupulous enough to
        // thieve the semaphore.
        //
        if (errobj->crash_dump) {            
            _be_ue_force_release_semaphore (errobj);
        }

    } while (errobj->crash_dump && status != BE_SUCCESS && loops++ < 2);

    if (status != BE_SUCCESS) {
        sa_tlog_trace (errobj->tlog, UE_TLOG_SEMAPHORE_TIMEOUT);
        sa_release_spinlock (&errobj->semaphore_lock, &errobj->irq);
    }

    return status;
}


u32
be_ue_chip_version (
    IN BE_ERROR_OBJECT *errobj)
{
    PCICFG_ID_F0_CSR id;

    SA_TOU32 (id) = BePciCfgRead (errobj->sa_dev, 
                                  SA_FIELD_OFFSET (PCICFG_PF0_CSRMAP, id), 
                                  NULL);

    if (id.deviceid == 0x211 ||
        id.deviceid == 0x212 ||
        id.deviceid == 0x215 ||
        id.deviceid == 0x700 ||
        id.deviceid == 0x702 ||
        id.deviceid == 0x703)
    {
        return BE_VERSION_BE2; 
    }

    if (id.deviceid == 0x710 ||
        id.deviceid == 0x712 ||
        id.deviceid == 0x221 ||
        id.deviceid == 0x222)
    {
        return BE_VERSION_BE3;
    }
    
    return BE_VERSION_UNKNOWN;
}

BESTATUS
be_ue_driver_load (
    IN BE_ERROR_OBJECT *errobj)
{
    BESTATUS status = BE_SUCCESS;
    
    SA_ASSERT_PASSIVE ();
    
    status = be_ue_check_hardware_ue (errobj);
    if (status == BE_SUCCESS) {
        // No UE, we are done.
        goto Error;
    }

    // Log UE during driver load.
    sa_tlog_trace (errobj->tlog, UE_TLOG_RECOVERY_DURING_LOAD); 

    if (!errobj->supports_recovery) {
        // BE2, no recovery possible.
        return BE_STATUS_RECOVERY_FAILED;
    }


    //
    // UE exists during load!
    // Try a reset.  Most likely we will be the master, but another device
    // may have loaded at the same time.
    //
    status = be_ue_recovery (errobj, FALSE);
    if (status == BE_SUCCESS) {
        // Reset successfull.
        sa_tlog_trace (errobj->tlog, UE_TLOG_INIT_RECOVERY);
    } else {
        sa_tlog_trace (errobj->tlog, UE_TLOG_RECOVERY_FAILED_DURING_LOAD);
    }

Error:
    return status;
}

//
// Queries if the FAT dump is still being written.
//
boolean
be_ue_fat_in_progress (
    BE_ERROR_OBJECT *errobj)
{
    MGMT_HBA_POST_STATUS_STRUCT postStat;

    if (BeCsrBarAvailable (errobj->sa_dev)) {

        SA_TOU32 (postStat) = 
            BeCsrRead (errobj->sa_dev, BE_UEREGISTER_MPU_EP_SEMAPHORE, NULL);

    } else {

        // VMs don't need this information.
        return FALSE;

    }

    //
    // ARM firmware asserts the UE after FAT completes.
    //
    if ((postStat.stage & POST_STAGE_ARMFW_UE) != POST_STAGE_ARMFW_UE) {
        return TRUE; // Fat dump still in progress
    }

    return FALSE;
}

BESTATUS
be_ue_recovery_status (
    BE_ERROR_OBJECT *errobj)
{
    MGMT_HBA_POST_STATUS_STRUCT postStat;

    if (BeCsrBarAvailable (errobj->sa_dev)) {

        SA_TOU32 (postStat) = 
            BeCsrRead (errobj->sa_dev, BE_UEREGISTER_MPU_EP_SEMAPHORE, NULL);

    } else {

        SA_TOU32 (postStat) = 
            BePdRead (errobj->sa_dev, BE_UEREGISTER_MPU_SCRATCHPAD0_DB, NULL);  

    }

    //
    // If the UE is no longer asserted by the ARM, the reset worked.
    //
    if ((postStat.stage & POST_STAGE_ARMFW_UE) == POST_STAGE_ARMFW_UE) {
        return BE_STATUS_RECOVERY_FAILED;
    }

    return BE_SUCCESS;
}

BESTATUS
be_ue_restore_scratchpad (
    BE_ERROR_OBJECT *errobj,
    UE_SCRATCHPAD scratchpad)
{
    UE_SCRATCHPAD restore;

    BESTATUS status = be_ue_acquire_semaphore (errobj);
    if (status  != BE_SUCCESS) {
        // We timed out out waiting for the semaphore
        goto Error;
    }

    SA_TOU32 (restore) = be_ue_read_scratchpad (errobj);

    //
    // Restore fields in scratchpad after reset. 
    // These fields may be modified by applicatinos or other drivers
    // and it is best to restore them immediately (possibly before the ARM reboots).
    //
    SA_TOU32 (restore) = 0;
    //restore.master_wait = scratchpad.master_wait;
    //restore.slave_wait = scratchpad.slave_wait;
    restore.no_soft_ue = scratchpad.no_soft_ue;

    be_ue_write_scratchpad (errobj, SA_TOU32 (restore));

    be_ue_release_semaphore (errobj);

Error:
    return status;
}

//
// Called by the UE_MASTER to verify that conditions still warrant
// the soft reset.  Returns TRUE if the reset should proceed, FALSE
// otherwise.  This will clear the MASTER setting in the scratchpad 
// if the reset is aborted.  This also returns the current scratchpad
// value for restoring after the reset.
//
boolean
_be_ue_verify_reset_required (
    BE_ERROR_OBJECT *errobj,
    UE_SCRATCHPAD *saved_scratchpad)
{
    // Assume we will perform the reset at this point.
    boolean do_reset = FALSE;
    UE_SCRATCHPAD scratchpad;
    BESTATUS status;

    SA_TOU32 (*saved_scratchpad) = 0;

    status = be_ue_acquire_semaphore (errobj);
    if (status != BE_SUCCESS) {
        goto Error;
    }

    SA_TOU32 (scratchpad) = be_ue_read_scratchpad (errobj);
    do_reset = TRUE;
    
    if (scratchpad.disable) {
        // Driver disabled recovery! 
        do_reset = FALSE;
        status = BE_STATUS_RECOVERY_DISABLED;
    }

    if (!scratchpad.master ||
        scratchpad.master_function != errobj->pci_function_number)
    {
        // Someone stole the master role!  Let someone else reset.
        do_reset = FALSE;

        // Report the slave overthrow.
        sa_tlog_trace64 (errobj->tlog, UE_TLOG_SLAVE_COOP, SA_TOU32(scratchpad));
    }

    //
    // Verify if another entity both stole MASTER role and 
    // already cleared the UE condition.
    //
    status = be_ue_check_hardware_ue (errobj);
    if (status == BE_SUCCESS) {
        do_reset = FALSE;
    }

    if (!do_reset) {

        //
        // Release the master if we are not performing the reset.
        //
        if (scratchpad.master &&
            scratchpad.master_function == errobj->pci_function_number)
        {
            scratchpad.master = 0;
            be_ue_write_scratchpad (errobj, SA_TOU32 (scratchpad));
        }

    }

    be_ue_release_semaphore (errobj);

    // Save this for restoring later.
    *saved_scratchpad = scratchpad;

Error:
    return do_reset;
}



BESTATUS 
be_ue_reset (
    BE_ERROR_OBJECT *errobj)
{
    BESTATUS status = BE_SUCCESS;
    boolean do_reset;
    UE_SCRATCHPAD scratchpad;

    ASSERT (errobj->master);

    //
    // Determine if reset is necessary and save the current scratchpad.
    //
    do_reset = _be_ue_verify_reset_required (errobj, &scratchpad);
   
    if (do_reset) {

        //
        // Perform the actual reset
        //
        status = be_ue_be3_reset_sequence (errobj);

        if (status == BE_SUCCESS) {

            sa_tlog_trace (errobj->tlog, UE_TLOG_RESET_SUCCESS);

            //
            // Restore the scratchpad.
            //
            be_ue_restore_scratchpad (errobj, scratchpad);

        } else {
            sa_tlog_trace (errobj->tlog, UE_TLOG_RESET_FAIL);
        }
    }

    return status;
}


// Low level reset code.
BESTATUS
_be_ue_be3_issue_soft_reset (
    BE_ERROR_OBJECT *errobj)
{
    PCICFG_SOFT_RESET_CSR pciSoftReset;    
    /* i32 timeout = errobj->pci_reset_poll_timeout; // total timeout in microseconds */
    /* i32 delay = 10; // per delay time microseconds */
    BESTATUS status = 0;

    SA_TOU32 (pciSoftReset) = 
        BePciCfgRead (errobj->sa_dev, 
                      SA_FIELD_OFFSET (PCICFG_PF0_CSRMAP, soft_reset), 
                      NULL);
        
    pciSoftReset.softreset = 1;
    
    BePciCfgWrite (errobj->sa_dev, 
                   SA_FIELD_OFFSET (PCICFG_PF0_CSRMAP, soft_reset), 
                   SA_TOU32 (pciSoftReset), 
                   NULL);

    // wait til soft reset is deasserted - hardware
    // deasserts after some time.
    // while (pciSoftReset.softreset && timeout > 0) {
    // 
    //     sa_dev_stall (errobj->sa_dev, delay); 
    // 
    //     timeout -= delay;
    // 
    //     SA_TOU32 (pciSoftReset) = 
    //         BePciCfgRead (errobj->sa_dev, 
    //                       SA_FIELD_OFFSET (PCICFG_PF0_CSRMAP, soft_reset), 
    //                       NULL);           
    // };
    // 
    // if (pciSoftReset.softreset != 0) {
    //     sa_tlog_trace (errobj->tlog, UE_TLOG_RESET_ERROR);
    //     status = BE_NOT_OK;        
    // }

    return status;
}


//
// Full reset sequence for BE3.
//
BESTATUS
be_ue_be3_reset_sequence (
    BE_ERROR_OBJECT *errobj)
{
    BESTATUS status = BE_SUCCESS;    
    /* PCICFG_ONLINE0_CSR pciOnline0;
    PCICFG_ONLINE1_CSR pciOnline1; */
    u32 i;

    for (i=0; i<1; i++) {

        // Stall for a bit.
        sa_dev_stall (errobj->sa_dev, 20000);  
    
        BePciCfgWrite (errobj->sa_dev, SA_FIELD_OFFSET (PCICFG_PF0_CSRMAP, online0), 0, NULL);
        BePciCfgWrite (errobj->sa_dev, SA_FIELD_OFFSET (PCICFG_PF0_CSRMAP, online1), 0, NULL);

        // Mask everything 
        BePciCfgWrite (errobj->sa_dev, SA_FIELD_OFFSET (PCICFG_PF0_CSRMAP, ue_status_low_mask), 0xFFFFFFFF, NULL);
        BePciCfgWrite (errobj->sa_dev, SA_FIELD_OFFSET (PCICFG_PF0_CSRMAP, ue_status_hi_mask),  0xFFFFFFFF, NULL);

        sa_dev_stall (errobj->sa_dev, 20000);  

        // Issue soft reset
        status = _be_ue_be3_issue_soft_reset (errobj);
        if (status != BE_SUCCESS) {
            goto Error;
        }

        // delay for a bit
        sa_dev_stall (errobj->sa_dev, 20000);  
    }

Error:
    return status;
}

//
// Drivers call this to start the UE recovery process. 
// This will determine if the driver is a master or slave.
// If this returns BE_PENDING, the driver should poll be_ue_recovery_poll
// to complete the algorithm.
//
BESTATUS
be_ue_recovery_start (
    BE_ERROR_OBJECT *errobj,
    boolean crash_dump)
{
    BESTATUS status;
    SA_IRQ irq;

    if (!errobj->supports_recovery) {
        return BE_STATUS_RECOVERY_FAILED;
    }

    // Cache the UE registers at the start of recovery for logging.  
    // Assert that a UE is expected at this point.
    status = be_ue_check_hardware_ue (errobj);
    ASSERT (status != BE_SUCCESS);

    if (!BeCsrBarAvailable (errobj->sa_dev)) {
        
        //
        // Virtual functions are always slaves.
        //
        errobj->master = FALSE;

    } else {

        UE_SCRATCHPAD scratchpad;

        // This will allow the errobj to steal the semaphore after a timeout.
        // If the crash dump driver is truly the only one running, this is overkill.
        // However, this flag tries to safely steal the master privilege so it could
        // potentially be used by a high priority boot driver as well.
        errobj->crash_dump = crash_dump;

        //
        // Physical function elect a master.
        //
        status = be_ue_acquire_semaphore (errobj);
        if (status != BE_SUCCESS && !errobj->crash_dump) {        
            // We timed out out waiting for the semaphore        
            goto Error;        
        }

        SA_TOU32 (scratchpad) = be_ue_read_scratchpad (errobj);

        // Check if recovery is disabled for any reason.
        if (scratchpad.disable != 0)
        {
            status = BE_STATUS_RECOVERY_DISABLED;
            be_ue_release_semaphore (errobj);
            goto Error;
        }

        if (scratchpad.master == 0 || crash_dump) {
        
            sa_tlog_trace (errobj->tlog, UE_TLOG_MASTER);

            // Elect myself master.
            errobj->master = TRUE;

            scratchpad.master = 1;
            scratchpad.master_function = errobj->pci_function_number;

            be_ue_write_scratchpad (errobj, SA_TOU32 (scratchpad));

        } else {

            errobj->master = FALSE;

        }

        be_ue_release_semaphore (errobj);

        //
        // Save the timeout values.
        //
        //errobj->master_timeout = (5  * (scratchpad.master_wait + 1)) * 1000000ULL; // microseconds
        //
        //errobj->slave_timeout  = (10 * (scratchpad.slave_wait + 1)) * 1000000ULL + // microseconds
        //                         errobj->master_timeout;

    }
    

    //
    // Log the start of recovery and the last UE detected.
    //
    sa_tlog_trace64 (errobj->tlog, UE_TLOG_RECOVERY_START, be_ue_get_last_ue_status (errobj));

    //
    // Acquire the recovery 
    //
    sa_acquire_spinlock (&errobj->recovery_lock, &irq);    

    be_ue_state_transition (errobj, RSTATE_START);
    
    sa_release_spinlock (&errobj->recovery_lock, &irq);

    status = BE_PENDING;

Error:

    return status;
}

// Internal state transition.
void
be_ue_state_transition (
    BE_ERROR_OBJECT *errobj,
    u32 state)
{
    u32 recovery_microseconds;

    // Update the state variable.
    errobj->recovery_state = state;

    // Record the time we enter this state.
    errobj->start_time = sa_get_system_time ();

    // For the final states we log the total recovery time, precomputed here.
    recovery_microseconds = sa_diff_system_time (errobj->original_time, sa_get_system_time ());
    
    // Log a nice state transition message.
    switch (state)
    {
        case RSTATE_SLAVE_TIMEOUT:        sa_tlog_trace64 (errobj->tlog, UE_TLOG_SLAVE_TIMEOUT, recovery_microseconds);  break;    
        case RSTATE_MASTER_TIMEOUT:       sa_tlog_trace64 (errobj->tlog, UE_TLOG_MASTER_TIMEOUT, recovery_microseconds)  ; break;    
        case RSTATE_IDLE:                 sa_tlog_trace (errobj->tlog, UE_TLOG_IDLE)            ; break;    
        case RSTATE_START:                sa_tlog_trace (errobj->tlog, UE_TLOG_START)           ; break;
        case RSTATE_VERIFY_RECOVERY:      sa_tlog_trace64 (errobj->tlog, UE_TLOG_VERIFY_RECOVERY, recovery_microseconds) ; break;
        case RSTATE_WAIT_FAT:             sa_tlog_trace64 (errobj->tlog, UE_TLOG_WAIT_FAT, recovery_microseconds);  break;
        case RSTATE_RESET:                sa_tlog_trace64 (errobj->tlog, UE_TLOG_RESET, recovery_microseconds)           ; break;
        case RSTATE_AFTER_RESET_TIMEOUT:  sa_tlog_trace64 (errobj->tlog, UE_TLOG_POST_RESET, recovery_microseconds)  ; break;
        case RSTATE_ABORTED:              sa_tlog_trace64 (errobj->tlog, UE_TLOG_ABORTED, recovery_microseconds); break;
        case RSTATE_SUCCESS:              sa_tlog_trace64 (errobj->tlog, UE_TLOG_DONE, recovery_microseconds);  break;
        default:
            break;
    }
}

// Returns TRUE if the current timeout for the state is exceeded.
STATIC INLINE boolean
_be_ue_state_timeout (
    BE_ERROR_OBJECT *errobj,
    u64 timeout)
{
    u64 now = sa_get_system_time ();
    return (boolean)(sa_diff_system_time (errobj->start_time, now) > timeout);    
}

//
// Drivers poll this (either at passive or dispatch) to complete the UE
// recovery. This return BE_PENDING while the recovery is in progress, and 
// either BE_SUCCESS or BE_STATUS_RECOVERY_FAILED when it finishes.
//
BESTATUS
be_ue_recovery_poll (
    BE_ERROR_OBJECT *errobj)
{
    BESTATUS status = BE_PENDING;
    u32 state;
    SA_IRQ irq;
    /* UE_SCRATCHPAD scratchpad; */

    if (!errobj->supports_recovery) {
        return BE_STATUS_RECOVERY_FAILED;
    }

    sa_acquire_spinlock (&errobj->recovery_lock, &irq);

    do {
       
        state = errobj->recovery_state;

        switch (state)
        {
        case RSTATE_IDLE:
            status = BE_SUCCESS;
            break;

        case RSTATE_ABORTED:            
            status = BE_STATUS_RECOVERY_FAILED;
            break;

        case RSTATE_SUCCESS:
            status = BE_SUCCESS;
            break;

        case RSTATE_START:

            // Save the original start time to see how long recovery takes.
            errobj->original_time = errobj->start_time;
            if (errobj->master) {
                be_ue_state_transition (errobj, RSTATE_MASTER_TIMEOUT);
            } else {
                be_ue_state_transition (errobj, RSTATE_SLAVE_TIMEOUT);                    
            }

            break;

        case RSTATE_SLAVE_TIMEOUT:
            // Using the simple timeout mode, slaves sleep for a fixed timeout to
            // avoid touching hardware registers during the reset sequence.  After
            // the timeout, they poll for the reset status.
            if (_be_ue_state_timeout (errobj, errobj->slave_timeout)) {
                be_ue_state_transition (errobj, RSTATE_VERIFY_RECOVERY);
            }
            break;

        case RSTATE_MASTER_TIMEOUT:
            //
            // Wait for the master timeout.
            //
            ASSERT (errobj->master);
            if (_be_ue_state_timeout (errobj, errobj->master_timeout)) {
                be_ue_state_transition (errobj, RSTATE_WAIT_FAT);
            }
            break;
            
        case RSTATE_VERIFY_RECOVERY:
            if (be_ue_recovery_status (errobj) == BE_SUCCESS) {
                //
                // If the hardware UE condition is cleared, we report success for
                // this algorithm.
                be_ue_state_transition (errobj, RSTATE_SUCCESS);
                
            } else {
                //
                // UE condition still exists. Report failure.
                //
                be_ue_state_transition (errobj, RSTATE_ABORTED);
            }
            break;

        case RSTATE_WAIT_FAT:
            ASSERT (errobj->master);
            if (be_ue_fat_in_progress (errobj)) {
                if (_be_ue_state_timeout (errobj, errobj->fat_timeout))
                {
                    // FAT timeout.  Not fatal, so log an extra tlog event.
                    sa_tlog_trace (errobj->tlog, UE_TLOG_FAT_TIMEOUT);
                    be_ue_state_transition (errobj, RSTATE_RESET);
                }
            } else {
                // fat done
                be_ue_state_transition (errobj, RSTATE_RESET);
            }
            break;

        case RSTATE_AFTER_RESET_TIMEOUT:
            if (_be_ue_state_timeout (errobj, errobj->after_reset_timeout))
            {
                be_ue_state_transition (errobj, RSTATE_VERIFY_RECOVERY);
            }
            break;

        case RSTATE_RESET:
            // Reset the chip, the master returns RECOVERY_DISABLED if another
            // function steals the master privilege, so ignore the status for
            // now. The post master wait allows the other function to finish
            // the reset.
            ASSERT (errobj->master);
            be_ue_reset (errobj);
            
            // Either the reset failed, or the master privilege was stolen.
            // Wait for the slave timeout in case the later is true before
            // verifying the result.
            be_ue_state_transition (errobj, RSTATE_AFTER_RESET_TIMEOUT);
            break;

        default:
            BREAKPOINT (); // bad state
            status = BE_NOT_OK;
            break;
        }

    } while (state != errobj->recovery_state);

    sa_release_spinlock (&errobj->recovery_lock, &irq);    

    return status;
}

//
// This function must be called at passive.  It combines
// the be_ue_recovery_start and be_ue_recovery_poll to implement
// the full recovery algorithm.
//
BESTATUS
be_ue_recovery (
    BE_ERROR_OBJECT *errobj,
    boolean crash_dump)
{
    BESTATUS status;

    SA_ASSERT_PASSIVE ();

    status = be_ue_recovery_start (errobj, crash_dump);
    if (status < 0) {
        goto Error;
    }

    while (status == BE_PENDING) 
    {
        sa_sleep (500000); // delay in microseconds

        status = be_ue_recovery_poll (errobj);
    }

Error:
    return status;
}

//
// Checks hardware for an asserted UE.  This 
// never caches the UE result -- only the UE masks.
//
BESTATUS
be_ue_check_hardware_ue (
    BE_ERROR_OBJECT *errobj)
{
    BESTATUS status = BE_SUCCESS;
    u32 ue0, ue1, ue0mask, ue1mask;

    if (!errobj) {
        // UE queries are not supported without an error object.
        TRACE (DL_WARN, "Warning: Check hardware UE not possible without UE object.");
        return BE_SUCCESS;
    }

    ue0mask = errobj->ue0mask;
    ue1mask = errobj->ue1mask;
    ue0 = errobj->ue0 = BePciCfgRead (errobj->sa_dev, BE_UEREGISTER_PCICFG0_UESTATUS_LOW, NULL);
    ue1 = errobj->ue1 = BePciCfgRead (errobj->sa_dev, BE_UEREGISTER_PCICFG0_UESTATUS_HI, NULL);

    if ((ue0 & ~ue0mask) || (ue1 & ~ue1mask)) {

        TRACE (DL_ERR, 
               "UE Detected. ue0:0x%x ue1:0x%x", 
               (ue0 & ~ue0mask), 
               (ue1 & ~ue1mask));

        status = BE_STATUS_UE_DETECTED;
    }

    return status;
}

// Returns the last UE value cached by be_ue_check_hardware_ue.
u64
be_ue_get_last_ue_status (
    BE_ERROR_OBJECT *errobj)
{
    u64 value = ((u64)(errobj->ue1 & ~errobj->ue1mask) << 32) |
                ((u64)(errobj->ue0 & ~errobj->ue0mask));
    return value;
}


//
// Generates a software UE, which is actually a TRUE hardware UE.
// This is only possible from the physical function.
//
BESTATUS
be_ue_generate_software_ue (
    BE_ERROR_OBJECT *errobj,
    u64 optional_code)
{
    UE_SCRATCHPAD scratchpad;
    BESTATUS status = BE_SUCCESS;

    if (!errobj->supports_recovery) {
        return BE_STATUS_WARN_NO_RECOVERY; // BE2
    }

    if (BeCsrBarAvailable (errobj->sa_dev)) {

        SA_TOU32 (scratchpad) = be_ue_read_scratchpad (errobj);
        
        if (scratchpad.no_soft_ue) {
            sa_tlog_trace64 (errobj->tlog, UE_TLOG_SOFTWARE_UE_DISABLED, optional_code);
            status = BE_STATUS_SOFTWARE_UE_IGNORED;
            goto Error;
        }

        sa_tlog_trace64 (errobj->tlog, UE_TLOG_SOFTWARE_UE, optional_code);

        // Enable all UEs for the MPU (mpu.ue_enable).
        BeCsrWrite (errobj->sa_dev, 0x138, 0xFFFFFFFF, NULL);

        // Write to last register in the CSR bar.
        BeCsrWrite (errobj->sa_dev, 0x1FFFC, 0, NULL);

        TRACE (DL_ERR, "Generate software UE. Function:%d", errobj->pci_function_number);

    } else {

        TRACE (DL_ERR, "Software UE ignored. No CSR access.");

        sa_tlog_trace64 (errobj->tlog, UE_TLOG_SOFTWARE_UE_IGNORED, optional_code);

        status = BE_STATUS_SOFTWARE_UE_IGNORED;
    }

Error:

    return status;
}


//
// Sets timeouts for the UE algorithm.
//
BESTATUS
be_ue_set_timeouts (
    BE_ERROR_OBJECT *errobj,
    BE_UE_TIMEOUT_ENUM which_value,
    u64 value)
{
    BESTATUS status = BE_SUCCESS;
    /* u32 wait; 
    UE_SCRATCHPAD scratchpad; */

    //
    // Modify the scratchpad for the MASTER and SLAVE wait values.
    //
    /*
    if (which_value == UE_TIMEOUT_SLAVE ||
        which_value == UE_TIMEOUT_MASTER)
    {
        status = be_ue_acquire_semaphore (errobj);
        if (status != BE_SUCCESS) {
            // We timed out out waiting for the semaphore.
            goto Error;
        }

        SA_TOU32 (scratchpad) = be_ue_read_scratchpad (errobj);

        if (which_value == UE_TIMEOUT_SLAVE) {

            wait = (u32)(value / BE_SLAVE_WAIT_UNITS);

            // This truncates to the maximum value supported in scratchpad.
            wait = MIN (0x7, wait); 

            // This uses the existing setting if another driver requested a longer wait.
            wait = MAX (wait, scratchpad.slave_wait);

            scratchpad.slave_wait = wait;
        } else {
            
            wait = (u32)(value / BE_MASTER_WAIT_UNITS);

            // This truncates to the maximum value supported in scratchpad.
            wait = MIN (0x7, wait); 

            // This uses the existing setting if another driver requested a longer wait.
            wait = MAX (wait, scratchpad.master_wait);

            scratchpad.master_wait = wait;
        }

        be_ue_write_scratchpad (errobj, SA_TOU32 (scratchpad));

        be_ue_release_semaphore (errobj);
    }
    */

    // Save the software values.  This is not really useful for the master/slave waits
    // since these are modified during UE recovery to the register values.
    switch (which_value)
    {
    case UE_WHICH_TIMEOUT_SLAVE          : errobj->slave_timeout          = value ; break;
    case UE_WHICH_TIMEOUT_SEMAPHORE      : errobj->semaphore_timeout      = value ; break;
    case UE_WHICH_TIMEOUT_MASTER         : errobj->master_timeout         = value ; break;
    case UE_WHICH_TIMEOUT_FAT            : errobj->fat_timeout            = value ; break;
    case UE_WHICH_TIMEOUT_PCI_RESET_POLL : errobj->pci_reset_poll_timeout = value ; break;
    case UE_WHICH_TIMEOUT_AFTER_RESET    : errobj->after_reset_timeout    = value ; break;
    default:
        BREAKPOINT ();
        return BE_NOT_OK;
        break;
    }

    return status;
}

//
// Creates and initializes the UE object at driver load time.
//
BESTATUS
be_ue_create (
    SA_DEV *sa_dev,
    BE_ERROR_OBJECT *errobj)
{
    PCICFG_MEMBAR_CTRL_INT_CTRL_CSR membar_ctrl_int_ctrl;

    sa_zero_mem (errobj, sizeof (BE_ERROR_OBJECT));

    sa_init_spinlock (&errobj->semaphore_lock);
    sa_init_spinlock (&errobj->recovery_lock);
    
    ASSERT (sa_dev);
    errobj->sa_dev = sa_dev;
    errobj->recovery_state = RSTATE_IDLE;

    // Timeouts in microseconds.
    if (BeCsrBarAvailable (errobj->sa_dev)) {
        errobj->slave_timeout          = UE_TIMEOUT_SLAVE;   // 25 seconds
    } else {
        errobj->slave_timeout          = UE_TIMEOUT_VM;      // 30 seconds
    }
    errobj->semaphore_timeout      = UE_TIMEOUT_SEMAPHORE;  // 20 milliseconds spinning at dispatch 
    errobj->fat_timeout            = UE_TIMEOUT_FAT;        // 10 seconds
    errobj->master_timeout         = UE_TIMEOUT_MASTER;     // 10 seconds    
    errobj->pci_reset_poll_timeout = UE_TIMEOUT_RESET_POLL; // 50 milliseconds
    errobj->after_reset_timeout    = UE_TIMEOUT_AFTER_RESET; // 20 seconds

    // Cache the UE mask values.
    errobj->ue0mask = BePciCfgRead (errobj->sa_dev, BE_UEREGISTER_PCICFG0_UEMASK_LOW, NULL);
    errobj->ue1mask = BePciCfgRead (errobj->sa_dev, BE_UEREGISTER_PCICFG0_UEMASK_HI,  NULL);

    // Read the PCI function number
    SA_TOU32 (membar_ctrl_int_ctrl) = 
        BePciCfgRead (sa_dev, SA_FIELD_OFFSET (PCICFG_PF0_CSRMAP, membar_ctrl_intr_ctrl), NULL);
    errobj->pci_function_number = membar_ctrl_int_ctrl.PFNUM;

    // TODO: Get the chip version.
    errobj->chip_version = be_ue_chip_version (errobj); // BE3   

    // Determine if recovery is supported.
    errobj->supports_recovery = (boolean)(errobj->chip_version == BE_VERSION_BE3);

    return BE_SUCCESS;
}

//
// Destroys the UE object during driver unload.
//
void
be_ue_destroy (
    BE_ERROR_OBJECT *errobj)
{
    //
    // Reenable UE recovery on the system when this driver unloads.
    //
    if (errobj->sa_dev) {
        be_ue_disable_recovery (errobj, FALSE);
    }

    sa_zero_mem (errobj, sizeof (BE_ERROR_OBJECT));
}

//
// Links a tiny log object to this UE object for tracking
// the sequence of operations.
//
void
be_ue_link_tiny_log (
    BE_ERROR_OBJECT *errobj,
    SA_TLOG *tlog)
{
    ASSERT (errobj);
    errobj->tlog = tlog;
}

u32
_be_ue_get_disable_bit (
    u32 pci_function_number)
{
    u32 disable;

    // Choose the bit to set.
    switch (pci_function_number) {
        case 0: disable = UE_DISABLE_PF0; break;
        case 1: disable = UE_DISABLE_PF1; break;
        case 2: disable = UE_DISABLE_PF2; break;
        case 3: disable = UE_DISABLE_PF3; break;
        case 4: disable = UE_DISABLE_PF4; break;
        case 5: disable = UE_DISABLE_PF5; break;
        case 6: disable = UE_DISABLE_PF6; break;
        case 7: disable = UE_DISABLE_PF7; break;
        default:
            disable = 0;
            break;
    }

    return disable;
}

BESTATUS
be_ue_disable_recovery (
    BE_ERROR_OBJECT *errobj,
    boolean disable)
{
    u32 bit;
    BESTATUS status;
    UE_SCRATCHPAD scratchpad;

    if (!errobj->supports_recovery) {
        return BE_SUCCESS; // BE2
    }

    if (!BeCsrBarAvailable (errobj->sa_dev)) {
        return BE_SUCCESS;
    }

    status = be_ue_acquire_semaphore (errobj);
    if (status != BE_SUCCESS) {
        // We timed out out waiting for the semaphore.
        goto Error;
    }

    SA_TOU32 (scratchpad) = be_ue_read_scratchpad (errobj);

    bit = _be_ue_get_disable_bit (errobj->pci_function_number);

    if (disable) {
        scratchpad.disable |= bit;
    } else {
        scratchpad.disable &= ~bit;
    }

    be_ue_write_scratchpad (errobj, SA_TOU32 (scratchpad));

    be_ue_release_semaphore (errobj);

Error:
    return status;
}


